! $Id: AllocateStmts.f90 2815 2019-03-24 02:47:45Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the AllocateStmts module.


!*******************************************************************************
!!
!> Procedures and types for parsing /allocate-stmts/'s (R626).

MODULE AllocateStmts
  
  USE DeclarationTypeSpecs
  USE Designators
  USE Expressions
  USE Identifiers
  USE ParseNodes
  USE SubscriptLists
  USE Variables
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseAllocateObject
  PUBLIC :: ParseAllocateStmt
  
  !-----------------------------------------------------------------------------
  
  !> Abstract parent class for an /allocate-object/.
  TYPE, EXTENDS(TermParseNode), PUBLIC, ABSTRACT :: AllocateObject
  CONTAINS
    PROCEDURE, NOPASS :: What => ao_What
  END TYPE AllocateObject
  
  !> An item in an array of /allocate-object/'s.
  TYPE, EXTENDS(TermParseNode), PUBLIC :: Allocation
    !> The /allocate-object/.
    !!
    !! Always allocated.
    CLASS(AllocateObject), ALLOCATABLE :: object
    !> Optional /allocate-shape-spec-list/.
    !!
    !! Not allocated if there was no shape spec in the /allocation/.
    TYPE(DimSpec), ALLOCATABLE :: shape_spec(:)
    !> Optional /allocate-coarray-spec/.
    !!
    !! Not allocated if there was no coarray spec in the /allocation/.
    TYPE(DimSpec), ALLOCATABLE :: coarray_spec(:)
  CONTAINS
    PROCEDURE :: Write => al_Write
    PROCEDURE :: Query => al_Query
    PROCEDURE :: Modify => al_Modify
    PROCEDURE, NOPASS :: What => al_What
    PROCEDURE :: GetLocation => al_GetLocation
  END TYPE Allocation
  
  !> Represents the results of parsing an /allocate-stmt/.
  TYPE, EXTENDS(SCParseNode), PUBLIC ::  AllocateStmt
    !> The optional /type-spec/ that may come before the  allocatable 
    !! objects in the allocate statement.
    !!
    !! Not allocated if the /type-spec/ isn't present.  Will not be allocated 
    !! if mold or source are allocated.
    TYPE(TypeSpec), ALLOCATABLE :: type_spec
    !> The things being allocated.
    !!
    !! Always allocated, always size one or greater.
    TYPE(Allocation), ALLOCATABLE :: allocations(:)
    !> The expression nominated in a SOURCE= option.
    !!
    !! Not allocated if there was no SOURCE= option.  Will not be allocated 
    !! if type_spec or mold is allocated.
    CLASS(Expression), ALLOCATABLE :: source
    !> The expression nominated in a MOLD= option.
    !!
    !! Not allocated if there was no MOLD= option.  Will not be allocated if 
    !! type_spec or source is allocated.
    CLASS(Expression), ALLOCATABLE :: mold
    !> The variable nominated in a STAT= option.
    !!
    !! Not allocated if there was no STAT= option.
    CLASS(Variable), ALLOCATABLE :: stat
    !> The variable nominated in a ERRMSG= option.
    !!
    !! Not allocated if there was no ERRMSG= option.
    CLASS(Variable), ALLOCATABLE :: errmsg
  CONTAINS
    PROCEDURE :: Dump => alloc_Dump
    PROCEDURE :: Query => alloc_Query
    PROCEDURE :: Modify => alloc_Modify
    PROCEDURE, NOPASS :: What => alloc_What
  END TYPE AllocateStmt
  
  !-----------------------------------------------------------------------------
  ! Types that implement the various /allocate-object/ extensions.
  
  !> The result of parsing an /allocate-object/ that is a /variable-name/.
  TYPE, PUBLIC, EXTENDS(AllocateObject) :: AllocateObjectVariable
    !> The /variable-name/.
    TYPE(Identifier) :: name
  CONTAINS
    PROCEDURE :: Write => aov_Write
    PROCEDURE :: Query => aov_Query
    PROCEDURE :: Modify => aov_Modify
    PROCEDURE :: Advertise => aov_Advertise
    PROCEDURE :: GetLocation => aov_GetLocation
  END TYPE AllocateObjectVariable
  
  !> The result of parsing an /allocate-object/ that is a 
  !! /structure-component/.
  TYPE, PUBLIC, EXTENDS(AllocateObject) :: AllocateObjectStructureComponent
    !> The /part-ref/'s that make up the /structure-component/.
    !!
    !! Always allocated with size greater than one.
    TYPE(PartRef), ALLOCATABLE :: parts(:)
  CONTAINS
    PROCEDURE :: Write => aosc_Write
    PROCEDURE :: Query => aosc_Query
    PROCEDURE :: Modify => aosc_Modify
    PROCEDURE :: Advertise => aosc_Advertise
    PROCEDURE :: GetLocation => aosc_GetLocation
  END TYPE AllocateObjectStructureComponent
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /allocate-stmt/ (R626).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    alloc_stmt        Results of parsing the /allocate-stmt/.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseAllocateStmt(tlist, alloc_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(AllocateStmt), INTENT(OUT) :: alloc_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule for an /allocate-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R626'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect* calls.
    INTEGER :: closing        ! Index of the closing delimiter.
    INTEGER :: it             ! Current token index.
    INTEGER :: pass           ! Current pass number.
    INTEGER :: alloc_count    ! Number of allocations in the list.
    INTEGER :: declare        ! Index of the declare (::) token.
    INTEGER :: assign         ! Index of the assign (=) token.
    INTEGER :: list_count     ! Number of items in the list.
    INTEGER :: item           ! List item index.
    INTEGER :: item_start     ! Start index in tlist of a particular item.
    INTEGER :: item_finish    ! Finish index in tlist of a particular item.
    
    ! Flag to indicate that at least one /alloc-opt/ has been seen - all 
    ! items in the list from that point on must be /alloc-opt/'s.
    LOGICAL :: seen_opt
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /allocate-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! ALLOCATE ( ... )
    
    CALL ExpectKeyword(tlist, 1, scck_'ALLOCATE', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL Expect(tlist, 2, ittDelimiter, scck_'(', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL ExpectClosing(tlist, 2, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! The combination of the /allocation-list/ and /alloc-opt-list/.
    !
    ! Previously we dynamically expanded the allocation component, but an 
    ! ifort 12.1.2 bug with assignment of polymorphic components made us change 
    ! to the two pass approach.
    
    ! Some items inside the () are /allocation/', some are /alloc-opt/'s.  
    ! list_count is the total number of both.
    list_count = ListItemCount(tlist(3:closing-1))
    
    ! /alloc-opt/'s must follow /allocation/'s by the syntax rules.  Set a 
    ! flag to track this.  Flag only valid on the first pass.
    seen_opt = .FALSE.
    
    ! First pass count and generate errors relating to position, second pass 
    ! parse.
    pass_loop: DO pass = 1, 2
      
      alloc_count = 0
      item_finish = 0
      
      DO item = 1, list_count
        CALL GetNextListItem(tlist, 3, closing - 1, item_start, item_finish)
        
        IF (item == 1) THEN
          ! Do we have a type-spec?  Note the index returned is not relative 
          ! to tlist, but to the item's slice out of tlist.
          declare = FindBare(  &
              tlist(item_start:item_finish),  &
              ittDeclare,  &
              scck_'::' )
          IF (declare /= 0) THEN  ! There is a /type-spec/.
            IF (pass == 2) THEN
              ALLOCATE(alloc_stmt%type_spec)
              CALL ParseTypeSpec(  &
                  tlist(item_start:item_start+declare-2),  &
                  alloc_stmt%type_spec,  &
                  sub_err_list )
              CALL Add(err_list, sub_err_list, COMPONENT=comp)
              ! Continue on error.
            END IF
            it = item_start + declare   ! Index of token after ::.
          ELSE
            it = item_start
          END IF
        ELSE
          it = item_start
        END IF
        
        ! Do we have an /alloc-opt/?
        assign = FindBare(tlist(it:item_finish), ittAssign, scck_'=')
        IF (assign == 0) THEN
          ! Not an /alloc-opt/
          alloc_count = alloc_count + 1
          IF (seen_opt .AND. (pass == 1)) THEN
            ! Cannot have /allocation/'s after we've started seeing 
            ! /alloc-opt/'s.
            CALL Add( err_list,  &
                CODE=errSyntax,  &
                COMPONENT=comp,  &
                LOCATION=QueryLocation(tlist(it)),  &
                MSG='All the /allocations/ in an /allocate-stmt/ must come &
                  &before the /alloc-opt/ specifications.' )
            ! Continue on this error.
          ELSE IF (pass == 2) THEN
            ! Parse the /allocation/
            CALL parse_allocation(  &
                tlist,  &
                it, item_finish,  &
                alloc_stmt%allocations(alloc_count),  &
                sub_err_list )
            CALL Add(err_list, sub_err_list, COMPONENT=comp)
            ! Continue on error.
          END IF
        ELSE
          ! Looks like an /alloc-opt/.
          seen_opt = .TRUE.
          IF ((pass == 1) .AND. (alloc_count == 0)) THEN
            ! Need to have at least one allocation.
            CALL Add( err_list,  &
                CODE=errSyntax,  &
                COMPONENT=comp,  &
                LOCATION=QueryLocation(tlist(it)),  &
                MSG='An /allocate-stmt/ must have at least one /allocation/.' )
            ! Continue on this error.
          END IF
          IF (pass == 2) THEN
            CALL parse_alloc_opt( &
                tlist,  &
                it, item_finish,  &
                alloc_stmt,  &
                sub_err_list )
            CALL Add(err_list, sub_err_list, COMPONENT=comp)
            ! Continue on error.
          END IF
        END IF
      END DO
      
      IF (pass == 1) ALLOCATE(alloc_stmt%allocations(alloc_count))
      
    END DO pass_loop
    
    !---------------------------------------------------------------------------
    ! That's it.
    
    CALL ExpectEOS(tlist, closing + 1, comp, err_list)
    
  END SUBROUTINE ParseAllocateStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /alloc-opt/ (R627).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the first token of the 
  !! /alloc-opt/ in @a tlist.
  !!
  !! @param[in]     finish            Index of the last token of the 
  !! /alloc-opt/ in @a tlist.
  !!
  !! @param[in,out] alloc_stmt        Object for storing the results of 
  !! parsing the parent /allocate-stmt/.  Note that this is INOUT.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE parse_alloc_opt(tlist, start, finish, alloc_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(AllocateStmt), INTENT(INOUT) :: alloc_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for an /alloc-opt/.
    CHARACTER(*), PARAMETER :: comp = 'R627'
    
    ! Keywords for each alloc-opt.
    CHARACTER(*,KIND=scck), PARAMETER :: keywords(4) = [  &
        scck_'ERRMSG', scck_'MOLD  ', scck_'SOURCE', scck_'STAT  ' ]
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: opt            ! Expect option.
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    ! The argument keyword.
    
    CALL ExpectKeyword(tlist, start, keywords, '/alloc-opt/', opt, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    
    CALL Expect(tlist, start + 1, ittAssign, scck_'=', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    
    SELECT CASE (opt)
    CASE (1)  ! ERRMSG
      IF (ALLOCATED(alloc_stmt%errmsg)) THEN
        CALL already('ERRMSG')
      ELSE
        CALL ParseVariable(  &
            tlist,  &
            start + 2, finish,  &
            alloc_stmt%errmsg,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (2)  ! MOLD
      IF (ALLOCATED(alloc_stmt%mold)) THEN
        CALL already('MOLD')
      ELSE IF ( ALLOCATED(alloc_stmt%type_spec)  &
          .OR. ALLOCATED(alloc_stmt%mold) ) THEN
        CALL Add( err_list,  &
            CODE=errSyntax,  &
            COMPONENT='C637 on ' // comp,  &
            LOCATION=QueryLocation(tlist(start)),  &
            MSG='At most one of /source-expr/ and /type-spec/ shall appear.' )
      ELSE
        CALL ParseExpr(  &
            tlist,  &
            start + 2, finish,  &
            alloc_stmt%mold,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (3)  ! SOURCE
      IF (ALLOCATED(alloc_stmt%source)) THEN
        CALL already('SOURCE')
      ELSE IF ( ALLOCATED(alloc_stmt%type_spec)  &
          .OR. ALLOCATED(alloc_stmt%mold) ) THEN
        CALL Add( err_list,  &
            CODE=errSyntax,  &
            COMPONENT='C637 on ' // comp,  &
            LOCATION=QueryLocation(tlist(start)),  &
            MSG='At most one of /source-expr/ and /type-spec/ shall appear.' )
      ELSE
        CALL ParseExpr(  &
            tlist,  &
            start + 2, finish,  &
            alloc_stmt%source,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (4)  ! STAT
      IF (ALLOCATED(alloc_stmt%stat)) THEN
        CALL already('STAT')
      ELSE
        CALL ParseVariable(  &
            tlist,  &
            start + 2, finish,  &
            alloc_stmt%stat,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE DEFAULT
      STOP 'Internal error in AllocateStmts%%parse_alloc_opt, &
          &unknown /alloc-opt/.'
      
    END SELECT
    
  CONTAINS
    
    SUBROUTINE already(text)
      USE ErrorCodes
      !-------------------------------------------------------------------------
      ! Arguments
      CHARACTER(*), INTENT(IN) :: text
      !*************************************************************************
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT='C636 on ' // comp,  &
          LOCATION=QueryLocation(tlist(start)),  &
          MSG=text // ' has already been specified for this &
            &/allocate-stmt/.' )
    END SUBROUTINE already
    
  END SUBROUTINE parse_alloc_opt
  
  
  !*****************************************************************************
  !!
  !> Implementation of AllocateStmt%Dump - write a debugging representation of 
  !! the results of parsing an allocate statement to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE alloc_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(AllocateStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Allocation index.
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%type_spec)) THEN
      WRITE (unit, "(A,'/type-spec/ = ')",ADVANCE='NO') pfx
      CALL WriteTypeSpec(unit, pnode%type_spec)
      WRITE (unit, "()")
    END IF
    WRITE (unit, "(A,'/allocation-list/')") pfx
    DO i = 1, SIZE(pnode%allocations)
      WRITE (unit, "(A)", ADVANCE='NO') pfx // '  '
      CALL pnode%allocations(i)%Write(unit)
      WRITE (unit, "()")
    END DO
    IF (ALLOCATED(pnode%stat)) THEN
      WRITE (unit, "(A,'STAT = ')", ADVANCE='NO') pfx
      CALL WriteExpr(unit, pnode%stat)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%errmsg)) THEN
      WRITE (unit, "(A,'ERRMSG = ')", ADVANCE='NO') pfx
      CALL WriteExpr(unit, pnode%errmsg)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%source)) THEN
      WRITE (unit, "(A,'SOURCE = ')", ADVANCE='NO') pfx
      CALL WriteExpr(unit, pnode%source)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%mold)) THEN
      WRITE (unit, "(A,'MOLD = ')", ADVANCE='NO') pfx
      CALL WriteExpr(unit, pnode%mold)
      WRITE (unit, "()")
    END IF
    
  END SUBROUTINE alloc_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of AllocateStmt%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE alloc_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(AllocateStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Allocation index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%type_spec)) CALL pnode%type_spec%Query(visitor)
    
    DO i = 1, SIZE(pnode%allocations)
      CALL pnode%allocations(i)%Query(visitor)
    END DO
    
    IF (ALLOCATED(pnode%source)) CALL pnode%source%Query(visitor)
    IF (ALLOCATED(pnode%mold)) CALL pnode%mold%Query(visitor)
    IF (ALLOCATED(pnode%stat)) CALL pnode%stat%Query(visitor)
    IF (ALLOCATED(pnode%errmsg)) CALL pnode%errmsg%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE alloc_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of AllocateStmt%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[inout] pnode             The parse node element to modify.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE alloc_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(AllocateStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Allocation index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%type_spec)) CALL pnode%type_spec%Modify(visitor)
    
    DO i = 1, SIZE(pnode%allocations)
      CALL pnode%allocations(i)%Modify(visitor)
    END DO
    
    IF (ALLOCATED(pnode%source)) CALL pnode%source%Modify(visitor)
    IF (ALLOCATED(pnode%mold)) CALL pnode%mold%Modify(visitor)
    IF (ALLOCATED(pnode%stat)) CALL pnode%stat%Modify(visitor)
    IF (ALLOCATED(pnode%errmsg)) CALL pnode%errmsg%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE alloc_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of AllocateStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'allocate-stmt'.
  
  FUNCTION alloc_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'allocate-stmt'
    
  END FUNCTION alloc_What
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for Allocation
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /allocation/ (R626).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the first token of the 
  !! /allocation/ in @a tlist.
  !!
  !! @param[in]     finish            Index of the last token of the 
  !! /allocation/ in @a tlist.
  !!
  !! @param[out]    alloc             Results of parsing the /allocation/.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE parse_allocation(tlist, start, finish, alloc, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(Allocation), INTENT(OUT) :: alloc
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! The syntax rule for an /allocation/.
    CHARACTER(*), PARAMETER :: comp = 'R631'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(error) :: err        ! Error object for Expect* calls.
    INTEGER :: it             ! Current token index.
    INTEGER :: opt            ! Expect option.
    INTEGER :: closing        ! Index of closing parenthesis.
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    ! The structure component or variable name.
    
    CALL ParseAllocateObject(  &
        tlist,  &
        start, finish,  &
        alloc%object,  &
        it,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
    ! Next bits are optional.
    IF (it > finish) RETURN
    
    !---------------------------------------------------------------------------
    ! Do we have an ( /allocate-shape-spec-list/ ) or a 
    ! [ /allocate-coarray-spec ]?
    
    CALL Expect(  &
        tlist,  &
        it,  &
        [ittDelimiter, ittDelimiter],  &
        [scck_'(', scck_'['],  &
        opt, err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    IF (opt == 1) THEN
      ! Failure not tested by StmtParsingTests because a missing closing 
      ! parenthesis causes a failure in the caller.
      CALL ExpectClosing(tlist, it, closing, err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      ! Parse the /allocate-shape-spec-list/.  This shares syntax (but not 
      ! constraints) with an /explicit-shape-spec-list/.
      CALL ParseExprBasedShapeSpecList(  &
          tlist,  &
          it + 1, closing - 1,  &
          'R633', '/allocate-shape-spec/',  &
          'R634', '/lower-bound-expr/',  &
          'R635', '/upper-bound-expr/',  &
          alloc%shape_spec,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      ! Continue on error.
      
      it = closing + 1
      IF (it > finish) RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! [ /allocate-coarray-spec/ ]
    
    CALL ExpectClosing(tlist, it, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL ParseAllocateCoarraySpec(  &
        tlist,  &
        it + 1,  &
        closing - 1,  &
        alloc%coarray_spec,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    
    ! That's it.
    CALL ExpectEnd(tlist(:finish), closing + 1, comp, err_list)
    
  END SUBROUTINE parse_allocation
  
  
  !*****************************************************************************
  !!
  !> Implementation of Allocation%Write - write a text representation of the 
  !! parse node to a logical unit.
  !!
  !! @param[in]     pnode             The format.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output for the text represenation.
  
  SUBROUTINE al_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(Allocation), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Array/coarray index.
    
    !***************************************************************************
    
    WRITE (unit, "('/allocate-object/ = ')", ADVANCE='NO')
    CALL pnode%object%Write(unit)
    
    IF (ALLOCATED(pnode%shape_spec)) THEN
      WRITE (unit, "(' ( ')", ADVANCE='NO')
      DO i = 1, SIZE(pnode%shape_spec) - 1
        CALL pnode%shape_spec(i)%Write(unit)
        WRITE (unit, "(' , ')", ADVANCE='NO')
      END DO
      CALL pnode%shape_spec(i)%Write(unit)
      WRITE (unit, "(' ) ')", ADVANCE='NO')
    END IF
    
    IF (ALLOCATED(pnode%coarray_spec)) THEN
      WRITE (unit, "(' [ ')", ADVANCE='NO')
      DO i = 1, SIZE(pnode%coarray_spec) - 1
        CALL pnode%coarray_spec(i)%Write(unit)
        WRITE (unit, "(' , ')", ADVANCE='NO')
      END DO
      CALL pnode%coarray_spec(i)%Write(unit)
      WRITE (unit, "(' ] ')", ADVANCE='NO')
    END IF
    
  END SUBROUTINE al_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of Allocation%Query - traverse the node with a 
  !! query operation.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  SUBROUTINE al_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(Allocation), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Array/coarray spec index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%object%Query(visitor)
    
    IF (ALLOCATED(pnode%shape_spec)) THEN
      DO i = 1, SIZE(pnode%shape_spec)
        CALL pnode%shape_spec(i)%Query(visitor)
      END DO
    END IF
    
    IF (ALLOCATED(pnode%coarray_spec)) THEN
      DO i = 1, SIZE(pnode%coarray_spec)
        CALL pnode%coarray_spec(i)%Query(visitor)
      END DO
    END IF
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE al_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of Allocation%Modify - traverse the node with a 
  !! modify operation.
  !!
  !! @param[in]     pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  SUBROUTINE al_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(Allocation), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Array/coarray spec index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%object%Modify(visitor)
    
    IF (ALLOCATED(pnode%shape_spec)) THEN
      DO i = 1, SIZE(pnode%shape_spec)
        CALL pnode%shape_spec(i)%Modify(visitor)
      END DO
    END IF
    
    IF (ALLOCATED(pnode%coarray_spec)) THEN
      DO i = 1, SIZE(pnode%coarray_spec)
        CALL pnode%coarray_spec(i)%Modify(visitor)
      END DO
    END IF
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE al_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of Allocation%What - return a descriptive 
  !! string for the parse node.
  !!
  !! @returns The syntax rule name for the nodes.
  
  FUNCTION al_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'allocation'
    
  END FUNCTION al_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of Allocation%GetLocation - get the start of the 
  !! source that corresponds to the ParseNode.
  !!
  !! @param[in]     pnode           The parse node of interest
  !!
  !! @returns The location in source of the start of the ParseNode.
  
  FUNCTION al_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !-------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(Allocation), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    loc = .LocationOf. pnode%object
    
  END FUNCTION al_GetLocation
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for AllocateObject.
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /allocate-object/ (R632).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the first token of the 
  !! /allocation/ in @a tlist.
  !!
  !! @param[in]     finish            Index of the last token of the 
  !! /allocation/ in @a tlist.  Note that this is the finish of the 
  !! parent clause, not the clause that is actually parsed by this 
  !! procedure.
  !!
  !! @param[out]    object            Results of parsing the /allocation/.
  !!
  !! @param[out]    after             Index of the first token immediately 
  !! after the /allocate-object/.  This might be an opening parenthesis or 
  !! /rbracket/ if the parent /allocation/ contains an array or coarray 
  !! specification.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseAllocateObject( tlist, start, finish, object, after,  &
      err_list )
    
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    CLASS(AllocateObject), INTENT(OUT), ALLOCATABLE :: object
    INTEGER, INTENT(OUT) :: after
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule for an /allocate-object/.
    CHARACTER(*), PARAMETER :: comp = 'R632'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: num_parts      ! Number of parts in the object.
    
    ! Temporary for when an /allocate-object/ is a /variable-name/.
    CLASS(AllocateObjectVariable), ALLOCATABLE :: object_var
    
    ! Temporary for when an /allocate-object/ is a /structure-component/.
    CLASS(AllocateObjectStructureComponent), ALLOCATABLE :: object_comp
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (finish < start) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          LOCATION=SafeLocation(tlist, finish),  &
          MSG='Expected /allocate-object/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    ! See how many parts we have.
    num_parts = CountParts(tlist, start, finish)
    
    IF (num_parts == 1) THEN
      ! We have a variable name.
      CALL Expect(tlist, start, '/variable-name/', err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      ALLOCATE(object_var)
      object_var%name = IdentifierIfortBug(tlist(start))
      CALL MOVE_ALLOC(object_var, object)
      after = start + 1
    ELSE
      ! We have a /structure-component/.
      ALLOCATE(object_comp)
      CALL ParseStructureComponent(  &
          tlist(:finish),  &
          start,  &
          object_comp%parts,  &
          after,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
      
      CALL MOVE_ALLOC(object_comp, object)
    END IF
    
  END SUBROUTINE ParseAllocateObject
  
  
  !*****************************************************************************
  !!
  !> Implementation of AllocateObject%What - return a descriptive 
  !! string for the parse node.
  !!
  !! @returns The syntax rule name for the nodes.
  
  FUNCTION ao_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'allocate-object'
    
  END FUNCTION ao_What
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for AllocateObjectVariable
  
  
  !*****************************************************************************
  !!
  !> Implementation of AllocateObjectVariable%Write - write a 
  !! text representation of the parse node to a logical unit.
  !!
  !! @param[in]     pnode             The format.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output for the text represenation.
  
  SUBROUTINE aov_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(AllocateObjectVariable), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    WRITE (unit, "(A)", ADVANCE='NO') .NameOf. pnode%name
    
  END SUBROUTINE aov_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of AllocateObjectVariable%Query - traverse 
  !! the node with a query operation.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  SUBROUTINE aov_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(AllocateObjectVariable), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE aov_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of AllocateObjectVariable%Modify - traverse 
  !! the node with a modify operation.
  !!
  !! @param[in]     pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  SUBROUTINE aov_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(AllocateObjectVariable), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE aov_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of AllocateObjectVariable%Advertise - advertise the 
  !! identifiers referenced or defined by the construct.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE aov_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(AllocateObjectVariable), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(1))
    
    ident_list(1) = IdentifierNature(  &
        NAME=.NameOf. pnode%name,  &
        NATURES=[NatureRefDataObject],  &
        LOCATIONS=[.LocationOf. pnode%name] )
    
  END SUBROUTINE aov_Advertise
  
  
  !*****************************************************************************
  !!
  !> Implementation of AllocateObjectVariable%GetLocation - get the start 
  !! of the source that corresponds to the ParseNode.
  !!
  !! @param[in]     pnode           The parse node of interest
  !!
  !! @returns The location in source of the start of the ParseNode.
  
  FUNCTION aov_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !-------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(AllocateObjectVariable), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    loc = .LocationOf. pnode%name
    
  END FUNCTION aov_GetLocation
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for AllocateObjectStructureComponent
  
  
  !*****************************************************************************
  !!
  !> Implementation of AllocateObjectStructureComponent%Write - write a 
  !! text representation of the parse node to a logical unit.
  !!
  !! @param[in]     pnode             The format.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output for the text represenation.
  
  SUBROUTINE aosc_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(AllocateObjectStructureComponent), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Part index.
    
    !***************************************************************************
    
    DO i = 1, SIZE(pnode%parts) - 1
      CALL WritePartRef(unit, pnode%parts(i))
      WRITE (unit, "(' % ')", ADVANCE='NO')
    END DO
    CALL WritePartRef(unit, pnode%parts(SIZE(pnode%parts)))
    
  END SUBROUTINE aosc_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of AllocateObjectStructureComponent%Query - traverse 
  !! the node with a query operation.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  SUBROUTINE aosc_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(AllocateObjectStructureComponent), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Part index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%parts)
      CALL pnode%parts(i)%Query(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE aosc_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of AllocateObjectStructureComponent%Modify - traverse 
  !! the node with a modify operation.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  SUBROUTINE aosc_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(AllocateObjectStructureComponent), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Part index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%parts)
      CALL pnode%parts(i)%Modify(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE aosc_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of AllocateObjectStructureComponent%Advertise - advertise 
  !! the identifiers referenced or defined by the construct.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE aosc_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(AllocateObjectStructureComponent), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(1))
    
    ident_list(1) = IdentifierNature(  &
        NAME=.NameOf. pnode%parts(1)%part_name,  &
        NATURES=[NatureRefDataObject],  &
        LOCATIONS=[.LocationOf. pnode%parts(1)%part_name] )
    
  END SUBROUTINE aosc_Advertise
  
  
  !*****************************************************************************
  !!
  !> Implementation of AllocateObjectStructureComponent%GetLocation - get the 
  !! start of the source that corresponds to the ParseNode.
  !!
  !! @param[in]     pnode           The parse node of interest
  !!
  !! @returns The location in source of the start of the ParseNode.
  
  FUNCTION aosc_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !-------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(AllocateObjectStructureComponent), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    loc = .LocationOf. pnode%parts(1)
    
  END FUNCTION aosc_GetLocation
  
END MODULE AllocateStmts
