! $Id: BindStmts.f90 2159 2016-05-31 13:08:04Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the BindStmts module.


!*******************************************************************************
!!
!> Procedures and types for parsing /bind-stmts/.

MODULE BindStmts
  
  USE Identifiers
  USE LanguageBindingSpecs
  USE ParseNodes
  USE SubscriptLists
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseBindStmt
  
  !-----------------------------------------------------------------------------
  
  !> Stores the results of parsing a /bind-entity/.
  TYPE, PUBLIC :: BindEntity
    !> The name of the object being declared.
    TYPE(Identifier) :: name
    !> Flag to indicate that the entity is a common block name.
    LOGICAL :: common_block
  END TYPE BindEntity
  
  !> Represents the results of parsing a bind statement.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: BindStmt
    !> The language specification.
    TYPE(LangSpec) :: ls
    !> The list of /bind-entity/'s.
    TYPE(BindEntity), ALLOCATABLE :: entities(:)
  CONTAINS
    PROCEDURE :: Dump => bind_Dump
    PROCEDURE :: Query => bind_Query
    PROCEDURE :: Modify => bind_Modify
    PROCEDURE, NOPASS :: What => bind_What
  END TYPE BindStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /bind-stmt/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    bind_stmt         The results of parsing the statement.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! This procedure has been written assuming that the slash comes back as an 
  !! operator type token.  @todo Consider modifying tlist to have the 
  !! correct token type.
  
  SUBROUTINE ParseBindStmt(tlist, bind_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(BindStmt), INTENT(OUT) :: bind_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! The syntax rule number for a /bind-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R529'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err      ! Error object for Expect calls.
    INTEGER :: opt          ! Expect option.
    INTEGER :: it           ! Current token
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    INTEGER :: list_count     ! Number of items in the /bind-entity-list/.
    INTEGER :: item           ! List item index.
    INTEGER :: item_start     ! Start index in tlist of a particular item.
    INTEGER :: item_finish    ! Finish index in tlist of a particular item.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /bind-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! BIND...
    
    CALL ExpectKeyword(tlist, 1, scck_'BIND', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL Parse(tlist, 2, bind_stmt%ls, it, sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
    !---------------------------------------------------------------------------
    ! Optional :: before entity-list.
    
    CALL Expect(  &
        tlist, it,  &
        [ittDeclare, ittName, ittOperator],  &
        [scck_'::', scck_'  ', scck_'/ '],  &
        '"::" or /object-name/',  &
        opt, err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    IF (opt == 1) it = it + 1
    
    !---------------------------------------------------------------------------
    ! /bind-entity-list/
    
    list_count = ListItemCount(tlist(it:))
    ALLOCATE(bind_stmt%entities(list_count))
    item_finish = 0
    DO item = 1, list_count
      CALL GetNextListItem(tlist, it, SIZE(tlist), item_start, item_finish)
      CALL parse_bind_entity(  &
          tlist,  &
          item_start, item_finish,  &
          bind_stmt%entities(item),  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
    END DO
    
    ! That's it!
    
  END SUBROUTINE ParseBindStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /bind-entity/ (R530).
  
  SUBROUTINE parse_bind_entity(tlist, start, finish, entity, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(BindEntity), INTENT(OUT) :: entity
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /bind-entity/.
    CHARACTER(*), PARAMETER :: comp = 'R530'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err          ! Error object for Expect calls.
    INTEGER :: opt              ! Expect option
    INTEGER :: it               ! Current token index.
    
    !***************************************************************************
  
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    ! / /common-block-name/ /  or  /entity-name/
    
    CALL Expect(  &
        tlist, start,  &
        [ittOperator, ittName],  &
        [scck_'/', scck_' '],  &
        '"/" or /entity-name',  &
        opt, err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
        
    IF (opt == 1) THEN
      entity%common_block = .TRUE.
      
      CALL Expect(tlist, start + 1, '/common-block-name/', err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      entity%name = IdentifierIfortBug(tlist(start + 1))
      
      CALL Expect(tlist, start + 2, ittOperator, scck_'/', err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      it = start + 3
    ELSE
      entity%common_block = .FALSE.
      entity%name = IdentifierIfortBug(tlist(start))
      it = start + 1
    END IF
    
    ! That's all folks.
    CALL ExpectEnd(tlist(:finish), it, comp, err_list)
    
  END SUBROUTINE parse_bind_entity
  
  
  !*****************************************************************************
  !!
  !> Implementation of BindStmt%Dump - write a debugging representation 
  !! of the results of parsing a bind statement to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE bind_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(BindStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! List index.
    
    !***************************************************************************
    
    WRITE (unit, "(A)", ADVANCE='NO') pfx
    CALL WriteLangSpec(unit, pnode%ls)
    WRITE (unit, "()")
      
    WRITE (unit, "(A,'/bind-entity-list/')") pfx
    DO i = 1, SIZE(pnode%entities)
      IF (pnode%entities(i)%common_block) THEN
        WRITE (unit, "(A,'/ ',A,' /')")  &
            pfx // '  ', .NameOf. pnode%entities(i)%name
      ELSE
        WRITE (unit, "(A,A)") pfx // '  ', .NameOf. pnode%entities(i)%name
      END IF
    END DO
    
  END SUBROUTINE bind_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of BindStmt%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE bind_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(BindStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%ls%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE bind_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of BindStmt%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[inout] pnode             The parse node element to modify.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE bind_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(BindStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%ls%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE bind_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of BindStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'bind-stmt'.
  
  FUNCTION bind_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'bind-stmt'
    
  END FUNCTION bind_What
  
END MODULE BindStmts
