! $Id: CommonStmts.f90 2829 2019-03-24 09:46:01Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the CommonStmts module.


!*******************************************************************************
!!
!> Procedures and types for parsing /common-stmt/'s (R568).

MODULE CommonStmts
  
  USE CompilerKinds
  USE DimensionStmts
  USE Identifiers
  USE ParseNodes
  USE SubscriptLists
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseCommonStmt
  
  !-----------------------------------------------------------------------------
  
  !> A /common-block-object/.
  !!
  !! This is similar to DimensionDecl, but we decided to duplicate it anyway 
  !! as the DimensionDecl always has the array_spec component allocated.
  TYPE, EXTENDS(TermParseNode), PUBLIC :: CommonBlockObject
    !> The name of the object being declared.
    TYPE(Identifier) :: name
    !> The /array-spec/.
    !!
    !! Not allocated if the common block object is not an array.  If 
    !! allocated, then always size greater than zero.
    TYPE(DimSpec), ALLOCATABLE :: array_spec(:)
  CONTAINS
    PROCEDURE :: Write => cbo_Write
    PROCEDURE :: Query => cbo_Query
    PROCEDURE :: Modify => cbo_Modify
    PROCEDURE, NOPASS :: What => cbo_What
    PROCEDURE :: Advertise => cbo_Advertise
    PROCEDURE :: GetLocation => cbo_GetLocation
  END TYPE CommonBlockObject
  
  !> A common block set - the combination of the name of the common block and 
  !! the objects that are defined as being part of the common block.
  TYPE, EXTENDS(TermParseNode), PUBLIC :: CommonBlockSet
    !> The name of the common block.
    !!
    !! Not allocated if the common block had no name.
    TYPE(Identifier), ALLOCATABLE :: name
    !> The objects that are part of the common block.
    !!
    !! Always allocated, always size greater than zero.
    TYPE(CommonBlockObject), ALLOCATABLE :: objects(:)
  CONTAINS
    PROCEDURE :: Write => cbs_Write
    PROCEDURE :: Query => cbs_Query
    PROCEDURE :: Modify => cbs_Modify
    PROCEDURE, NOPASS :: What => cbs_What
    PROCEDURE :: GetLocation => cbs_GetLocation
  END TYPE CommonBlockSet
  
  !> The results of parsing a common statement.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: CommonStmt
    !> The common block sets defined by the statement.
    !!
    !! Always allocated, always size greater than zero.
    TYPE(CommonBlockSet), ALLOCATABLE :: sets(:)
  CONTAINS
    PROCEDURE :: Dump => common_Dump
    PROCEDURE :: Query => common_Query
    PROCEDURE :: Modify => common_Modify
    PROCEDURE, NOPASS :: What => common_What
  END TYPE CommonStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /common-stmt/ (R568)
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    common_stmt       The results of parsing the statement.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! The procedure has been written to be tolerant of the scanner not being 
  !! able to distinguish two slashes together (two delimiters in the context 
  !! of a common statement) from the concatenation token.  If child procedures 
  !! are modified to modify tlist when this is detected then the loop logic 
  !! below will need to be modified.
  !!
  !! This procedure has been written assuming that the slash comes back as an 
  !! operator type token.  @todo Consider modifying tlist to have the 
  !! correct token type.
  
  SUBROUTINE ParseCommonStmt(tlist, common_stmt, err_list)
    
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE MatchUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(CommonStmt), INTENT(OUT) :: common_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule for an /common-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R568'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect* calls.
    INTEGER :: it             ! Current token index.
    INTEGER :: it_end         ! End of the current syntax part.
    
    ! Temporary for growing the list of sets.
    TYPE(CommonBlockSet), ALLOCATABLE :: tmp_sets(:)
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /common-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! COMMON...
    
    CALL ExpectKeyword(tlist, 1, scck_'COMMON', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! The common block sets.
    
    ALLOCATE(common_stmt%sets(1))
    it = 2
    DO
      ! Progressive parse - it_end gets set to the token one past the end 
      ! of the namelist group.
      CALL parse_common_block_set(  &
          tlist,  &
          it,  &
          SIZE(common_stmt%sets) == 1,  &
          common_stmt%sets(SIZE(common_stmt%sets)),  &
          it_end,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
      
      IF (it_end > SIZE(tlist)) EXIT
      it = it_end
      
      ! Create space for the next item
      ALLOCATE(tmp_sets(SIZE(common_stmt%sets) + 1))
      tmp_sets(:SIZE(common_stmt%sets)) = common_stmt%sets
      
      CALL MOVE_ALLOC(tmp_sets, common_stmt%sets)
      
      ! A comma is optional.  parse_common_block_set points it_end 
      ! to the comma if it is there.
      IF (Match(tlist, it, ittComma)) it = it + 1
    END DO
    
  END SUBROUTINE ParseCommonStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a common block set.
  !!
  !! A common block set is a convenience defined by us as:
  !! 
  !! [ / [ /common-block-name/ ] / / ] /common-block-object-list/
  !!
  !! Because it is a convenience definition this construct has no syntax rule.
  !!
  !! @param[in]     tlist             The list of tokens.
  !!
  !! @param[in]     start             Index of the first token of the common 
  !! block set in @a tlist.
  !!
  !! @param[in]     first             Flag to indicate that this is the 
  !! first common block set to be parsed for the statement.  The first set 
  !! has relaxed syntax rules.
  !!
  !! @param[out]    set               Results of parsing the common block 
  !! set.
  !!
  !! @param[out]    after            Index of the token after the common 
  !! block set in @a tlist.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! The procedure has been written to be tolerant of the scanner not being 
  !! able to distinguish two slashes together (two delimiters in the context 
  !! of a common statement) from the concatenation token.  @todo Consider 
  !! modifying tlist (includes modifying its size) to split and reclassify 
  !! that token.
  !!
  !! This procedure has been written assuming that the slash comes back as an 
  !! operator type token.  @todo Consider modifying tlist to have the 
  !! correct token type - its use in the context of a common statement is as 
  !! a delimiter.
  
  SUBROUTINE parse_common_block_set(tlist, start, first, set, after, err_list)
    
    USE Errors
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    LOGICAL, INTENT(IN) :: first
    TYPE(CommonBlockSet), INTENT(OUT) :: set
    INTEGER, INTENT(OUT) :: after
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect* calls.
    INTEGER :: opt            ! Expect option.
    INTEGER :: it             ! Current token index.
    INTEGER :: slash          ! Index of the next slash.
    INTEGER :: concat         ! Index of the next concatenation operator.
    INTEGER :: list_finish    ! Index of the end of the object list.
    INTEGER :: list_count     ! Number of items in the object list.
    INTEGER :: item           ! List item index.
    INTEGER :: item_start     ! Start index in tlist of a particular item.
    INTEGER :: item_finish    ! Finish index in tlist of a particular item.
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    ! If this is the first set then we can have / [ name ] ... or go straight 
    ! into the object list (a name).  Later sets must start with a slash.
    
    ! We have to handle the ambiguity of // or / / coming out of the scanner.
    IF (first) THEN
      CALL Expect(  &
          tlist, start,  &
          [ittOperator, ittOperator, ittName],  &
          [scck_'/ ', scck_'//', scck_'  '],  &
          '"/" or /common-block-object/',  &
          opt,  &
          err )
      CALL Add(err_list, err)
      IF (Failure(err)) RETURN
    ELSE
      CALL Expect(  &
          tlist, start,  &
          [ittOperator, ittOperator],  &
          [scck_'/ ', scck_'//'],  &
          '"/"',  &
          opt,  &
          err )
      CALL Add(err_list, err)
      IF (Failure(err)) RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! The common block name.
    
    SELECT CASE (opt)
    CASE (1)  ! Single slash
      ! Was a name provided?
      CALL Expect(  &
          tlist, start + 1,  &
          [ittOperator, ittName],  &
          [scck_'/', scck_' '],  &
          '"/" or /common-block-name/',  &
          opt,  &
          err )
      CALL Add(err_list, err)
      IF (Failure(err)) RETURN
      
      IF (opt == 1) THEN    ! No block name present.
        it = start + 2
      ELSE                  ! Block name present.
        set%name = IdentifierIfortbug(tlist(start+1))
        
        CALL Expect(tlist, start + 2, ittOperator, scck_'/', err)
        CALL Add(err_list, err)
        IF (Failure(err)) RETURN
      
        ! Skip closing delimiter.
        it = start + 3
      END IF
      
    CASE (2)  ! // - same as /  /
      it = start + 1
      
    CASE (3)  ! A /name/ - we are straight into the object list.
      it = start
      
    END SELECT
    ! it here needs to be the start of the object list.
    
    !---------------------------------------------------------------------------
    ! The common block object list.  This is terminated either by a slash, 
    ! or by a comma immediately followed by a slash, or by the end of the 
    ! statement.
    
    ! Find the next / or //.  If we don't find one we process the remainder 
    ! of the statement as the list.  // is valid here - because the next 
    ! common block definition may not be named.
    slash = FindBare(tlist, ittOperator, scck_'/', it)
    concat = FindBare(tlist, ittOperator, scck_'//', it)
    IF (concat /= 0) THEN
      IF (slash == 0) THEN
        slash = concat
      ELSE IF (concat < slash) THEN
        slash = concat
      END IF
    END IF
    
    IF (slash == 0) THEN
      ! If this is the end of the statement (no slash to delineate the next 
      ! group) then a trailing comma is not allowed.  If one is there we will 
      ! error out in the processing of the object list.
      list_finish = SIZE(tlist)
    ELSE
      ! Do we have a comma immediately preceding the terminating slash?
      IF (tlist(slash-1) == ittComma) THEN
        list_finish = slash - 2
      ELSE
        ! No trailing comma.
        list_finish = slash - 1
      END IF
    END IF
    
    ! The object list itself
    list_count = ListItemCount(tlist(it:list_finish))
    ALLOCATE(set%objects(list_count))
    item_finish = 0
    DO item = 1, list_count
      CALL GetNextListItem(tlist, it, list_finish, item_start, item_finish)
      CALL parse_common_block_object( tlist, item_start, item_finish,  &
          set%objects(item), sub_err_list )
      CALL Add(err_list, sub_err_list)
      IF (Failure(sub_err_list)) RETURN
    END DO
    
    ! list_finish here is the end of the list of objects.  The end of the 
    ! common block set is one past it - which may be a comma.
    after = list_finish + 1
    
  END SUBROUTINE parse_common_block_set
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /common-block-object/ (R569).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the first token of the 
  !! /common-block-object/ in @a tlist.
  !!
  !! @param[in]     finish            Index of the last token of the 
  !! /common-block-object/ in @a tlist.
  !!
  !! @param[out]    object            Results of parsing the 
  !! /common-block-object/.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE parse_common_block_object(tlist, start, finish, object, err_list)
    
    USE Errors
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(CommonBlockObject), INTENT(OUT) :: object
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule for a /common-block-object/.
    CHARACTER(*), PARAMETER :: comp = 'R569'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect* calls.
    INTEGER :: closing        ! Index of the closing delimiter.
    
    ! Error list for procedure calls
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    ! /variable-name/
    
    CALL ExpectName(tlist, start, '/variable-name/', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    object%name = IdentifierIfortBug(tlist(start))
    
    !---------------------------------------------------------------------------
    ! Optional ( /array-spec/ )
    
    IF (finish == start) RETURN
    
    CALL Expect(tlist, start + 1, ittDelimiter, scck_'(', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL ExpectClosing(tlist, start + 1, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL ParseArraySpec(  &
        tlist(start+2:closing-1),  &
        .FALSE.,  &
        object%array_spec,  &
        sub_err_list )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    !  That's it.
    
    CALL ExpectEnd(tlist(:finish), closing + 1, comp, err_list)
    
  END SUBROUTINE parse_common_block_object
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for bindings in CommonStmt.
  
  
  !*****************************************************************************
  !!
  !> Implementation of CommonStmt%Dump - write a debugging representation 
  !! of the results of parsing a common statement to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE common_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(CommonStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: is             ! Set index.
    
    !***************************************************************************
    
    DO is = 1, SIZE(pnode%sets)
      WRITE (unit, "(A)", ADVANCE='NO') pfx
      CALL pnode%sets(is)%Write(unit)
      WRITE (unit, "()")
    END DO
    
  END SUBROUTINE common_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of CommonStmt%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE common_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(CommonStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Set index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%sets)
      CALL pnode%sets(i)%Query(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE common_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of CommonStmt%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[in,out] pnode             The parse node element to modify.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE common_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(CommonStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Set index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%sets)
      CALL pnode%sets(i)%Modify(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE common_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of CallStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'common-stmt'.
  
  FUNCTION common_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'common-stmt'
    
  END FUNCTION common_What
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for bindings in CommonBlockSet.
  
  
  !*****************************************************************************
  !!
  !> Implementation of CommonBlockSet%Write - write a debugging representation 
  !! of the node to a logical unit.
  !!
  !1 @param[in]     pnode             The node to write.
  !!
  !! @param[in]     unit              Unit connected for formatted output.
  
  SUBROUTINE cbs_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(CommonBlockSet), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: io             ! Object index.
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%name)) THEN
      WRITE (unit, "('/common-block-name/ = ',A,' ')", ADVANCE='NO')  &
        .NameOf. pnode%name
    END IF
    WRITE (unit, "('/common-block-object-list/ = ')", ADVANCE='NO')
    
    DO io = 1, SIZE(pnode%objects) - 1
      CALL pnode%objects(io)%Write(unit)
      WRITE (unit, "(', ')", ADVANCE='NO')
    END DO
    CALL pnode%objects(io)%Write(unit)
    
  END SUBROUTINE cbs_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of CommonStmt%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE cbs_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(CommonBlockSet), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Declaration index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%objects)
      CALL pnode%objects(i)%Query(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE cbs_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of CommonStmt%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[in,out] pnode             The parse node element to modify.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE cbs_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(CommonBlockSet), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Declaration index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%objects)
      CALL pnode%objects(i)%Modify(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE cbs_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of CommonBlockSet%What - return a descriptive string 
  !! for the parse node.
  !!
  !! @returns '(common-block-set)'.
  
  FUNCTION cbs_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = '(common-block-set)'
    
  END FUNCTION cbs_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of CommonBlockSet%GetLocation - get the start of the 
  !! source that corresponds to the ParseNode.
  !!
  !! @param[in]     pnode           The parse node of interest
  !!
  !! @returns The location in source of the start of the ParseNode.
  
  FUNCTION cbs_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !-------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(CommonBlockSet), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%name)) THEN
      loc = .LocationOf. pnode%name
    ELSE
      loc = .LocationOf. pnode%objects(1)
    END IF
    
  END FUNCTION cbs_GetLocation
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for bindings in CommonBlockObject.
  
  
  !*****************************************************************************
  !!
  !> Implementation of CommonBlockObject%Write - write a debugging 
  !! representation of the node to a logical unit.
  !!
  !1 @param[in]     pnode             The node to write.
  !!
  !! @param[in]     unit              Unit connected for formatted output.
  
  SUBROUTINE cbo_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(CommonBlockObject), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%array_spec)) THEN
      WRITE (unit, "('/variable-name/ = ',A,' (')",  ADVANCE='NO') &
          .NameOf. pnode%name
      CALL WriteDimSpec(unit, pnode%array_spec)
      WRITE (unit, "(')')", ADVANCE='NO')
    ELSE
      WRITE (unit, "('/variable-name/ = ',A)", ADVANCE='NO')  &
          .NameOf. pnode%name
    END IF
    
  END SUBROUTINE cbo_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of CommonBlockObject%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE cbo_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(CommonBlockObject), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Array spec index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%array_spec)) THEN
      DO i = 1, SIZE(pnode%array_spec)
        CALL pnode%array_spec(i)%Query(visitor)
      END DO
    END IF
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE cbo_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of CommonBlockObject%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[in,out] pnode             The parse node element to modify.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE cbo_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(CommonBlockObject), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Array spec index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%array_spec)) THEN
      DO i = 1, SIZE(pnode%array_spec)
        CALL pnode%array_spec(i)%Modify(visitor)
      END DO
    END IF
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE cbo_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of CommonBlockObject%What - return a descriptive string 
  !! for the parse node.
  !!
  !! @returns 'common-block-object'.
  
  FUNCTION cbo_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'common-block-object'
    
  END FUNCTION cbo_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of CommonBlockObject%Advertise - advertise the 
  !! identifiers in a CommonBlockObject that are not child nodes.
  !!
  !! @param[in]     pnode             The statement to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE cbo_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(CommonBlockObject), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(1))
    
    IF (ALLOCATED(pnode%array_spec)) THEN
      ident_list(1) = IdentifierNature(  &
          NAME=.NameOf. pnode%name,  &
          NATURES=[NatureDefCommonObject, NatureDefDimension],  &
          LOCATIONS=[.LocationOf. pnode%name, .LocationOf. pnode%name] )
    ELSE
      ident_list(1) = IdentifierNature(  &
          NAME=.NameOf. pnode%name,  &
          NATURES=[NatureDefCommonObject],  &
          LOCATIONS=[.LocationOf. pnode%name] )
    END IF
    
  END SUBROUTINE cbo_Advertise
  
  
  !*****************************************************************************
  !!
  !> Implementation of CommonBlockObject%GetLocation - get the start of the 
  !! source that corresponds to the ParseNode.
  !!
  !! @param[in]     pnode           The parse node of interest
  !!
  !! @returns The location in source of the start of the ParseNode.
  
  FUNCTION cbo_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !-------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(CommonBlockObject), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    loc = .LocationOf. pnode%name
    
  END FUNCTION cbo_GetLocation
  
END MODULE CommonStmts
