! $Id: ConstructNames.f90 2796 2019-03-22 09:39:36Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the ConstructNames module.


!*******************************************************************************
!!
!> Utility routines for processing of construct names.
!!
!! Construct names are the things that appear ahead of a colon in things like 
!! /if-stmts/, /do-stmt/'s, etc.  The rules on consistency of names have 
!! some similarity to those on program unit names, but we handle them 
!! separately here.
!!
!! Testing implicit in the testing of code that uses this module.
!!
!! @todo Consider putting in the code that extracts a construct name from 
!! the leading tokens in a statement in here too, if anything just to enforce 
!! consistency.

MODULE ConstructNames
  
  IMPLICIT NONE
  
  PRIVATE
  
  PUBLIC :: CheckConstructNameEnd
  PUBLIC :: CheckConstructNameIntermediate
  
  !-----------------------------------------------------------------------------
  
  !> Check for errors in the specification of a construct name on a statement 
  !! that ends a construct.
  INTERFACE CheckConstructNameEnd
    MODULE PROCEDURE CheckConstructNameEnd_
  END INTERFACE CheckConstructNameEnd
  
  !> Check for errors in the specification of a construct name on an 
  !! intermediate statement in a construct.
  INTERFACE CheckConstructNameIntermediate
    MODULE PROCEDURE CheckConstructNameIntermediate_
  END INTERFACE CheckConstructNameIntermediate
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Check that the construct name part of a statement that ends a construct 
  !! matches the typical requirements for that name.
  !!
  !! @param[in]     tlist             Sequence of tokens that make up the 
  !! end statement for the construct.  Must have size greater than zero.
  !!
  !! @param[in]     it                Index of the token that hypothetically 
  !! would be the construct name, if it is present.  May be beyond the end 
  !! of the token sequence (in cases where the construct name is not present 
  !! in the end statement for the construct).
  !!
  !! @param[in]     construct_name    Construct name on the opening statement 
  !! of the construct, not allocated if no construct name was provided.  If 
  !! allocated it should be in upper case.
  !!
  !! @param[in]     initial_stmt      Syntax rule name for the initial 
  !! statement, used for error reporting.  Include the '/'.
  !!
  !! @param[in]     end_stmt          Syntax rule name for the end statement 
  !! of the construct - used for error reporting.  Include the '/'.
  !!
  !! @param[in]     comp              Component for error reporting.  
  !! Typically this would include the constraint and the syntax rule number 
  !! for the end statement.
  !!
  !! @param[in,out] err_list          List of errors.  Note this is 
  !! INTENT(INOUT) - the idea being that a mismatch of construct names does 
  !! not terminate parsing.
  !!
  !! We check:
  !! - If the opening statement had a construct name, then the end 
  !! statement must have a construct name.
  !! - If the opening statement does not have a construct name, then 
  !! the end statement must not have a construct name.
  !! - If the opening statement has a construct name and the end statement 
  !! has a construct name, then the names must match.
  !! - That if no construct name is present, it is beyond the end of the 
  !! statement, and if a construct name is present, that it is the last 
  !! thing in the statement.
  
  SUBROUTINE CheckConstructNameEnd_( tlist, it, construct_name,  &
      initial_stmt, end_stmt, comp, err_list )
    
    USE Errors
    USE ErrorCodes
    USE MatchUtils
    USE Identifiers
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: it
    TYPE(Identifier), INTENT(IN), ALLOCATABLE :: construct_name
    CHARACTER(*), INTENT(IN) :: initial_stmt
    CHARACTER(*), INTENT(IN) :: end_stmt
    CHARACTER(*), INTENT(IN) :: comp
    TYPE(Error), INTENT(INOUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    
    !***************************************************************************
    
    IF (ALLOCATED(construct_name)) THEN
      ! The opening statement has a construct name.  The end statement must 
      ! have a name, and it must match.
      IF (SIZE(tlist) < it) THEN
        CALL Add( err_list, CODE=errNameMismatch, COMPONENT=comp,  & 
            LOCATION=QueryLocation(tlist(SIZE(tlist))),  &
            MSG='If a construct name is provided in '  & 
              // Articalise(initial_stmt) // ' then the same construct name &
              &must be specified in the matching ' // end_stmt // '.' )
      ELSE
        ! The name must be a name.
        CALL Expect(tlist, it, ittName, err)
        CALL Add(err_list, err, COMPONENT=comp)
        IF (Failure(err)) RETURN
        
        ! The name must match the construct name from the initial statement.
        IF (tlist(it) /= .NameOf. construct_name)  &
            CALL Add( err_list, CODE=errNameMismatch, COMPONENT=comp,  & 
                LOCATION=QueryLocation(tlist(it)),  &
                MSG='The construct name in the ' // end_stmt  &
                  // '("' // QueryRaw(tlist(it))  &
                  // '") does not match the construct name in the opening '  &
                  // initial_stmt // '("' // .NameOf. construct_name  &
                  // '") at ' // .LocationStringOf. construct_name  &
                  // '.' )
        
        ! There must be nothing after the name.
        CALL ExpectEOS(tlist, it + 1, comp, err_list)
      END IF
    ELSE
      ! Then opening statement has no construct name.  The end statement 
      ! must also not have a name.
      IF (SIZE(tlist) < it) RETURN
      IF (Match(tlist, it, ittName)) THEN
        CALL Add( err_list, CODE=errNameMismatch, COMPONENT=comp,  & 
            LOCATION=QueryLocation(tlist(it)),  &
            MSG='A construct name may only be provided in '  &
              // Articalise(initial_stmt)  &
              // ' if a construct name is specified in the opening '  &
              // end_stmt // '.' )
      ELSE
        CALL ExpectEOS(tlist, it, comp, err_list)
      END IF
    END IF
    
  END SUBROUTINE CheckConstructNameEnd_
  
  
  !*****************************************************************************
  !!
  !> Check that the construct name part of a intermediate statement in a 
  !! construct matches the typical requirements on the construct name.
  !!
  !! An intermediate statement is something like an ELSE statement - 
  !! a fundamental part of the if construct that appears in the middle of 
  !! the construct.
  !!
  !! @param[in]     tlist             Sequence of tokens that make up the 
  !! intermediate statement for the construct.  Must have size greater than 
  !! zero.
  !!
  !! @param[in]     it                Index of the token that hypothetically 
  !! would be the construct name, if it is present.  May be beyond the end 
  !! of the token sequence (in cases where the construct name is not present 
  !! in the intermediate statement for the construct).
  !!
  !! @param[in]     initial_name      Construct name on the opening statement 
  !! of the construct, not allocated if no construct name was provided.
  !!
  !! @param[in]     initial_stmt      Syntax rule name for the initial 
  !! statement, used for error reporting.  Include the '/'.
  !!
  !! @param[in]     intermediate_stmt  Syntax rule name for the intermediate 
  !! statement, used for error reporting.  Include the '/'.
  !!
  !! @param[in]     comp              Component for error reporting.  
  !! Typically this would include the constraint and the syntax rule number 
  !! for the end statement.
  !!
  !! @param[in,out] err_list          List of errors.  Note this is 
  !! INTENT(INOUT) - the idea being that a mismatch of construct names does 
  !! not terminate parsing.
  !!
  !! We check:
  !! - If the opening statement does not have a construct name, then 
  !! the intermediate statement must not have a construct name.
  !! - If the opening statement has a construct name and the intermediate 
  !! statement has a construct name, then the names must match.
  !! - That if no construct name is present, it is beyond the end of the 
  !! statement, and if a construct name is present, that it is the last 
  !! thing in the statement.
  !!
  !! The difference with the End check is that the presence of the construct 
  !! name on the intermediate statement is optional.
  
  SUBROUTINE CheckConstructNameIntermediate_( tlist, it, construct_name,  &
      initial_stmt, intermediate_stmt, comp, err_list )
    
    USE Errors
    USE ErrorCodes
    USE Identifiers
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: it
    TYPE(Identifier), INTENT(IN), ALLOCATABLE :: construct_name
    CHARACTER(*), INTENT(IN) :: initial_stmt
    CHARACTER(*), INTENT(IN) :: intermediate_stmt
    CHARACTER(*), INTENT(IN) :: comp
    TYPE(Error), INTENT(INOUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    
    !***************************************************************************
    
    ! The construct name is optional on an intermediate statement.
    IF (SIZE(tlist) < it) RETURN
    
    ! Past here, the intermediate statement has a name.
    
    ! A construct name is a name.
    CALL Expect(tlist, it, ittName, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    IF (ALLOCATED(construct_name)) THEN
      ! The opening statement has a name.  The intermediate statement may 
      ! have a name, but if it has a name it must match.
      IF (tlist(it) /= .NameOf. construct_name)  &
          CALL Add( err_list, CODE=errNameMismatch, COMPONENT=comp,  & 
              LOCATION=QueryLocation(tlist(it)),  &
              MSG='The construct name in the ' // intermediate_stmt  &
                // ' ("' // QueryRaw(tlist(it))  &
                // '") does not match the construct name in the opening '  &
                // initial_stmt // ' ("' // .NameOf. construct_name  &
                // '") at ' // .LocationStringOf. construct_name // '.' )
    ELSE
      ! The opening statement had no name.  The intermediate statement must 
      ! not have a name.
      CALL Add( err_list, CODE=errNameMismatch, COMPONENT=comp,  & 
          LOCATION=QueryLocation(tlist(it)),  &
          MSG='A construct name may only be provided in '  &
            // Articalise(intermediate_stmt)  &
            // ' if a construct name is specified in the opening '  &
            // initial_stmt // '.' )
    END IF
    
    CALL ExpectEnd(tlist, it + 1, comp, err_list)
    
  END SUBROUTINE CheckConstructNameIntermediate_
  
END MODULE ConstructNames
