! $Id: ContiguousStmts.f90 1863 2015-09-14 23:45:00Z ian $
!> @file
!! Defines the ContiguousStmts module.


!*******************************************************************************
!!
!> Procedures for parsing /contiguous-stmt/'s.

MODULE ContiguousStmts
  
  USE ParseNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseContiguousStmt
  
  !> Stores the results of parsing a /contiguous-stmt/.
  TYPE, EXTENDS(NameListPNode), PUBLIC :: ContiguousStmt
  CONTAINS
    PROCEDURE, NOPASS :: GetListName => contiguous_GetListName
    PROCEDURE, NOPASS :: What => contiguous_What
  END TYPE ContiguousStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /contiguous-stmt/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    object_name_list  List of the object names declared 
  !! in the statement.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseContiguousStmt(tlist, cont_stmt, err_list)
  
    USE Strings
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(ContiguousStmt), INTENT(OUT) :: cont_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /contiguous-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R533'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err      ! Error object for Expect calls.
    INTEGER :: it           ! Current token index.
    INTEGER :: opt          ! Expect option return.
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /contiguous-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    CALL ExpectKeyword(tlist, 1, scck_'CONTIGUOUS', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! :: or /object-name-list/
    
    CALL Expect(  &
        tlist, 2,  &
        [ittDeclare, ittName],  &
        [scck_'::', scck_'  '],  & 
        '"::" or /object-name/',  &
        opt, err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    IF (opt == 1) THEN
      it = 3
    ELSE
      it = 2
    END IF
    
    !---------------------------------------------------------------------------
    ! The /object-name-list/
    
    CALL ParseNameList(  &
        tlist,  &
        it, SIZE(tlist),  &
        '/object-name/',  &
        cont_stmt%name_list,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
    ! That's it!
    
  END SUBROUTINE ParseContiguousStmt
  
  
  !*****************************************************************************
  !!
  !> Implementation of ContiguousStmt%GetListName - return a name for the 
  !! stored list of names.
  !!
  !! @returns '/object-name-list/'
  
  FUNCTION contiguous_GetListName() RESULT(name)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: name
    
    !***************************************************************************
    
    name = '/object-name-list/'
    
  END FUNCTION contiguous_GetListName
  
  
  !*****************************************************************************
  !!
  !> Implementation of ContiguousStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'contiguous-stmt'.
  
  FUNCTION contiguous_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'contiguous-stmt'
    
  END FUNCTION contiguous_What
  
END MODULE ContiguousStmts
