! $Id: DeallocateStmts.f90 2795 2019-03-22 09:15:08Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the AllocateStmts module.


!*******************************************************************************
!!
!> Procedures and types for parsing /deallocate-stmts/'s (R640).

MODULE DeallocateStmts
  
  USE AllocateStmts
  USE CompilerKinds
  USE DeclarationTypeSpecs
  USE Designators
  USE Expressions
  USE ParseNodes
  USE SubscriptLists
  USE Variables
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseDeallocateStmt
  
  !-----------------------------------------------------------------------------
  
  !> Type to store a list of allocate objects.
  TYPE AllocateObjectList
    CLASS(AllocateObject), ALLOCATABLE :: item
  END TYPE AllocateObjectList
  
  !> Represents the results of parsing a /deallocate-stmt/.
  TYPE, EXTENDS(SCParseNode), PUBLIC ::  DeallocateStmt
    !> The things being deallocated.
    !!
    !! Always allocated, always size one or greater.
    TYPE(AllocateObjectList), ALLOCATABLE :: objects(:)
    !> The variable nominated in a STAT= option.
    !!
    !! Not allocated if there was no STAT= option.
    CLASS(Variable), ALLOCATABLE :: stat
    !> The variable nominated in a ERRMSG= option.
    !!
    !! Not allocated if there was no ERRMSG= option.
    CLASS(Variable), ALLOCATABLE :: errmsg
  CONTAINS
    PROCEDURE :: Dump => dealloc_Dump
    PROCEDURE :: Query => dealloc_Query
    PROCEDURE :: Modify => dealloc_Modify
    PROCEDURE, NOPASS :: What => dealloc_What
  END TYPE DeallocateStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /deallocate-stmt/ (R640).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    dealloc_stmt      Results of parsing the /allocate-stmt/.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseDeallocateStmt(tlist, dealloc_stmt, err_list)
    
    USE Errors
    USE ErrorCodes
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(DeallocateStmt), INTENT(OUT) :: dealloc_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule for an /deallocate-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R640'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect* calls.
    INTEGER :: closing        ! Index of the closing delimiter.
    INTEGER :: pass           ! Current pass number.
    INTEGER :: dealloc_count  ! Number of deallocations in the list.
    INTEGER :: assign         ! Index of the assign (=) token.
    INTEGER :: list_count     ! Number of items in the list.
    INTEGER :: item           ! List item index.
    INTEGER :: item_start     ! Start index in tlist of a particular item.
    INTEGER :: item_finish    ! Finish index in tlist of a particular item.
    INTEGER :: after          ! Index of the token after the object.
    
    ! Flag to indicate that at least one /dealloc-opt/ has been seen - all 
    ! items in the list from that point on must be /dealloc-opt/'s.
    LOGICAL :: seen_opt
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
          MSG='Expected /deallocate-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! DEALLOCATE ( ... )
    
    CALL ExpectKeyword(tlist, 1, scck_'DEALLOCATE', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL Expect(tlist, 2, ittDelimiter, scck_'(', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL ExpectClosing(tlist, 2, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! The combination of the /allocate-object-list/ and /dealloc-opt-list/.
    
    ! Some items inside the () are /allocate-object/'s, some are 
    ! /dealloc-opt/'s.  list_count is the total number of both.
    list_count = ListItemCount(tlist(3:closing-1))
    
    ! /dealloc-opt/'s must follow /allocate-objects/'s by the syntax rules.  
    ! Set a flag to track this.  Flag only valid on the first pass.
    seen_opt = .FALSE.
    
    ! First pass count and generate errors relating to position, second pass 
    ! parse.
    pass_loop: DO pass = 1, 2
      
      dealloc_count = 0
      item_finish = 0
      
      DO item = 1, list_count
        CALL GetNextListItem(tlist, 3, closing - 1, item_start, item_finish)
        
        ! Do we have a /dealloc-opt/?
        assign = FindBare(tlist(item_start:item_finish), ittAssign, scck_'=')
        IF (assign == 0) THEN
          ! Not a /dealloc-opt/
          dealloc_count = dealloc_count + 1
          IF (seen_opt .AND. (pass == 1)) THEN
            ! Cannot have /allocate-object/'s after we've started seeing 
            ! /dealloc-opt/'s.
            CALL Add( err_list,  &
                CODE=errSyntax,  &
                COMPONENT=comp,  &
                LOCATION=QueryLocation(tlist(item_start)),  &
                MSG='All the /allocate-object/''s in a /deallocate-stmt/ &
                  &must come before the /dealloc-opt/ specifications.' )
            ! Continue on this error.
          ELSE IF (pass == 2) THEN
            ! Parse the /allocate-object/
            CALL ParseAllocateObject(  &
                tlist,  &
                item_start, item_finish,  &
                dealloc_stmt%objects(dealloc_count)%item,  &
                after, &
                sub_err_list )
            CALL Add(err_list, sub_err_list, COMPONENT=comp)
            ! Continue on error, but don't rely on after.
            IF (.NOT. Failure(sub_err_list)) THEN
              ! Test that there was nothing following the /allocate-object/.
              CALL ExpectEnd(tlist(:item_finish), after, comp, err_list)
            END IF
          END IF
        ELSE
          ! Looks like a /dealloc-opt/.
          seen_opt = .TRUE.
          IF ((pass == 1) .AND. (dealloc_count == 0)) THEN
            ! Need to have at least one deallocation.
            CALL Add( err_list,  &
                CODE=errSyntax,  &
                COMPONENT=comp,  &
                LOCATION=QueryLocation(tlist(item_start)),  &
                MSG='A /deallocate-stmt/ must have at least one &
                  &/allocate-object/.' )
            ! Continue on this error.
          END IF
          IF (pass == 2) THEN
            CALL parse_dealloc_opt( tlist, item_start, item_finish,  &
                dealloc_stmt, sub_err_list )
            CALL Add(err_list, sub_err_list, COMPONENT=comp)
            ! Continue on error.
          END IF
        END IF
      END DO
      
      IF (pass == 1) ALLOCATE(dealloc_stmt%objects(dealloc_count))
      
    END DO pass_loop
    
    !---------------------------------------------------------------------------
    ! That's it.
    
    CALL ExpectEOS(tlist, closing + 1, comp, err_list)
    
  END SUBROUTINE ParseDeallocateStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /dealloc-opt/ (R641).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the first token of the 
  !! /dealloc-opt/ in @a tlist.
  !!
  !! @param[in]     finish            Index of the last token of the 
  !! /dealloc-opt/ in @a tlist.
  !!
  !! @param[in,out] alloc_stmt        Object for storing the results of 
  !! parsing the parent /deallocate-stmt/.  Note that this is INOUT.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE parse_dealloc_opt(tlist, start, finish, dealloc_stmt, err_list)
    
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(DeallocateStmt), INTENT(INOUT) :: dealloc_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /dealloc-opt/.
    CHARACTER(*), PARAMETER :: comp = 'R641'
    
    ! Keywords for each /dealloc-opt/.
    CHARACTER(*,KIND=scck), PARAMETER :: keywords(2) = [  &
        scck_'ERRMSG', scck_'STAT  ' ]
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: opt            ! Expect option.
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    ! The argument keyword.
    
    CALL ExpectKeyword(tlist, start, keywords, '/dealloc-opt/', opt, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    
    CALL Expect(tlist, start + 1, ittAssign, scck_'=', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    
    SELECT CASE (opt)
    CASE (1)  ! ERRMSG
      IF (ALLOCATED(dealloc_stmt%errmsg)) THEN
        CALL already('ERRMSG')
      ELSE
        CALL ParseVariable(  &
            tlist,  &
            start + 2, finish,  &
            dealloc_stmt%errmsg,  &
            sub_err_list )
      END IF
      
    CASE (2)  ! STAT
      IF (ALLOCATED(dealloc_stmt%stat)) THEN
        CALL already('STAT')
      ELSE
        CALL ParseVariable(  &
            tlist,  &
            start + 2, finish,  &
            dealloc_stmt%stat,  &
            sub_err_list )
      END IF
      
    CASE DEFAULT
      STOP 'Unknown /dealloc-opt/.'
      
    END SELECT
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    
  CONTAINS
    
    SUBROUTINE already(text)
      USE ErrorCodes
      !-------------------------------------------------------------------------
      ! Arguments
      CHARACTER(*), INTENT(IN) :: text
      !*************************************************************************
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT='C646 on ' // comp,  &
          LOCATION=QueryLocation(tlist(start)),  &
          MSG=text // ' has already been specified for this &
            &/deallocate-stmt/.' )
    END SUBROUTINE already
    
  END SUBROUTINE parse_dealloc_opt
  
  
  !*****************************************************************************
  !!
  !> Implementation of DeallocateStmt%Dump - write a debugging representation 
  !! of the results of parsing a deallocate statement to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE dealloc_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DeallocateStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Allocation index.
    
    !***************************************************************************
    
    WRITE (unit, "(A,'/allocate-object-list/')") pfx
    DO i = 1, SIZE(pnode%objects)
      WRITE (unit, "(A)", ADVANCE='NO') pfx // '  '
      CALL pnode%objects(i)%item%Write(unit)
      WRITE (unit, "()")
    END DO
    IF (ALLOCATED(pnode%stat)) THEN
      WRITE (unit, "(A,'stat variable = ')", ADVANCE='NO') pfx
      CALL WriteExpr(unit, pnode%stat)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%errmsg)) THEN
      WRITE (unit, "(A,'errmsg variable = ')", ADVANCE='NO') pfx
      CALL WriteExpr(unit, pnode%errmsg)
      WRITE (unit, "()")
    END IF
    
  END SUBROUTINE dealloc_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of DeallocateStmt%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE dealloc_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DeallocateStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Allocate object index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%objects)
      CALL pnode%objects(i)%item%Query(visitor)
    END DO
    
    IF (ALLOCATED(pnode%stat)) THEN
      CALL pnode%stat%Query(visitor)
    END IF
    
    IF (ALLOCATED(pnode%errmsg)) THEN
      CALL pnode%errmsg%Query(visitor)
    END IF
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE dealloc_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of DeallocateStmt%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[in,out] pnode             The parse node element to modify.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE dealloc_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DeallocateStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Allocate object index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%objects)
      CALL pnode%objects(i)%item%Modify(visitor)
    END DO
    
    IF (ALLOCATED(pnode%stat)) THEN
      CALL pnode%stat%Modify(visitor)
    END IF
    
    IF (ALLOCATED(pnode%errmsg)) THEN
      CALL pnode%errmsg%Modify(visitor)
    END IF
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE dealloc_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of DeallocateStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'deallocate-stmt'.
  
  FUNCTION dealloc_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'deallocate-stmt'
    
  END FUNCTION dealloc_What
  
END MODULE DeallocateStmts
