! $Id: DeclarationTypeSpecs.f90 2947 2019-09-25 21:05:13Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the DeclarationTypeSpecs module.


!*******************************************************************************
!!
!> Routines for parsing and a type for storing /declaration-type-specs/
!! and /type-specs/.  This also includes parsing of /intrinsic-type-specs/.
!!
!! Perhaps this module should be renamed to TypeSpecs given it parses
!! both sorts.
!!
!! R403: /declaration-type-spec/ is /intrinsic-type-spec/
!!                               or TYPE ( /intrinsic-type-spec/ )
!!                               or TYPE ( /derived-type-spec/ )
!!                               or CLASS ( /derived-type-spec/ )
!!                               or CLASS ( * )
!!
!! R404: /intrinsic-type-spec/ is INTEGER [ /kind-selector/ ]
!!                             or REAL [ /kind-selector/ ]
!!                             or DOUBLE PRECISION
!!                             or COMPLEX [ /kind-selector/ ]
!!                             or CHARACTER [ /char-selector/ ]
!!                             or LOGICAL [ /kind-selector/ ]
!!
!! R405: /kind-selector/ is ( [ KIND = ] /scalar-int-constant-expr/ )
!!
!! R420: /char-selector/ is /length-selector/
!!                       or ( LEN = type-param-value , 
!!                              KIND = scalar-int-constant-expr )
!!                       or ( type-param-value , 
!!                              [ KIND = ] scalar-int-constant-expr )
!!                       or ( KIND = scalar-int-constant=expr
!!                              [ , LEN = type-param-value ] )
!!
!! (This is a fancy way of saying that the default type parameter for 
!! CHARACTER variables is its LEN parameter, but it can have a KIND
!! parameter too.)
!!
!! R421: /length-selector/ is ( [ LEN = ] type-param-value )
!!                         or * char-length [ , ]
!!
!! (The second form is deprecated - which is a good thing given the problems
!! that it causes in scanning.)
!!
!! Note that we currently consider the optional trailing comma to be part of 
!! a /declaration-type-spec/ that is an /intrinsic-type-spec/, rather than 
!! part of the /length-selector/.  See C421 on R421 for more details.
!!
!! R422: /char-length/ is ( type-param-value )
!!                     or int-literal-constant
!!
!! R401; /type-param-value/ is scalar-int-expr
!!                          or *
!!                          or :

MODULE DeclarationTypeSpecs
  
  USE CompilerKinds, ONLY: scck
  USE Expressions
  USE Identifiers
  USE ParseNodes
  USE SourceLocations
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseDeclarationTypeSpec
  PUBLIC :: WriteDeclTypeSpec
  
  PUBLIC :: ParseTypeSpec
  PUBLIC :: WriteTypeSpec
  
  PUBLIC :: MatchIntrinsic
  PUBLIC :: MatchDeclarationTypeSpecStart
  PUBLIC :: GetDeclarationTypeSpecFinish
  
  !-----------------------------------------------------------------------------
  
  !> Represents a /type-param-value/, as described by R401, and also the 
  !! variants for some of the intrinsic kinds.
  TYPE, EXTENDS(TermParseNode), PUBLIC :: TypeParamValue
    !> .TRUE. if the type-param-value was a colon.
    !!
    !! Mutually exclusive with is_star being .TRUE. or with int_expr being
    !! allocated.
    LOGICAL :: is_colon
    
    !> .TRUE. if the type-param-value was a star.
    !!
    !! Mutually exclusive with is_colon being .TRUE. or with int_expr being
    !! allocated.
    LOGICAL :: is_star
    
    !> The location of the colon or star, if either of those is present.
    !!
    !! Not allocated if is_colon or is_star is false.  In this case the 
    !! location information can be queried out of the expression.
    TYPE(SourceLocation), ALLOCATABLE :: my_location
    
    !> Any other type-param-value.
    !!
    !! Not allocated if either is_colon or is_star is .TRUE..
    CLASS(Expression), ALLOCATABLE :: int_expr
  CONTAINS
    PROCEDURE :: Write => tpv_Write
    PROCEDURE :: Query => tpv_Query
    PROCEDURE :: Modify => tpv_Modify
    PROCEDURE, NOPASS :: What => tpv_What
    PROCEDURE :: GetLocation => tpv_GetLocation
    PROCEDURE :: Advertise => tpv_Advertise
  END TYPE TypeParamValue
  
  !> Represents a /type-param spec/, as described by R454, and also the 
  !! equivalent for the intrinsic kinds.
  TYPE, EXTENDS(TypeParamValue), PUBLIC :: TypeParamSpec
    !> Name of the parameter, if present.
    !!
    !! Not allocated if no parameter name has been specified.
    TYPE(Identifier), ALLOCATABLE :: name
  CONTAINS
    PROCEDURE :: Write => tps_Write
    PROCEDURE :: Query => tps_Query
    PROCEDURE :: Modify => tps_Modify
    PROCEDURE, NOPASS :: What => tps_What
    PROCEDURE :: GetLocation => tps_GetLocation
  END TYPE TypeParamSpec
  
  !> Represents the result of parsing a type-spec (R402), which could be an 
  !! intrinsic-type-spec or a derived-type-spec.
  !!
  !! A type-spec is basically a declaration-type-spec without the CLASS or
  !! type keyword (the type keyword being optional for intrinsic types).
  TYPE, EXTENDS(TermParseNode), PUBLIC :: TypeSpec
    !> The type name.
    TYPE(Identifier) :: type_name
    !> The type parameters.
    !!
    !! Always allocated, may be zero size.
    TYPE(TypeParamSpec), ALLOCATABLE :: params(:)
    !> Indicates that the /type-spec/ was a ye-olde CHARATCER*xxx style.
    !!
    !! There are restrictions on this style, that are relevant during 
    !! semantic analysis.
    LOGICAL :: has_char_length = .FALSE.
  CONTAINS
    PROCEDURE :: Write => ts_Write
    PROCEDURE :: Query => ts_Query
    PROCEDURE :: Modify => ts_Modify
    PROCEDURE :: GetLocation => ts_GetLocation
    PROCEDURE, NOPASS :: What => ts_What
    PROCEDURE :: Advertise => ts_Advertise
  END TYPE TypeSpec
  
  !> Represents the result of parsing a declaration-type-spec (R403).
  TYPE, EXTENDS(TermParseNode), PUBLIC :: DeclTypeSpec
    !> Whether the declaration is for a polymorphic (CLASS) or non-polymoprhic 
    !! entity (TYPE).
    LOGICAL :: is_polymorphic
    
    !> The specified type of the declaration, if is_unlimited is .FALSE.
    !!
    !! Not allocated if the specifier is for an unlimited polymorphic 
    !! object.
    TYPE(TypeSpec), ALLOCATABLE :: type_spec
    
    !> The location of the first token in the declaration type spec, if 
    !! the declaration type spec uses TYPE or CLASS.
    TYPE(SourceLocation), PRIVATE, ALLOCATABLE :: my_location
  CONTAINS
    PROCEDURE :: Write => dts_Write
    PROCEDURE :: Query => dts_Query
    PROCEDURE :: Modify => dts_Modify
    PROCEDURE :: GetLocation => dts_GetLocation
    PROCEDURE, NOPASS :: What => dts_What
  END TYPE DeclTypeSpec
  
  !> Get the final token of a /declaration-type-spec/.
  INTERFACE GetDeclarationTypeSpecFinish
    MODULE PROCEDURE GetDeclarationTypeSpecFinish_
  END INTERFACE GetDeclarationTypeSpecFinish
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a /declaration-type-spec/ (R403).
  !!
  !! @param[in]     tlist             List of tokens.
  !!
  !! @param[in]     comma_permitted   Flag to indicate that the 
  !! /declaration-type-spec/ is at the front of a /type-declaration-stmt/ 
  !! that does not have a ::. This is used for constraint checking around the 
  !! use of an optional separating comma after the obsolescent form of 
  !! CHARACTER type specifications.  See C421 and C422 on R421.
  !!
  !! @param[out]    ds                Object that describes the results of 
  !! parsing the /declaration-type-spec/.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseDeclarationTypeSpec(tlist, comma_permitted, ds, err_list)
    
    USE Errors
    USE ErrorCodes
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    LOGICAL, INTENT(IN) :: comma_permitted
    TYPE(DeclTypeSpec), INTENT(OUT) :: ds
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /declaration-type-spec/.
    CHARACTER(*), PARAMETER :: comp = 'R403'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    
    ! Error object for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    INTEGER :: opt            ! Expect option.
    INTEGER :: closing        ! Index of the closing parenthesis.
    
    ! Procedures for token matching.
    TYPE(MatchProcList) :: proc_list(3)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /declaration-type-spec/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    proc_list(1)%item => MatchIntrinsic
    proc_list(2)%item => match_type
    proc_list(3)%item => match_class
    
    CALL Expect( tlist,  &
        1,  &
        proc_list,  &
        '"INTEGER", "REAL", "LOGICAL", "COMPLEX", "CHARACTER", "TYPE", &
          &"CLASS" or "DOUBLE PRECISION"',  &
        opt,  &
        err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    SELECT CASE (opt)
    CASE (1)    ! Intrinsic type name.
      ds%is_polymorphic = .FALSE.
      ALLOCATE(ds%type_spec)
      CALL parse_intrinsic_type_spec(  &
          tlist,  &
          MERGE(0, 1, comma_permitted),  &
          ds%type_spec,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      
    CASE (2)    ! TYPE...
      ds%is_polymorphic = .FALSE.
      ALLOCATE(ds%type_spec)
      ds%my_location = QueryLocation(tlist(1))
      
      CALL Expect(tlist, 2, ittDelimiter, scck_'(', err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      CALL ExpectClosing(tlist, 2, closing, err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      IF (MatchIntrinsic(tlist, 3)) THEN
        ! F08/0097 in TC 3 deleted the option of a comma in an 
        ! intrinsic type spec inside TYPE(...).
        CALL parse_intrinsic_type_spec(  &
            tlist(3:closing-1),  &
            2,  &
            ds%type_spec,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      ELSE
        CALL parse_derived_type_spec(  &
            tlist(3:closing-1), &
            ds%type_spec,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
      CALL ExpectEnd(tlist, closing + 1, comp, err_list)
      
    CASE (3)    ! CLASS
      ds%is_polymorphic = .TRUE.
      
      CALL Expect(tlist, 2, ittDelimiter, scck_'(', err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      CALL ExpectClosing(tlist, 2, closing, err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      IF (Match(tlist, 3, ittOperator, scck_'*')) THEN
        IF (closing /= 4) THEN
          CALL Add( err_list,  &
              CODE=errSyntax,  &
              COMPONENT=comp,  &
              LOCATION=QueryLocation(tlist(closing)),  &
              MSG='Unexpected tokens following "*".' )
        END IF
      ELSE IF (MatchIntrinsic(tlist, 3)) THEN
        CALL Add( err_list,  &
            CODE=errSyntax,  &
            COMPONENT=comp,  &
            LOCATION=QueryLocation(tlist(3)),  &
            MSG='An /intrinsic-type-spec/ is not permitted in a &
              &polymorphic (CLASS) /declaration-type-spec/.' )
      ELSE
        ALLOCATE(ds%type_spec)
        CALL parse_derived_type_spec(  &
            tlist(3:closing-1),  &
            ds%type_spec,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
      CALL ExpectEnd(tlist, closing + 1, comp, err_list)
      
    END SELECT
    
  END SUBROUTINE ParseDeclarationTypeSpec
  
  
  !*****************************************************************************
  !!
  !> Parses sequence of tokens that is a type specification (R402).
  !!
  !! This is defined as an /intrinsic-type-spec/ or a /derived-type-spec/.
  !!
  !! A type specification is a /declaration-type-spec/ less the CLASS or
  !! TYPE keywords and without the unlimited polymorphic option (*).  In 
  !! addition to /type-declaration-stmt/'s it may be used in ALLOCATE 
  !! statements and SELECT TYPE guard statements.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    type_spec         Object describing the results of parsing 
  !! the /type-spec/.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseTypeSpec(tlist, type_spec, err_list)
    
    USE Errors
    USE ErrorCodes
    USE MatchUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(TypeSpec), INTENT(OUT) :: type_spec
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /type-spec/.
    CHARACTER(*), PARAMETER :: comp = 'R402'
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /type-spec/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! Work out whether we have an /intrinsic-type-spec/ or a 
    ! /derived-type-spec/
    
    IF (MatchIntrinsic(tlist, 1)) THEN
      ! /type-spec/'s cannot appear in /type-declaration-stmt/'s.
      CALL parse_intrinsic_type_spec(tlist, 3, type_spec, err_list)
    ELSE IF (Match(tlist, 1, ittName)) THEN
      CALL parse_derived_type_spec(tlist, type_spec, err_list)
    ELSE
      IF (SIZE(tlist) > 0) THEN
        CALL Add( err_list,  &
            CODE=errSyntax,  &
            COMPONENT=comp,  &
            LOCATION=QueryLocation(tlist(1)),  &
            MSG='Expected /intrinsic-type-spec/ or /derived-type-spec/.' )
      ELSE
        CALL Add( err_list,  &
            CODE=errSyntax,  &
            COMPONENT=comp,  &
            MSG='Expected /intrinsic-type-spec/ or /derived-type-spec/.' )
      END IF
    END IF
    
  END SUBROUTINE ParseTypeSpec
  
  
  !*****************************************************************************
  !!
  !> Get the final token of a /declaration-type-spec/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     comma_permitted   Flag to indicate that the 
  !! /declaration-type-spec/ is at the front of a /type-declaration-stmt/ 
  !! that does not have a ::.
  !!
  !! @returns The index of the final token of the declaration type spec 
  !! in @a tlist.
  
  FUNCTION GetDeclarationTypeSpecFinish_(tlist, comma_permitted) RESULT(it)
    
    USE MatchUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    LOGICAL, INTENT(IN) :: comma_permitted
    
    ! Function result
    INTEGER :: it
    
    !***************************************************************************
    
    ! Locate the tokens associated with the declaration-type-spec.  This 
    ! always starts from token 1, but we need to work out where it finishes.  
    ! There are three options that involve parentheses that must start from 
    ! the second token:
    !
    ! INTEGER(kinds...) attributes-and-vars
    ! TYPE(type-name...) attributes-and-vars
    ! CLASS(type-name...) attributes-and-vars
    !
    ! The DOUBLE PRECISION series of tokens implicitly specifies the kind,
    ! so you can't have DOUBLE PRECISION (xxx).
    IF (match_dbl_prec(tlist, 1)) THEN
      it = 2
    ELSE IF (match_dblprec(tlist, 1)) THEN
      it = 1
    ELSE IF (Match(tlist, 2, ittDelimiter, scck_'(')) THEN 
      ! xxxx(...
      it = FindClosingParens(tlist, 2)
      
    ! There are no parentheses.  This means we are dealing with an 
    ! intrinsic type, something like INTEGER attributes-and-vars, 
    ! or CHARACTER * /char-length/ [,] (a deprecated form only permitted 
    ! in a /type-declaration-stmt/).
    ELSE IF (Match(tlist, 2, ittOperator, scck_'*')) THEN 
      ! If REAL*8 or somesuch platform specific nonsense is specified we 
      ! still pass that through to the decl_spec parsing code.
      !
      ! The comma after the int literal constant is optional, but it is 
      ! considered part of the /length-selector/ that is part of the 
      ! /intrinsic-type-spec/ that is part of the /declaration-type-spec/.
      IF (Match(tlist, 3, ittDelimiter, scck_'(')) THEN
        it = FindClosingParens(tlist, 3)
        IF (it == 0) RETURN
        
        IF ( comma_permitted  &
            .AND. Match(tlist, it + 1, ittComma, scck_',') )  &
              it = it + 1
      ELSE IF ( comma_permitted  &
          .AND. Match(tlist, 4, ittComma, scck_',') ) THEN  ! '
        it = 4
      ELSE
        it = MIN(SIZE(tlist), 3)
      END IF
    ELSE 
      it = MIN(SIZE(tlist), 1)
    END IF
    
  END FUNCTION GetDeclarationTypeSpecFinish_
  
  
  !*****************************************************************************
  !!
  !> Test for the start (first token or two) of a /declaration-type-spec/.
  !!
  !! @param[in]     tlist             List of tokens.
  !!
  !! @param[in]     start             Position in @a tlist from which to test.
  !!
  !! @returns .TRUE. if the tokens starting at @a tlist ( @a start ) look like
  !! the start of a /declaration-type-spec/, .FALSE. otherwise.
  
  FUNCTION MatchDeclarationTypeSpecStart(tlist, start) RESULT(b)
    
    USE CompilerKinds
    USE MatchUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    
    ! Function result
    LOGICAL :: b
    
    !***************************************************************************
    
    b = MatchIntrinsic(tlist, start)  &
        .OR. Match(tlist, start, ittName, scck_'CLASS')  &
        .OR. Match(tlist, start, ittName, scck_'TYPE')
    
  END FUNCTION MatchDeclarationTypeSpecStart
  
  
  !*****************************************************************************
  !!
  !> Parse an /intrinsic-type-spec/ (R404).
  !!
  !! @param[in]     tlist             List of tokens.
  !!
  !! @param[in]     no_comma          Indicates whether a trailing comma is 
  !! permitted.  The value affects the constraint used in error reporting.
  !!  - 0: A comma is permitted.
  !!  - 1: No comma is permitted, a double colon was present, C422.
  !!  - 2: No comma is permitted, we are inside TYPE(...), from C406a in TC3.
  !!  - 3: No comma is permitted, this is not in a  /declaration-type-spec/.
  !!
  !! @param[out]    ts                Object that describes the results of 
  !! parsing the /intrinsic-type-spec/.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE parse_intrinsic_type_spec(tlist, no_comma, ts, err_list)
    
    USE Errors
    USE ErrorCodes
    USE ErrorLevels
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: no_comma
    TYPE(TypeSpec), INTENT(OUT) :: ts
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for an /intrinsic-type-spec/.
    CHARACTER(*), PARAMETER :: comp = 'R404'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(MatchProcList) :: proc_list(4)
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    
    ! Error list for child parse calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    INTEGER :: opt            ! Expect option.
    INTEGER :: comma          ! Index of the comma.
    INTEGER :: it             ! Current token index.
    INTEGER :: it_end         ! Last token in current part.
    INTEGER :: closing        ! Index of the closing parenthesis.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    ! Use the leading token to work out what sort of spec we have.
    
    proc_list(1)%item => match_intrinsic_lirc
    proc_list(2)%item => match_intrinsic_c
    proc_list(3)%item => match_dbl_prec
    proc_list(4)%item => match_dblprec
    
    CALL Expect(tlist, 1, proc_list,  &
        '"INTEGER", "REAL", "LOGICAL", "COMPLEX", "CHARACTER" or &
          &"DOUBLE PRECISION"', opt, err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    SELECT CASE (opt)
    CASE (1)    ! LOGICAL, INTEGER, REAL or COMPLEX
      ts%type_name = IdentifierIfortBug(tlist(1))
      
      IF (Match(tlist, 2, ittDelimiter, scck_'(')) THEN 
        CALL ExpectClosing(tlist, 2, closing, err)
        CALL Add(err_list, err, COMPONENT=comp)
        IF (Failure(err)) RETURN
              
        CALL parse_type_param_spec_list(  &
            tlist(3:closing-1),  &
            ts%params,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
        it = closing + 1
      ELSE
        ALLOCATE(ts%params(0))
        it = 2
      END IF
      CALL ExpectEnd(tlist, it, comp, err_list)
      
    CASE (2)    ! CHARACTER - can have archaic length specifications.
      ts%type_name = IdentifierIfortBug(tlist(1))
      
      IF (Match(tlist, 2, ittOperator, scck_'*')) THEN
        ! Expect a /char-length/ followed by an optional comma.
        
        ! Is there a trailing comma at the end?
        comma = FindBare(tlist, ittComma, scck_',', 3)
        IF (comma == 0) THEN
          it_end = SIZE(tlist)
        ELSE
          it_end = comma - 1
        END IF
        
        ! Deprecated feature warning
        CALL Add( err_list,  &
            CODE=errObsolescentCharLength,  &
            LEVEL=errLevelWarning,  &
            COMPONENT=comp,  &
            LOCATION=QueryLocation(tlist(2)),  &
            MSG='The * /char-length/ form of character variable length &
              &specification is an obsolescent feature.' )
        
        ts%has_char_length = .TRUE.
        
        ! Parse the /char-length/.
        ALLOCATE(ts%params(1))
        CALL parse_char_length(  &
            tlist(3:it_end),  &
            ts%params(1)%TypeParamValue,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
        IF (Failure(sub_err_list)) RETURN
        
        IF (comma /= 0) THEN
          SELECT CASE (no_comma)
          CASE (0)    ! trailing comma ok.
          CASE (1)    ! double colon present
            CALL Add( err_list,  &
                CODE=errSyntax,  &
                COMPONENT='C422 via ' // comp,  &
                LOCATION=QueryLocation(tlist(comma)),  &
                MSG='A comma after the /char-length/ of a CHARACTER*nnn &
                  &style declaration is only permitted if no double-colon &
                  &separator appears in the /type-decl-stmt/.' )
          CASE (2)    ! inside TYPE(...)   F08/0097 TC3.
            CALL Add( err_list,  &
                CODE=errSyntax,  &
                COMPONENT='C406a via ' // comp,  &
                LOCATION=QueryLocation(tlist(comma)),  &
                MSG='A comma after the /char-length/ of a CHARACTER*nnn &
                  &style declaration is not permitted inside a &
                  &/declaration-type-spec/ using TYPE(...).' )
          CASE (3)    ! not in a /declaration-type-spec/.
            CALL Add( err_list,  &
                CODE=errSyntax,  &
                COMPONENT='C421 via ' // comp,  &
                LOCATION=QueryLocation(tlist(comma)),  &
                MSG='A comma after the /char-length/ of a CHARACTER*nnn &
                  &style declaration is not permitted outside of a &
                  &/type-decl-stmt/.' )
          END SELECT
          it = comma + 1
        ELSE
          ! No need to check for extra tokens.
          RETURN
        END IF
      
      ELSE IF (Match(tlist, 2, ittDelimiter, scck_'(')) THEN
        CALL ExpectClosing(tlist, 2, closing, err)
        CALL Add(err_list, err, COMPONENT=comp)
        IF (Failure(err)) RETURN
        
        CALL parse_type_param_spec_list(  &
            tlist(3:closing-1),  &
            ts%params,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
        it = closing + 1
        
      ELSE
        
        ALLOCATE(ts%params(0))
        it = 2
        
      END IF
      
      ! Check for extra tokens.
      CALL ExpectEnd(tlist, it, comp, err_list)
      
     CASE (3)   ! DOUBLE PRECISION
      CALL do_double_precision_ts(QueryLocation(tlist(1)), ts)
      CALL ExpectEnd(tlist, 3, comp, err_list)
        
    CASE (4)    ! DOUBLEPRECISION
      CALL do_double_precision_ts(QueryLocation(tlist(1)), ts)
      CALL ExpectEnd(tlist, 2, comp, err_list)
    END SELECT
    
  END SUBROUTINE parse_intrinsic_type_spec
  
  
  !*****************************************************************************
  !!
  !> Parse a derived-type-spec (R453).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    ts                The results of parsing the type spec.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! This is defined as:
  !!
  !! /type-name/ [ ( /type-param-spec-list/ ) ]
  
  SUBROUTINE parse_derived_type_spec(tlist, ts, err_list)
    
    USE CompilerKinds
    USE Errors
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(TypeSpec), INTENT(OUT) :: ts
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /derived-type-spec/.
    CHARACTER(*), PARAMETER :: comp = 'R453'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    
    ! Error list for child parse calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    INTEGER :: it             ! Current token index.
    INTEGER :: closing        ! Index of closing parenthesis.
    
    !***************************************************************************
  
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    ! Derived type name
    
    CALL ExpectName(tlist, 1, '/type-name/', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ts%type_name = IdentifierIfortBug(tlist(1))
    
    !---------------------------------------------------------------------------
    ! Optional type parameters
    
    IF (Match(tlist, 2, ittDelimiter, scck_'(')) THEN 
      CALL ExpectClosing(tlist, 2, closing, err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      CALL parse_type_param_spec_list(  &
          tlist(3:closing-1),  &
          ts%params,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
      it = closing + 1
    ELSE
      ALLOCATE(ts%params(0))
      it = 2
    END IF
    
    !---------------------------------------------------------------------------
    
    CALL ExpectEnd(tlist, it, comp, err_list)
    
  END SUBROUTINE parse_derived_type_spec
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /type-param-spec-list/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    params            The type param spec list.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE parse_type_param_spec_list(tlist, params, err_list)
    
    USE Errors
    USE ErrorCodes
    USE Tokens
    USE ParseUtils
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(TypeParamSpec), INTENT(OUT), ALLOCATABLE :: params(:)
    TYPE(Error), INTENT(INOUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: list_count   ! Number of items in the /type-param-spec-list/.
    INTEGER :: item_start   ! Starting index of an item.
    INTEGER :: item_finish  ! Finishing index of an item.
    INTEGER :: item         ! Item index.
    
    ! Flag that indicates we've seen a KEYWORD= specifier in the list.
    LOGICAL :: seen_keyword
    
    ! Error list for child procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          MSG='Expected /type-param-spec/.' )
      RETURN
    END IF
    
    seen_keyword = .FALSE.
    
    !---------------------------------------------------------------------------
    
    list_count = ListItemCount(tlist)
    ALLOCATE(params(list_count))
    item_finish = 0
    DO item = 1, list_count
      CALL GetNextListItem(tlist, 1, SIZE(tlist), item_start, item_finish)
      CALL parse_type_param_spec(  &
          tlist(item_start:item_finish),  &
          params(item), sub_err_list )
      CALL Add(err_list, sub_err_list)
      IF (Failure(sub_err_list)) RETURN
      
      IF (seen_keyword) THEN
        IF (.NOT. ALLOCATED(params(item)%name)) THEN
          CALL Add( err_list,  &
              CODE=errSyntax,  &
              COMPONENT='C486 on R454',  &
              LOCATION=QueryLocation(tlist(item_start)),  &
              MSG='The keyword= may be omitted from a /type-param-spec/ &
                &only if the keyword= has been omitted from each &
                &preceding /type-param-spec/.' )
        END IF
      ELSE IF (ALLOCATED(params(item)%name)) THEN
        seen_keyword = .TRUE.
      END IF
    END DO
    
  END SUBROUTINE parse_type_param_spec_list
  
  
  !*****************************************************************************
  !!
  !> Parse a /char-length/ (the archaic form of the length specification).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    tpv               The type param value.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE parse_char_length(tlist, tpv, err_list)
    
    USE Errors
    USE ErrorCodes
    USE Expressions
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(TypeParamValue), INTENT(OUT) :: tpv
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /char-length/.
    CHARACTER(*), PARAMETER :: comp = 'R422'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: it             ! Token index.
    TYPE(Error) :: err        ! Error object for Expect call.
    INTEGER :: opt            ! Expect options.
    
    ! Token sequence match functions for an Expect call.
    TYPE(MatchProcList) :: procs(2)
    
    ! Error list for child parse calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    procs(1)%item => match_int_literal
    procs(2)%item => match_parens
    CALL Expect(tlist, 1, procs, '"/int-literal/" or "("', opt, err)
    CALL Add(err_list, err, COMPONENT=comp)
    
    IF (opt == 1) THEN
      IF (HasKind(tlist(1))) THEN
        CALL Add( err_list,  &
            CODE=errSyntax,  &
            COMPONENT='C416 via ' // comp,  &
!            LOCATION=tlist(1)%position,  &
            MSG='The /int-literal-constant/ in the /char-length/ of a &
              &CHARACTER*nnn style declaration may not have a &
              &/kind-param/.' )
      END IF
      CALL ParseExpr(tlist, 1, 1, tpv%int_expr, sub_err_list)
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      tpv%is_colon = .FALSE.
      tpv%is_star = .FALSE.
      
      CALL ExpectEnd(tlist, 2, comp, err_list)
    ELSE
      ! ( /type-param-value/ )
      CALL ExpectClosing(tlist, 1, it, err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      CALL parse_type_param_value(  &
          tlist(2:it-1),  &
          tpv%is_colon,  &
          tpv%is_star,  &
          tpv%int_expr,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      
      CALL ExpectEnd(tlist, it + 1, comp, err_list)
    END IF
    
  END SUBROUTINE parse_char_length
  
  
  !*****************************************************************************
  !!
  !> Parse the combination of an optional "keyword =" followed by a 
  !! /type-param-value/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    param             The results of parsing the 
  !! type parameter specification.
  !!
  !! @parma[out]    err_list          List of errors.
  
  SUBROUTINE parse_type_param_spec(tlist, param, err_list)
    
    USE Errors
    USE MatchUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(TypeParamSpec), INTENT(OUT) :: param
    TYPE(Error), INTENT(INOUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule numner for a /type-param-spec/.
    CHARACTER(*), PARAMETER :: comp = 'R454'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: it             ! Token index.
    
    ! Error list for child parse calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ! See if we have a NAME= thing happening.
    IF ( Match(tlist, 1, ittName)  &
        .AND. Match(tlist, 2, ittAssign) ) THEN
      param%name = IdentifierIfortbug(tlist(1))
      it = 3
    ELSE
      it = 1
    END IF
    
    CALL parse_type_param_value(  &
        tlist(it:),  &
        param%is_colon,  &
        param%is_star,  &
        param%int_expr,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
    IF (param%is_colon .OR. param%is_star) THEN
      param%my_location = QueryLocation(tlist(it))
    END IF
    
  END SUBROUTINE parse_type_param_spec
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /type-param-value/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    is_colon          Flag to indicate whether the type 
  !! parameter value was a colon (a deferred parameter).
  !!
  !! @param[out]    is_star           Flag to indicate whether the type 
  !! parameter value was a star (an assumed parameter).
  !!
  !! @param[out]    expr              If the type parameter value was 
  !! a colon or a star, this is not allocated.  Otherwise this is the 
  !! expression that specified the value.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! The three separate result arguments are the components of the 
  !! TypeParamValue derived type.  It's not clear why we provide them 
  !! as separate arguments.
  
  SUBROUTINE parse_type_param_value(tlist, is_colon, is_star, expr, err_list)
    
    USE Errors
    USE ErrorCodes
    USE Expressions
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    LOGICAL, INTENT(OUT) :: is_colon
    LOGICAL, INTENT(OUT) :: is_star
    CLASS(Expression), INTENT(OUT), ALLOCATABLE :: expr
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /type-param-value/.
    CHARACTER(*), PARAMETER :: comp = 'R401'
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Missing /type-param-value/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    IF (Match(tlist, 1, ittColon, scck_':')) THEN
      is_colon = .TRUE.
      CALL ExpectEnd(tlist, 2, comp, err_list)
    ELSE IF (Match(tlist, 1, ittOperator, scck_'*')) THEN
      is_star = .TRUE.
      CALL ExpectEnd(tlist, 2, comp, err_list)
    ELSE
      CALL ParseExpr(tlist, 1, SIZE(tlist), expr, err_list)
    END IF
    
  END SUBROUTINE parse_type_param_value
  
  
  !*****************************************************************************
  !!
  !> Handle a type-spec (which would be an intrinsic-type-spec) that is
  !! for "DOUBLE PRECISION".
  !!
  !! @param[out]    type_spec         A type specifier that describes a 
  !! type equivalent to REAL(KIND(0.0D0)).  For now we set the type 
  !! name to DOUBLEPRECISION and rely upon later stages that are aware 
  !! of the relevant kind of REAL to specify the details.
  
  SUBROUTINE do_double_precision_ts(location, type_spec)
    
    USE SourceLocations
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(TypeSpec), INTENT(OUT) :: type_spec
    TYPE(SourceLocation), INTENT(IN) :: location
    
    !***************************************************************************
    
    type_spec%type_name = IdentifierIfortBug('DOUBLEPRECISION', location)
    
  END SUBROUTINE do_double_precision_ts
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Specific functions that can be used as match procedures in an Expect call.
  
  
  !*****************************************************************************
  !!
  !> Test whether a token sequence matches the first keywords (the type name)
  !! of an intrinsic type spec.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Position at which to start the match.
  !!
  !! @returns True if the token sequence in @a tlist starting at @a start 
  !! looks like an intrinsic type name.  The double precision intrinsic 
  !! type specifier means that multiple tokens might be tested.
  
  FUNCTION MatchIntrinsic(tlist, start) RESULT(b)
    
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    
    ! Function result
    LOGICAL :: b
    
    !***************************************************************************
  
    b = match_intrinsic_lirc(tlist, start)
    IF (b) RETURN
    
    b = match_intrinsic_c(tlist, start)
    IF (b) RETURN
    
    b = match_dbl_prec(tlist, start)
    IF (b) RETURN
    
    b = match_dblprec(tlist, start)
    
  END FUNCTION MatchIntrinsic
  
  
  !***************************************************************************** 
  !! 
  !> Test a token to see if it matches LOGICAL, INTEGER, REAL or COMPLEX.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Position at which to start the match.
  !!
  !! @returns True if the token at start matches one of the given keywords, 
  !! false otherwise.
   
  FUNCTION match_intrinsic_lirc(tlist, start) RESULT(b)
    
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    
    ! Function result
    LOGICAL :: b
    
    !***************************************************************************
    
    IF (start <= SIZE(tlist)) THEN
      IF (tlist(start) == ittName) THEN
        b = tlist(start) == scck_'INTEGER'
        IF (b) RETURN
        
        b = tlist(start) == scck_'LOGICAL'
        IF (b) RETURN
        
        b = tlist(start) == scck_'REAL'
        IF (b) RETURN
        
        b = tlist(start) == scck_'COMPLEX'
        IF (b) RETURN
      END IF
    END IF
    
    b = .FALSE.
    
  END FUNCTION match_intrinsic_lirc
  
  
  !*****************************************************************************
  !!
  !> Test a token to see if it matches CHARACTER.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Position at which to start the match.
  !!
  !! @returns True if the token in @a tlist at @a start is CHARACTER, 
  !! false otherwise.
  
  FUNCTION match_intrinsic_c(tlist, start) RESULT(b)
    
    USE CompilerKinds
    USE MatchUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    
    ! Function result
    LOGICAL :: b
    
    !***************************************************************************
    
    b = Match(tlist, start, ittName, scck_'CHARACTER')
    
  END FUNCTION match_intrinsic_c
  
  
  !*****************************************************************************
  !!
  !> Test a token to see if it matches TYPE.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Position at which to start the match.
  !!
  !! @returns True if the token in @a tlist at @a start is TYPE, 
  !! false otherwise.
  
  FUNCTION match_type(tlist, start) RESULT(b)
    
    USE CompilerKinds
    USE MatchUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    
    ! Function result
    LOGICAL :: b
    
    !***************************************************************************
    
    b = Match(tlist, start, ittName, scck_'TYPE')
    
  END FUNCTION match_type
  
  
  !*****************************************************************************
  !!
  !> Test a token to see if it matches CLASS.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Position at which to start the match.
  !!
  !! @returns True if the token in @a tlist at @a start is CLASS, 
  !! false otherwise.
  
  FUNCTION match_class(tlist, start) RESULT(b)
    
    USE CompilerKinds
    USE MatchUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    
    ! Function result
    LOGICAL :: b
    
    !***************************************************************************
    
    b = Match(tlist, start, ittName, scck_'CLASS')
    
  END FUNCTION match_class
  
  
  !*****************************************************************************
  !!
  !> Test a token sequence to see if it matches DOUBLE PRECISION.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Position at which to start the match.
  !!
  !! @returns True if the token sequence in @a tlist beginning at @a start 
  !! matches DOUBLE PRECISION, false otherwise.
  
  FUNCTION match_dbl_prec(tlist, start) RESULT(b)
    
    USE CompilerKinds
    USE MatchUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    
    ! Function result
    LOGICAL :: b
    
    !***************************************************************************
    
    b = Match(tlist, start, ittName, scck_'DOUBLE')  &
      .AND. Match(tlist, start + 1, ittName, scck_'PRECISION')
    
  END FUNCTION match_dbl_prec
  
  
  !*****************************************************************************
  !!
  !> Test a token to see if it matches DOUBLEPRECISION.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Position at which to start the match.
  !!
  !! @returns True if the single token in @a tlist at @a start is 
  !! DOUBLEPRECISION, false otherwise.
  
  FUNCTION match_dblprec(tlist, start) RESULT(b)
    
    USE CompilerKinds
    USE MatchUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    
    ! Function result
    LOGICAL :: b
    
    !***************************************************************************
    
    b = Match(tlist, start, ittName, scck_'DOUBLEPRECISION')
    
  END FUNCTION match_dblprec
  
  
  !*****************************************************************************
  !!
  !> Test a token to see if it matches an /integer-literal-constant/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Position at which to start the match.
  !!
  !! @returns True if the single token in @a tlist at @a start is an 
  !! ittIntegerLiteral token, false otherwise.
  
  FUNCTION match_int_literal(tlist, start) RESULT(b)
    
    USE MatchUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    
    ! Function result
    LOGICAL :: b
    
    !***************************************************************************
    
    b = Match(tlist, start, ittIntegerLiteral)
    
  END FUNCTION match_int_literal
  
  
  !*****************************************************************************
  !!
  !> Test a token to see if it matches an opening parenthesis.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Position at which to start the match.
  !!
  !! @returns True if the single token in @a tlist at @a start is an 
  !! ittDelimiter token with value "(" (i.e. an open parenthesis), 
  !! false otherwise.
  
  FUNCTION match_parens(tlist, start) RESULT(b)
    
    USE MatchUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    
    ! Function result
    LOGICAL :: b
    
    !***************************************************************************
    
    b = Match(tlist, start, ittDelimiter, scck_'(')
    
  END FUNCTION match_parens
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for DeclTypeSpec.
  
  
  !*****************************************************************************
  !!
  !> Implementation of TypeSpec%Write - dump a debugging representation of 
  !! the loop control to a logical unit.
  !!
  !! @param[in]     loop              The loop control.
  !!
  !! @param[in]     unit              The logical unit.
  
  SUBROUTINE dts_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DeclTypeSpec), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    CALL WriteDeclTypeSpec(unit, pnode)
    
  END SUBROUTINE dts_Write
  
  
  !*****************************************************************************
  !!
  !> Write a debugging representation of a declaration type spec to a unit.
  !!
  !! @param[in]     log_unit          The unit connected for formatted output
  !! to write the representation of the declaration type spec to.
  !!
  !! @param[in]     type_spec         The type spec to write out.
  
  RECURSIVE SUBROUTINE WriteDeclTypeSpec(log_unit, ds)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    INTEGER, INTENT(IN) :: log_unit
    TYPE(DeclTypeSpec), INTENT(IN) :: ds
    
    !***************************************************************************
    
    IF (ds%is_polymorphic) THEN
      WRITE (log_unit, "('CLASS(')", ADVANCE='NO')
    ELSE
      WRITE (log_unit, "('TYPE(')", ADVANCE='NO')
    END IF
    IF (ALLOCATED(ds%type_spec)) THEN
      CALL ds%type_spec%Write(log_unit)
    ELSE
      WRITE (log_unit, "('*')", ADVANCE='NO')
    END IF
    WRITE (log_unit, "(')')", ADVANCE='NO')
    
  END SUBROUTINE WriteDeclTypeSpec
  
  
  !*****************************************************************************
  !!
  !> Implementation of DeclTypeSpec%Query - traverse the node with a query 
  !! operation.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  SUBROUTINE dts_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(DeclTypeSpec), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%type_spec)) THEN
      CALL pnode%type_spec%Query(visitor)
    END IF
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE dts_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of DeclTypeSpec%Modify - traverse the node with a modify 
  !! operation.
  !!
  !! @param[in]     pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  SUBROUTINE dts_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(DeclTypeSpec), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%type_spec)) THEN
      CALL pnode%type_spec%Modify(visitor)
    END IF
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE dts_Modify
  
  
  !*****************************************************************************
  !!
  !> Get the location of the declaration type spec.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @return The location in source of the node.
  
  FUNCTION dts_GetLocation(pnode) RESULT(loc)
    
    !---------------------------------------------------------------------------
    ! Chraacteristics
    
    CLASS(DeclTypeSpec), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%my_location)) THEN
      loc = pnode%my_location
    ELSE
      loc = pnode%type_spec%GetLocation()
    END IF
    
  END FUNCTION dts_GetLocation
  
  
  !*****************************************************************************
  !!
  !> Implementation of DeclTypeSpec%What - get a text description of the 
  !! relevant syntax rule for the parse node.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @returns 'decl-type-spec'
  
  FUNCTION dts_What() RESULT(what)
      
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'decl-type-spec'
    
  END FUNCTION dts_What
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for bindings in TypeSpec.
  
  
  !*****************************************************************************
  !!
  !> Implementation of TypeSpec%Write - dump a debugging representation of 
  !! the loop control to a logical unit.
  !!
  !! @param[in]     loop              The loop control.
  !!
  !! @param[in]     unit              The logical unit.
  
  SUBROUTINE ts_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(TypeSpec), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    CALL WriteTypeSpec(unit, pnode)
    
  END SUBROUTINE ts_Write
  
  
  !*****************************************************************************
  !!
  !> Write a debugging representation of a type spec to a unit.
  !!
  !! @param[in]     log_unit          The unit connected for formatted output 
  !! to write the representation of the type spec to.
  !!
  !! @param[in]     type_spec         The type spec to write out.
  
  RECURSIVE SUBROUTINE WriteTypeSpec(log_unit, type_spec)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    INTEGER, INTENT(IN) :: log_unit
    TYPE(TypeSpec), INTENT(IN) :: type_spec
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i            ! Type parameter index.
    
    !***************************************************************************
    
    WRITE (log_unit, "(A)", ADVANCE='NO') .NameOf. type_spec%type_name
    IF (type_spec%has_char_length) THEN
      WRITE (log_unit, "('*')", ADVANCE='NO')
    END IF
    IF (SIZE(type_spec%params) /= 0) THEN
      WRITE (log_unit, "('(')", ADVANCE='NO')
      DO i = 1, SIZE(type_spec%params)
        CALL type_spec%params(i)%Write(log_unit)
      END DO
      WRITE (log_unit, "(')')", ADVANCE='NO')
    END IF
    
  END SUBROUTINE WriteTypeSpec
  
  
  !*****************************************************************************
  !!
  !> Implementation of TypeSpec%Query - traverse the node with a query 
  !! operation.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  ! A type spec could contain another type spec in one of the expressions 
  ! for a parameter value.
  RECURSIVE SUBROUTINE ts_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(TypeSpec), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: ip             ! param index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO ip = 1, SIZE(pnode%params)
      CALL pnode%params(ip)%Query(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE ts_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of TypeSpec%Modify - traverse the node with a modify 
  !! operation.
  !!
  !! @param[in]     pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  ! A type spec could contain another type spec in one of the expressions 
  ! for a parameter value.
  RECURSIVE SUBROUTINE ts_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(TypeSpec), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: ip             ! param index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO ip = 1, SIZE(pnode%params)
      CALL pnode%params(ip)%Modify(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE ts_Modify
  
  
  !*****************************************************************************
  !!
  !> Get the location of the type spec.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @return The location in source of the node.
  !!
  !! Simply forwards to the location of the type identifier.
  
  FUNCTION ts_GetLocation(pnode) RESULT(loc)
    
    !---------------------------------------------------------------------------
    ! Chraacteristics
    
    CLASS(TypeSpec), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    loc = pnode%type_name%GetLocation()
    
  END FUNCTION ts_GetLocation
  
  
  !*****************************************************************************
  !!
  !> Implementation of TypeSpec%What - get a text description of the 
  !! relevant syntax rule for the parse node.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @returns 'type-spec'
  
  FUNCTION ts_What() RESULT(what)
      
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'type-spec'
    
  END FUNCTION ts_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of TypeSpec%Advertise - advertise the identifiers 
  !! referenced or defined by the specifier.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE ts_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(TypeSpec), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(1))
    
    ident_list(1) = IdentifierNature(  &
        NAME=.NameOf. pnode%type_name,  &
        NATURES=[NatureRefType],  &
        LOCATIONS=[.LocationOf. pnode%type_name] )
    
  END SUBROUTINE ts_Advertise
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for TypeParamSpec.
  
  
  !*****************************************************************************
  !!
  !> Implementation of TypeSpec%Write - dump a debugging representation of 
  !! the loop control to a logical unit.
  !!
  !! @param[in]     loop              The loop control.
  !!
  !! @param[in]     unit              The logical unit.
  
  SUBROUTINE tps_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(TypeParamSpec), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%name)) THEN
      WRITE (unit, "(A,'=')", ADVANCE='NO') .NameOf. pnode%name
    END IF
    
    CALL pnode%TypeParamValue%Write(unit)
    
  END SUBROUTINE tps_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of TypeParamValue%Query - traverse the node with a query 
  !! operation.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  ! A type spec could contain another type spec in one of the expressions 
  ! for a parameter value.
  RECURSIVE SUBROUTINE tps_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(TypeParamSpec), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%TypeParamValue%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE tps_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of TypeParamValue%Modify - traverse the node with a modify 
  !! operation.
  !!
  !! @param[in]     pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  ! A type spec could contain another type spec in one of the expressions 
  ! for a parameter value.
  RECURSIVE SUBROUTINE tps_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(TypeParamSpec), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%TypeParamValue%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE tps_Modify
  
  
  !*****************************************************************************
  !!
  !> Get the location of a type param spec.
  !!
  !! @param[in]     pnode             The type param value to query.
  !!
  !! @returns The location in source of the start of the value.
  !!
  !! Interface to match that of ParseNode, if we should ever go that way.
  
  FUNCTION tps_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(TypeParamSpec), INTENT(IN) :: pnode
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%name)) THEN
      loc = .LocationOf. pnode%name
    ELSE
      loc = pnode%TypeParamValue%GetLocation()
    END IF
    
  END FUNCTION tps_GetLocation
  
  
  !*****************************************************************************
  !!
  !> Implementation of TypeParamValue%Advertise - advertise the identifiers 
  !! referenced or defined by the specifier.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  !!
  !! If the type param value involves an expression, then the Advertise call 
  !! is forwarded to that expression.
  
  ! Recursive because a TypeParamValue may contain another TypeParamValue via 
  ! a structure constructor in a parameter expression or similar.
  RECURSIVE SUBROUTINE tps_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(TypeParamSpec), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%name)) THEN
      ALLOCATE(ident_list(1))
      ident_list(1) = IdentifierNature(  &
          NAME=.NameOf. pnode%name,  &
          NATURES=[NatureRefTypeParameter],  &
          LOCATIONS=[.LocationOf. pnode%name] )
    ELSE
      ALLOCATE(ident_list(0))
    END IF
    
  END SUBROUTINE tps_Advertise
  
  
  !*****************************************************************************
  !!
  !> Implementation of TypeParamSpec%What - get a text description of the 
  !! relevant syntax rule for the parse node.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @returns 'type-param-spec'
  
  FUNCTION tps_What() RESULT(what)
      
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'type-param-spec'
    
  END FUNCTION tps_What
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for TypeParamValue.
  
  
  !*****************************************************************************
  !!
  !> Implementation of TypeSpec%Write - dump a debugging representation of 
  !! the loop control to a logical unit.
  !!
  !! @param[in]     loop              The loop control.
  !!
  !! @param[in]     unit              The logical unit.
  
  SUBROUTINE tpv_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(TypeParamValue), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    IF (pnode%is_colon) WRITE (unit, "(':')", ADVANCE='NO')
    IF (pnode%is_star) WRITE (unit, "('*')", ADVANCE='NO')
    IF (ALLOCATED(pnode%int_expr)) CALL pnode%int_expr%Write(unit)
    
  END SUBROUTINE tpv_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of TypeParamValue%Query - traverse the node with a query 
  !! operation.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  ! A type spec could contain another type spec in one of the expressions 
  ! for a parameter value.
  RECURSIVE SUBROUTINE tpv_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(TypeParamValue), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%int_expr)) THEN
      CALL pnode%int_expr%Query(visitor)
    END IF
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE tpv_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of TypeParamValue%Modify - traverse the node with a modify 
  !! operation.
  !!
  !! @param[in]     pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  ! A type spec could contain another type spec in one of the expressions 
  ! for a parameter value.
  RECURSIVE SUBROUTINE tpv_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(TypeParamValue), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%int_expr)) THEN
      CALL pnode%int_expr%Modify(visitor)
    END IF
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE tpv_Modify
  
  
  !*****************************************************************************
  !!
  !> Get the location of a type param value.
  !!
  !! @param[in]      pnode             The type param value to query.
  !!
  !! @returns The location in source of the start of the value.
  !!
  !! Interface to match that of ParseNode, if we should ever go that way.
  
  FUNCTION tpv_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(TypeParamValue), INTENT(IN) :: pnode
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%my_location)) THEN
      loc = pnode%my_location
    ELSE IF (ALLOCATED(pnode%int_expr)) THEN
      loc = pnode%int_expr%GetLocation()
    ELSE
      ! Shouldn't happen.
      loc = SourceLocationIfortBug()
    END IF
    
  END FUNCTION tpv_GetLocation
  
  
  !*****************************************************************************
  !!
  !> Implementation of TypeParamValue%Advertise - advertise the identifiers 
  !! referenced or defined by the specifier.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  !!
  !! If the type param value involves an expression, then the Advertise call 
  !! is forwarded to that expression.
  
  ! Recursive because a TypeParamValue may contain another TypeParamValue via 
  ! a structure constructor in a parameter expression or similar.
  RECURSIVE SUBROUTINE tpv_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(TypeParamValue), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(0))
    
  END SUBROUTINE tpv_Advertise
  
  
  !*****************************************************************************
  !!
  !> Implementation of TypeParamValue%What - get a text description of the 
  !! relevant syntax rule for the parse node.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @returns 'type-param-value'
  
  FUNCTION tpv_What() RESULT(what)
      
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'type-param-value'
    
  END FUNCTION tpv_What
  
END MODULE DeclarationTypeSpecs
