! $Id: DerivedTypeStmts.f90 2793 2019-03-20 08:07:59Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the DerivedTypeStmts module.


!*******************************************************************************
!!
!> Procedures and types for parsing /derived-type-def-stmt/'s, 
!! /sequence-stmt/'s and /private-components-stmt/'s.

MODULE DerivedTypeStmts
  
  USE Identifiers
  USE ParseNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseDerivedTypeStmt
  PUBLIC :: ParseEndTypeStmt
  PUBLIC :: ParseSequenceStmt
  PUBLIC :: ParsePrivateComponentsStmt
  PUBLIC :: ParseBindingPrivateStmt
  
  PUBLIC :: FindParam
  
  !-----------------------------------------------------------------------------
  
  !> Stores the results of parsing a /derived-type-def-stmt/.
  TYPE, EXTENDS(SCParseNode), PUBLIC ::  DerivedTypeStmt
    !> The name of the type being defined.
    TYPE(Identifier) :: name
    !> .TRUE. if the type is an abstract type.
    LOGICAL :: abstract = .FALSE.
    !> The name of the parent type if this is an extension, or not allocated 
    !! otherwise.
    TYPE(Identifier), ALLOCATABLE :: extends
    !> An access spec code - see constants defined in the 
    !! AccessSpecifications module.
    INTEGER :: access = 0
    !> The binding language for the type, or zero if no binding language 
    !! specified.  See the LanguageBinding module for more information.
    !!
    !! The only valid binding language is C.
    INTEGER :: lang = 0
    !> The type parameter names.
    !!
    !! Always allocated, but maybe zero length.
    TYPE(Identifier), ALLOCATABLE :: params(:)
  CONTAINS
    PROCEDURE :: Dump => dt_Dump
    PROCEDURE :: Query => dt_Query
    PROCEDURE :: Modify => dt_Modify
    PROCEDURE, NOPASS :: What => dt_What
    PROCEDURE :: Advertise => dt_Advertise
  END TYPE DerivedTypeStmt
  
  !> Get the index of a type parameter.
  INTERFACE FindParam
    MODULE PROCEDURE FindParam_
  END INTERFACE FindParam
  
CONTAINS
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for DerivedTypeStmt.

  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is supposedly a /derived-type-stmt/ 
  !! (R426).
  !!
  !! A /derived-type-stmt/ is defined as:
  !!
  !! TYPE [ [ , /type-attr-spec-list/ ] :: ] /type-name/ 
  !!     [ ( type-param-name-list ) ]
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    dt                Object with the results of processing 
  !! the /derived-type-stmt/.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseDerivedTypeStmt(tlist, dt, err_list)
    
    USE AccessStmts
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE LanguageBindingSpecs
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(DerivedTypeStmt), INTENT(OUT) :: dt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /derived-type-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R426'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    TYPE(Error) :: err
    
    INTEGER :: it
    INTEGER :: it_end
    INTEGER :: comma
    INTEGER :: declare
    INTEGER :: finish
    
    TYPE(LangSpec) :: ls      ! Language specification temporary
    INTEGER :: opt            ! Expect option.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
          MSG='Expected /derived-type-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------    
    ! TYPE ...
    
    CALL ExpectKeyword(tlist, 1, scck_'TYPE', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! [ [ , type-attr-spec-list ] :: ]
    
    IF (Match(tlist, 2, ittComma)) THEN
      
      ! The list of attributes finishes at a double colon.
      declare = FindBare(tlist, ittDeclare, 2)
      IF (declare == 0) THEN
        ! We'll pick up an error for the missing :: later.
        finish = SIZE(tlist)
      ELSE
        finish = declare - 1
      END IF
      
      ! List of items
      it = 3
      DO
        comma = FindBare(tlist, ittComma, scck_',', it)
        IF ((comma == 0) .OR. (comma > finish)) THEN
          it_end = finish
          comma = 0
        ELSE
          it_end = comma - 1
        END IF
        
        IF (Match(tlist, it, ittName, scck_'PUBLIC')) THEN
          
          CALL do_access(iaccPublic)
          
        ELSE IF (Match(tlist, it, ittName, scck_'PRIVATE')) THEN
          
          CALL do_access(iaccPrivate)
          
        ELSE IF (Match(tlist, it, ittName, scck_'BIND')) THEN
          
          IF (dt%lang /= 0) THEN
            CALL Add( err_list, CODE=errSyntax,  &
                COMPONENT='C427 via ' // comp,  &
!                LOCATION=tlist(it)%position,  &
                MSG='The BIND attribute has already been specified for this &
                  &/derived-type-stmt/.' )
          END IF
          CALL Parse(tlist, it, it_end, ls, sub_err_list)
          CALL Add(err_list, sub_err_list, COMPONENT=comp)
          IF (.NOT. Failure(sub_err_list)) THEN
            IF (ALLOCATED(ls%name)) THEN
              CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
!                 LOCATION=tlist(it+3)%position,  &
                  MSG='The "NAME=" specifier is not permitted in a BIND &
                    &clause in a /derived-type-stmt/.' )
            END IF
            IF (dt%lang == 0) dt%lang = ls%lang
          END IF
          
          IF (dt%abstract) THEN
            CALL Add( err_list, CODE=errNotExtensible, &
                COMPONENT='C430',  & ! LOCATION=tlist(it)%position,  &
                MSG='If "ABSTRACT" appears the type must be extensible &
                  &(it cannot have the "BIND" attribute).' )
          END IF
          
        ELSE IF (Match(tlist, it, ittName, scck_'EXTENDS')) THEN
          
          IF (ALLOCATED(dt%extends)) THEN
            CALL Add( err_list,  &
                CODE=errSyntax,  &
                COMPONENT='C427 via ' // comp,  &
                LOCATION=.LocationOf. dt%extends,  &
                MSG='EXTENDS has already been specified for this &
                  &/derived-type-stmt/.' )
          END IF
          
          CALL Expect(tlist, it + 1, ittDelimiter, scck_'(', err)
          CALL Add(err_list, err, COMPONENT=comp)
          IF (.NOT. Failure(err)) THEN
            CALL Expect(tlist, it + 2, ittName, err)
            CALL Add(err_list, err, COMPONENT=comp)
            IF (.NOT. Failure(err)) THEN
              CALL Expect(tlist, it + 3, ittDelimiter, scck_')', err)
              CALL Add(err_list, err, COMPONENT=comp)
              IF (.NOT. Failure(err)) THEN
                IF (.NOT. ALLOCATED(dt%extends))  &
                    dt%extends = IdentifierIfortBug(tlist(it+2))
                CALL ExpectEnd(tlist(:it_end), it + 4, comp, err_list)
              END IF
            END IF
          END IF
          
        ELSE IF (Match(tlist, it, ittName, scck_'ABSTRACT')) THEN
          
          IF (dt%abstract) THEN
            CALL Add( err_list, CODE=errSyntax,  &
                COMPONENT='C427 via ' // comp,  &
!                LOCATION=tlist(it)%position,  &
                MSG='ABSTRACT has already been specified for this &
                  &/derived-type-stmt/.' )
          ELSE
            dt%abstract = .TRUE.
          END IF
          
          IF (dt%lang /= 0) THEN
            CALL Add( err_list, CODE=errNotExtensible, &
                COMPONENT='C430',  & ! LOCATION=tlist(it)%position,  &
                MSG='If "ABSTRACT" appears the type must be extensible &
                  &(it cannot have the "BIND" attribute).' )
          END IF
          
          CALL ExpectEnd(tlist(:it_end), it+1, comp, err_list)
          
        ELSE
        
          IF (it <= SIZE(tlist)) THEN
            CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
    !            LOCATION=tlist(2)%position,  &
                MSG='Expected /type-attr-spec/ but "'  &
                  // QueryRaw(tlist(it)) // '" was encountered instead.' )
          ELSE
            CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
    !            LOCATION=tlist(2)%position,  &
                MSG='Expected /type-attr-spec/ but the end of the statement &
                  & was encountered instead.' )
          END IF
          
        END IF
        
        IF (comma == 0) EXIT
        
        ! Skip the comma
        it = comma + 1
        
      END DO
      
      it = finish + 1
      
      CALL Expect(tlist, it, ittDeclare, err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      it = it + 1
      
      CALL ExpectName(tlist, it, '/type-name/', err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
    ELSE IF (Match(tlist, 2, ittDeclare)) THEN
      it = 3
      CALL ExpectName(tlist, it, '/type-name/', err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
    ELSE IF (Match(tlist, 2, ittName)) THEN
      it = 2
    ELSE
      IF (SIZE(tlist) >= 2) THEN
        CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
!            LOCATION=tlist(2)%position,  &
            MSG='Expected /name/, "," or "::" after "TYPE" but "'  &
              // QueryRaw(tlist(2)) // '" was encountered instead.' )
      ELSE
        CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
!            LOCATION=tlist(2)%position,  &
            MSG='Expected /name/, "," or "::" after "TYPE" but the end of the &
              &statement was encountered instead.' )
      END IF
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! Copy across the name.
    
    ! All paths above have validated that tlist(it) will be a name token.
    
    ! The name cannot be the same as an intrinsic type name or DOUBLEPRECISION.
    IF ( (tlist(it) == scck_'DOUBLEPRECISION')  &
        .OR. (tlist(it) == scck_'INTEGER')  &
        .OR. (tlist(it) == scck_'REAL')  &
        .OR. (tlist(it) == scck_'LOGICAL')  &
        .OR. (tlist(it) == scck_'CHARACTER')  &
        .OR. (tlist(it) == scck_'COMPLEX') ) THEN
      CALL Add( err_list, CODE=errInvalidTypeName,  &
          COMPONENT='C426 via ' // comp,  & ! LOCATION=tlist(it)%position,  &
          MSG='The name of the type cannot be DOUBLEPRECISION or the same as &
            &the name of an intrinsic type.' )
    END IF
    dt%name = IdentifierIfortBug(tlist(it))
    it = it + 1
    
    !---------------------------------------------------------------------------
    ! Go looking for type parameters.
    
    CALL Expect(tlist, it, [ittEndStatement, ittDelimiter],  &
        [scck_' ', scck_'('],  opt, err ) 
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
        
    IF (opt == 1) THEN
      ! End of statement - no type parameters.
      ALLOCATE(dt%params(0))
      RETURN
    END IF
    
    ! Find the end of the /type-param-name-list/.
    CALL ExpectClosing(tlist, it, finish, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
        
    CALL ParseNameList( tlist, it + 1, finish - 1, '/type-param-name/',  &
        dt%params, sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    it = finish + 1
    
    CALL ExpectEnd(tlist, it, comp, err_list)
    
  CONTAINS
    
    !***************************************************************************
    
    SUBROUTINE do_access(iacc)
      
      !-------------------------------------------------------------------------
      ! Arguments
      
      INTEGER, INTENT(IN) :: iacc
      
      !*************************************************************************
      
      IF (dt%access /= 0) THEN
        CALL Add( err_list, CODE=errSyntax,  &
            COMPONENT='C427 via ' // comp,  &
!            LOCATION=tlist(it)%position,  &
            MSG='The PUBLIC or PRIVATE attribute has already been &
              &specified for this type.' )
      ELSE
        dt%access = iacc
      END IF
      
      CALL ExpectEnd(tlist(:it_end), it + 1, comp, err_list)
      
    END SUBROUTINE do_access
    
  END SUBROUTINE ParseDerivedTypeStmt
  
  
  !*****************************************************************************
  !!
  !> Find the index of a type parameter designated by name.
  !!
  !! @param[in]     dt_stmt           The results of parsing a derived type 
  !! statement.
  !!
  !! @param[in]     tp_name           The name of the type parameter of 
  !! interest.
  !!
  !! @returns The index of the type parameter in the parameters listed in 
  !! the statement (i.e. in the type parameter order), or zero if no type 
  !! parameter was found.
  
  FUNCTION FindParam_(dt_stmt, tp_name) RESULT(idx)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(DerivedTypeStmt), INTENT(IN) :: dt_stmt
    CHARACTER(*), INTENT(IN) :: tp_name
    
    ! Function result
    INTEGER :: idx
    
    !***************************************************************************
    
    DO idx = 1, SIZE(dt_stmt%params)
      IF (.NameOf. dt_stmt%params(idx) == tp_name) RETURN
    END DO
    
    idx = 0
    
  END FUNCTION FindParam_
  
  
  !*****************************************************************************
  !!
  !> Implementation of DerivedTypeStmt%Dump - write a debugging representation 
  !! of the results of parsing a Derived type statement to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE dt_Dump(pnode, unit, pfx)
    
    USE AccessStmts
    USE Expressions
    USE TypeBoundProcedureStmts
    USE TypeBoundGenericStmts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DerivedTypeStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Type parameter index.
    
    !***************************************************************************
    
    WRITE (unit, "(A,'abstract = ',L1,' access = ',I0,' lang = ',I0)") pfx,  &
        pnode%abstract, pnode%access, pnode%lang
    IF (ALLOCATED(pnode%extends))  &
      WRITE (unit, "(A,'extends = ',A)") pfx, .NameOf. pnode%extends
    WRITE (unit, "(A,'name = ',A)") pfx, .NameOf. pnode%name
    IF (SIZE(pnode%params) /= 0) THEN
      WRITE (unit, "(A,'params:')") pfx
      DO i = 1, SIZE(pnode%params)
        WRITE (unit, "(A,A)") pfx // '  ', .NameOf. pnode%params(i)
      END DO
    END IF
    
  END SUBROUTINE dt_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of DerivedTypeStmt%Query - traverse the node with a 
  !! query operation.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  RECURSIVE SUBROUTINE dt_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(DerivedTypeStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE dt_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of DerivedTypeStmt%Modify - traverse the node with a 
  !! modify operation.
  !!
  !! @param[in,out] pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  RECURSIVE SUBROUTINE dt_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(DerivedTypeStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE dt_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of DerivedTypeStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'derived-type-stmt'.
  
  FUNCTION dt_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'derived-type-stmt'
    
  END FUNCTION dt_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of DerivedTypeStmt%Advertise - advertise the identifiers 
  !! referenced or defined by the block.
  !!
  !! @param[in]     pnode             The node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE dt_Advertise(pnode, ident_list)
    
    USE AccessStmts
    USE IdentifierNatures
    USE SourceLocations
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DerivedTypeStmt), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! Type parameter index.
    INTEGER :: ip
    
    !***************************************************************************
    
    ALLOCATE(ident_list(0))
    
    ! Establish a reference to the type being defined.
    CALL MergeNatures(  &
        ident_list,  &
        IdentifierNature(  &
          NAME=.NameOf. pnode%name,  &
          NATURES=[NatureRefType],  &
          LOCATIONS=[.LocationOf. pnode%name] ) )
    
    ! Parent type, if present.
    IF (ALLOCATED(pnode%extends)) THEN
      CALL MergeNatures(  &
          ident_list,  &
          IdentifierNature(  &
            NAME=.NameOf. pnode%extends,  &
            NATURES=[NatureRefType],  &
            LOCATIONS=[.LocationOf. pnode%extends] ) )
    END IF
    
    DO ip = 1, SIZE(pnode%params)
      CALL MergeNatures(  &
          ident_list,  &
          IdentifierNature(  &
              NAME=.NameOf. pnode%params(ip),  &
              NATURES=[NatureDefTypeParameter],  &
              LOCATIONS=[.LocationOf. pnode%params(ip)] ) )
    END DO
    
  END SUBROUTINE dt_Advertise
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for handling /sequence-stmt/'s.
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that make up a /sequence-stmt/.
  !!
  !! @param[in]     tlist             The list of tokens.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseSequenceStmt(tlist, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /sequence-stmt/
    CHARACTER(*), PARAMETER :: comp = 'R430'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
          MSG='Expected /sequence-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    CALL ExpectKeyword(tlist, 1, scck_'SEQUENCE', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    
    CALL ExpectEOS(tlist, 2, comp, err_list)
    
  END SUBROUTINE ParseSequenceStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that make up a /private-components-stmt/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParsePrivateComponentsStmt(tlist, err_list)
    
    USE Errors
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !***************************************************************************
    
    CALL parse_private_stmt( tlist, 'R444', '/private-components-stmt/',  &
        err_list )
    
  END SUBROUTINE ParsePrivateComponentsStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /binding-private-stmt/.
  !!
  !! @param[in]     tlist             The sequence of tokens that make up the 
  !! statement.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseBindingPrivateStmt(tlist, err_list)
    
    USE Errors
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !***************************************************************************
    
    CALL parse_private_stmt(tlist, 'R446', '/binding-private-stmt/', err_list)
    
  END SUBROUTINE ParseBindingPrivateStmt
  
  
  !*****************************************************************************
  !!
  !> Worker routine for ParseBindingPrivateStmt_ and 
  !! ParsePrivateComponentsStmt_.
  !!
  !! @param[in]     tlist             The sequence of tokens that make up the 
  !! statement.
  !!
  !! @param[in]     comp              Component (syntax rule number) for 
  !! error reporting.
  !!
  !! @param[in]     syntax_name       Name of the syntax rule for the 
  !! statement for error reporting.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE parse_private_stmt(tlist, comp, syntax_name, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    CHARACTER(*), INTENT(IN) :: comp
    CHARACTER(*), INTENT(IN) :: syntax_name
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add(err_list, CODE=errSyntax, COMPONENT=comp,  &
          MSG='Expected ' // syntax_name // '.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    CALL ExpectKeyword(tlist, 1, scck_'PRIVATE', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    
    CALL ExpectEOS(tlist, 2, comp, err_list)
    
  END SUBROUTINE parse_private_stmt
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! /end-type-stmt/ processing
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that supposedly represent an /end-type-stmt/ 
  !! (R429).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     type_name         The name given to the derived type in 
  !! the opening /derived-type-stmt/.
  !!
  !! @param[out]    err_list          List of errors.
    
  SUBROUTINE ParseEndTypeStmt(tlist, type_name, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    CHARACTER(*,KIND=scck), INTENT(IN) :: type_name
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for an /end-type-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R429'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object from Expect etc.
    INTEGER :: after          ! Position after the END TYPE keywords. 
    
    TYPE(MatchProcList) :: procs(2)
    INTEGER :: opt            ! Expect option.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
          MSG='Expected /end-type-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! Either END TYPE or ENDTYPE
    
    CALL ExpectSplittableKeyword(tlist, 1, scck_'END', scck_'TYPE', after, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! Type name is optional
    
    procs(1)%item => MatchEOS
    procs(2)%item => MatchName
    CALL Expect( tlist, after, procs, 'end-of-statement or /type-name/',  &
        opt, err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! End of statement - nothing more to do.
    IF (opt == 1) RETURN
    
    !---------------------------------------------------------------------------
    ! Type name needs to match.
    
    IF (tlist(after) /= type_name) THEN
      CALL Add( err_list, CODE=errNameMismatch,  &
          COMPONENT='C435 via ' // comp,  & 
!          LOCATION=tlist(after)%position,  &
          MSG='The /type-name/ in an /end-type-stmt/ ("'  &
            // QueryRaw(tlist(after)) // '") shall match the &
            &/type-name/ in the corresponding /derived-type-stmt/ ("'  &
            // type_name // '").' )
    END IF
    
    !---------------------------------------------------------------------------
    
    CALL ExpectEOS(tlist, after+1, comp, err_list)
    
  END SUBROUTINE ParseEndTypeStmt
  
END MODULE DerivedTypeStmts
