! $Id: DoConstructs.f90 2363 2016-08-22 01:29:48Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the BlockConstructs module.


!*******************************************************************************
!!
!> Procedures and types for parsing /do-construct/'s.
!!
!! This includes both /block-do-construct/'s and /nonblock-do-construct/'s.

MODULE DoConstructs
  
  USE DoStmts
  USE ExecutionParts
  USE Identifiers
  USE Labels
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseDoConstruct
  
  !-----------------------------------------------------------------------------
  
  !> The results of parsing a do-construct.
  TYPE, EXTENDS(BlockStorage), PUBLIC :: DoConstruct
    !> The construct name, if present.
    TYPE(Identifier), ALLOCATABLE :: construct_name
    !> The label for a construct that starts with a labelled do statement.
    !!
    !! Not allocated if the do statement is not a labelled do statement.
    !!
    !! @todo Check whether we actually need to store this here - by 
    !! the time the statement tree has been built this is probably no 
    !! longer required.
    TYPE(Label), ALLOCATABLE :: do_label
    !> The loop control in the statement, if present, or not allocated if the 
    !! statement had no loop control.
    CLASS(LoopControl), ALLOCATABLE :: loop_control
    !> The label on the end do statement, if present.
    !!
    !! Not allocated if there was no label on the end do statement.
    !!
    !! If a labelled do construct was terminated by a label on a continue 
    !! statement, then that statement and its label are stored in the 
    !! child executable statements of the construct.  The following 
    !! label is specifically for END DO statements, which are 
    !! not stored as executable children.
    TYPE(Label), ALLOCATABLE :: end_label
  CONTAINS
    PROCEDURE :: Dump => doc_Dump
    PROCEDURE, NOPASS :: What => doc_What
    PROCEDURE :: Query => doc_Query
    PROCEDURE :: Modify => doc_Modify
    PROCEDURE :: Advertise => doc_Advertise
  END TYPE DoConstruct
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a /block-do-construct/ or /nonblock-do-construct/.
  !!
  !! @param[in]     st                The statement tree node for the 
  !! construct.
  !!
  !! @param[out]    if_construct      The results of parsing the 
  !! construct.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseDoConstruct(st, do_construct, err_list)
    
    USE DoConstructStNodes
    USE Errors
    USE ExecutionParts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(DoConstructStNode), INTENT(INOUT) :: st
    TYPE(DoConstruct), INTENT(OUT) :: do_construct
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    ! Opening statement.
    
    SELECT CASE (st%first%ist)
    CASE (istLabelDo)
      ALLOCATE(do_construct%do_label)
      CALL ParseLabelDoStmt(  &
          st%first%tlist,  &
          do_construct%loop_control,  &
          do_construct%construct_name,  &
          do_construct%do_label,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list)
      
    CASE (istNonlabelDo)
      CALL ParseNonlabelDoStmt(  &
          st%first%tlist,  &
          do_construct%loop_control,  &
          do_construct%construct_name,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list)
      
    CASE DEFAULT
      STOP 'Internal error in DoConstructs%%ParseDoConstruct, &
          &unknown opening statement in a do-construct.'
      
    END SELECT
    
    !---------------------------------------------------------------------------
    ! Child block.
    
    CALL ParseBlock(st%exec, do_construct%children, sub_err_list)
    CALL Add(err_list, sub_err_list)
    
    !---------------------------------------------------------------------------
    ! Closing statement.
    
    IF (ALLOCATED(st%last)) THEN
      ! The last statement in a do construct might not be an END DO, but in 
      ! that case the statement gets bundled into the rest of the executable 
      ! statements.
      CALL ParseEndDoStmt(  &
          st%last%tlist,  &
          do_construct%construct_name,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list)
      
      ! Save the statement label of the last statement, if any, in case 
      ! it is referenced by other statements (which is valid if those 
      ! statements are within the executable children of the do construct).
      IF (ALLOCATED(st%last%statement_label)) THEN
        do_construct%end_label = st%last%statement_label
      END IF
    END IF
    
  END SUBROUTINE ParseDoConstruct
  
  
  !*****************************************************************************
  !!
  !> Implementation of DoConstruct%Dump - write a debugging 
  !! representation of the ParseNode to a logical unit.
  !!
  !! @param[in]     pnode             The ParseNode to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix for each record.
  
  SUBROUTINE doc_Dump(pnode, unit, pfx)
    
    USE ExecutionParts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DoConstruct), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%do_label)) THEN
      WRITE (unit, "(A,'/label/ = ',I0)", ADVANCE='NO')  &
          pfx, .ValueOf. pnode%do_label
    END IF
    
    IF (ALLOCATED(pnode%loop_control)) THEN
      WRITE (unit, "(A)", ADVANCE='NO') pfx
      CALL pnode%loop_control%Write(unit)
    END IF
    WRITE (unit, "()")
    
    WRITE (unit, "(A,'child statements:')") pfx
    CALL DumpBlock(pnode%children, unit, pfx // '  ')
    
  END SUBROUTINE doc_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of DoConstruct%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'do-construct'.
  
  FUNCTION doc_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'do-construct'
    
  END FUNCTION doc_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of DoConstruct%Query - traverse the parse node 
  !! with a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  ! A do construct can contain another do construct - hence recursive.
  RECURSIVE SUBROUTINE doc_Query(pnode, visitor)
    
    USE ParseNodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DoConstruct), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%loop_control)) THEN
      CALL pnode%loop_control%Query(visitor)
    END IF
    
    CALL QueryBlock(pnode, visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE doc_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of DoConstruct%Modify - traverse the parse node 
  !! with a query operation.
  !!
  !! @param[in,out] pnode             The parse node element to modify.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  ! A do construct can contain another do construct - hence recursive.
  RECURSIVE SUBROUTINE doc_Modify(pnode, visitor)
    
    USE ParseNodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DoConstruct), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%loop_control)) THEN
      CALL pnode%loop_control%Modify(visitor)
    END IF
    
    CALL ModifyBlock(pnode, visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE doc_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of DoConstruct%Advertise - advertise the identifiers 
  !! referenced or defined by the construct.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE doc_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DoConstruct), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%construct_name)) THEN
      ALLOCATE(ident_list(1))
      ident_list(1) = IdentifierNature(  &
            NAME=.NameOf. pnode%construct_name,  &
            NATURES=[NatureDefConstruct], &
            LOCATIONS=[.LocationOf. pnode%construct_name] )
    ELSE
      ALLOCATE(ident_list(0))
    END IF
    
  END SUBROUTINE doc_Advertise
  
END MODULE DoConstructs