! $Id: EndProgramUnitStmts.f90 1847 2015-09-08 12:54:18Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the EndProgramUnitStmts module.


!*******************************************************************************
!!
!> Procedures and types for parsing the common arrangement for the end 
!! statement of program units and procedures.  
!!
!! This includes:
!! - /end-function-stmt/
!! - /end-subroutine-stmt/
!! - /end-module-stmt/
!! - /end-mp-subprogram-stmt/
!! - /end-submodule-stmt/
!!
!! It does not include /end-program-stmt/ parsing because that has slightly 
!! different behaviour due to the optional nature of the /program-stmt.  It 
!! does not include /end-block-data-stmt/ parsing as the BLOCKDATA keyword
!! can be two separate keywords.

MODULE EndProgramUnitStmts
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseEndProgramUnitStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /end-function-stmt/ (R1232).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     text              The token text for the end statement, 
  !! exluding the END keyword.
  !!
  !! @param[in]     comp              The syntax rule identifier for the 
  !! end-*-stmt.
  !!
  !! @param[in]     end_rule          The syntax rule name for the 
  !! end-*-stmt.
  !!
  !! @param[in]     name_rule         The syntax rule name for the name that 
  !! is optionally in the end-*-stmt.
  !!
  !! @param[out]    name              The name provided in the statement.  
  !! Not allocated if no name was provided.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseEndProgramUnitStmt( tlist, text, comp, end_rule,  &
      name_rule, name, err_list )
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE Identifiers
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    CHARACTER(*,KIND=scck), INTENT(IN) :: text
    CHARACTER(*), INTENT(IN) :: comp
    CHARACTER(*), INTENT(IN) :: end_rule
    CHARACTER(*), INTENT(IN) :: name_rule
    TYPE(Identifier), INTENT(OUT), ALLOCATABLE :: name
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
            
    TYPE(Error) :: err      ! Error object for Expect calls.
    INTEGER :: opt          ! Expect option.
    INTEGER :: it           ! Current token index.
    
    ! Keyword in default character
    CHARACTER(:), ALLOCATABLE :: def_text
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected ' // end_rule // '.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------    
    ! END [ TEXT [ name ] ] or ENDTEXT [ name ]
    
    def_text = text           ! Kind conversion.
    CALL ExpectKeyword(  &
        tlist,  &
        1,  &
        [scck_'END' // REPEAT(' ',LEN(text)), scck_'END' // text],  &
        '"END" or "END' // def_text //'"',  &
        opt,  &
        err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    IF (opt == 1) THEN
      ! TEXT could be next, or the statement could be an END on its own.
      CALL Expect( &
          tlist,  &
          2,  &
          [ittName, ittEndStatement],  &
          [text, REPEAT(' ', LEN(text))],  &
          opt,  &
          err )
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      IF (opt == 2) RETURN    ! end-of-statement - we are done.
      it = 3
    ELSE
      it = 2
    END IF
    ! Name is optional.
    IF (it > SIZE(tlist)) RETURN
    
    !---------------------------------------------------------------------------
    ! The name is optional
    
    CALL Expect(  &
        tlist,  &
        it,  &
        [ittName, ittEndStatement],  &
        [scck_'', scck_''],  &
        name_rule // ' or end-of-statement',  &
        opt,  &
        err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    IF (opt == 2) RETURN
    
    ALLOCATE(name)      ! ifort bug - should be automatically allocated.
    name = IdentifierIfortBug(tlist(it))
    
    CALL ExpectEOS(tlist, it + 1, comp, err_list)
    
  END SUBROUTINE ParseEndProgramUnitStmt
  
END MODULE EndProgramUnitStmts
