! $Id: EnumDefs.f90 2794 2019-03-21 10:28:48Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the BlockConstructs module.


!*******************************************************************************
!!
!> Procedures and types for parsing /enum-def/'s.

MODULE EnumDefs
  
  USE EnumDefStmts
  USE LanguageBindingSpecs
  USE ParseNodes
  
  IMPLICIT NONE
  
  !-----------------------------------------------------------------------------
  
  !> Stores the results of parsing an /enum-def/.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: EnumDef
    !> The language spec on the opening statement.
    TYPE(LangSpec) :: ls
    !> The list of /enumerator-def-stmt/'s.
    TYPE(EnumeratorDefStmt), ALLOCATABLE :: children(:)
  CONTAINS
    PROCEDURE :: Dump => ed_Dump
    PROCEDURE, NOPASS :: What => ed_What
    PROCEDURE :: Query => ed_Query
    PROCEDURE :: Modify => ed_Modify
  END TYPE EnumDef
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse the statement tree node for an /enum-def/.
  !!
  !! @param[in]     st                The statement tree node for the 
  !! /enum-def/.
  !!
  !! @param[out]    enum_def          The results of parsing the /enum-def/.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseEnumDef(st, enum_def, err_list)
    
    USE EnumDefStNodes
    USE Errors
    USE ErrorCodes
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(EnumDefStNode), INTENT(INOUT) :: st
    TYPE(EnumDef), INTENT(OUT) :: enum_def
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Error list for child parsing statements.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    INTEGER :: i              ! enumerator def index.
    INTEGER :: id             ! children index.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    ! The opening /enum-def-stmt/.
    
    CALL ParseEnumDefStmt(st%first%tlist, enum_def%ls, sub_err_list)
    CALL Add(err_list, sub_err_list)
    
    !---------------------------------------------------------------------------
    ! The child /enumerator-def-stmt/'s.
    
    ! Check that we have at least one statement.  Perhaps this is better 
    ! handled in tree building.
    IF (SIZE(st%defs) == 0) THEN
      ! Place the error against the closing statement.
      CALL Add( err_list, CODE=errSyntax, COMPONENT='R458',  &
          LOCATION=QueryLocation(st%last%tlist(1)),  &
          MSG='At least one /enumerator-def-stmt/ is required inside a &
            &/enum-def/.' )
    END IF
    
    ALLOCATE(enum_def%children(SIZE(st%defs)))
    id = 1
    DO i = 1, SIZE(st%defs)
      CALL ParseEnumeratorDefStmt( st%defs(i)%tlist, enum_def%children(id),  &
          sub_err_list )
      CALL Add(err_list, sub_err_list)
      IF (.NOT. Failure(sub_err_list)) THEN
        id = id + 1
      END IF
    END DO
    ! If there have been errors, then chop the resulting array back to size.
    IF (id /= SIZE(st%defs) + 1) enum_def%children = enum_def%children(:id-1)
    
    !---------------------------------------------------------------------------
    ! The closing /end-enum-stmt/
    
    CALL ParseEndEnumStmt(st%last%tlist, sub_err_list)
    CALL Add(err_list, sub_err_list)
    
  END SUBROUTINE ParseEnumDef
  
  
  !*****************************************************************************
  !!
  !> Implementation of EnumDef%Dump - write a debugging 
  !! representation of the results of parsing a /enum-def/ to a 
  !! logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE ed_Dump(pnode, unit, pfx)
    
    USE LanguageBindingSpecs
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(EnumDef), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: id             ! Child /enumerator-def-stmt/ index.
    
    !***************************************************************************
    
    WRITE (unit, "(A,'language spec = ')", ADVANCE='NO') pfx
    CALL WriteLangSpec(unit, pnode%ls)
    WRITE (unit, "()")
    
    WRITE (unit, "(A,'child /enum-def-stmt/s:')") pfx
    
    DO id = 1, SIZE(pnode%children)
      CALL pnode%children(id)%Dump(unit, pfx // '  ')
    END DO
    
  END SUBROUTINE ed_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of EnumDef%What - return a descriptive string 
  !! for the parse node.
  !!
  !! @returns 'enum-def'.
  
  FUNCTION ed_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'enum-def'
    
  END FUNCTION ed_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of EnumDef%Query.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  RECURSIVE SUBROUTINE ed_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(EnumDef), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%ls%Query(visitor)
    
    ! Child statements aren't currently parse nodes.  Perhaps they should be.
!    DO i = 1, SIZE(pnode%children)
!      CALL pnode%children(i)%Query(visitor, order)
!    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE ed_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of EnumDef%Modify.
  !!
  !! @param[in,out] pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  RECURSIVE SUBROUTINE ed_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(EnumDef), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%ls%Modify(visitor)
    
    ! Child statements aren't currently parse nodes.  Perhaps they should be.
!    DO i = 1, SIZE(pnode%children)
!      CALL pnode%children(i)%Modify(visitor, order)
!    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE ed_Modify
  
END MODULE EnumDefs
