! $Id: FunctionStmts.f90 2929 2019-04-27 22:38:46Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the FunctionStmts module.


!*******************************************************************************
!!
!> Procedures and types for parsing /function-stmt/'s and 
!! /end-function-stmt/'s.
!!
!! The /subroutine-stmt/ parsing code borrows some of the prefix parsing 
!! stuff.
!!
!! The /entry-stmt/ parsing code borrows some of the suffix parsing stuff.

MODULE FunctionStmts
  
  USE DeclarationTypeSpecs
  USE Identifiers
  USE LanguageBindingSpecs
  USE ParseNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseOtherPrefixSpec
  PUBLIC :: DumpOtherPrefixSpec
  
  PUBLIC :: ParseSuffix
  
  PUBLIC :: ParseFunctionStmt
  PUBLIC :: ParseEndFunctionStmt
  
  PUBLIC :: MatchOtherPrefixSpec
  
  !-----------------------------------------------------------------------------
  
  !> Attributes (prefixes) specified in a /function-stmt/ or a 
  !! /subroutine-stmt/, excluding the /declaration-type-spec/ that 
  !! may be in the prefixes for a /function-stmt/.
  TYPE, PUBLIC :: OtherPrefixSpecs
    !> Elemental prefix specified.
    LOGICAL :: elemental = .FALSE.
    !> Function purity.
    !!
    !! - 0: Not specified
    !! - 1: IMPURE
    !! - 2: PURE
    INTEGER :: purity = 0
    !> Separate module procedure prefix.
    LOGICAL :: module = .FALSE.
    !> Recursive prefix.
    LOGICAL :: recursive = .FALSE.
  END TYPE OtherPrefixSpecs
  
  !> Results of parsing a /function-stmt/.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: FunctionStmt
    !> Function type.
    !!
    !! Not allocated if there was no such prefix on the statement.
    TYPE(DeclTypeSpec), ALLOCATABLE :: ds
    !> Other prefixes.
    TYPE(OtherPrefixSpecs) :: prefixes
    !> Function name.
    TYPE(Identifier) :: function_name
    !> Dummy argument names.
    !!
    !! Always allocated but may be zero length.
    TYPE(Identifier), ALLOCATABLE :: args(:)
    !> Function result name.  May not be allocated.
    TYPE(Identifier), ALLOCATABLE :: result_name
    !> Language binding spec.  May not be allocated.
    TYPE(LangSpec), ALLOCATABLE :: ls
  CONTAINS
    PROCEDURE :: Dump => fun_Dump
    PROCEDURE :: Query => fun_Query
    PROCEDURE :: Modify => fun_Modify
    PROCEDURE, NOPASS :: What => fun_What
    PROCEDURE :: Advertise => fun_Advertise
  END TYPE FunctionStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /function-stmt/ (R1228).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    fn_stmt           Object describing the result of the 
  !! parse.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseFunctionStmt(tlist, fn_stmt, err_list)
    
    USE Strings
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE LanguageBindingSpecs
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(FunctionStmt), INTENT(OUT) :: fn_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /function-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R1228'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Procedures for matching particular token sequences.
    TYPE(MatchProcList) :: procs(3)
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    INTEGER :: opt            ! Expect option result.
    INTEGER :: it             ! Current token index.
    INTEGER :: it_end         ! End of current clause.
    INTEGER :: closing        ! Index of closing parenthesis.
    
    !***************************************************************************
  
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /function-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    procs(1)%item => MatchDeclarationTypeSpecStart
    procs(2)%item => MatchOtherPrefixSpec
    procs(3)%item => match_function
    it = 1
    DO
      CALL Expect(tlist, it, procs, '/prefix/ or "FUNCTION"', opt, err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      SELECT CASE (opt)
      CASE (1)
        ! Length of the prefix
        it_end = GetDeclarationTypeSpecFinish(tlist(it:), .FALSE.)
        IF (it_end == 0) THEN
          it_end = SIZE(tlist)
        ELSE
          it_end = it + it_end - 1
        END IF
        
        ! Already specified?
        IF (ALLOCATED(fn_stmt%ds)) THEN
          CALL Add( err_list,  &
              CODE=errSyntax,  &
              COMPONENT=comp,  &
              LOCATION=QueryLocation(tlist(it)),  &
              MSG='A /prefix-spec/ that is a /declaration-type-spec/ has &
                &already been provided.' )
        ELSE
          ALLOCATE(fn_stmt%ds)
          CALL ParseDeclarationTypeSpec(  &
              tlist(it:it_end),  &
              .FALSE.,  &
              fn_stmt%ds,  &
              sub_err_list )
          CALL Add(err_list, sub_err_list, COMPONENT=comp)
          IF (Failure(sub_err_list)) RETURN
        END IF
        it = it_end + 1
        
      CASE (2)
        CALL ParseOtherPrefixSpec(  &
            tlist(it:it),  &
            fn_stmt%prefixes,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
        IF (Failure(sub_err_list)) RETURN
        it = it + 1
        
      CASE (3)
        it = it + 1
        EXIT
        
      END SELECT
    END DO
    
    !---------------------------------------------------------------------------
    ! The /function-name/
    
    CALL ExpectName(tlist, it, '/function-name/', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    fn_stmt%function_name = IdentifierIfortBug(tlist(it))
    it = it + 1
    
    !---------------------------------------------------------------------------
    ! Arguments in parentheses
    
    CALL Expect(tlist, it, ittDelimiter, scck_'(', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL ExpectClosing(tlist, it, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! The argument list is optional.
    IF (closing - 1 >= it + 1) THEN
      CALL ParseNameList(  &
          tlist,  &
          it + 1, closing - 1,  &
          '/dummy-arg-name/',  &
          fn_stmt%args,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
    ELSE
      ALLOCATE(fn_stmt%args(0))
    END IF
    it = closing + 1
    
    !---------------------------------------------------------------------------
    ! suffix
    
    CALL Expect(  &
        tlist,  &
        it,  &
        [ittEndStatement, ittName, ittName],  &
        [scck_'      ', scck_'RESULT', scck_'BIND  '],  &
        opt,  &
        err ) 
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    IF (opt == 1) RETURN
    
    CALL ParseSuffix(  &
        tlist,  &
        it,  &
        fn_stmt%result_name,  &
        fn_stmt%ls,  &
        it_end,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
    CALL ExpectEOS(tlist, it_end, comp, err_list)
    
  END SUBROUTINE ParseFunctionStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /end-function-stmt/ (R1232).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     name              The /function-name/ from the statement, 
  !! or not allocated if there was no name.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseEndFunctionStmt(tlist, name, err_list)
    
    USE CompilerKinds
    USE Identifiers
    USE EndProgramUnitStmts
    USE Errors
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(Identifier), INTENT(OUT), ALLOCATABLE :: name
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !***************************************************************************
    
    CALL ParseEndProgramUnitStmt(  &
        tlist,  &
        scck_'FUNCTION',  &
        'R1232',  &
        '/end-function-stmt/',  &
        '/function-name/',  &
        name,  &
        err_list )
    
  END SUBROUTINE ParseEndFunctionStmt
  
  
  !*****************************************************************************
  !!
  !> Match a /prefix-spec/ that isn't a /declaration-type-spec/ in a 
  !! sequence of tokens.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             The position from which to start the 
  !! match.
  !!
  !! @returns True if the token sequence in @a tlist beginning at @a start 
  !! looks like a /prefix-spec/, false otherwise.
  
  FUNCTION MatchOtherPrefixSpec(tlist, start) RESULT(b)
    
    USE CompilerKinds
    USE DeclarationTypeSpecs
    USE MatchUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    
    ! Function result
    LOGICAL :: b
    
    !***************************************************************************
    
    b = Match(tlist, start, ittName, scck_'ELEMENTAL')  &
        .OR. Match(tlist, start, ittName, scck_'IMPURE')  &
        .OR. Match(tlist, start, ittName, scck_'MODULE')  &
        .OR. Match(tlist, start, ittName, scck_'PURE')  &
        .OR. Match(tlist, start, ittName, scck_'RECURSIVE')
    
  END FUNCTION MatchOtherPrefixSpec
  
  
  !*****************************************************************************
  !!
  !> Parse a /prefix-spec/ other than a /declaration-type-spec/, i.e. the 
  !! prefixes that can be shared between function statements and subroutine 
  !! statements.
  !!
  !! (This is not a standard syntax rule in its own right - the prohibition 
  !! on having a declaration type spec in a subroutine statement is a 
  !! constraint.)
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    attrs             The results of parsing the particular 
  !! /prefix-spec/.  Note this is an INTENT(INOUT) argument - each 
  !! /prefix-spec/ parsed modifies the structure.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! This procedure is also called by the subroutine statement parsing 
  !! procedures.
  
  SUBROUTINE ParseOtherPrefixSpec(tlist, attrs, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(OtherPrefixSpecs), INTENT(INOUT) :: attrs
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /prefix-spec/.
    CHARACTER(*), PARAMETER :: comp = 'R1226'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    ! Failure not tested in StmtParsingTests because of prefix 
    ! matching logic.
    CALL Expect(tlist, 1, '/prefix-spec/', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    
    SELECT CASE (QueryValue(tlist(1)))
    CASE ('ELEMENTAL')
      IF (attrs%elemental) CALL already('ELEMENTAL')
      IF (attrs%recursive) THEN
        CALL Add( err_list,  &
            CODE=errSyntax,  &
            COMPONENT='C1245 via ' // comp,  &
            LOCATION=QueryLocation(tlist(1)),  &
            MSG='A prefix shall not specify both RECURSIVE and ELEMENTAL.' )
      END IF
      attrs%elemental = .TRUE.
      
    CASE (scck_'IMPURE')
      IF (attrs%purity == 1) THEN
        CALL already('IMPURE')
      ELSE IF (attrs%purity == 2) THEN
        CALL Add( err_list,  &
            CODE=errSyntax,  &
            COMPONENT='C1244 via ' // comp,  &
            LOCATION=QueryLocation(tlist(1)),  &
            MSG='A prefix shall not specify both PURE and IMPURE.' )
      ELSE
        attrs%purity = 1
      END IF
      
    CASE (scck_'MODULE')
      IF (attrs%module) CALL already('MODULE')
      attrs%module = .TRUE.
      
    CASE (scck_'PURE')
      IF (attrs%purity == 2) THEN
        CALL already('PURE')
      ELSE IF (attrs%purity == 1) THEN
        CALL Add( err_list,  &
            CODE=errSyntax,  &
            COMPONENT='C1244 via ' // comp,  &
            LOCATION=QueryLocation(tlist(1)),  &
            MSG='A prefix shall not specify both PURE and IMPURE.' )
      ELSE
        attrs%purity = 2
      END IF
      
    CASE (scck_'RECURSIVE')
      IF (attrs%recursive) CALL already('RECURSIVE')
      IF (attrs%elemental) THEN
        CALL Add( err_list,  &
            CODE=errSyntax,  &
            COMPONENT='C1245 via ' // comp,  &
            LOCATION=QueryLocation(tlist(1)),  &
            MSG='A prefix shall not specify both RECURSIVE and ELEMENTAL.' )
      END IF
      attrs%recursive = .TRUE.
      
    CASE DEFAULT
      ! Not tested in StmtParsingTests because of prefix matching 
      ! logic.  Almost an internal error, but here in case we change 
      ! prefix matching logic.
      CALL Add( err_list,  &
            CODE=errSyntax,  &
            COMPONENT=comp,  &
            LOCATION=QueryLocation(tlist(1)),  &
            MSG='"' // QueryRaw(tlist(1)) // '" is not a valid /prefix-spec/.' )
      
    END SELECT
    
  CONTAINS
    
    SUBROUTINE already(text)
      CHARACTER(*), INTENT(IN) :: text
      !-------------------------------------------------------------------------
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT='C1243 via ' // comp,  &
          LOCATION=QueryLocation(tlist(1)),  &
          MSG='The ' // text // ' prefix has already been specified.' )
    END SUBROUTINE already
    
  END SUBROUTINE ParseOtherPrefixSpec
  
  
  !*****************************************************************************
  !!
  !> Match the FUNCTION keyword.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             First token in the sequence from 
  !! which to start the match.
  !!
  !! @returns True if the given token is the FUNCTION keyword, false 
  !! otherwise.
  
  FUNCTION match_function(tlist, start) RESULt(b)
    
    USE CompilerKinds
    USE MatchUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    
    ! Function result
    LOGICAL :: b
    
    !***************************************************************************
    
    b = Match(tlist, start, ittName, scck_'FUNCTION')
    
  END FUNCTION match_function
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that makes up a /suffix/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             The first token in tlist that makes up
  !!the suffux
  !!
  !! @param[out]    result_name       The name of the result.  Not allocated
  !! if the suffix did not contain a RESULT clause.
  !!
  !! @param[out]    ls                The language specification.  Not 
  !! allocated if the suffix did not contain a BIND clause.
  !!
  !! @param[out]    after             The index of the token after the suffix.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseSuffix(tlist, start, result_name, ls, after, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    TYPE(Identifier), INTENT(OUT), ALLOCATABLE :: result_name
    TYPE(LangSpec), INTENT(OUT), ALLOCATABLE :: ls
    INTEGER, INTENT(OUT) :: after
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /suffix/.
    CHARACTER(*), PARAMETER :: comp = 'R1231'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err    ! Error object for Expect.
    INTEGER :: it         ! Current token index.
    INTEGER :: opt        ! Expect option.
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    CALL ExpectKeyword(  &
        tlist,  &
        start,  &
        [scck_'RESULT', scck_'BIND  '],  &
        '/suffix/',  &
        opt,  &
        err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    IF (opt == 1) THEN
      ! Progressive parse of the result clause.
      ALLOCATE(result_name)
      CALL parse_result(tlist, start + 1, result_name, after, sub_err_list)
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
      it = after
      
      CALL Expect(  &
          tlist,  &
          it,  &
          [ittEndStatement, ittName],  &
          [scck_'    ', scck_'BIND'],  &
          opt,  &
          err ) 
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      IF (opt == 1) RETURN      ! End of statement after the RESULT clause.
      
      ! Progressive parse of the language spec.
      ALLOCATE(ls)
      CALL Parse(tlist, it + 1, ls, after, sub_err_list)
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
    ELSE
      ! Progressive parse of the language spec.
      ALLOCATE(ls)
      CALL Parse(tlist, start + 1, ls, after, sub_err_list)
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
      it = after
      
      CALL Expect( tlist, it, [ittEndStatement, ittName],  &
          [scck_'      ', scck_'RESULT'], opt, err ) 
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      IF (opt == 1) RETURN      ! End of statement after the BIND clause.
      
      ! Progressive parse of the result clause.
      ALLOCATE(result_name)
      CALL parse_result(tlist, it + 1, result_name, after, sub_err_list)
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
    END IF
    
  END SUBROUTINE ParseSuffix
  
  
  !*****************************************************************************
  !!
  !> Parse the parenthesised part of the RESULT clause.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the token that follows the
  !! RESULT keyword.  This should be the opening parenthesis.
  !!
  
  SUBROUTINE parse_result(tlist, start, result_name, after, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    TYPE(Identifier), INTENT(OUT) :: result_name
    INTEGER, INTENT(OUT) :: after
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /suffix/.
    CHARACTER(*), PARAMETER :: comp = 'R1231'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    CALL Expect(tlist, start, ittDelimiter, scck_'(', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL ExpectName(tlist, start + 1, '/result-name/', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    result_name = IdentifierIfortBug(tlist(start + 1))
    
    CALL Expect(tlist, start + 2, ittDelimiter, scck_')', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    after = start + 3
    
  END SUBROUTINE parse_result
  
  
  !*****************************************************************************
  !!
  !> Write a debugging list of the other prefix spec's to a logical unit.
  !!
  !! @param[in]     attrs             The prefix specifications.
  !!
  !! @param[in]     log_unit          Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix for each record.
  
  SUBROUTINE DumpOtherPrefixSpec(attrs, log_unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(OtherPrefixSpecs), INTENT(IN) :: attrs
    INTEGER, INTENT(IN) :: log_unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    IF (attrs%elemental) WRITE (log_unit, "(A,'ELEMENTAL')") pfx
    
    ! purity of zero means not specified.
    IF (attrs%purity == 1) THEN
      WRITE (log_unit, "(A,'IMPURE')") pfx
    ELSE IF (attrs%purity == 2) THEN
      WRITE (log_unit, "(A,'PURE')") pfx
    END IF
    
    IF (attrs%module) WRITE (log_unit, "(A,'MODULE')") pfx
    
    IF (attrs%recursive) WRITE (log_unit, "(A,'RECURSIVE')") pfx
    
  END SUBROUTINE DumpOtherPrefixSpec
  
  
  !*****************************************************************************
  !!
  !> Implementation of FunctionStmt%Dump - write a debugging 
  !! representation of the results of parsing the statement 
  !! to a logical unit.
  !!
  !! @param[in]     pnode             The FunctionSubprogram to dump.
  !!
  !! @param[in]     unit              Logical unit connected for 
  !! formatted output.
  !!
  !! @param[in]     pfx               Prefix for each record.
  
  ! As separate module procedures cannot contain separate module procedures, 
  ! this doesn't need to be recursive.
  SUBROUTINE fun_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(FunctionStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Argument index.
    
    !***************************************************************************
    
    WRITE (unit, "(A,'/function-name/ = ',A)") pfx,  &
        .NameOf. pnode%function_name
    
    IF (ALLOCATED(pnode%ds)) THEN
      WRITE (unit, "(A,'/declaraion-type-spec/ = ')", ADVANCE='NO') pfx
      CALL WriteDeclTypeSpec(unit, pnode%ds)
      WRITE (unit, "()")
    END IF
    
    CALL DumpOtherPrefixSpec(pnode%prefixes, unit, pfx)
    
    ! Arguments.
    WRITE (unit, "(A,'/dummy-arg-name-list/')") pfx
    DO i = 1, SIZE(pnode%args)
      WRITE (unit, "(A,A)") pfx // '  ', .NameOf. pnode%args(i)
    END DO
    
    ! Function result.
    IF (ALLOCATED(pnode%result_name)) THEN
      WRITE (unit, "(A,'/result-name/ = ',A)") pfx, .NameOf. pnode%result_name
    END IF
    
    ! Language spec (BIND attribute).
    IF (ALLOCATED(pnode%ls)) THEN
      WRITE (unit, "(A,'/proc-language-binding-spec/ = ')", ADVANCE='NO')  &
          pfx
      CALL WriteLangSpec(unit, pnode%ls)
      WRITE (unit, "()")
    END IF
    
  END SUBROUTINE fun_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of FunctionSubprogram%Query.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  SUBROUTINE fun_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(FunctionStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%ds)) THEN
      CALL pnode%ds%Query(visitor)
    END IF
    
    IF (ALLOCATED(pnode%ls)) CALL pnode%ls%Query(visitor)

    CALL visitor%Finish(pnode)
    
  END SUBROUTINE fun_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of FunctionSubprogram%Modify.
  !!
  !! @param[in,out] pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  SUBROUTINE fun_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(FunctionStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%ds)) THEN
      CALL pnode%ds%Modify(visitor)
    END IF
    
    IF (ALLOCATED(pnode%ls)) CALL pnode%ls%Modify(visitor)

    CALL visitor%Finish(pnode)
    
  END SUBROUTINE fun_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of FunctionStmt%What - return a descriptive 
  !! string for the parse node.
  !!
  !! @returns 'function-stmt'
  
  FUNCTION fun_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'function-stmt'
    
  END FUNCTION fun_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of FunctionStmt%Advertise - advertise the identifiers 
  !! referenced or defined by the statement.
  !!
  !! @param[in]     pnode             The statement to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE fun_Advertise(pnode, ident_list)
    
    USE Debug
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(FunctionStmt), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: ident_size     ! number of identifiers.
    INTEGER :: in             ! last nature index.
    INTEGER :: iarg           ! argument index.
    
    !***************************************************************************
    
    ! If the function statement has no result clause, then we mark the 
    ! identifier for the function name as being both a function (which 
    ! it is) and a result (which it is).  Perhaps another view would 
    ! be that because the name of the function is the result name in 
    ! the scope of the function body, it is not a function, or perhaps 
    ! what we are doing now is best (are there any contexts in the 
    ! local scope where the name of a function that is also the name 
    ! of the function result is treated as the name of the function?) 
    ident_size = SIZE(pnode%args) + MERGE(2, 1, ALLOCATED(pnode%result_name))
    ALLOCATE(ident_list(ident_size))
    IF (ALLOCATED(pnode%result_name)) THEN
      ! There is a RESULT suffix, the name of the function is the function only.
      ident_list(1) = IdentifierNature(  &
          NAME=.NameOf. pnode%function_name,  &
          NATURES=[NatureDefFunction, NatureDefProcedureDef],  &
          LOCATIONS=[  &
            .LocationOf. pnode%function_name,  &
            .LocationOf. pnode%function_name ] )
    ELSE
      ! There is no RESULT suffix, the name of the function is 
      ! both a function and a result variable.
      IF (ALLOCATED(pnode%ds)) THEN
        ident_list(1) = IdentifierNature(  &
            NAME=.NameOf. pnode%function_name,  &
            NATURES=[  &
              NatureDefResult,  &
              NatureDefFunction,  &
              NatureDefProcedureDef,  &
              NatureDefHasType ],  &
            LOCATIONS=[  &
              .LocationOf. pnode%function_name,  &
              .LocationOf. pnode%function_name,  &
              .LocationOf. pnode%function_name,  &
              .LocationOf. pnode%function_name ] )
      ELSE
        ident_list(1) = IdentifierNature(  &
            NAME=.NameOf. pnode%function_name,  &
            NATURES=[  &
              NatureDefResult,  &
              NatureDefFunction,  &
              NatureDefProcedureDef ],  &
            LOCATIONS=[  &
              .LocationOf. pnode%function_name,  &
              .LocationOf. pnode%function_name,  &
              .LocationOf. pnode%function_name ] )
      END IF
    END IF
    in = 1
    
    ! Identifiers in the arguments.
    DO iarg = 1, SIZE(pnode%args)
      ident_list(in+iarg) = IdentifierNature(  &
          NAME=.NameOf. pnode%args(iarg),  &
          NATURES=[NatureDefDummy],  &
          LOCATIONS=[.LocationOf. pnode%args(iarg)] )
    END DO
    in = in + iarg
    
    ! The identifier in the function result.
    IF (ALLOCATED(pnode%result_name)) THEN
      IF (ALLOCATED(pnode%ds)) THEN
        ident_list(in) = IdentifierNature(  &
            NAME=.NameOf. pnode%result_name,  &
            NATURES=[  &
              NatureDefResult,  &
              NatureDefHasType ],  &
            LOCATIONS=[  &
              .Locationof. pnode%result_name,  &
              .Locationof. pnode%result_name ] )
      ELSE
        ident_list(in) = IdentifierNature(  &
            NAME=.NameOf. pnode%result_name,  &
            NATURES=[NatureDefResult],  &
            LOCATIONS=[.Locationof. pnode%result_name] )
      END IF
    END IF
    
    ! @todo identifiers in any binding label.
    
  END SUBROUTINE fun_Advertise
  
END MODULE FunctionStmts
