! $Id: GenericSpecs.f90 2795 2019-03-22 09:15:08Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the GenericSpecs module.


!*******************************************************************************
!!
!> Procedures and types for parsing /generic-spec/'s and 
!! /defined-io-generic-spec/'s.
!!
!! These are used in /use-stmt/'s, /type-bound-generic-stmt/'s and 
!! /interface-stmt/'s.  They are defined by R1207 as:
!! - /generic-name/
!! - OPERATOR ( /defined-operator/ )
!! - ASSIGNMENT ( = )
!! - /defined-io-generic-spec/
!!
!! where /defined-io-generic-spec/ is defined by R1208 as:
!! - READ ( FORMATTED )
!! - READ ( UNFORMATTED )
!! - WRITE ( FORMATTED )
!! - WRITE ( UNFORMATTED )

MODULE GenericSpecs
  
  USE Identifiers
  USE IdentifierNatures
  USE IdentifierSymbols
  USE ParseNodes
  USE SourceLocations
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseGenericSpec
  PUBLIC :: WriteGenericSpec
  PUBLIC :: GetGenericSpecDescription
  PUBLIC :: GetGenericSpecIdentifierName
  PUBLIC :: IdentifierNature
  PUBLIC :: IdentifierSymbol
  
  !-----------------------------------------------------------------------------
  
  !> Storage of the results of parsing a /generic-spec/.
  TYPE, EXTENDS(TermParseNode), PUBLIC :: GenericSpec
    !> - 0: No keyword - the name is just the name.
    !! - 1: OPERATOR - the name will be the operator.
    !! - 2: ASSIGNMENT - the name will be '='.
    !! - 3: READ(FORMATTED) - name not used.
    !! - 4: WRITE(FORMATTED) - name not used.
    !! - 5: READ(UNFORMATTED) - name not used.
    !! - 6: WRITE(UNFORMATTED) - name not used.
    INTEGER :: keyword
    !> Name of the generic spec - see comments for keyword.
    !!
    !! Not allocated if keyword is 3 or more.
    TYPE(Identifier), ALLOCATABLE :: name
    !> The location of the specification.
    !!
    !! Ideally we would make this private, but we need some way of setting 
    !! it from the UseStmts module.
    !!
    !! This is allocatable because perhaps reasons not really sure.
    TYPE(SourceLocation), ALLOCATABLE :: location
  CONTAINS
    PROCEDURE :: GetLocation => gs_GetLocation
    PROCEDURE, NOPASS :: What => gs_What
    PROCEDURE :: Write => gs_Write
    PROCEDURE :: Query => gs_Query
    PROCEDURE :: Modify => gs_Modify
  END TYPE GenericSpec
  
  INTERFACE IdentifierSymbol
    MODULE PROCEDURE IdentifierSymbol_spec
  END INTERFACE IdentifierSymbol
  
  INTERFACE IdentifierNature
    MODULE PROCEDURE IdentifierNature_spec
  END INTERFACE IdentifierNature
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a /generic-spec/ from a sequence of tokens.
  !!
  !! @param[in]     tlist             The sequence of tokens that make up the 
  !! /generic-spec/.
  !!
  !! @param[out]    spec              The results of parsing the generic spec.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseGenericSpec(tlist, spec, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(GenericSpec), INTENT(OUT) :: spec
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /generic-spec/ for error reporting.
    CHARACTER(*), PARAMETER :: comp = 'R1207'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: next           ! Index after the last parsed token.
    
    ! FORM option in a /defined-io-generic-spec/ (see process_form).
    INTEGER :: form
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
          MSG='Expected /generic-spec/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    CALL Expect(tlist, 1, ittName, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! Save location.
    spec%location = QueryLocation(tlist(1))
      
    IF (Match(tlist, 2, ittDelimiter, scck_'(')) THEN
      ! Looks like we've got a generic-spec that's not just a name.
      SELECT CASE (QueryValue(tlist(1)))
      CASE (scck_'OPERATOR')
        CALL Expect(tlist, 3, ittOperator, err)
        CALL Add(err_list, err, COMPONENT=comp)
        IF (Failure(err)) RETURN
        
        CALL Expect(tlist, 4, ittDelimiter, scck_')', err)
        CALL Add(err_list, err, COMPONENT=comp)
        IF (Failure(err)) RETURN
        
        spec%keyword = 1
        spec%name = IdentifierIfortBug(tlist(3))
        next = 5
      
      CASE (scck_'ASSIGNMENT')
        CALL Expect(tlist, 3, ittAssign, err)
        CALL Add(err_list, err, COMPONENT=comp)
        IF (Failure(err)) RETURN
        
        CALL Expect(tlist, 4, ittDelimiter, scck_')', err)
        CALL Add(err_list, err, COMPONENT=comp)
        IF (Failure(err)) RETURN
        
        spec%keyword = 2
        spec%name = IdentifierIfortbug(tlist(3))
        next = 5
        
      CASE (scck_'READ')
        CALL process_form(tlist, 3, form, err)
        CALL Add(err_list, err, COMPONENT=comp)
        IF (Failure(err)) RETURN
        
        IF (form == 1) THEN
          spec%keyword = 3
        ELSE
          spec%keyword = 5
        END IF
        next = 5
        
      CASE (scck_'WRITE')
        CALL process_form(tlist, 3, form, err)
        CALL Add(err_list, err, COMPONENT=comp)
        IF (Failure(err)) RETURN
        IF (form == 1) THEN
          spec%keyword = 4
        ELSE
          spec%keyword = 6
        END IF
        next = 5
        
      END SELECT
    ELSE
      ! We've just got a name
      spec%keyword = 0
      spec%name = IdentifierIfortBug(tlist(1))
      next = 2
    END IF
    
    CALL ExpectEnd(tlist, next, comp, err_list)
    
  END SUBROUTINE ParseGenericSpec
  
  
  !*****************************************************************************
  !!
  !> Parses the FORMATTED or UNFORMATTED keyword and closing parentheses 
  !! after the "READ(" or "WRITE(" sequence in a /defined-io-generic-spec/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Location of the first token after the 
  !! opening parenthesis in @a tlist.
  !!
  !! @param[in]     form              Form of the /defined-io-generic-spec/:
  !! - 0: Not recognised.
  !! - 1: FORMATTED.
  !! - 2: UNFORMATTED.
  !!
  !! @param[out]    err               Error object.
  !!
  !! For error reporting this procedure may reference tlist(start - 1).
  
  SUBROUTINE process_form(tlist, start, form, err)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(OUT) :: form
    TYPE(Error), INTENT(OUT) :: err
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /defined-io-generic-spec/ for error reporting.
    CHARACTER(*), PARAMETER :: comp = 'R1208'
    
    !***************************************************************************
    
    CALL Expect( tlist, start, [ittName, ittName],  &
        [scck_'FORMATTED  ', scck_'UNFORMATTED'], form, err )
    IF (Failure(err)) RETURN
    
    CALL Expect(tlist, start + 1, ittDelimiter, scck_')', err)
    
  END SUBROUTINE process_form
  
  
  !*****************************************************************************
  !!
  !> Write a text representation of a GenericSpec to a logical unit.
  !!
  !! @param[in]     log_unit          Logical unit connected for formatted 
  !! output for the text represenation.
  !!
  !! @param[in]     spec              The generic spec.
  
  SUBROUTINE WriteGenericSpec(log_unit, spec)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    INTEGER, INTENT(IN) :: log_unit
    TYPE(GenericSpec), INTENT(IN) :: spec
    
    !***************************************************************************
    
    WRITE (log_unit, "('keyword = ',I0)", ADVANCE='NO') spec%keyword
    IF ((spec%keyword < 3) .OR. ALLOCATED(spec%name)) THEN
      WRITE (log_unit, "('  name = ',A)", ADVANCE='NO') .NameOf. spec%name
    END IF
    
  END SUBROUTINE WriteGenericSpec
  
  
  !*****************************************************************************
  !!
  !> Implementation of GenericSpec%Write - write a text representation of a 
  !! generic spec to a logical unit.
  !!
  !! @param[in]     pnode             The format.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output for the text represenation.
  
  SUBROUTINE gs_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(GenericSpec), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    CALL WriteGenericSpec(unit, pnode)
    
  END SUBROUTINE gs_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of GenericSpec%What - get a text description of the 
  !! relevant syntax rule for the parse node.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @returns 'generic-spec'
  
  FUNCTION gs_What() RESULT(what)
      
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'generic-spec'
    
  END FUNCTION gs_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of GenericSpec%Query - traverse the parse node 
  !! with a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE gs_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(GenericSpec), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE gs_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of GenericSpec%Modify - traverse the parse node 
  !! with a modify operation.
  !!
  !! @param[inout] pnode             The parse node element to modify.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE gs_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(GenericSpec), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE gs_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of GenericSpec%GetLocation - get the location of the 
  !! generic specification.
  !!
  !! @param[in]     pnode             The generic spec.
  !!
  !! @returns The location of the generic spec.
  
  FUNCTION gs_GetLocation(pnode) RESULT(loc)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(GenericSpec), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    loc = pnode%location
    
  END FUNCTION gs_GetLocation
  
  
  !*****************************************************************************
  !!
  !> Get descriptive text for a generic spec.
  !!
  !! @param[in]     spec              The generic spec to describe.
  !!
  !! @returns A string that resembles the original source.
  
  PURE FUNCTION GetGenericSpecDescription(spec) RESULT(str)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(GenericSpec), INTENT(IN) :: spec
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: str
    
    !***************************************************************************
    
    SELECT CASE (spec%keyword)
    CASE (0)      ; str = .NameOf. spec%name
    CASE (1)      ; str = 'OPERATOR(' // .NameOf. spec%name // ')'
    CASE (2)      ; str = 'ASSIGNMENT(=)'
    CASE (3)      ; str = 'READ(FORMATTED)'
    CASE (4)      ; str = 'WRITE(FORMATTED)'
    CASE (5)      ; str = 'READ(UNFORMATTED)'
    CASE (6)      ; str = 'WRITE(UNFORMATTED)'
    CASE DEFAULT  ; str = '(invalid)'
    END SELECT
    
  END FUNCTION GetGenericSpecDescription
  
  
  !*****************************************************************************
  !!
  !> Get the internal name used for the generic identifier.
  !!
  !! @param[in]     spec              The generic spec to query.
  !!
  !! @returns The internal name for the generic identifier.
  
  PURE FUNCTION GetGenericSpecIdentifierName(spec) RESULT(str)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(GenericSpec), INTENT(IN) :: spec
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: str
    
    !---------------------------------------------------------------------------
    ! Locals
    
    CHARACTER(:), ALLOCATABLE :: name_text
    
    !***************************************************************************
    
    IF (ALLOCATED(spec%name)) name_text = .NameOf. spec%name
    
    str = GetIdentifierName(name_text, spec%keyword)
    
  END FUNCTION GetGenericSpecIdentifierName
  
  
  !*****************************************************************************
  !!
  !> Convenience constructor to create an IdentifierSymbol from a 
  !! generic spec.
  
  FUNCTION IdentifierSymbol_spec(spec) RESULT(sym)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(GenericSpec), INTENT(IN) :: spec
    
    ! Function result.
    TYPE(IdentifierSymbol) :: sym
    
    !***************************************************************************
    
    SELECT CASE (spec%keyword)
    CASE (0)      ; sym = NameSymbol(.NameOf. spec%name)
    CASE (1)      ; sym = OperatorSymbol(.NameOf. spec%name)
    CASE (2)      ; sym = AssignmentSymbol()
    CASE (3)      ; sym = ReadFormattedSymbol()
    CASE (4)      ; sym = WriteFormattedSymbol()
    CASE (5)      ; sym = ReadUnformattedSymbol()
    CASE (6)      ; sym = WriteUnformattedSymbol()
    CASE DEFAULT  ; sym = NameSymbol('???')
    END SELECT
    
  END FUNCTION IdentifierSymbol_spec
  
  
  !*****************************************************************************
  !!
  !> Convenience constructor for an IdentifierNature object that 
  !! takes a generic spec.
  !!
  !! @param[in]     spec              The generic spec for the identifier.
  !!
  !! @param[in]     natures           The natures for the identifier.
  !!
  !! @param[in]     locations         The locations in source responsible 
  !! for each nature.  This must be the same size as @a natures.
  !!
  !! @returns A constructed IdentifierNature object.
  
  FUNCTION IdentifierNature_spec(spec, natures, locations) RESULT(r)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(GenericSpec), INTENT(IN) :: spec
    INTEGER, INTENT(IN) :: natures(:)
    TYPE(SourceLocation), INTENT(IN) :: locations(:)
    
    ! Function result.
    TYPE(IdentifierNature) :: r
    
    !***************************************************************************
    
    ! ifort bug?  Have to use specific procedure here.
    r = IdentifierNature(IdentifierSymbol_spec(spec), natures, locations)
    
  END FUNCTION IdentifierNature_spec
  
END MODULE GenericSpecs
