! $Id: GotoStmts.f90 2795 2019-03-22 09:15:08Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the ContinueStmts module.


!*******************************************************************************
!!
!> Procedures for parsing /goto-stmt/'s and /computed-goto-stmt/'s.

MODULE GotoStmts
  
  USE Expressions
  USE Labels
  USE ParseNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseGotoStmt
  PUBLIC :: ParseComputedGotoStmt
  
  !> Stores the results of parsing a /goto-stmt/.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: GotoStmt
    !> The label to jump to.
    TYPE(Label) :: goto_label
  CONTAINS
    PROCEDURE :: Dump => goto_Dump
    PROCEDURE :: Query => goto_Query
    PROCEDURE :: Modify => goto_Modify
    PROCEDURE, NOPASS :: What => goto_What
  END TYPE GotoStmt
  
  !> Stores the results of parsing a /computed-goto-stmt/.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: ComputedGotoStmt
    !> The labels in side the parentheses of the computed goto statement.
    !!
    !! Always allocated, will not be zero sized.
    TYPE(Label), ALLOCATABLE :: labels(:)
    !> The expression after the parentheses.
    CLASS(Expression), ALLOCATABLE :: expr
  CONTAINS
    PROCEDURE :: Dump => cgoto_Dump
    PROCEDURE :: Query => cgoto_Query
    PROCEDURE :: Modify => cgoto_Modify
    PROCEDURE, NOPASS :: What => cgoto_What
  END TYPE ComputedGotoStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a /goto-stmt/ (R854) from a sequence of tokens.
  !!
  !! @param[in]     tlist             The sequence of tokens that make up the 
  !! statement.
  !!
  !! @param[out]    goto_stmt         The results of parsing the statement.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! A /goto-stmt/ consists of the keyword "GOTO" followed by a label.
  
  SUBROUTINE ParseGotoStmt(tlist, goto_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Scanner
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(GotoStmt), INTENT(OUT) :: goto_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /goto-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R851'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: it             ! Token index.
    TYPE(Error) :: err        ! Error object for Expect calls.
    
    ! Error list for child Parse calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /goto-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! GOTO...
    
    CALL ExpectSplittableKeyword(tlist, 1, scck_'GO', scck_'TO', it, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! it here should point to the label.
    
    !---------------------------------------------------------------------------
    ! ... /label/
    
    CALL ParseLabel(  &
        tlist,  &
        it,  &
        MIN(SIZE(tlist), it),  &
        goto_stmt%goto_label,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
    !---------------------------------------------------------------------------
    
    ! Nothing else!
    CALL ExpectEOS(tlist, it + 1, comp, err_list)
    
  END SUBROUTINE ParseGotoStmt
  
  
  !*****************************************************************************
  !!
  !> Implementation of GotoStmt%Dump - write a debugging 
  !! representation of the results of parsing a goto statement to a 
  !! logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE goto_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(GotoStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    WRITE (unit, "(A,'/label/ = ',I0)") pfx, .ValueOf. pnode%goto_label
    
  END SUBROUTINE goto_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of GotoStmt%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE goto_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(GotoStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE goto_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of GotoStmt%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[in,out] pnode             The parse node element to modify.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE goto_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(GotoStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE goto_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of GotoStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'goto-stmt'.
  
  FUNCTION goto_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'goto-stmt'
    
  END FUNCTION goto_What
  
 
  !*****************************************************************************
  !!
  !> Parse a /computed-goto-stmt/ (R854) from a sequence of tokens.
  !!
  !! @param[in]     tlist             The sequence of tokens that make up the 
  !! statement.
  !!
  !! @param[out]    cgoto_stmt        The results of parsing the statement.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! A /computed-goto-stmt/ consists of the keyword "GOTO" followed by a 
  !! list of labels in parentheses, followed by a scalar integer expression.
  !!
  !! This is a deprecated language feature.
  
  SUBROUTINE ParseComputedGotoStmt(tlist, cgoto_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE Expressions
    USE MatchUtils
    USE ParseUtils
    USE Scanner
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(ComputedGotoStmt), INTENT(OUT) :: cgoto_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /computed-goto-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R852'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: it             ! Token index.
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: closing        ! Index of closing parenthesis.
    INTEGER :: item_count     ! Number of items in the label list.
    INTEGER :: item           ! Label list item index.
    INTEGER :: item_start     ! Index of the first token in a label list item.
    INTEGER :: item_finish    ! Index of the last token in a label list item.
    
    ! Error list for child Parse calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
          MSG='Expected /computed-goto-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! GOTO or GO TO...
    
    CALL ExpectSplittableKeyword(tlist, 1, scck_'GO', scck_'TO', it, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! it here should point to the opening parenthesis.
    
    !---------------------------------------------------------------------------
    ! ... ( /label-list/ ) ...
    
    CALL Expect(tlist, it, ittDelimiter, scck_'(', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL ExpectClosing(tlist, it, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    item_count = ListItemCount(tlist(it+1:closing-1))
    ALLOCATE(cgoto_stmt%labels(item_count))
    item_finish = 0
    DO item = 1, item_count
      CALL GetNextListItem(tlist, it+1, closing - 1, item_start, item_finish)
      
      CALL ParseLabel(  &
          tlist,  &
          item_start,  &
          item_finish,  &
          cgoto_stmt%labels(item),  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
      
      CALL ExpectEnd(tlist(:item_finish), item_start + 1, comp, err_list)
      ! Carry on...
    END DO
    
    ! Skip closing parenthesis.
    it = closing + 1
    
    !---------------------------------------------------------------------------
    ! [,] scalar-int-expression.
    
    IF (it > SIZE(tlist)) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          LOCATION=QueryLocation(tlist(SIZE(tlist))),  &
          MSG='Expected "," or /scalar-int-expr/.' )
      RETURN
    END IF
    
    ! if the comma is there, swallow it.
    IF (Match(tlist, it, ittComma, scck_',')) it = it + 1
    
    CALL ParseExpr(tlist, it, SIZE(tlist), cgoto_stmt%expr, sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    ! The expression parse deals with superfluous cruft.
    
  END SUBROUTINE ParseComputedGotoStmt
  
  
  !*****************************************************************************
  !!
  !> Implementation of ComputedGotoStmt%Dump - write a debugging 
  !! representation of the results of parsing a computed goto statement to a 
  !! logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE cgoto_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ComputedGotoStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    WRITE (unit, "(A,'/label-list/ = ',999(I0,:,','))") pfx,  &
        .ValueOf. pnode%labels
    
    WRITE (unit, "(A,'/scalar-int-expr/ = ')", ADVANCE='NO' ) pfx
    CALL WriteExpr(unit, pnode%expr)
    WRITE (unit, "()")
    
  END SUBROUTINE cgoto_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of ComputedGotoStmt%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE cgoto_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ComputedGotoStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%expr%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE cgoto_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of ComputedGotoStmt%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[in,out] pnode             The parse node element to query.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE cgoto_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ComputedGotoStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%expr%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE cgoto_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of ComputedGotoStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'computed-goto-stmt'.
  
  FUNCTION cgoto_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'computed-goto-stmt'
    
  END FUNCTION cgoto_What
  
END MODULE GotoStmts
