! $Id: IdentifierUtils.f90 2795 2019-03-22 09:15:08Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the Identifiers module.


!*******************************************************************************
!!
!> Defines a type that can act as common storage for identifier like things 
!! in parse nodes.
!!
!! Split out from the Identifiers module to help cut down on 
!! module dependencies.  See the Identifiers module for more 
!! information.

MODULE IdentifierUtils
  
  USE SourceLocations
  
  IMPLICIT NONE
  
  PRIVATE
  
  PUBLIC :: Identifier
  PUBLIC :: IdentifierIfortBug
  
  !-----------------------------------------------------------------------------
  
  !> Store an identifier of some sort.
  !!
  !! Identifier here may have a slightly different meaning from that in the 
  !! standard, particularly in the context of generic identifiers that are 
  !! not names.  Here we are really just dealing with the single token.
  !!
  !! Note that the name doesn't have to be a name in a syntax sense - it could 
  !! be an operator or similar too.    The concept is just some text and a 
  !! location.
  TYPE :: Identifier
    PRIVATE
    !> The text for the identifier.  This is normally just its name.
    CHARACTER(:), ALLOCATABLE :: name
    !> The location in source of the identifier.
    TYPE(SourceLocation) :: location
  CONTAINS
    PROCEDURE, PRIVATE :: get_name
    GENERIC :: OPERATOR(.NameOf.) => get_name
    
    PROCEDURE :: GetLocation => identifier_GetLocation
    GENERIC :: OPERATOR(.LocationOf.) => GetLocation
    
    PROCEDURE, PRIVATE :: get_location_string
    GENERIC :: OPERATOR(.LocationStringOf.) => get_location_string
    
    PROCEDURE, PRIVATE :: write_formatted
    GENERIC :: WRITE(FORMATTED) => write_formatted
  END TYPE Identifier
  
  !> Constructors for Identifier.
  !!
  !! Rename when the relevant ifort bug is fixed.  No good as of 15.0.
  INTERFACE IdentifierIfortBug
    MODULE PROCEDURE Identifier_token
    MODULE PROCEDURE Identifier_components
  END INTERFACE IdentifierIfortBug
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Overloaded structure constructor that constructs an Identifier object 
  !! given a Token.
  !!
  !! @param[in]     tok               The token to use to construct the 
  !! identifier.
  !!
  !! @returns An identifier.
  
  FUNCTION Identifier_token(tok) RESULT(r)
    
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(Token), INTENT(IN) :: tok
    
    ! Function result.
    TYPE(Identifier) :: r
    
    !***************************************************************************
    
    r%name = QueryValue(tok)
    r%location = QueryLocation(tok)
    
  END FUNCTION Identifier_token
  
  
  !*****************************************************************************
  !!
  !> Overloaded structure constructor that constructs an Identifier object 
  !! given a Token.
  !!
  !! @param[in]     name              The name of the identifier.
  !!
  !! @param[in]     location          The location of the identifier.
  !!
  !! @returns An identifier.
  
  FUNCTION Identifier_components(name, location) RESULT(r)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CHARACTER(*), INTENT(IN) :: name
    TYPE(SourceLocation), INTENT(IN) :: location
    
    ! Function result.
    TYPE(Identifier) :: r
    
    !***************************************************************************
    
    r%name = name
    r%location = location
    
  END FUNCTION Identifier_components
  
  
  !*****************************************************************************
  !!
  !> Get the name (or similar text) of the identifier.
  !!
  !! @param[in]     ident             The identifier to be queried.
  !!
  !! @returns The value of the name component.
  
  PURE FUNCTION get_name(ident) RESULT(name)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(Identifier), INTENT(IN) :: ident
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: name
    
    !***************************************************************************
    
    name = ident%name
    
  END FUNCTION get_name
  
  
  !*****************************************************************************
  !!
  !> Get the location of the identifier.
  !!
  !! @param[in]     ident             The identifier to be queried.
  !!
  !! @returns The value of the location component.
  
  FUNCTION identifier_GetLocation(ident) RESULT(location)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(Identifier), INTENT(IN) :: ident
    
    ! Function result.
    TYPE(SourceLocation) :: location
    
    !***************************************************************************
    
    location = ident%location
    
  END FUNCTION identifier_GetLocation
  
  
  !*****************************************************************************
  !!
  !> Get the location of the identifier.
  !!
  !! @param[in]     ident             The identifier to be queried.
  !!
  !! @returns The value of the location component.
  
  FUNCTION get_location_string(ident) RESULT(s)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(Identifier), INTENT(IN) :: ident
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: s
    
    !***************************************************************************
    
    s = .ToString. ident%location
    
  END FUNCTION get_location_string
  
  
  !*****************************************************************************
  !!
  !> Write the identifier to a logical unit.
  !!
  !! @param[in]     dtv               The identifier to write.
  !!
  !! @param[in]     unit              The logical unit.
  !!
  !! @param[in]     v_list            List of integer specifiers in the 
  !! data edit descriptor.
  !!
  !! @param[out]    iostat            Error code.
  !!
  !! @param[in,out] iomsg             Error message - redefined if 
  !! @a iostat is non-zero.
  
  SUBROUTINE write_formatted(dtv, unit, iotype, v_list, iostat, iomsg)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(Identifier), INTENT(IN) :: dtv
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: iotype
    INTEGER, INTENT(IN) :: v_list(:)
    INTEGER, INTENT(OUT) :: iostat
    CHARACTER(*), INTENT(INOUT) :: iomsg
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: width          ! Field width.
    CHARACTER(10) :: fmt      ! Format buffer.
    
    !***************************************************************************
    
    IF (iotype == 'DTFULL') THEN
      ! The location.
      IF (SIZE(v_list) >= 1) THEN
        width = v_list(1)
      ELSE
        width = 0
      END IF
      IF (width == 0) THEN
        fmt = '(A,1X)'
      ELSE
        WRITE (fmt, "('A',I0,',1X')") width
      END IF
      WRITE (unit, fmt, IOSTAT=iostat, IOMSG=iomsg) .ToString. dtv%location
      IF (iostat /= 0) RETURN
      
      ! The name.
      IF (SIZE(v_list) >= 2) THEN
        width = v_list(2)
      ELSE
        width = 0
      END IF
      IF (width == 0) THEN
        fmt = '(A)'
      ELSE
        WRITE (fmt, "('A',I0)") width
      END IF
      WRITE (unit, fmt, IOSTAT=iostat, IOMSG=iomsg) dtv%name
      IF (iostat /= 0) RETURN
    ELSE
      ! When we are not being asked for anything special, just write 
      ! out the name.
      WRITE (unit, "(A)", IOSTAT=iostat, IOMSG=iomsg) dtv%name
      IF (iostat /= 0) RETURN
    END IF
    
  END SUBROUTINE write_formatted
  
END MODULE IdentifierUtils
