! $Id: InquireStmts.f90 2795 2019-03-22 09:15:08Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the InquireStmts module.


!*******************************************************************************
!!
!> Types and procedures for parsing /inquire-stmt/'s (R930).

MODULE InquireStmts
  
  USE Expressions
  USE InputOutputItems
  USE Labels
  USE ParseNodes
  USE Variables
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseInquireStmt
  
  !-----------------------------------------------------------------------------
  
  !> The results of parsing an /inquire-stmt/.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: InquireStmt
    !> The UNIT specification.
    CLASS(Expression), ALLOCATABLE :: unit
    !> The FILE specification.
    !!
    !! This is a /file-name-expr/.
    CLASS(Expression), ALLOCATABLE :: file
    !> The ACCESS specification.
    CLASS(Variable), ALLOCATABLE :: access
    !> The ACTION specification.
    CLASS(Variable), ALLOCATABLE :: action
    !> The ASYNCHRONOUS specification.
    CLASS(Variable), ALLOCATABLE :: asynchronous
    !> The BLANK specification.
    CLASS(Variable), ALLOCATABLE :: blank
    !> The DECIMAL specification.
    CLASS(Variable), ALLOCATABLE :: decimal
    !> The DELIM specification.
    CLASS(Variable), ALLOCATABLE :: delim
    !> The DIRECT specification.
    CLASS(Variable), ALLOCATABLE :: direct
    !> The ENCODING specification.
    CLASS(Variable), ALLOCATABLE :: encoding
    !> The ERR specification.
    TYPE(Label), ALLOCATABLE :: err
    !> The EXIST specification.
    CLASS(Variable), ALLOCATABLE :: exist
    !> The FORM specification.
    CLASS(Variable), ALLOCATABLE :: form
    !> The FORMATTED specification.
    CLASS(Variable), ALLOCATABLE :: formatted
    !> The ID specification.
    !!
    !! Check tests that if this is allocated then pending must also be 
    !! allocated.
    CLASS(Expression), ALLOCATABLE :: id
    !> The IOMSG specification.
    !!
    !! This is an /iomsg-variable/.
    CLASS(Variable), ALLOCATABLE :: iomsg
    !> The IOSTAT specification.
    CLASS(Variable), ALLOCATABLE :: iostat
    !> The NAME specification.
    CLASS(Variable), ALLOCATABLE :: name
    !> The NAMED specification.
    CLASS(Variable), ALLOCATABLE :: named
    !> The NEXTREC specification.
    CLASS(Variable), ALLOCATABLE :: nextrec
    !> The NUMBER specification.
    CLASS(Variable), ALLOCATABLE :: number
    !> The OPENED specification.
    CLASS(Variable), ALLOCATABLE :: opened
    !> The PAD specification.
    CLASS(Variable), ALLOCATABLE :: pad
    !> The PENDING specification.
    !!
    !! Check tests that if this is allocated then id must also be allocated.
    CLASS(Variable), ALLOCATABLE :: pending
    !> The POS specification.
    CLASS(Variable), ALLOCATABLE :: pos
    !> The POSITION specification.
    CLASS(Variable), ALLOCATABLE :: position
    !> The READ specification.
    CLASS(Variable), ALLOCATABLE :: read
    !> The READWRITE specification.
    CLASS(Variable), ALLOCATABLE :: readwrite
    !> The RECL specification.
    CLASS(Variable), ALLOCATABLE :: recl
    !> The ROUND specification.
    CLASS(Variable), ALLOCATABLE :: round
    !> The SEQUENTIAL specification.
    CLASS(Variable), ALLOCATABLE :: sequential
    !> The SIGN specification.
    CLASS(Variable), ALLOCATABLE :: sign
    !> The SIZE specification.
    CLASS(Variable), ALLOCATABLE :: size
    !> The STREAM specification.
    CLASS(Variable), ALLOCATABLE :: stream
    !> The UNFORMATTED specification.
    CLASS(Variable), ALLOCATABLE :: unformatted
    !> The WRITE specification.
    CLASS(Variable), ALLOCATABLE :: write
    !> The IOLENGTH specification.  
    !!
    !! This is mutually exclusive with all others.
    CLASS(Variable), ALLOCATABLE :: iolength
    !> The /output-item-list/ that corresponds to an IOLENGTH style INQUIRE.
    !!
    !! Not allocated if the inquire statement is a UNIT= or FILE= form.
    TYPE(IOItemList), ALLOCATABLE :: output_list(:)
  CONTAINS
    PROCEDURE :: Dump => inquire_Dump
    PROCEDURE :: Query => inquire_Query
    PROCEDURE :: Modify => inquire_Modify
    PROCEDURE, NOPASS :: What => inquire_What
    PROCEDURE :: Check => inquire_Check
  END TYPE InquireStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /inquire-stmt/ (R930).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    inquire_stmt      The results of parsing the 
  !! /inquire-stmt/.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseInquireStmt(tlist, inquire_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(InquireStmt), INTENT(OUT) :: inquire_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for an /inquire-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R930'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: closing        ! Index of the closing delimiter.
    INTEGER :: list_count     ! Number of items in the /inquire-spec-list/.
    INTEGER :: item           ! List item index.
    INTEGER :: item_start     ! Start index in tlist of a particular item.
    INTEGER :: item_finish    ! Finish index in tlist of a particular item.
    LOGICAL :: no_spec        ! Flag to indicate UNIT= was absent.
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /inquire-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! INQUIRE...
    
    CALL ExpectKeyword(tlist, 1, scck_'INQUIRE', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! INQUIRE(...)
    
    CALL Expect(tlist, 2, ittDelimiter, scck_'(', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL ExpectClosing(tlist, 2, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! Determine the form of INQUIRE statement.
    
    ! Do we have anything following the closing parenthesis?
    IF (SIZE(tlist) > closing) THEN
      CALL ExpectKeyword(tlist, 3, scck_'IOLENGTH', err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      CALL Expect(tlist, 4, ittAssign, scck_'=', err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      CALL ParseExpr(  &
          tlist,  &
          5, closing - 1,  &
          inquire_stmt%iolength,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      ! Continue on error.
      
      CALL ParseIOItemList(  &
          tlist,  &
          closing + 1,  &
          SIZE(tlist),  &
          .TRUE.,  &          ! Specify output item list.
          inquire_stmt%output_list,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
    ELSE
      ! /inquire-spec-list/
      list_count = ListItemCount(tlist(3:closing-1))
      item_finish = 0
      DO item = 1, list_count
        CALL GetNextListItem(tlist, 3, closing - 1, item_start, item_finish)
        CALL parse_inquire_spec(  &
            tlist,  &
            item_start,  &
            item_finish,  &
            inquire_stmt,  &
            no_spec,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
        ! Continue on error.
        IF ((item /= 1) .AND. no_spec) THEN
          CALL Add( err_list,  &
              CODE=errSyntax,  &
              COMPONENT='C949 on ' // comp,  &
              LOCATION=QueryLocation(tlist(item_start)),  &
              MSG='If the optional "UNIT=" is omitted, the /file-unit-number/ &
                &shall be the first item in the /inquire-spec-list/.' )
          ! Continue on error.
        END IF
      END DO
      
      ! That's it.
      CALL ExpectEOS(tlist, closing + 1, comp, err_list)
    END IF
    
  END SUBROUTINE ParseInquireStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /inquire-spec/ (R931).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the first token of the 
  !! /inquire-spec/ in @a tlist.
  !!
  !! @param[in]     finish            Index of the last token of the 
  !! /inquire-spec/ in @a tlist.
  !!
  !! @param[in,out] inquire_stmt      Results of parsing the /inquire-spec/.  
  !! Note that this is an INTENT(INOUT) argument - each spec sets a particular 
  !! component
  !!
  !! @param[out]    no_spec           Flag to indicate that the "XXX=" part of 
  !! the spec was absent.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE parse_inquire_spec( tlist, start, finish, inquire_stmt, no_spec,  &
      err_list )
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE IoControlSpecs
    USE MatchUtils
    USE ParseUtils
    USE Scanner
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(InquireStmt), INTENT(INOUT) :: inquire_stmt
    LOGICAL, INTENT(OUT) :: no_spec
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /inquire-spec/.
    CHARACTER(*), PARAMETER :: comp = 'R931'
    
    ! Keywords for each /connect-spec/.
    CHARACTER(*,KIND=scck), PARAMETER :: keywords(36) = [  &
        scck_'UNIT        ', scck_'FILE        ', scck_'ACCESS      ',  &
        scck_'ACTION      ', scck_'ASYNCHRONOUS', scck_'BLANK       ',  &
        scck_'DECIMAL     ', scck_'DELIM       ', scck_'DIRECT      ',  &
        scck_'ENCODING    ', scck_'ERR         ', scck_'EXIST       ',  &
        scck_'FORM        ', scck_'FORMATTED   ', scck_'ID          ',  &
        scck_'IOMSG       ', scck_'IOSTAT      ', scck_'NAME        ',  &
        scck_'NAMED       ', scck_'NEXTREC     ', scck_'NUMBER      ',  &
        scck_'OPENED      ', scck_'PAD         ', scck_'PENDING     ',  &
        scck_'POS         ', scck_'POSITION    ', scck_'READ        ',  &
        scck_'READWRITE   ', scck_'RECL        ', scck_'ROUND       ',  &
        scck_'SEQUENTIAL  ', scck_'SIGN        ', scck_'SIZE        ',  &
        scck_'STREAM      ', scck_'UNFORMATTED ', scck_'WRITE       ' ]
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: opt            ! Expect option.
    INTEGER :: assign         ! Index of the =.
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    IF (finish < start) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          LOCATION=SafeLocation(tlist, finish),  &
          MSG='Expected /inquire-spec/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! Do we have an  argument keyword?
    
    assign = FindBare(tlist(:finish), ittAssign, scck_'=', start)
    IF (assign == 0) THEN
      ! No - this is a unit spec.
      no_spec = .TRUE.
      opt = 1
    ELSE
      no_spec = .FALSE.
      CALL ExpectKeyword(tlist, start, keywords, '/inquire-spec/', opt, err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
    
      CALL Expect(tlist, start + 1, ittAssign, scck_'=', err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    SELECT CASE (opt)
    CASE (1)      ! UNIT
      IF (ALLOCATED(inquire_stmt%unit)) THEN
        CALL already('UNIT')
      ELSE IF (ALLOCATED(inquire_stmt%file)) THEN
        CALL add_file_unit_err
      ELSE
        IF (assign == 0) THEN
          CALL ParseFileUnitNumber(  &
              tlist,  &
              start,  &
              finish,  &
              inquire_stmt%unit,  &
              sub_err_list )
        ELSE
          CALL ParseFileUnitNumber(  &
              tlist,  &
              assign + 1,  &
              finish,  &
              inquire_stmt%unit,  &
              sub_err_list )
        END IF
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
       
    CASE (2)      ! FILE
      IF (ALLOCATED(inquire_stmt%file)) THEN
        CALL already('FILE')
      ELSE IF (ALLOCATED(inquire_stmt%unit)) THEN
        CALL add_file_unit_err
      ELSE
        ! The syntax term is actually a /file-name-expr/.
        CALL ParseExpr(  &
            tlist,  &
            assign + 1, finish,  &
            inquire_stmt%file,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (3)      ! ACCESS
      IF (ALLOCATED(inquire_stmt%access)) THEN
        CALL already('ACCESS')
      ELSE
        CALL ParseVariable(  &
            tlist,  &
            assign + 1, finish,  &
            inquire_stmt%access,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (4)      ! ACTION
      IF (ALLOCATED(inquire_stmt%action)) THEN
        CALL already('ACTION')
      ELSE
        CALL ParseVariable(  &
            tlist,  &
            assign + 1, finish,  &
            inquire_stmt%action,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (5)      ! ASYNCHRONOUS
      IF (ALLOCATED(inquire_stmt%asynchronous)) THEN
        CALL already('ASYNCHRONOUS')
      ELSE
        CALL ParseVariable(  &
            tlist,  &
            assign + 1, finish,  &
            inquire_stmt%asynchronous,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (6)      ! BLANK
      IF (ALLOCATED(inquire_stmt%blank)) THEN
        CALL already('BLANK')
      ELSE
        CALL ParseVariable(  &
            tlist,  &
            assign + 1, finish,  &
            inquire_stmt%blank,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (7)      ! DECIMAL
      IF (ALLOCATED(inquire_stmt%decimal)) THEN
        CALL already('DECIMAL')
      ELSE
        CALL ParseVariable(  &
            tlist,  &
            assign + 1, finish,  &
            inquire_stmt%decimal,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (8)      ! DELIM
      IF (ALLOCATED(inquire_stmt%delim)) THEN
        CALL already('DELIM')
      ELSE
        CALL ParseVariable(  &
            tlist,  &
            assign + 1, finish,  &
            inquire_stmt%delim,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (9)      ! DIRECT
      IF (ALLOCATED(inquire_stmt%direct)) THEN
        CALL already('DIRECT')
      ELSE
        CALL ParseVariable(  &
            tlist,  &
            assign + 1, finish,  &
            inquire_stmt%direct,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (10)     ! ENCODING
      IF (ALLOCATED(inquire_stmt%encoding)) THEN
        CALL already('ENCODING')
      ELSE
        CALL ParseVariable(  &
            tlist,  &
            assign + 1, finish,  &
            inquire_stmt%encoding,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (11)     ! ERR
      IF (ALLOCATED(inquire_stmt%err)) THEN
        CALL already('ERR')
      ELSE
        ALLOCATE(inquire_stmt%err)
        CALL ParseLabel(  &
            tlist,  &
            assign + 1,  &
            finish,  &
            inquire_stmt%err,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
        IF (Failure(sub_err_list)) RETURN
      END IF
      
    CASE (12)     ! EXIST
      IF (ALLOCATED(inquire_stmt%exist)) THEN
        CALL already('EXIST')
      ELSE
        CALL ParseVariable(  &
            tlist,  &
            assign + 1, finish,  &
            inquire_stmt%exist,  &
            sub_err_list)
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (13)     ! FORM
      IF (ALLOCATED(inquire_stmt%form)) THEN
        CALL already('FORM')
      ELSE
        CALL ParseVariable(  &
            tlist,  &
            assign + 1, finish,  &
            inquire_stmt%form,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (14)     ! FORMATTED
      IF (ALLOCATED(inquire_stmt%formatted)) THEN
        CALL already('FORM')
      ELSE
        CALL ParseVariable(  &
            tlist,  &
            assign + 1, finish,  &
            inquire_stmt%formatted,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (15)     ! ID
      IF (ALLOCATED(inquire_stmt%id)) THEN
        CALL already('ID')
      ELSE
        CALL ParseVariable(  &
            tlist,  &
            assign + 1, finish,  &
            inquire_stmt%id,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (16)     ! IOMSG
      IF (ALLOCATED(inquire_stmt%iomsg)) THEN
        CALL already('IOMSG')
      ELSE
        ! The syntax term is actually an /iomsg-variable/.
        CALL ParseVariable(  &
            tlist,  &
            assign + 1, finish,  &
            inquire_stmt%iomsg,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (17)     ! IOSTAT
      IF (ALLOCATED(inquire_stmt%iostat)) THEN
        CALL already('IOSTAT')
      ELSE
        CALL ParseVariable(  &
            tlist,  &
            assign + 1, finish,  &
            inquire_stmt%iostat,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (18)     ! NAME
      IF (ALLOCATED(inquire_stmt%name)) THEN
        CALL already('NAME')
      ELSE
        CALL ParseVariable(  &
            tlist,  &
            assign + 1, finish,  &
            inquire_stmt%name,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (19)     ! NAMED
      IF (ALLOCATED(inquire_stmt%named)) THEN
        CALL already('NAMED')
      ELSE
        CALL ParseVariable(  &
            tlist,  &
            assign + 1, finish,  &
            inquire_stmt%named,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (20)     ! NEXTREC
      IF (ALLOCATED(inquire_stmt%nextrec)) THEN
        CALL already('NEXTREC')
      ELSE
        CALL ParseVariable(  &
            tlist,  &
            assign + 1, finish,  &
            inquire_stmt%nextrec,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (21)     ! NUMBER
      IF (ALLOCATED(inquire_stmt%number)) THEN
        CALL already('NAMED')
      ELSE
        CALL ParseVariable(  &
            tlist,  &
            assign + 1, finish,  &
            inquire_stmt%number,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (22)     ! OPENED
      IF (ALLOCATED(inquire_stmt%opened)) THEN
        CALL already('OPENED')
      ELSE
        CALL ParseVariable(  &
            tlist,  &
            assign + 1, finish,  &
            inquire_stmt%opened,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (23)     ! PAD
      IF (ALLOCATED(inquire_stmt%pad)) THEN
        CALL already('PAD')
      ELSE
        CALL ParseVariable(  &
            tlist,  &
            assign + 1, finish,  &
            inquire_stmt%pad,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (24)     ! PENDING
      IF (ALLOCATED(inquire_stmt%pending)) THEN
        CALL already('PENDING')
      ELSE
        CALL ParseVariable(  &
            tlist,  &
            assign + 1, finish,  &
            inquire_stmt%pending,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (25)     ! POS
      IF (ALLOCATED(inquire_stmt%pos)) THEN
        CALL already('POS')
      ELSE
        CALL ParseVariable(  &
            tlist,  &
            assign + 1, finish,  &
            inquire_stmt%pos,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (26)     ! POSITION
      IF (ALLOCATED(inquire_stmt%position)) THEN
        CALL already('POSITION')
      ELSE
        CALL ParseVariable(  &
            tlist,  &
            assign + 1, finish,  &
            inquire_stmt%position,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (27)     ! READ
      IF (ALLOCATED(inquire_stmt%read)) THEN
        CALL already('READ')
      ELSE
        CALL ParseVariable(  &
            tlist,  &
            assign + 1, finish,  &
            inquire_stmt%read,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (28)     ! READWRITE
      IF (ALLOCATED(inquire_stmt%readwrite)) THEN
        CALL already('READWRITE')
      ELSE
        CALL ParseVariable(  &
            tlist,  &
            assign + 1, finish,  &
            inquire_stmt%readwrite,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (29)     ! RECL
      IF (ALLOCATED(inquire_stmt%recl)) THEN
        CALL already('RECL')
      ELSE
        CALL ParseVariable(  &
            tlist,  &
            assign + 1, finish,  &
            inquire_stmt%recl,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (30)     ! ROUND
      IF (ALLOCATED(inquire_stmt%round)) THEN
        CALL already('ROUND')
      ELSE
        CALL ParseVariable(  &
            tlist,  &
            assign + 1, finish,  &
            inquire_stmt%round,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (31)     ! SEQUENTIAL
      IF (ALLOCATED(inquire_stmt%sequential)) THEN
        CALL already('SEQUENTIAL')
      ELSE
        CALL ParseVariable(  &
            tlist,  &
            assign + 1, finish,  &
            inquire_stmt%sequential,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (32)     ! SIGN
      IF (ALLOCATED(inquire_stmt%sign)) THEN
        CALL already('SIGN')
      ELSE
        CALL ParseVariable(  &
            tlist,  &
            assign + 1, finish,  &
            inquire_stmt%sign,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (33)     ! SIZE
      IF (ALLOCATED(inquire_stmt%size)) THEN
        CALL already('SIZE')
      ELSE
        CALL ParseVariable(  &
            tlist,  &
            assign + 1, finish,  &
            inquire_stmt%size,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (34)     ! STREAM
      IF (ALLOCATED(inquire_stmt%stream)) THEN
        CALL already('STREAM')
      ELSE
        CALL ParseVariable(  &
            tlist,  &
            assign + 1, finish,  &
            inquire_stmt%stream,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (35)     ! UNFORMATTED
      IF (ALLOCATED(inquire_stmt%unformatted)) THEN
        CALL already('UNFORMATTED')
      ELSE
        CALL ParseVariable(  &
            tlist,  &
            assign + 1, finish,  &
            inquire_stmt%unformatted,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (36)     ! WRITE
      IF (ALLOCATED(inquire_stmt%write)) THEN
        CALL already('WRITE')
      ELSE
        CALL ParseVariable(  &
            tlist,  &
            assign + 1, finish,  &
            inquire_stmt%write,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE DEFAULT
      STOP 'Unknown /inquire-spec/.'
      
    END SELECT
    
  CONTAINS
    
    SUBROUTINE already(text)
      USE ErrorCodes
      !-------------------------------------------------------------------------
      ! Arguments
      CHARACTER(*), INTENT(IN) :: text
      !*************************************************************************
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT='C947',  &
          LOCATION=QueryLocation(tlist(start)),  &
          MSG=text // ' has already been specified for this /inquire-stmt/.' )
    END SUBROUTINE already
    
    SUBROUTINE add_file_unit_err
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT='C948',  &
          LOCATION=QueryLocation(tlist(start)),  &
          MSG='An /inquire-spec-list/ shall contain one FILE= specifier &
            &or one /file-unit-number/, but not both.' )
    END SUBROUTINE add_file_unit_err
    
  END SUBROUTINE parse_inquire_spec
  
  
  !*****************************************************************************
  !!
  !> Implementation of InquireStmt%Dump - write a debugging representation 
  !! of the results of parsing an inquire statement to a logical unit.
  !!
  !! @param[in]     pnode             The ParseNode to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix for each record.
  
  SUBROUTINE inquire_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(InquireStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%unit)) THEN
      WRITE (unit, "(A,'UNIT = ')", ADVANCE='NO') pfx
      CALL WriteExpr(unit, pnode%unit)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%file)) THEN
      WRITE (unit, "(A,'FILE = ')", ADVANCE='NO') pfx
      CALL WriteExpr(unit, pnode%file)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%access)) THEN
      WRITE (unit, "(A,'ACCESS = ')", ADVANCE='NO') pfx
      CALL WriteVariable(unit, pnode%access)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%action)) THEN
      WRITE (unit, "(A,'ACTION = ')", ADVANCE='NO') pfx
      CALL WriteVariable(unit, pnode%action)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%asynchronous)) THEN
      WRITE (unit, "(A,'ASYNCHRONOUS = ')", ADVANCE='NO') pfx
      CALL WriteVariable(unit, pnode%asynchronous)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%blank)) THEN
      WRITE (unit, "(A,'BLANK = ')", ADVANCE='NO') pfx
      CALL WriteVariable(unit, pnode%blank)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%decimal)) THEN
      WRITE (unit, "(A,'DECIMAL = ')", ADVANCE='NO') pfx
      CALL WriteVariable(unit, pnode%decimal)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%delim)) THEN
      WRITE (unit, "(A,'DELIM = ')", ADVANCE='NO') pfx
      CALL WriteVariable(unit, pnode%delim)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%direct)) THEN
      WRITE (unit, "(A,'DIRECT = ')", ADVANCE='NO') pfx
      CALL WriteVariable(unit, pnode%direct)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%encoding)) THEN
      WRITE (unit, "(A,'ENCODING = ')", ADVANCE='NO') pfx
      CALL WriteVariable(unit, pnode%encoding)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%err)) THEN
      WRITE (unit, "(A,'ERR = ',I0)") pfx, .ValueOf. pnode%err
    END IF
    IF (ALLOCATED(pnode%exist)) THEN
      WRITE (unit, "(A,'EXIST = ')", ADVANCE='NO') pfx
      CALL WriteVariable(unit, pnode%exist)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%form)) THEN
      WRITE (unit, "(A,'FORM = ')", ADVANCE='NO') pfx
      CALL WriteVariable(unit, pnode%form)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%formatted)) THEN
      WRITE (unit, "(A,'FORMATTED = ')", ADVANCE='NO') pfx
      CALL WriteVariable(unit, pnode%formatted)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%id)) THEN
      WRITE (unit, "(A,'ID = ')", ADVANCE='NO') pfx
      CALL WriteVariable(unit, pnode%id)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%iomsg)) THEN
      WRITE (unit, "(A,'IOMSG = ')", ADVANCE='NO') pfx
      CALL WriteVariable(unit, pnode%iomsg)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%iostat)) THEN
      WRITE (unit, "(A,'IOSTAT = ')", ADVANCE='NO') pfx
      CALL WriteVariable(unit, pnode%iostat)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%name)) THEN
      WRITE (unit, "(A,'NAME = ')", ADVANCE='NO') pfx
      CALL WriteVariable(unit, pnode%name)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%named)) THEN
      WRITE (unit, "(A,'NAMED = ')", ADVANCE='NO') pfx
      CALL WriteVariable(unit, pnode%named)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%nextrec)) THEN
      WRITE (unit, "(A,'NEXTREC = ')", ADVANCE='NO') pfx
      CALL WriteVariable(unit, pnode%nextrec)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%number)) THEN
      WRITE (unit, "(A,'NUMBER = ')", ADVANCE='NO') pfx
      CALL WriteVariable(unit, pnode%number)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%opened)) THEN
      WRITE (unit, "(A,'OPENED = ')", ADVANCE='NO') pfx
      CALL WriteVariable(unit, pnode%opened)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%pad)) THEN
      WRITE (unit, "(A,'PAD = ')", ADVANCE='NO') pfx
      CALL WriteVariable(unit, pnode%pad)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%pending)) THEN
      WRITE (unit, "(A,'PENDING = ')", ADVANCE='NO') pfx
      CALL WriteVariable(unit, pnode%pending)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%pos)) THEN
      WRITE (unit, "(A,'POS = ')", ADVANCE='NO') pfx
      CALL WriteVariable(unit, pnode%pos)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%position)) THEN
      WRITE (unit, "(A,'POSITION = ')", ADVANCE='NO') pfx
      CALL WriteVariable(unit, pnode%position)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%read)) THEN
      WRITE (unit, "(A,'READ = ')", ADVANCE='NO') pfx
      CALL WriteVariable(unit, pnode%read)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%readwrite)) THEN
      WRITE (unit, "(A,'READWRITE = ')", ADVANCE='NO') pfx
      CALL WriteVariable(unit, pnode%readwrite)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%recl)) THEN
      WRITE (unit, "(A,'RECL = ')", ADVANCE='NO') pfx
      CALL WriteVariable(unit, pnode%recl)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%round)) THEN
      WRITE (unit, "(A,'ROUND = ')", ADVANCE='NO') pfx
      CALL WriteVariable(unit, pnode%round)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%sequential)) THEN
      WRITE (unit, "(A,'SEQUENTIAL = ')", ADVANCE='NO') pfx
      CALL WriteVariable(unit, pnode%sequential)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%sign)) THEN
      WRITE (unit, "(A,'SIGN = ')", ADVANCE='NO') pfx
      CALL WriteVariable(unit, pnode%sign)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%size)) THEN
      WRITE (unit, "(A,'SIZE = ')", ADVANCE='NO') pfx
      CALL WriteVariable(unit, pnode%size)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%stream)) THEN
      WRITE (unit, "(A,'STREAM = ')", ADVANCE='NO') pfx
      CALL WriteVariable(unit, pnode%stream)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%unformatted)) THEN
      WRITE (unit, "(A,'UNFORMATTED = ')", ADVANCE='NO') pfx
      CALL WriteVariable(unit, pnode%unformatted)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%write)) THEN
      WRITE (unit, "(A,'WRITE = ')", ADVANCE='NO') pfx
      CALL WriteVariable(unit, pnode%write)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%iolength)) THEN
      WRITE (unit, "(A,'STATUS = ')", ADVANCE='NO') pfx
      CALL WriteVariable(unit, pnode%iolength)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%output_list)) THEN
      WRITE (unit, "(A,'/output-item-list/':)") pfx
      CALL DumpIOItemList(pnode%output_list, unit, pfx // '  ')
    END IF
    
  END SUBROUTINE inquire_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of InquireStmt%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE inquire_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(InquireStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%unit)) CALL pnode%unit%Query(visitor)
    IF (ALLOCATED(pnode%file)) CALL pnode%file%Query(visitor)
    IF (ALLOCATED(pnode%access)) CALL pnode%access%Query(visitor)
    IF (ALLOCATED(pnode%action)) CALL pnode%action%Query(visitor)
    IF (ALLOCATED(pnode%asynchronous)) CALL pnode%asynchronous%Query(visitor)
    IF (ALLOCATED(pnode%blank)) CALL pnode%blank%Query(visitor)
    IF (ALLOCATED(pnode%decimal)) CALL pnode%decimal%Query(visitor)
    IF (ALLOCATED(pnode%delim)) CALL pnode%delim%Query(visitor)
    IF (ALLOCATED(pnode%direct)) CALL pnode%direct%Query(visitor)
    IF (ALLOCATED(pnode%encoding)) CALL pnode%encoding%Query(visitor)
    ! err
    IF (ALLOCATED(pnode%exist)) CALL pnode%exist%Query(visitor)
    IF (ALLOCATED(pnode%form)) CALL pnode%form%Query(visitor)
    IF (ALLOCATED(pnode%formatted)) CALL pnode%formatted%Query(visitor)
    IF (ALLOCATED(pnode%id)) CALL pnode%id%Query(visitor)
    IF (ALLOCATED(pnode%iomsg)) CALL pnode%iomsg%Query(visitor)
    IF (ALLOCATED(pnode%iostat)) CALL pnode%iostat%Query(visitor)
    IF (ALLOCATED(pnode%name)) CALL pnode%name%Query(visitor)
    IF (ALLOCATED(pnode%named)) CALL pnode%named%Query(visitor)
    IF (ALLOCATED(pnode%nextrec)) CALL pnode%nextrec%Query(visitor)
    IF (ALLOCATED(pnode%number)) CALL pnode%number%Query(visitor)
    IF (ALLOCATED(pnode%opened)) CALL pnode%opened%Query(visitor)
    IF (ALLOCATED(pnode%pad)) CALL pnode%pad%Query(visitor)
    IF (ALLOCATED(pnode%pending)) CALL pnode%pending%Query(visitor)
    IF (ALLOCATED(pnode%pos)) CALL pnode%pos%Query(visitor)
    IF (ALLOCATED(pnode%position)) CALL pnode%position%Query(visitor)
    IF (ALLOCATED(pnode%read)) CALL pnode%read%Query(visitor)
    IF (ALLOCATED(pnode%readwrite)) CALL pnode%readwrite%Query(visitor)
    IF (ALLOCATED(pnode%recl)) CALL pnode%recl%Query(visitor)
    IF (ALLOCATED(pnode%round)) CALL pnode%round%Query(visitor)
    IF (ALLOCATED(pnode%sequential)) CALL pnode%sequential%Query(visitor)
    IF (ALLOCATED(pnode%sign)) CALL pnode%sign%Query(visitor)
    IF (ALLOCATED(pnode%size)) CALL pnode%size%Query(visitor)
    IF (ALLOCATED(pnode%stream)) CALL pnode%stream%Query(visitor)
    IF (ALLOCATED(pnode%unformatted)) CALL pnode%unformatted%Query(visitor)
    IF (ALLOCATED(pnode%write)) CALL pnode%write%Query(visitor)
    IF (ALLOCATED(pnode%iolength)) CALL pnode%iolength%Query(visitor)
    
    IF (ALLOCATED(pnode%output_list)) THEN
      CALL QueryIOItemList(pnode%output_list, visitor)
    END IF
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE inquire_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of InquireStmt%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[in,out] pnode             The parse node element to modify.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE inquire_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(InquireStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%unit)) CALL pnode%unit%Modify(visitor)
    IF (ALLOCATED(pnode%file)) CALL pnode%file%Modify(visitor)
    IF (ALLOCATED(pnode%access)) CALL pnode%access%Modify(visitor)
    IF (ALLOCATED(pnode%action)) CALL pnode%action%Modify(visitor)
    IF (ALLOCATED(pnode%asynchronous)) CALL pnode%asynchronous%Modify(visitor)
    IF (ALLOCATED(pnode%blank)) CALL pnode%blank%Modify(visitor)
    IF (ALLOCATED(pnode%decimal)) CALL pnode%decimal%Modify(visitor)
    IF (ALLOCATED(pnode%delim)) CALL pnode%delim%Modify(visitor)
    IF (ALLOCATED(pnode%direct)) CALL pnode%direct%Modify(visitor)
    IF (ALLOCATED(pnode%encoding)) CALL pnode%encoding%Modify(visitor)
    ! err
    IF (ALLOCATED(pnode%exist)) CALL pnode%exist%Modify(visitor)
    IF (ALLOCATED(pnode%form)) CALL pnode%form%Modify(visitor)
    IF (ALLOCATED(pnode%formatted)) CALL pnode%formatted%Modify(visitor)
    IF (ALLOCATED(pnode%id)) CALL pnode%id%Modify(visitor)
    IF (ALLOCATED(pnode%iomsg)) CALL pnode%iomsg%Modify(visitor)
    IF (ALLOCATED(pnode%iostat)) CALL pnode%iostat%Modify(visitor)
    IF (ALLOCATED(pnode%name)) CALL pnode%name%Modify(visitor)
    IF (ALLOCATED(pnode%named)) CALL pnode%named%Modify(visitor)
    IF (ALLOCATED(pnode%nextrec)) CALL pnode%nextrec%Modify(visitor)
    IF (ALLOCATED(pnode%number)) CALL pnode%number%Modify(visitor)
    IF (ALLOCATED(pnode%opened)) CALL pnode%opened%Modify(visitor)
    IF (ALLOCATED(pnode%pad)) CALL pnode%pad%Modify(visitor)
    IF (ALLOCATED(pnode%pending)) CALL pnode%pending%Modify(visitor)
    IF (ALLOCATED(pnode%pos)) CALL pnode%pos%Modify(visitor)
    IF (ALLOCATED(pnode%position)) CALL pnode%position%Modify(visitor)
    IF (ALLOCATED(pnode%read)) CALL pnode%read%Modify(visitor)
    IF (ALLOCATED(pnode%readwrite)) CALL pnode%readwrite%Modify(visitor)
    IF (ALLOCATED(pnode%recl)) CALL pnode%recl%Modify(visitor)
    IF (ALLOCATED(pnode%round)) CALL pnode%round%Modify(visitor)
    IF (ALLOCATED(pnode%sequential)) CALL pnode%sequential%Modify(visitor)
    IF (ALLOCATED(pnode%sign)) CALL pnode%sign%Modify(visitor)
    IF (ALLOCATED(pnode%size)) CALL pnode%size%Modify(visitor)
    IF (ALLOCATED(pnode%stream)) CALL pnode%stream%Modify(visitor)
    IF (ALLOCATED(pnode%unformatted)) CALL pnode%unformatted%Modify(visitor)
    IF (ALLOCATED(pnode%write)) CALL pnode%write%Modify(visitor)
    IF (ALLOCATED(pnode%iolength)) CALL pnode%iolength%Modify(visitor)
    
    IF (ALLOCATED(pnode%output_list)) THEN
      CALL modifyIOItemList(pnode%output_list, visitor)
    END IF
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE inquire_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of InquireStmt%What - return a descriptive string 
  !! for the parse node.
  !!
  !! @returns 'inquire-stmt'
  
  FUNCTION inquire_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'inquire-stmt'
    
  END FUNCTION inquire_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of InquireStmt%Check - check the parse node for 
  !! conformance with language standards.
  !!
  !! @param[in]     pnode             The parse node to check.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE inquire_Check(pnode, err_list)
    
    USE Errors
    USE ErrorCodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(InquireStmt), INTENT(IN) :: pnode
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (ALLOCATED(pnode%id)) THEN
      IF (.NOT. ALLOCATED(pnode%pending)) THEN
        CALL Add( err_list,  &
            CODE=errSyntax,  &
            COMPONENT='C950',  &
            LOCATION=.Locationof. pnode,  &
            MSG='If "ID=" appears then "PENDING=" shall appear.' )
      END IF
    END IF
    
  END SUBROUTINE inquire_Check
  
END MODULE InquireStmts
