! $Id: IntentStmts.f90 2363 2016-08-22 01:29:48Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the IntentStmts module.


!*******************************************************************************
!!
!> Procedures for parsing /intent-stmt/'s.

MODULE IntentStmts
  
  USE Identifiers
  USE ParseNodes
  USE Strings
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseIntentStmt
  
  !-----------------------------------------------------------------------------
  
  !> Stores the results of parsing an /intent-stmt/.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: IntentStmt
    !> The names in the statement.
    TYPE(Identifier), ALLOCATABLE :: name_list(:)
    !> The intent specification - an iint constant from the IntentSpecs module.
    INTEGER :: intent
  CONTAINS
    PROCEDURE :: Dump => intent_Dump
    PROCEDURE, NOPASS :: What => intent_What
    PROCEDURE :: Query => intent_Query
    PROCEDURE :: Modify => intent_Modify
  END TYPE IntentStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /intent-stmt/ (R546). 
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    intent_stmt       The results of parsing the statement.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseIntentStmt(tlist, intent_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE IntentSpecs
    USE ParseUtils
    USE Tokens
  
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(IntentStmt), INTENT(OUT) :: intent_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule for an /intent-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R546'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: closing        ! Index of the closing delimiter.
    INTEGER :: opt            ! Expect option.
    INTEGER :: it             ! Current token index.
    
    ! Error list for child procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /intent-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------  
    ! INTENT...
    
    CALL ExpectKeyword(tlist, 1, scck_'INTENT', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! Locate the range of the () bit.
    
    CALL Expect(tlist, 2, ittDelimiter, scck_'(', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL ExpectClosing(tlist, 2, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! Parse the intent specification
    
    CALL ParseIntentSpec(tlist(3:closing-1), intent_stmt%intent, sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
    !---------------------------------------------------------------------------
    ! :: or /dummy-arg-name/
    
    CALL Expect(  &
        tlist,  &
        closing + 1,  &
        [ittDeclare, ittName],  &
        [scck_'::', scck_'  '],  &
        '"::" or /dummy-arg-name/',  &
        opt,  &
        err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    IF (opt == 1) THEN
      ! Skip over declare token.
      it = closing + 2
    ELSE
      ! /dummy-arg-name-list/ starts immediately after the INTENT clause.
      it = closing + 1
    END IF    
    ! it points to the start of the /dummy-arg-name-list/ here.
    
    !---------------------------------------------------------------------------
    ! The /dummy-arg-name-list.
    
    CALL ParseNameList(  &
        tlist,  &
        it, SIZE(tlist),  &
        '/dummy-arg-name/',  &
        intent_stmt%name_list,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
    ! That's it!
    
  END SUBROUTINE ParseIntentStmt
  
  
  !*****************************************************************************
  !!
  !> Implementation of IntentStmt%Dump - write a debugging representation 
  !! of the results of parsing an intent statement to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE intent_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(IntentStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Name index.
    
    !***************************************************************************
    
    WRITE (unit, "(A,'/intent-spec/ = ')", ADVANCE='NO') pfx
    SELECT CASE (pnode%intent)
    CASE (1) ; WRITE (unit, "('IN')")
    CASE (2) ; WRITE (unit, "('OUT')")
    CASE (3) ; WRITE (unit, "('INOUT')")
    CASE DEFAULT
      STOP 'Internal error in IntentStmts%%intent_Dump, &
          &bad /intent-spec/.'
    END SELECT
      
    WRITE (unit, "(A,'/dummy-arg-name-list/')") pfx
    DO i = 1, SIZE(pnode%name_list)
      WRITE (unit, "(A,A)") pfx // '  ', .NameOf. pnode%name_list(i)
    END DO
    
  END SUBROUTINE intent_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of IntentStmt%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE intent_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(IntentStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE intent_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of IntentStmt%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[in,out] pnode             The parse node element to query.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE intent_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(IntentStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE intent_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of IntentStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'intent-stmt'.
  
  FUNCTION intent_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'intent-stmt'
    
  END FUNCTION intent_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of IntrinsicStmt%Advertise - advertise the identifiers 
  !! referenced or defined by the construct.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE intent_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    USE IntentSpecs
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(IntentStmt), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: nature         ! Identifier nature.
    INTEGER :: i              ! Identifier index.
    
    !***************************************************************************
    
    ALLOCATE(ident_list(SIZE(pnode%name_list)))
    
    SELECT CASE (pnode%intent)
    CASE (iintIn)           ; nature = NatureDefIntentIn
    CASE (iintOut)          ; nature = NatureDefIntentOut
    CASE (iintInOut)        ; nature = NatureDefIntentInOut
    END SELECT
    
    DO i = 1, SIZE(pnode%name_list)
      ident_list(i) = IdentifierNature(  &
          NAME=.NameOf. pnode%name_list(i),  &
          NATURES=[nature], &
          LOCATIONS=[.LocationOf. pnode%name_list(i)] )
    END DO
    
  END SUBROUTINE intent_Advertise
  
END MODULE IntentStmts
