! $Id: InterfaceBlocks.f90 2887 2019-04-06 22:52:29Z ian $
! ff08 source code copyright 2014 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the InterfaceBlocks module


!*******************************************************************************
!!
!> Procedures and types for parsing /interface-blocks/.

MODULE InterfaceBlocks
  
  USE FunctionStmts
  USE GenericSpecs
  USE Identifiers
  USE ParseNodes
  USE SpecificationParts
  USE SubroutineStmts
  
  IMPLICIT NONE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseInterfaceBlock
  PUBLIC :: ParseFunctionInterfaceBody
  PUBLIC :: ParseSubroutineInterfaceBody
  
  !-----------------------------------------------------------------------------
  
  !> An interface specification.
  TYPE, EXTENDS(SCParseNode), PUBLIC, ABSTRACT :: InterfaceSpecification
  END TYPE InterfaceSpecification
  
  ! Interfaces for deferred bindings in interface_specification
  ABSTRACT INTERFACE
    !> Interface for interface_specification%Advertise - report the 
    !! identifiers defined or referenced by the node, and the nature of that 
    !! definition or reference.
    !!
    !! @param[in]     pnode           The interface specification to advertise.
    !!
    !! @param[out]    ident_list      The identifiers advertise by the node.
    SUBROUTINE is_Advertise(pnode, ident_list)
      USE IdentifierNatures
      IMPORT :: InterfaceSpecification
      IMPLICIT NONE
      !-------------------------------------------------------------------------
      CLASS(InterfaceSpecification), INTENT(IN) :: pnode
      TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    END SUBROUTINE is_Advertise
  END INTERFACE
  
  !> An item in an array of interface specifications.
  TYPE interface_spec_element
    !> The item.
    CLASS(InterfaceSpecification), ALLOCATABLE :: item
  CONTAINS
    PROCEDURE :: Advertise => ise_Advertise
  END TYPE interface_spec_element
  
  !> An interface specification that is a procedure statement.
  TYPE, EXTENDS(InterfaceSpecification), PUBLIC :: ProcedureStmt
    !> Flag to indicate that the procedure statement had the MODULE keyword.
    LOGICAL :: has_module
    !> The procedure names.
    !!
    !! Always allocated, always size greater than zero.
    TYPE(Identifier), ALLOCATABLE :: procedure_names(:)
  CONTAINS
    PROCEDURE :: Dump => ps_Dump
    PROCEDURE :: Query => ps_Query
    PROCEDURE :: Modify => ps_Modify
    PROCEDURE, NOPASS :: What => ps_What
    PROCEDURE :: Advertise => ps_Advertise
  END TYPE ProcedureStmt
  
  !> An interface specification that is an interface body.
  TYPE, EXTENDS(InterfaceSpecification), PUBLIC, ABSTRACT :: InterfaceBody
    !> The specification part of the program unit.
    TYPE(SpecificationPart) :: spec
    !> The name in the end statement of the interface body.  Not 
    !! allocated if there was no such name.
    TYPE(Identifier), ALLOCATABLE, PRIVATE :: end_name
    !> Indicates whether the containing interface block had the 
    !! ABSTRACT keyword in the opening statement.
    !!
    !! This object is owned by the parent interface block.
    LOGICAL, POINTER, PRIVATE :: abstract => NULL()
  CONTAINS
    PROCEDURE(ib_Who), DEFERRED :: Who
  END TYPE InterfaceBody
  
  ! Interfaces for bindings in InterfaceBody.
  ABSTRACT INTERFACE
    !> Query the identifier for the interface body.
    !!
    !! @param[in]     pnode           The parse node for the interface body.
    !!
    !! @returns The identifier of the function or subroutine declared by 
    !! the interface body.
    FUNCTION ib_Who(pnode) RESULT(name)
      IMPORT :: InterfaceBody
      IMPLICIT NONE
      !-------------------------------------------------------------------------
      CLASS(InterfaceBody), INTENT(IN) :: pnode
      CHARACTER(:), ALLOCATABLE :: name
    END FUNCTION ib_Who
  END INTERFACE
  
  !> An interface specification that is a function body.
  TYPE, EXTENDS(InterfaceBody), PUBLIC :: FunctionInterfaceBody
    !> The results of parsing the function statement.
    TYPE(FunctionStmt) :: fn_stmt
  CONTAINS
    PROCEDURE :: Dump => fun_Dump
    PROCEDURE :: Query => fun_Query
    PROCEDURE :: Modify => fun_Modify
    PROCEDURE, NOPASS :: What => fun_What
    PROCEDURE :: Advertise => fun_Advertise
    PROCEDURE :: Check => fun_Check
    
    PROCEDURE :: Who => fun_Who
  END TYPE FunctionInterfaceBody
  
  !> An interface specification that is a subroutine body.
  TYPE, EXTENDS(InterfaceBody) :: SubroutineInterfaceBody
    !> The results of parsing the opening subroutine statement.
    TYPE(SubroutineStmt) :: sub_stmt
  CONTAINS
    PROCEDURE :: Dump => sub_Dump
    PROCEDURE :: Query => sub_Query
    PROCEDURE :: Modify => sub_Modify
    PROCEDURE, NOPASS :: What => sub_What
    PROCEDURE :: Advertise => sub_Advertise
    PROCEDURE :: Check => sub_Check
    
    PROCEDURE :: Who => sub_Who
  END TYPE SubroutineInterfaceBody
  
  !-----------------------------------------------------------------------------
  
  !> Stores the results of parsing an interface block.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: InterfaceBlock
    !> Flag to indicate that the ABSTRACT keyword was supplied on the 
    !! opening statement.
    !!
    !! This is managed through a pointer so that we can reliably 
    !! share it with the interface bodies.  The object is owned by 
    !! this type.  Using a pointer object for this is serious overkill 
    !! as we only have a single logical attribute (we could just pass 
    !! the value of the attribute to the parsing routines for the 
    !! various interface bodies), but this approach is consistent 
    !! with how we have handled other attributes.
    !!
    !! Always associated after construction.
    LOGICAL, POINTER :: abstract => NULL()
    
    !> The generic spec, if any, provided on the interface block.
    !!
    !! Not allocated if there was no generic spec.
    TYPE(GenericSpec), ALLOCATABLE :: generic_spec
    
    !> The procedure statments and interface bodies.
    TYPE(interface_spec_element), ALLOCATABLE :: specs(:)
  CONTAINS
    PROCEDURE :: Dump => intf_Dump
    PROCEDURE, NOPASS :: What => intf_What
    PROCEDURE :: Query => intf_Query
    PROCEDURE :: Modify => intf_Modify
    PROCEDURE :: Advertise => intf_Advertise
    
    FINAL :: interfaceblock_final_scalar
  END TYPE InterfaceBlock
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse the statement tree node for an interface block.
  !!
  !! @param[in]     st                The statement tree node for the 
  !! interface block.
  !!
  !! @param[out]    intf_block        The results of parsing the block.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! Recursive because a specification inside an interface body could 
  !! itself be an interface block.
  
  RECURSIVE SUBROUTINE ParseInterfaceBlock(st, intf_block, err_list)
    
    USE Errors
    USE InterfaceStmts
    USE InterfaceStNodes
    USE InterfaceBodyStNodes
    USE SingleStNodes
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(InterfaceStNode), INTENT(INOUT) :: st
    TYPE(InterfaceBlock), INTENT(OUT) :: intf_block
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Error list for child parse calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    INTEGER :: ib             ! Interface body (or procedure statement) index.
    
    ! Temporary for procedure statement interface specifications.
    TYPE(ProcedureStmt), ALLOCATABLE :: proc_stmt
    
    ! Temporary for function interface body interface specifications.
    TYPE(FunctionInterfaceBody), ALLOCATABLE :: fn_body
    
    ! Temporary for subroutine interface body interface specifications.
    TYPE(SubroutineInterfaceBody), ALLOCATABLE :: sub_body
    
    ! General temporary for interface specifications.
    CLASS(InterfaceSpecification), ALLOCATABLE :: intf_spec
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    ! Set up attributes of the interface block that are shared with the 
    ! contained bodies.
    ALLOCATE(intf_block%abstract)
    intf_block%abstract = .FALSE.
    
    !---------------------------------------------------------------------------
    ! Opening statement
    
    CALL ParseInterfaceStmt(  &
        st%first%tlist,  &
        intf_block%abstract,  &
        intf_block%generic_spec,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list)
    
    !---------------------------------------------------------------------------
    ! The bodies and procedure statements.
    
    ALLOCATE(intf_block%specs(0))
    DO ib = 1, SIZE(st%bodies)
      SELECT TYPE (item => st%bodies(ib)%item)
      TYPE IS (FunctionInterfaceBodyStNode)
        ALLOCATE(fn_body)
        CALL ParseFunctionInterfaceBody(  &
            item,  &
            intf_block%abstract,  &
            fn_body,  &
            sub_err_list )
        IF (Failure(sub_err_list)) THEN
          DEALLOCATE(fn_body)
        ELSE
          CALL MOVE_ALLOC(fn_body, intf_spec)
        END IF
        
      TYPE IS (SubroutineInterfaceBodyStNode)
        ALLOCATE(sub_body)
        CALL ParseSubroutineInterfaceBody(  &
            item,  &
            intf_block%abstract,  &
            sub_body,  &
            sub_err_list )
        IF (Failure(sub_err_list)) THEN
          DEALLOCATE(sub_body)
        ELSE
          CALL MOVE_ALLOC(sub_body, intf_spec)
        END IF
        
      TYPE IS (SingleStNode)
        IF (item%stmt%ist /= istProcedure)  &
            STOP 'Internal error in InterfaceBlocks%%ParseInterfaceBlock, &
              &unknown type of interface body (single statement).'
        ALLOCATE(proc_stmt)
        CALL ParseProcedureStmt(  &
            item%stmt%tlist,  &
            proc_stmt%has_module,  &
            proc_stmt%procedure_names,  &
            sub_err_list )
        IF (Failure(sub_err_list)) THEN
          DEALLOCATE(proc_stmt)
        ELSE
          CALL MOVE_ALLOC(proc_stmt, intf_spec)
        END IF
        
      CLASS DEFAULT
        STOP 'Internal error in InterfaceBlocks%%ParseInterfaceBlock, &
            &unknown type of interface body.'
        
      END SELECT
      CALL Add(err_list, sub_err_list)
      IF (.NOT. Failure(sub_err_list)) THEN
        CALL intf_spec%Init(st%bodies(ib)%item)
        CALL add_spec(intf_block%specs, intf_spec)
      END IF
    END DO
    
    !---------------------------------------------------------------------------
    ! Closing statement.
    
    CALL ParseEndInterfaceStmt(  &
        st%last%tlist,  &
        intf_block%generic_spec,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list)
    
  END SUBROUTINE ParseInterfaceBlock
  
  
  !*****************************************************************************
  !!
  !> Implementation of InterfaceBlock%Dump - write a debugging 
  !! representation of the ParseNode to a logical unit.
  !!
  !! @param[in]     pnode             The ParseNode to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix for each record.
  
  ! Recursive because an interface body could contain another interface body.
  RECURSIVE SUBROUTINE intf_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(InterfaceBlock), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: is             ! Interface specification index.
    
    !***************************************************************************
    
    IF (pnode%abstract) WRITE (unit, "(A,'ABSTRACT')") pfx
    
    IF (ALLOCATED(pnode%generic_spec)) THEN
      WRITE (unit, "(A,'/generic-spec/ = ')", ADVANCE='NO') pfx
      CALL WriteGenericSpec(unit, pnode%generic_spec)
      WRITE (unit, "()")
    END IF
    
    !---------------------------------------------------------------------------
    ! Interface specifications.
    
    WRITE (unit, "(A,'interface specifications:')") pfx
    DO is = 1, SIZE(pnode%specs)
      CALL pnode%specs(is)%item%Dump(unit, pfx // '  ')
    END DO
    
  END SUBROUTINE intf_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of InterfaceBlock%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'interface-block'.
  
  FUNCTION intf_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'interface-block'
    
  END FUNCTION intf_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of InterfaceBlock%Query.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  RECURSIVE SUBROUTINE intf_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(InterfaceBlock), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: is             ! interface specification index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%generic_spec)) THEN
      CALL pnode%generic_spec%Query(visitor)
    END IF
    
    DO is = 1, SIZE(pnode%specs)
      CALL pnode%specs(is)%item%Query(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE intf_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of InterfaceBlock%Modify.
  !!
  !! @param[in,out] pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  RECURSIVE SUBROUTINE intf_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(InterfaceBlock), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: is             ! interface specification index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%generic_spec)) THEN
      CALL pnode%generic_spec%Modify(visitor)
    END IF
    
    DO is = 1, SIZE(pnode%specs)
      CALL pnode%specs(is)%item%Modify(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE intf_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of InterfaceBlock%Advertise - advertise the identifiers 
  !! referenced or defined by the interface block into the host scope.
  !!
  !! @param[in]     pnode             The node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  ! An interface block may contain an interface body that has an interface 
  ! block in its specification part, hence this is RECURSIVE.
  RECURSIVE SUBROUTINE intf_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(InterfaceBlock), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%generic_spec)) THEN
      ALLOCATE(ident_list(1))
      ident_list(1) = IdentifierNature(  &
              pnode%generic_spec,  &
              NATURES=[NatureDefGeneric],  &
              LOCATIONS=[pnode%generic_spec%GetLocation()] )
    ELSE
      ALLOCATE(ident_list(0))
    END IF
    
  END SUBROUTINE intf_Advertise
  
  
  !*****************************************************************************
  !!
  !> Scalar finalizer for an InterfaceBlock object.
  
  SUBROUTINE interfaceblock_final_scalar(intf_block)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(InterfaceBlock), INTENT(INOUT) :: intf_block
    
    !***************************************************************************
    
    IF (ASSOCIATED(intf_block%abstract)) DEALLOCATE(intf_block%abstract)
    
  END SUBROUTINE interfaceblock_final_scalar
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for interface_spec_element.
  
  
  !*****************************************************************************
  !!
  !> Add a new specification (by moving) to a list of interface specifications.
  !!
  !! @param[in]     specs             The list of interface specifications 
  !! to be added to.  Must be allocated on input.
  !!
  !! @param[in]     item              The specification to add.
  
  SUBROUTINE add_spec(specs, item)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(interface_spec_element), INTENT(INOUT), ALLOCATABLE :: specs(:)
    CLASS(InterfaceSpecification), INTENT(INOUT), ALLOCATABLE :: item
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Temporary for growing array.
    TYPE(interface_spec_element), ALLOCATABLE :: tmp(:)
    
    INTEGER :: i              ! Array index.
    
    !***************************************************************************
    
    ALLOCATE(tmp(LBOUND(specs,1):UBOUND(specs,1)+1))
    DO i = LBOUND(specs, 1), UBOUND(specs, 1)
      CALL MOVE_ALLOC(specs(i)%item, tmp(i)%item)
    END DO
    
    CALL MOVE_ALLOC(item, tmp(UBOUND(tmp,1))%item)
    CALL MOVE_ALLOC(tmp, specs)
    
  END SUBROUTINE add_spec
  
  
  !*****************************************************************************
  !!
  !> Implementation of interface_spec_element%Advertise - advertise the 
  !! identifiers referenced or defined by the interface block element into 
  !! the host scope.
  !!
  !! @param[in]     pnode             The node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  ! An interface block may contain an interface body that has an interface 
  ! block in its specification part, hence this is RECURSIVE.
  RECURSIVE SUBROUTINE ise_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(interface_spec_element), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    CALL pnode%item%Advertise(ident_list)
    
  END SUBROUTINE ise_Advertise
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for ProcedureStmt.
  
  
  !*****************************************************************************
  !!
  !> Implementation of ProcedureStmt%Dump - write a debugging 
  !! representation of the results of parsing the statement to a 
  !! logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE ps_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ProcedureStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Procedure name index.
    
    !***************************************************************************
    
    IF (pnode%has_module) THEN
      WRITE (unit, "(A,'MODULE procedures')", ADVANCE='NO') pfx
    ELSE
      WRITE (unit, "(A,'procedures')", ADVANCE='NO') pfx
    END IF
    DO i = 1, SIZE(pnode%procedure_names)
      WRITE (unit, "(1X,A)", ADVANCE='NO') .NameOf. pnode%procedure_names(i)
    END DO
    WRITE (unit, "()")
    
  END SUBROUTINE ps_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of ProcedureStmt%Query.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  SUBROUTINE ps_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(ProcedureStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE ps_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of ProcedureStmt%Modify.
  !!
  !! @param[in,out] pnode             The parse node being MOdified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  SUBROUTINE ps_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(ProcedureStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE ps_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of ProcedureStmt%What - return a descriptive string 
  !! for the parse node.
  !!
  !! @returns The syntax rule name for the node.
  
  FUNCTION ps_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'procedure-stmt'
    
  END FUNCTION ps_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of ProcedureStmt%Advertise - advertise the 
  !! identifiers referenced or defined by the procedure statement into 
  !! the host scope.
  !!
  !! @param[in]     pnode             The node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE ps_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ProcedureStmt), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! procedure name index.
    
    !***************************************************************************
    
    ALLOCATE(ident_list(SIZE(pnode%procedure_names)))
    
    DO i = 1, SIZE(pnode%procedure_names)
      ident_list(i) = IdentifierNature(  &
          NAME=.NameOf. pnode%procedure_names(i),  &
          NATURES=[NatureRefProcedure],  &
          LOCATIONS=[.LocationOf. pnode%procedure_names(i)] )
    END DO
    
  END SUBROUTINE ps_Advertise
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for FunctionInterfaceBody.
  
  
  !*****************************************************************************
  !!
  !> Parse the statement tree node for an interface body for a function.
  !!
  !! @param[out]    st                The statement tree node.
  !!
  !! @param[out]    fn_body           The results of the parse.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! Recursive because a specification inside a function interface body could 
  !! itself be a function interface body.
  
  RECURSIVE SUBROUTINE ParseFunctionInterfaceBody( st, abstract, fn_body,  &
      err_list )
    
    USE Errors
    USE InterfaceBodyStNodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(FunctionInterfaceBodyStNode), INTENT(INOUT) :: st
    LOGICAL, INTENT(IN), TARGET :: abstract
    TYPE(FunctionInterfaceBody), INTENT(OUT) :: fn_body
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    ! Save the "attributes" from the containing interface block.
    fn_body%abstract => abstract
    
    CALL ParseFunctionStmt(st%first%tlist, fn_body%fn_stmt, sub_err_list)
    CALL Add(err_list, sub_err_list)
    
    IF (ALLOCATED(st%spec)) THEN
      CALL ParseSpecificationPart(st%spec, fn_body%spec, sub_err_list)
      CALL Add(err_list, sub_err_list)
    END IF
    
    CALL ParseEndFunctionStmt(  &
        st%last%tlist,  &
        fn_body%end_name,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list)
    
  END SUBROUTINE ParseFunctionInterfaceBody
  
  
  !*****************************************************************************
  !!
  !> Implementation of FunctionInterfaceBody%Dump - write a debugging 
  !! representation of the results of parsing the interface body.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE fun_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(FunctionInterfaceBody), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    WRITE (unit, "(A,'function /interface-body/')") pfx
    WRITE (unit, "(A,'/',A,'/')") pfx // '  ', pnode%fn_stmt%What()
    CALL pnode%fn_stmt%Dump(unit, pfx // '    ')
    WRITE (unit, "(A,'/specification-part/')") pfx // '  '
    CALL pnode%spec%Dump(unit, pfx // '    ')
    
  END SUBROUTINE fun_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of FunctionInterfaceBody%Query.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  SUBROUTINE fun_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(FunctionInterfaceBody), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%fn_stmt%Query(visitor)
    
    CALL pnode%spec%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE fun_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of FunctionInterfaceBody%Modify.
  !!
  !! @param[in,out] pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  SUBROUTINE fun_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(FunctionInterfaceBody), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%fn_stmt%Modify(visitor)
    
    CALL pnode%spec%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE fun_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of FunctionInterfaceBody%What - return a descriptive 
  !! string for the parse node.
  !!
  !! @returns The syntax rule name for the node.
  
  FUNCTION fun_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'interface-body (function)'
    
  END FUNCTION fun_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of FunctionInterfaceBody%Advertise - advertise the 
  !! identifiers referenced or defined by the interface body into 
  !! the host scope.
  !!
  !! @param[in]     pnode             The node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  ! A function interface body may contain a function interface body 
  ! in its specification part, hence this is RECURSIVE.
  RECURSIVE SUBROUTINE fun_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(FunctionInterfaceBody), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(1))
    
    ! An interface body that is not for a separate module procedure 
    ! and that is not for an abstract interface also gives the external 
    ! attribute to the procedure.
    IF (pnode%fn_stmt%prefixes%module .OR. pnode%abstract) THEN
      ident_list(1) = IdentifierNature(  &
          NAME=.NameOf. pnode%fn_stmt%function_name,  &
          NATURES=[NatureDefFunction],  &
          LOCATIONS=[.Locationof. pnode%fn_stmt%function_name] )
    ELSE
      ident_list(1) = IdentifierNature(  &
          NAME=.NameOf. pnode%fn_stmt%function_name,  &
          NATURES=[NatureDefFunction, NatureDefExternal],  &
          LOCATIONS=[  &
            .Locationof. pnode%fn_stmt%function_name,  &
            .Locationof. pnode%fn_stmt%function_name ] )
    END IF
    
  END SUBROUTINE fun_Advertise
  
  
  !*****************************************************************************
  !!
  !> Implementation of FunctionInterfaceBody%Check - check the parse node for 
  !! conformance with language standards.
  !!
  !! @param[in]     pnode             The parse node to check.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE fun_Check(pnode, err_list)
    
    USE Errors
    USE ErrorCodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(FunctionInterfaceBody), INTENT(IN) :: pnode
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (ALLOCATED(pnode%end_name)) THEN
      IF ( .NameOf. pnode%fn_stmt%function_name  &
          /= .NameOf. pnode%end_name ) THEN
        CALL Add( err_list, CODE=errNameMismatch,  &
            COMPONENT='C1258 on R1232',  & 
            LOCATION=.LocationOf. pnode%end_name,  &
            MSG='The name in the /end-function-stmt/ ("'  &
              // .NameOf. pnode%end_name  &
              // '") must match the name given in the /function-stmt/ ("'  &
              // .NameOf. pnode%fn_stmt%function_name // '").' )
      END IF
    END IF
    
  END SUBROUTINE fun_Check
  
  
  !*****************************************************************************
  !!
  !> Implementation of FunctionInterfaceBody%Who - return the identifier text 
  !! for the parse node.
  !!
  !! @param[in]     pnode             The parse node to query.
  !!
  !! @returns The name of the function.
  
  FUNCTION fun_Who(pnode) RESULT(name)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(FunctionInterfaceBody), INTENT(IN) :: pnode
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: name
    
    !***************************************************************************
    
    name = .NameOf. pnode%fn_stmt%function_name
    
  END FUNCTION fun_Who
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for SubroutineInterfaceBody
  
  
  !*****************************************************************************
  !!
  !> Parse the statement tree node for an interface body for a subroutine.
  !!
  !! @param[out]    st                The statement tree node.
  !!
  !! @param[out]    sub_body          The results of the parse.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! Recursive because a specification inside a subroutine interface body 
  !! could itself be a subroutine interface body.
  
  RECURSIVE SUBROUTINE ParseSubroutineInterfaceBody( st, abstract, sub_body,  &
      err_list )
    
    USE Errors
    USE InterfaceBodyStNodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(SubroutineInterfaceBodyStNode), INTENT(INOUT) :: st
    LOGICAL, INTENT(IN), TARGET :: abstract
    TYPE(SubroutineInterfaceBody), INTENT(OUT) :: sub_body
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    ! Save the "attributes" from the containing interface block.
    sub_body%abstract => abstract
    
    CALL ParseSubroutineStmt(st%first%tlist, sub_body%sub_stmt, sub_err_list)
    CALL Add(err_list, sub_err_list)
    
    IF (ALLOCATED(st%spec)) THEN
      CALL ParseSpecificationPart(st%spec, sub_body%spec, sub_err_list)
      CALL Add(err_list, sub_err_list)
    END IF
    
    CALL ParseEndSubroutineStmt(  &
        st%last%tlist,  &
        sub_body%end_name,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list)
    
  END SUBROUTINE ParseSubroutineInterfaceBody
  
  
  !*****************************************************************************
  !!
  !> Implementation of SubroutineInterfaceBody%Dump - write a debugging 
  !! representation of the results of parsing the interface body.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE sub_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(SubroutineInterfaceBody), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    WRITE (unit, "(A,'subroutine /interface-body/')") pfx
    WRITE (unit, "(A,'/',A,'/')") pfx // '  ', pnode%sub_stmt%What()
    CALL pnode%sub_stmt%Dump(unit, pfx // '    ')
    WRITE (unit, "(A,'/specification-part/')") pfx // '  '
    CALL pnode%spec%Dump(unit, pfx // '    ')
    
  END SUBROUTINE sub_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of SubroutineInterfaceBody%Query.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  SUBROUTINE sub_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(SubroutineInterfaceBody), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%sub_stmt%Query(visitor)
    
    CALL pnode%spec%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE sub_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of SubroutineInterfaceBody%Modify.
  !!
  !! @param[in,out] pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  SUBROUTINE sub_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(SubroutineInterfaceBody), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%sub_stmt%Modify(visitor)
    
    CALL pnode%spec%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE sub_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of SubroutineInterfaceBody%What - return a descriptive 
  !! string for the parse node.
  !!
  !! @returns The syntax rule name for the node.
  
  FUNCTION sub_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'interface-body (subroutine)'
    
  END FUNCTION sub_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of SubroutineInterfaceBody%Advertise - advertise the 
  !! identifiers referenced or defined by the interface body into 
  !! the host scope.
  !!
  !! @param[in]     pnode             The node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  ! A subroutine interface body may contain a subroutine interface body 
  ! in its specification part, hence this is RECURSIVE.
  RECURSIVE SUBROUTINE sub_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(SubroutineInterfaceBody), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(1))
    IF (pnode%sub_stmt%prefixes%module .OR. pnode%abstract) THEN
      ident_list(1) = IdentifierNature(  &
          NAME=.NameOf. pnode%sub_stmt%subroutine_name,  &
          NATURES=[NatureDefSubroutine],  &
          LOCATIONS=[.Locationof. pnode%sub_stmt%subroutine_name] )
    ELSE
      ident_list(1) = IdentifierNature(  &
          NAME=.NameOf. pnode%sub_stmt%subroutine_name,  &
          NATURES=[NatureDefSubroutine, NatureDefExternal],  &
          LOCATIONS=[  &
            .Locationof. pnode%sub_stmt%subroutine_name,  &
            .Locationof. pnode%sub_stmt%subroutine_name ] )
    END IF
    
  END SUBROUTINE sub_Advertise
  
  
  !*****************************************************************************
  !!
  !> Implementation of SubroutineInterfaceBody%Check - check the parse node 
  !! for conformance with language standards.
  !!
  !! @param[in]     pnode             The parse node to check.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE sub_Check(pnode, err_list)
    
    USE Errors
    USE ErrorCodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(SubroutineInterfaceBody), INTENT(IN) :: pnode
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (ALLOCATED(pnode%end_name)) THEN
      IF ( .NameOf. pnode%sub_stmt%subroutine_name  &
          /= .NameOf. pnode%end_name ) THEN
        CALL Add( err_list, CODE=errNameMismatch,  &
            COMPONENT='C1261 on R1236',  & 
            LOCATION=.LocationOf. pnode%end_name,  &
            MSG='The name in the /end-subroutine-stmt/ ("'  &
              // .NameOf. pnode%end_name  &
              // '") must match the name given in the /subroutine-stmt/ ("'  &
              // .NameOf. pnode%sub_stmt%subroutine_name // '").' )
      END IF
    END IF
    
  END SUBROUTINE sub_Check
  
  
  !*****************************************************************************
  !!
  !> Implementation of SubroutineInterfaceBody%Who - return the identifier text 
  !! for the parse node.
  !!
  !! @param[in]     pnode             The parse node to query.
  !!
  !! @returns The name of the subroutine.
  
  FUNCTION sub_Who(pnode) RESULT(name)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(SubroutineInterfaceBody), INTENT(IN) :: pnode
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: name
    
    !***************************************************************************
    
    name = .NameOf. pnode%sub_stmt%subroutine_name
    
  END FUNCTION sub_Who
  
END MODULE InterfaceBlocks
