! $Id: IoControlSpecs.f90 2795 2019-03-22 09:15:08Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the IOControlSpecs module.


!*******************************************************************************
!!
!> A type and procedures for parsing /io-control-spec-list/'s and 
!! /io-control-spec/'s.
!!
!! These appear in /write-stmt/'s and in one of the forms of /read-stmt/.

MODULE IoControlSpecs
  
  USE CompilerKinds
  USE Expressions
  USE Labels
  USE Formats
  USE Identifiers
  USE ParseNodes
  USE SourceLocations
  USE Variables
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseIOControlSpecList
  PUBLIC :: ParseFileUnitNumber
  
  PUBLIC :: DumpIOControlSpecList
  
  !-----------------------------------------------------------------------------
  
  !> The results of parsing an /io-unit/ (R901).
  TYPE, EXTENDS(TermParseNode), PUBLIC :: IOUnit
    !> The /io-unit/ specification was a "*".
    LOGICAL :: is_star
    !> The location of the star, if \a is_star is true, otherwise not 
    !! allocated.
    TYPE(SourceLocation), ALLOCATABLE :: star_location
    !> The /io-unit/ specification was an /internal-file-variable/ or 
    !! a /file-unit-number/ (which both look like expressions).
    !!
    !! This will not be allocated if \a is_star is .TRUE., and will be 
    !! allocated if \a is_star is .FALSE..
    CLASS(Expression), ALLOCATABLE :: expr
  CONTAINS
    PROCEDURE :: Write => iou_Write
    PROCEDURE :: Query => iou_Query
    PROCEDURE :: Modify => iou_Modify
    PROCEDURE, NOPASS :: What => iou_What
    PROCEDURE :: GetLocation => iou_GetLocation
  END TYPE IOUnit
  
  !> Results of parsing an /io-control-spec-list/.
  !!
  !! @todo we should probably track the locations of the specifications 
  !! that are not currently tracked.
  TYPE, EXTENDS(TermParseNode), PUBLIC :: IoControlSpecList
    !> The location of the start of the /io-control-spec-list/.
    TYPE(SourceLocation) :: location
    !> The UNIT specification.
    TYPE(IOUnit), ALLOCATABLE :: unit
    !> The FMT (format) specification or NML (namelist) specification.
    !!
    !! Mutually exclusive with fmt and nml.  We use this component if no
    !! FMT= or NML= specifier is provided and the single token after the 
    !! = is a /name/ - in this case we need to know what that name is 
    !! before we can decide what sort of spec that we are dealing with.
    TYPE(Identifier), ALLOCATABLE :: fmt_or_nml
    !> The FMT (format) specification.
    !!
    !! Mutually exclusive with fmt_or_nml and nml.  If an explicit FMT= 
    !! specifier is provided or we can deduce from the syntax that the only 
    !! valid option is a format specification, then we populate this 
    !! component.
    TYPE(Format), ALLOCATABLE :: fmt
    !> The NML (namelist) specification.
    !!
    !! Mutually exclusive with fmt_or_nml and fmt.  If an explicit NML=
    !! specifier is provided then we use this component.
    TYPE(Identifier), ALLOCATABLE :: nml
    !> The ADVANCE specification.
    CLASS(Expression), ALLOCATABLE :: advance
    !> The ASYNCHRONOUS specification
    !!
    !! This is required to be a constant expression.
    CLASS(Expression), ALLOCATABLE :: asynchronous
    !> The BLANK specification
    CLASS(Expression), ALLOCATABLE :: blank
    !> The DECIMAL specification
    CLASS(Expression), ALLOCATABLE :: decimal
    !> The DELIM specification
    CLASS(Expression), ALLOCATABLE :: delim
    !> The END specification.
    TYPE(Label), ALLOCATABLE :: end
    !> The EOR specification.
    TYPE(Label), ALLOCATABLE :: eor
    !> The ERR specification.
    TYPE(Label), ALLOCATABLE :: err
    !> The ID specification.
    !!
    !! This is a variable.  It actually has it's own syntax rule /id-variable/ 
    !! (R914) to constrain its kind.
    CLASS(Variable), ALLOCATABLE :: id
    !> The IOMSG specification
    !!
    !! This is a variable (/iomsg-variable/).
    CLASS(Variable), ALLOCATABLE :: iomsg
    !> The IOSTAT specification.
    !!
    !! This is a variable.
    CLASS(Variable), ALLOCATABLE :: iostat
    !> The PAD specification.
    CLASS(Expression), ALLOCATABLE :: pad
    !> The POS specification.
    CLASS(Expression), ALLOCATABLE :: pos
    !> The REC specification.
    CLASS(Expression), ALLOCATABLE :: rec
    !> The ROUND specification.
    CLASS(Expression), ALLOCATABLE :: round
    !> The SIGN specification.
    CLASS(Expression), ALLOCATABLE :: sign
    !> The SIZE specification.
    !!
    !! This is a variable.
    CLASS(Variable), ALLOCATABLE :: size
  CONTAINS
    PROCEDURE :: Write => iocs_Write
    PROCEDURE :: Query => iocs_Query
    PROCEDURE :: Modify => iocs_Modify
    PROCEDURE, NOPASS :: What => iocs_What
    PROCEDURE :: GetLocation => iocs_GetLocation
    PROCEDURE :: Check => iocs_Check
  END TYPE IoControlSpecList
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /file-unit-number/ (R902).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the first token of the 
  !! /file-unit-number/ in @a tlist.
  !!
  !! @param[in]     finish            Index of the last token of the 
  !! /file-unit-number/ in @a tlist.
  !!
  !! @param[out]    expr              Results of parsing the 
  !! /file-unit-number/.  
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! This is just a utility routine that helps get the component chaining 
  !! in any error messages right.
  
  SUBROUTINE ParseFileUnitNumber(tlist, start, finish, expr, err_list)
    
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    CLASS(Expression), INTENT(OUT), ALLOCATABLE :: expr
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /file-unit-number/.
    CHARACTER(*), PARAMETER :: comp = 'R902'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    CALL ParseExpr(tlist, start, finish, expr, sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    
  END SUBROUTINE ParseFileUnitNumber
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /io-control-spec-list/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the first token in @a tlist of 
  !! the list.
  !!
  !! @param[in]     finish            Index of the last token in @a tlist of 
  !! the list.
  !!
  !! @param[in]     is_read           Flag to indicate (.TRUE.) that this 
  !! spec list is part of a read statement, otherwise it is part of a 
  !! write statement.
  !!
  !! @param[out]    io_control_spec   The results of parsing the list.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! The token sequence for the list does not include the delimiting "(" and 
  !! ")".
  
  SUBROUTINE ParseIOControlSpecList( tlist, start, finish, is_read,  &
      io_control_spec, err_list )
    
    USE Errors
    USE ErrorCodes
    USE Scanner
    USE Tokens
    USE ParseUtils
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    LOGICAL, INTENT(IN) :: is_read
    TYPE(IoControlSpecList), INTENT(OUT) :: io_control_spec
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for an /io-control-spec/
    CHARACTER(*), PARAMETER :: comp = 'R913'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: list_count     ! Number of items in the /close-spec-list/.
    INTEGER :: item           ! List item index.
    INTEGER :: item_start     ! Start index in tlist of a particular item.
    INTEGER :: item_finish    ! Finish index in tlist of a particular item.   
    LOGICAL :: no_spec        ! Flag to indicate UNIT= or FMT= or NML= absent.
    LOGICAL :: unit_no_spec   ! Flag to indicate UNIT= was absent.
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
  
    ALLOCATE(err_list(0))
    
    IF (finish < start) THEN
      ! Note no component - that will be provided by the parent syntax rule.
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          LOCATION=SafeLocation(tlist, finish),  &
          MSG='Expected /io-control-spec-list/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    ! Save location.
    io_control_spec%location = QueryLocation(tlist(start))
    
    list_count = ListItemCount(tlist(start:finish))
    item_finish = 0
    DO item = 1, list_count
      CALL GetNextListItem(tlist, start, finish, item_start, item_finish)
      CALL parse_io_control_spec(  &
          tlist,  &
          item_start, item_finish,  &
          item,  &
          io_control_spec,  &
          no_spec,  &
          sub_err_list )
      ! Note no component chaining.
      CALL Add(err_list, sub_err_list)
      ! Continue on error.
      IF ((item == 2) .AND. no_spec) THEN
        IF (.NOT. unit_no_spec) THEN
          CALL Add( err_list,  &
              CODE=errSyntax,  &
              COMPONENT=comp,  &
              LOCATION=QueryLocation(tlist(item_start)),  &
              MSG='If /format/ appears without a preceeding "FMT=" or &
                &/namelist-group-name/ appears without a preceeding "NML=" &
                &then the first item shall be /io-unit/.' )
        END IF
      ELSE
        unit_no_spec = (item == 1) .AND. no_spec
      END IF
    END DO
        
    !---------------------------------------------------------------------------
    ! Check constraints
    
    ! C913 and C914 (/read-stmt/ and /write-stmt/ specifics) checks.
    IF (is_read) THEN
    ELSE
    END IF
    
  END SUBROUTINE ParseIOControlSpecList
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /io-control-spec/ (R913).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the first token in @a tlist of 
  !! the spec.
  !!
  !! @param[in]     finish            Index of the last token in @a tlist of 
  !! the spec.
  !!
  !! @param[in]     idx               Item index in the parent 
  !! /io-control-spec-list/.
  !!
  !! @param[out]    io_control_spec   The results of parsing the spec.  Note 
  !! that this is an INTENT(INOUT) argument - each spec sets a particular 
  !! component.
  !!
  !! @param[out]    no_spec           Flag to indicate that the "XXX=" part of 
  !! the spec was absent.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! This procedure checks a number of constraints (or parts of constraints 
  !! where not all information is available).  A number of these constraints 
  !! are around incompatible combinations of specifiers.  An alternative 
  !! approach would be to check the constraints once the entire list 
  !! has been parsed (and this approach is necessary for some constraints, 
  !! particularly those that require combinations of specifiers rather than 
  !! prohibiting combinations), but the advantage of doing it at this level 
  !! is that the error message can easily be reported against the first 
  !! "violating token", rather than against the first token of the whole spec 
  !! list.
  !!
  !! The disadvantage is that the checking needs to be distributed amongst the 
  !! relevant specifiers for each constraint.
  
  SUBROUTINE parse_io_control_spec( tlist, start, finish, idx,  &
      io_control_spec, no_spec, err_list )
    
    USE Errors
    USE ErrorCodes
    USE MatchUtils
    USE ParseUtils
    USE Scanner
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
      
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    INTEGER, INTENT(IN) :: idx
    TYPE(IoControlSpecList), INTENT(INOUT) :: io_control_spec
    LOGICAL, INTENT(OUT) :: no_spec
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Keywords for each /io-control-spec/.
    CHARACTER(*,KIND=scck), PARAMETER :: keywords(20) = [  &
        scck_'UNIT        ', scck_'FMT         ', scck_'NML         ',  &
        scck_'ADVANCE     ', scck_'ASYNCHRONOUS', scck_'BLANK       ',  &
        scck_'DECIMAL     ', scck_'DELIM       ', scck_'END         ',  &
        scck_'EOR         ', scck_'ERR         ', scck_'ID          ',  &
        scck_'IOMSG       ', scck_'IOSTAT      ', scck_'PAD         ',  &
        scck_'POS         ', scck_'REC         ', scck_'ROUND       ',  &
        scck_'SIGN        ', scck_'SIZE        ' ]
    
    ! Syntax rule number for an /io-control-spec/
    CHARACTER(*), PARAMETER :: comp = 'R913'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: opt            ! Expect option for the keyword test.
    INTEGER :: assign         ! Index of the =.
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    IF (finish < start) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          LOCATION=SafeLocation(tlist, finish),  &
          MSG='Expected /io-control-spec/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! Do we have an  argument keyword?
    
    assign = FindBare(tlist(:finish), ittAssign, scck_'=', start)
    IF (assign == 0) THEN
      ! No.
      no_spec = .TRUE.
      SELECT CASE (idx)
      CASE (1)    ! This is a unit spec.
        opt = 1
      CASE (2)    ! This is either a FMT or NML spec.
        opt = -1
      CASE DEFAULT
        CALL Add( err_list,  &
            CODE=errSyntax,  &
            COMPONENT='C911/C918/C919',  &
            LOCATION=QueryLocation(tlist(start)),  &
            MSG='If the optional "UNIT=" is omitted, the /io-unit/ &
              &shall be the first item in the /io-control-spec-list/; if the&
              &optional "FMT=" or "NML=" is omitted, the /format/ or &
              &/namelist-group-name/ shall be the second item in the &
              &/io-control-spec-list/.' )
        RETURN
      END SELECT
    ELSE
      ! Yes.  What was the keyword?
      no_spec = .FALSE.
      CALL ExpectKeyword(  &
          tlist,  &
          start,  &
          keywords,  &
          '/io-control-spec/',  &
          opt, err )
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      CALL Expect(tlist, start + 1, ittAssign, scck_'=', err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    SELECT CASE (opt)
    CASE (-1)     ! Unspecified second item.  There will not be an =.
      ! This is only triggered for the second token.  There can only be one 
      ! second token.  Therefore io_control_spec%fmt_or_nml will not be 
      ! allocated here.  We need to make sure that the 
      IF ( ALLOCATED(io_control_spec%fmt)  &
          .OR. ALLOCATED(io_control_spec%nml) ) THEN
        CALL add_nml_fmt_err
      ELSE
        ! If the sequence of tokens that make up the second name is more than 
        ! a single /name/, then it cannot validly be a namelist specification, 
        ! so we will assume that it is a format specification.
        IF (looks_like_nml()) THEN
          ! Checks on validity inherent in the IF test above.
          io_control_spec%fmt_or_nml = IdentifierIfortBug(tlist(start))
        ELSE
          ALLOCATE(io_control_spec%fmt)
          CALL ParseFormat(  &
              tlist,  &
              start, finish,  &
              io_control_spec%fmt,  &
              sub_err_list )
          CALL Add(err_list, sub_err_list, COMPONENT=comp)
          ! Check some contraints.
          IF (.NOT. Failure(sub_err_list)) CALL check_fmt
        END IF
      END IF
      
    CASE (1)      ! UNIT
      IF (ALLOCATED(io_control_spec%unit)) THEN
        CALL already('UNIT')
      ELSE
        ALLOCATE(io_control_spec%unit)
        IF (assign == 0) THEN
          CALL parse_io_unit(  &
              tlist,  &
              start, finish,  &
              io_control_spec%unit,  &
              sub_err_list )
        ELSE
          CALL parse_io_unit(  &
              tlist,  &
              assign + 1, finish,  &
              io_control_spec%unit,  &
              sub_err_list )
        END IF
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
        ! Check some constraints.
        IF (.NOT. Failure(sub_err_list)) THEN
          IF (io_control_spec%unit%is_star) THEN
            IF (ALLOCATED(io_control_spec%rec)) CALL add_unit_star_rec_err
            IF (ALLOCATED(io_control_spec%pos)) CALL add_unit_star_pos_err
          END IF
        END IF
      END IF
      
    CASE (2)      ! FMT (explicit)
      IF (ALLOCATED(io_control_spec%fmt)) THEN
        CALL already('FMT')
      ELSE 
        ALLOCATE(io_control_spec%fmt)
        IF ( ALLOCATED(io_control_spec%nml)  &
            .OR. ALLOCATED(io_control_spec%fmt_or_nml) ) THEN
          CALL add_nml_fmt_err
        ELSE
          CALL ParseFormat(  &
              tlist,  &
              assign + 1, finish,  &
              io_control_spec%fmt,  &
              sub_err_list )
          CALL Add(err_list, sub_err_list, COMPONENT=comp)
          ! Check some contraints.
          IF (.NOT. Failure(sub_err_list)) CALL check_fmt
        END IF
      END IF
      
    CASE (3)      ! NML (explicit)
      IF (ALLOCATED(io_control_spec%nml)) THEN
        CALL already('UNIT')
      ELSE IF ( ALLOCATED(io_control_spec%fmt)  &
          .OR. ALLOCATED(io_control_spec%fmt_or_nml) ) THEN
        CALL add_nml_fmt_err
      ELSE IF (ALLOCATED(io_control_spec%rec)) THEN
        CALL add_nml_rec_err
      ELSE
        CALL Expect(tlist, assign + 1, '/namelist-group-name/', err)
        CALL Add(err_list, err, COMPONENT=comp)
        io_control_spec%nml = IdentifierIfortBug(tlist(assign+1))
        
        ! Don't check for extra tokens if the first token was in error.
        IF (.NOT. Failure(err)) THEN
          CALL ExpectEnd(tlist(:finish), assign + 2, comp, err_list)
        END IF
      END IF
      
    CASE (4)      ! ADVANCE
      IF (ALLOCATED(io_control_spec%advance)) THEN
        CALL already('ADVANCE')
      ELSE
        CALL ParseExpr(  &
            tlist,  &
            assign + 1, finish,  &
            io_control_spec%advance,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (5)      ! ASYNCHRONOUS
      IF (ALLOCATED(io_control_spec%asynchronous)) THEN
        CALL already('ASYNCHRONOUS')
      ELSE
        CALL ParseExpr(  &
            tlist,  &
            assign + 1, finish,  &
            io_control_spec%asynchronous,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (6)      ! BLANK
      IF (ALLOCATED(io_control_spec%blank)) THEN
        CALL already('BLANK')
      ELSE
        CALL ParseExpr(  &
            tlist,  &
            assign + 1, finish,  &
            io_control_spec%blank,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (7)      ! DECIMAL
      IF (ALLOCATED(io_control_spec%decimal)) THEN
        CALL already('DECIMAL')
      ELSE
        CALL ParseExpr(  &
            tlist,  &
            assign + 1, finish,  &
            io_control_spec%decimal,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (8)      ! DELIM
      IF (ALLOCATED(io_control_spec%delim)) THEN
        CALL already('DELIM')
      ELSE
        CALL ParseExpr(  &
            tlist,  &
            assign + 1, finish,  &
            io_control_spec%delim,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (9)      ! END
      IF (ALLOCATED(io_control_spec%end)) THEN
        CALL already('END')
      ELSE IF (ALLOCATED(io_control_spec%rec)) THEN
        CALL add_rec_end_err
      ELSE
        ALLOCATE(io_control_spec%end)
        CALL ParseLabel(  &
            tlist,  &
            assign + 1,  &
            finish,  &
            io_control_spec%end,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
        IF (Failure(sub_err_list)) RETURN
      END IF
      
    CASE (10)     ! EOR
      IF (ALLOCATED(io_control_spec%eor)) THEN
        CALL already('EOR')
      ELSE
        ALLOCATE(io_control_spec%eor)
        CALL ParseLabel(  &
            tlist,  &
            assign + 1,  &
            finish,  &
            io_control_spec%eor,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
        IF (Failure(sub_err_list)) RETURN
      END IF
      
    CASE (11)     ! ERR
      IF (ALLOCATED(io_control_spec%err)) THEN
        CALL already('ERR')
      ELSE
        ALLOCATE(io_control_spec%err)
        CALL ParseLabel(  &
            tlist,  &
            assign + 1,  &
            finish,  &
            io_control_spec%err,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
        IF (Failure(sub_err_list)) RETURN
      END IF
      
    CASE (12)     ! ID
      IF (ALLOCATED(io_control_spec%id)) THEN
        CALL already('ID')
      ELSE
        ! The syntax term is actually an /id-variable/.
        CALL ParseVariable(  &
            tlist,  &
            assign + 1, finish,  &
            io_control_spec%id,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (13)     ! IOMSG
      IF (ALLOCATED(io_control_spec%iomsg)) THEN
        CALL already('IOMSG')
      ELSE
        ! The syntax term is actually an /iomsg-variable/.
        CALL ParseVariable(  &
            tlist,  &
            assign + 1, finish,  &
            io_control_spec%iomsg,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (14)     ! IOSTAT
      IF (ALLOCATED(io_control_spec%iostat)) THEN
        CALL already('IOSTAT')
      ELSE
        CALL ParseVariable(  &
            tlist,  &
            assign + 1, finish,  &
            io_control_spec%iostat,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (15)     ! PAD
      IF (ALLOCATED(io_control_spec%pad)) THEN
        CALL already('PAD')
      ELSE
        CALL ParseExpr(  &
            tlist,  &
            assign + 1, finish,  &
            io_control_spec%pad,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (16)     ! POS
      IF (ALLOCATED(io_control_spec%pos)) THEN
        CALL already('POS')
      ELSE IF (unit_star()) THEN
        CALL add_unit_star_pos_err
      ELSE IF (ALLOCATED(io_control_spec%rec)) THEN
        CALL add_pos_rec_err
      ELSE
        CALL ParseExpr(  &
            tlist,  &
            assign + 1, finish,  &
            io_control_spec%pos,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (17)     ! REC
      IF (ALLOCATED(io_control_spec%rec)) THEN
        CALL already('REC')
      ELSE IF (ALLOCATED(io_control_spec%nml)) THEN
        ! Note that this doesn't catch the case when the NML is not explicit.
        CALL add_nml_rec_err
      ELSE IF (unit_star()) THEN
        CALL add_unit_star_rec_err
      ELSE IF (ALLOCATED(io_control_spec%end)) THEN
        CALL add_rec_end_err
      ELSE IF (fmt_star()) THEN
        CALL add_fmt_star_rec_err
      ELSE IF (ALLOCATED(io_control_spec%pos)) THEN
        CALL add_pos_rec_err
      ELSE
        CALL ParseExpr(  &
            tlist,  &
            assign + 1, finish,  &
            io_control_spec%rec,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (18)     ! ROUND
      IF (ALLOCATED(io_control_spec%round)) THEN
        CALL already('ROUND')
      ELSE
        CALL ParseExpr(  &
            tlist,  &
            assign + 1, finish,  &
            io_control_spec%round,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (19)     ! SIGN
      IF (ALLOCATED(io_control_spec%sign)) THEN
        CALL already('SIGN')
      ELSE
        CALL ParseExpr(  &
            tlist,  &
            assign + 1, finish,  &
            io_control_spec%sign,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (20)     ! SIZE
      IF (ALLOCATED(io_control_spec%size)) THEN
        CALL already('SIZE')
      ELSE
        CALL ParseVariable(  &
            tlist,  &
            assign + 1, finish,  &
            io_control_spec%size,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE DEFAULT
      STOP 'Unknown /io-control-spec/.'
      
    END SELECT
    
  CONTAINS
    
    SUBROUTINE already(text)
      !-------------------------------------------------------------------------
      ! Arguments
      CHARACTER(*), INTENT(IN) :: text
      !*************************************************************************
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT='C910',  &
          LOCATION=QueryLocation(tlist(start)),  &
          MSG=text // ' has already been specified for this statement.' )
    END SUBROUTINE already
        
    SUBROUTINE add_nml_fmt_err
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT='C917 on ' // comp,  &
          LOCATION=QueryLocation(tlist(start)),  &
          MSG='A /io-control-spec-list/ shall not contain both a /format/ &
            &and a /namelist-group-name/.' )
    END SUBROUTINE add_nml_fmt_err
    
    PURE FUNCTION looks_like_nml()
      LOGICAL :: looks_like_nml
      !*************************************************************************
      looks_like_nml = start == finish
      IF (.NOT. looks_like_nml) RETURN
      looks_like_nml = tlist(start) == ittName
    END FUNCTION looks_like_nml
    
    PURE FUNCTION unit_star()
      LOGICAL :: unit_star
      !*************************************************************************
      unit_star = ALLOCATED(io_control_spec%unit)
      IF (.NOT. unit_star) RETURN
      unit_star = io_control_spec%unit%is_star
    END FUNCTION unit_star
    
    PURE FUNCTION fmt_star()
      LOGICAL :: fmt_star
      !*************************************************************************
      fmt_star = ALLOCATED(io_control_spec%fmt)
      IF (.NOT. fmt_star) RETURN
      fmt_star = io_control_spec%fmt%is_star
    END FUNCTION fmt_star
    
    SUBROUTINE check_fmt()
      IF (io_control_spec%fmt%is_star) THEN
        IF (ALLOCATED(io_control_spec%rec)) CALL add_fmt_star_rec_err
        IF (ALLOCATED(io_control_spec%advance)) CALL add_fmt_star_advance_err
      END IF
    END SUBROUTINE check_fmt
    
    SUBROUTINE add_nml_rec_err
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT='C917 on ' // comp,  &
          LOCATION=QueryLocation(tlist(start)),  &
          MSG='A /namelist-group-name/ shall not appear if a "REC=" specifier &
            &appears.' )
    END SUBROUTINE add_nml_rec_err
    
    SUBROUTINE add_unit_star_rec_err
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT='C920 on ' // comp,  &
          LOCATION=QueryLocation(tlist(start)),  &
          MSG='If /io-unit/ is "*" the /io-control-spec-list/ shall not &
            &contain a "REC=" specifier.' )
    END SUBROUTINE add_unit_star_rec_err
    
    SUBROUTINE add_unit_star_pos_err
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT='C920 on ' // comp,  &
          LOCATION=QueryLocation(tlist(start)),  &
          MSG='If /io-unit/ is "*" the /io-control-spec-list/ shall not &
            &contain a "POS=" specifier.' )
    END SUBROUTINE add_unit_star_pos_err
    
    SUBROUTINE add_rec_end_err
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT='C921 on ' // comp,  &
          LOCATION=QueryLocation(tlist(start)),  &
          MSG='If a "REC=" specifier appears the /io-control-spec-list/ &
            &shall not contain an "END=" specifier.' )
    END SUBROUTINE add_rec_end_err
    
    SUBROUTINE add_fmt_star_rec_err
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT='C921 on ' // comp,  &
          LOCATION=QueryLocation(tlist(start)),  &
          MSG='If "REC=" appears the /format/ shall not be "*".' )
    END SUBROUTINE add_fmt_star_rec_err
    
    SUBROUTINE add_fmt_star_advance_err
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT='C922 on ' // comp,  &
          LOCATION=QueryLocation(tlist(start)),  &
          MSG='If "ADVANCE=" appears the /format/ shall not be "*".' )
    END SUBROUTINE add_fmt_star_advance_err
    
    SUBROUTINE add_pos_rec_err
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT='C927 on ' // comp,  &
          LOCATION=QueryLocation(tlist(start)),  &
          MSG='If "POS=" appears then "REC=" shall not appear.' )
    END SUBROUTINE add_pos_rec_err
    
  END SUBROUTINE parse_io_control_spec
  
  
  !*****************************************************************************
  !!
  !> Write a text representation of a IoControlSpecList to a logical unit.
  !!
  !! @param[in]     log_unit          Logical unit connected for formatted 
  !! output for the text represenation.
  !!
  !! @param[in]     spec              The /io-control-spec-list/.
  !!
  !! @param[in]     indent            Left hand side indent in characters.
  !!
  !! Unlike many of the other Write* routines in this module this procedure 
  !! uses non-advancing output and writes one item per record.
  
  SUBROUTINE DumpIOControlSpecList(log_unit, spec, indent)
    
    USE Expressions
    USE Variables
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    INTEGER, INTENT(IN) :: log_unit
    TYPE(IoControlSpecList), INTENT(IN) :: spec
    INTEGER, INTENT(IN) :: indent
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Indent prefix.
    CHARACTER(indent) :: prefix
    
    !***************************************************************************
    
    prefix = ''   ! Fixed length string - set to all blanks.
    
    IF (ALLOCATED(spec%unit)) THEN
      WRITE (log_unit, "(A,'UNIT = ')", ADVANCE='NO') prefix
      IF (spec%unit%is_star) THEN
        WRITE (log_unit, "('*')")
      ELSE
        CALL WriteExpr(log_unit, spec%unit%expr)
        WRITE (log_unit, "()")
      END IF
    END IF
    
    IF (ALLOCATED(spec%fmt)) THEN
      WRITE (log_unit, "(A,'FMT = ')", ADVANCE='NO') prefix
      CALL spec%fmt%Write(log_unit)
      WRITE (log_unit, "()")
    END IF
    
    IF (ALLOCATED(spec%nml)) THEN
      WRITE (log_unit, "(A,'NML = ')", ADVANCE='NO') prefix
      WRITE (log_unit, "(A)") .NameOf. spec%nml
    END IF
    
    IF (ALLOCATED(spec%fmt_or_nml)) THEN
      WRITE (log_unit, "(A,'FMT or NML = ')", ADVANCE='NO') prefix
      WRITE (log_unit, "(A)") .NameOf. spec%fmt_or_nml
    END IF
    
    IF (ALLOCATED(spec%advance)) THEN
      WRITE (log_unit, "(A,'ADVANCE = ')", ADVANCE='NO') prefix
      CALL WriteExpr(log_unit, spec%advance)
      WRITE (log_unit, "()")
    END IF
    
    IF (ALLOCATED(spec%asynchronous)) THEN
      WRITE (log_unit, "(A,'ASYNCHRONOUS = ')", ADVANCE='NO') prefix
      CALL WriteExpr(log_unit, spec%asynchronous)
      WRITE (log_unit, "()")
    END IF
    
    IF (ALLOCATED(spec%blank)) THEN
      WRITE (log_unit, "(A,'BLANK = ')", ADVANCE='NO') prefix
      CALL WriteExpr(log_unit, spec%blank)
      WRITE (log_unit, "()")
    END IF
    
    IF (ALLOCATED(spec%decimal)) THEN
      WRITE (log_unit, "(A,'DECIMAL = ')", ADVANCE='NO') prefix
      CALL WriteExpr(log_unit, spec%decimal)
      WRITE (log_unit, "()")
    END IF
    
    IF (ALLOCATED(spec%delim)) THEN
      WRITE (log_unit, "(A,'DELIM = ')", ADVANCE='NO') prefix
      CALL WriteExpr(log_unit, spec%delim)
      WRITE (log_unit, "()")
    END IF
    
    IF (ALLOCATED(spec%end)) THEN
      WRITE (log_unit, "(A,'END = ',I0)") prefix, .ValueOf. spec%end
    END IF
    
    IF (ALLOCATED(spec%eor)) THEN
      WRITE (log_unit, "(A,'EOR = ',I0)") prefix, .ValueOf. spec%eor
    END IF
    
    IF (ALLOCATED(spec%err)) THEN
      WRITE (log_unit, "(A,'ERR = ',I0)") prefix, .ValueOf. spec%err
    END IF
    
    IF (ALLOCATED(spec%id)) THEN
      WRITE (log_unit, "(A,'ID = ')", ADVANCE='NO') prefix
      CALL WriteVariable(log_unit, spec%id)
      WRITE (log_unit, "()")
    END IF
    
    IF (ALLOCATED(spec%iomsg)) THEN
      WRITE (log_unit, "(A,'IOMSG = ')", ADVANCE='NO') prefix
      CALL WriteVariable(log_unit, spec%iomsg)
      WRITE (log_unit, "()")
    END IF
    
    IF (ALLOCATED(spec%iostat)) THEN
      WRITE (log_unit, "(A,'IOSTAT = ')", ADVANCE='NO') prefix
      CALL WriteVariable(log_unit, spec%iostat)
      WRITE (log_unit, "()")
    END IF
    
    IF (ALLOCATED(spec%pad)) THEN
      WRITE (log_unit, "(A,'PAD = ')", ADVANCE='NO') prefix
      CALL WriteExpr(log_unit, spec%pad)
      WRITE (log_unit, "()")
    END IF
    
    IF (ALLOCATED(spec%pos)) THEN
      WRITE (log_unit, "(A,'POS = ')", ADVANCE='NO') prefix
      CALL WriteExpr(log_unit, spec%pos)
      WRITE (log_unit, "()")
    END IF
    
    IF (ALLOCATED(spec%rec)) THEN
      WRITE (log_unit, "(A,'REC = ')", ADVANCE='NO') prefix
      CALL WriteExpr(log_unit, spec%rec)
      WRITE (log_unit, "()")
    END IF
    
    IF (ALLOCATED(spec%round)) THEN
      WRITE (log_unit, "(A,'ROUND = ')", ADVANCE='NO') prefix
      CALL WriteExpr(log_unit, spec%round)
      WRITE (log_unit, "()")
    END IF
    
    IF (ALLOCATED(spec%sign)) THEN
      WRITE (log_unit, "(A,'SIGN = ')", ADVANCE='NO') prefix
      CALL WriteExpr(log_unit, spec%sign)
      WRITE (log_unit, "()")
    END IF
    
    IF (ALLOCATED(spec%size)) THEN
      WRITE (log_unit, "(A,'SIZE = ')", ADVANCE='NO') prefix
      CALL WriteVariable(log_unit, spec%size)
      WRITE (log_unit, "()")
    END IF
    
  END SUBROUTINE DumpIOControlSpecList
  
  
  !*****************************************************************************
  !!
  !> Implementation of IoControlSpecList%Write - write a text representation 
  !! of the parse node to a logical unit.
  !!
  !! @param[in]     pnode             The format.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output for the text represenation.
  
  SUBROUTINE iocs_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(IoControlSpecList), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! Indicates that a particular spec is the first.  This is used for comma 
    ! output control.
    LOGICAL :: first
    
    !***************************************************************************
    
    ! This is set to false in the do_sep internal procedure.
    first = .TRUE.
    
    IF (ALLOCATED(pnode%unit)) THEN
      CALL do_sep
        
      WRITE (unit, "('UNIT = ')", ADVANCE='NO')
      IF (pnode%unit%is_star) THEN
        WRITE (unit, "('*')", ADVANCE='NO')
      ELSE
        CALL WriteExpr(unit, pnode%unit%expr)
      END IF
    END IF
    
    IF (ALLOCATED(pnode%fmt)) THEN
      CALL do_sep
      
      WRITE (unit, "('FMT = ')", ADVANCE='NO')
      CALL pnode%fmt%Write(unit)
    END IF
    
    IF (ALLOCATED(pnode%nml)) THEN
      CALL do_sep
      
      WRITE (unit, "('NML = ')", ADVANCE='NO')
      WRITE (unit, "(A)", ADVANCE='NO') .NameOf. pnode%nml
    END IF
    
    IF (ALLOCATED(pnode%fmt_or_nml)) THEN
      CALL do_sep
      
      WRITE (unit, "('FMT or NML = ')", ADVANCE='NO')
      WRITE (unit, "(A)", ADVANCE='NO') .NameOf. pnode%fmt_or_nml
    END IF
    
    IF (ALLOCATED(pnode%advance)) THEN
      CALL do_sep
      
      WRITE (unit, "('ADVANCE = ')", ADVANCE='NO')
      CALL WriteExpr(unit, pnode%advance)
    END IF
    
    IF (ALLOCATED(pnode%asynchronous)) THEN
      CALL do_sep
      
      WRITE (unit, "('ASYNCHRONOUS = ')", ADVANCE='NO')
      CALL WriteExpr(unit, pnode%asynchronous)
    END IF
    
    IF (ALLOCATED(pnode%blank)) THEN
      CALL do_sep
      
      WRITE (unit, "('BLANK = ')", ADVANCE='NO')
      CALL WriteExpr(unit, pnode%blank)
    END IF
    
    IF (ALLOCATED(pnode%decimal)) THEN
      CALL do_sep
      
      WRITE (unit, "('DECIMAL = ')", ADVANCE='NO')
      CALL WriteExpr(unit, pnode%decimal)
    END IF
    
    IF (ALLOCATED(pnode%delim)) THEN
      CALL do_sep
      
      WRITE (unit, "('DELIM = ')", ADVANCE='NO')
      CALL WriteExpr(unit, pnode%delim)
    END IF
    
    IF (ALLOCATED(pnode%end)) THEN
      CALL do_sep
      
      WRITE (unit, "('END = ',I0)", ADVANCE='NO') .ValueOf. pnode%end
    END IF
    
    IF (ALLOCATED(pnode%eor)) THEN
      CALL do_sep
      
      WRITE (unit, "('EOR = ',I0)", ADVANCE='NO') .ValueOf. pnode%eor
    END IF
    
    IF (ALLOCATED(pnode%err)) THEN
      CALL do_sep
      
      WRITE (unit, "('ERR = ',I0)", ADVANCE='NO') .ValueOf. pnode%err
    END IF
    
    IF (ALLOCATED(pnode%id)) THEN
      CALL do_sep
      
      WRITE (unit, "('ID = ')", ADVANCE='NO')
      CALL WriteVariable(unit, pnode%id)
    END IF
    
    IF (ALLOCATED(pnode%iomsg)) THEN
      CALL do_sep
      
      WRITE (unit, "('IOMSG = ')", ADVANCE='NO')
      CALL WriteVariable(unit, pnode%iomsg)
    END IF
    
    IF (ALLOCATED(pnode%iostat)) THEN
      CALL do_sep
      
      WRITE (unit, "('IOSTAT = ')", ADVANCE='NO')
      CALL WriteVariable(unit, pnode%iostat)
    END IF
    
    IF (ALLOCATED(pnode%pad)) THEN
      CALL do_sep
      
      WRITE (unit, "('PAD = ')", ADVANCE='NO')
      CALL WriteExpr(unit, pnode%pad)
    END IF
    
    IF (ALLOCATED(pnode%pos)) THEN
      CALL do_sep
      
      WRITE (unit, "('POS = ')", ADVANCE='NO')
      CALL WriteExpr(unit, pnode%pos)
    END IF
    
    IF (ALLOCATED(pnode%rec)) THEN
      CALL do_sep
      
      WRITE (unit, "('REC = ')", ADVANCE='NO')
      CALL WriteExpr(unit, pnode%rec)
    END IF
    
    IF (ALLOCATED(pnode%round)) THEN
      CALL do_sep
      
      WRITE (unit, "('ROUND = ')", ADVANCE='NO')
      CALL WriteExpr(unit, pnode%round)
    END IF
    
    IF (ALLOCATED(pnode%sign)) THEN
      CALL do_sep
      
      WRITE (unit, "('SIGN = ')", ADVANCE='NO')
      CALL WriteExpr(unit, pnode%sign)
    END IF
    
    IF (ALLOCATED(pnode%size)) THEN
      CALL do_sep
      
      WRITE (unit, "('SIZE = ')", ADVANCE='NO')
      CALL WriteVariable(unit, pnode%size)
    END IF
    
  CONTAINS
    
    ! Test/sets the first flag, and if necessary outputs the comma that 
    ! separates list items.
    SUBROUTINE do_sep
      IF (first) THEN
        first = .FALSE.
      ELSE
        WRITE (unit, "(', ')", ADVANCE='NO')
      END IF
    END SUBROUTINE do_sep
    
  END SUBROUTINE iocs_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of IoControlSpecList%Query - traverse the node with a 
  !! query operation.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  SUBROUTINE iocs_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(IoControlSpecList), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%unit)) CALL pnode%unit%Query(visitor)
    ! fmt_or_nml??
    IF (ALLOCATED(pnode%fmt)) CALL pnode%fmt%Query(visitor)
    ! nml.
    IF (ALLOCATED(pnode%advance)) CALL pnode%advance%Query(visitor)
    IF (ALLOCATED(pnode%asynchronous)) CALL pnode%asynchronous%Query(visitor)
    IF (ALLOCATED(pnode%blank)) CALL pnode%blank%Query(visitor)
    IF (ALLOCATED(pnode%decimal)) CALL pnode%decimal%Query(visitor)
    IF (ALLOCATED(pnode%delim)) CALL pnode%delim%Query(visitor)
    ! end
    ! eor
    ! err
    IF (ALLOCATED(pnode%id)) CALL pnode%id%Query(visitor)
    IF (ALLOCATED(pnode%iomsg)) CALL pnode%iomsg%Query(visitor)
    IF (ALLOCATED(pnode%iostat)) CALL pnode%iostat%Query(visitor)
    IF (ALLOCATED(pnode%pad)) CALL pnode%pad%Query(visitor)
    IF (ALLOCATED(pnode%pos)) CALL pnode%pos%Query(visitor)
    IF (ALLOCATED(pnode%rec)) CALL pnode%rec%Query(visitor)
    IF (ALLOCATED(pnode%round)) CALL pnode%round%Query(visitor)
    IF (ALLOCATED(pnode%sign)) CALL pnode%sign%Query(visitor)
    IF (ALLOCATED(pnode%size)) CALL pnode%size%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE iocs_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of IoControlSpecList%Modify - traverse the node with a 
  !! modify operation.
  !!
  !! @param[in]     pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  SUBROUTINE iocs_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(IoControlSpecList), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%unit)) CALL pnode%unit%Modify(visitor)
    ! fmt_or_nml??
    IF (ALLOCATED(pnode%fmt)) CALL pnode%fmt%Modify(visitor)
    ! nml.
    IF (ALLOCATED(pnode%advance)) CALL pnode%advance%Modify(visitor)
    IF (ALLOCATED(pnode%asynchronous)) CALL pnode%asynchronous%Modify(visitor)
    IF (ALLOCATED(pnode%blank)) CALL pnode%blank%Modify(visitor)
    IF (ALLOCATED(pnode%decimal)) CALL pnode%decimal%Modify(visitor)
    IF (ALLOCATED(pnode%delim)) CALL pnode%delim%Modify(visitor)
    ! end
    ! eor
    ! err
    IF (ALLOCATED(pnode%id)) CALL pnode%id%Modify(visitor)
    IF (ALLOCATED(pnode%iomsg)) CALL pnode%iomsg%Modify(visitor)
    IF (ALLOCATED(pnode%iostat)) CALL pnode%iostat%Modify(visitor)
    IF (ALLOCATED(pnode%pad)) CALL pnode%pad%Modify(visitor)
    IF (ALLOCATED(pnode%pos)) CALL pnode%pos%Modify(visitor)
    IF (ALLOCATED(pnode%rec)) CALL pnode%rec%Modify(visitor)
    IF (ALLOCATED(pnode%round)) CALL pnode%round%Modify(visitor)
    IF (ALLOCATED(pnode%sign)) CALL pnode%sign%Modify(visitor)
    IF (ALLOCATED(pnode%size)) CALL pnode%size%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE iocs_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of IoControlSpecList%What - return a descriptive 
  !! string for the parse node.
  !!
  !! @returns The syntax rule name for the nodes.
  
  FUNCTION iocs_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'io-spec-control-list'
    
  END FUNCTION iocs_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of IoControlSpecList%GetLocation - get the start of the 
  !! source that corresponds to the ParseNode.
  !!
  !! @param[in]     pnode           The parse node of interest
  !!
  !! @returns The location in source of the start of the ParseNode.
  
  FUNCTION iocs_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !-------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(IoControlSpecList), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    loc = pnode%location
    
  END FUNCTION iocs_GetLocation
  
  
  !*****************************************************************************
  !!
  !> Implementation of IoControlSpecList%Check - check the parse node for 
  !! conformance with language standards.
  !!
  !! @param[in]     pnode             The parse node to check.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! @todo There are lots more errors that can be moved from the parsing 
  !! routine to here - particularly those that deal with incompatible 
  !! combinations.
  
  SUBROUTINE iocs_Check(pnode, err_list)
    
    USE Errors
    USE ErrorCodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(IoControlSpecList), INTENT(IN) :: pnode
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for an /io-control-spec/
    CHARACTER(*), PARAMETER :: comp = 'R913'
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (.NOT. ALLOCATED(pnode%unit)) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT='C911',  &
          LOCATION=.LocationOf. pnode,  &
          MSG='An /io-unit/ shall be specified in an /io-control-spec-list/.' )
    END IF
    
    IF (.NOT. ALLOCATED(pnode%advance)) THEN
      IF (ALLOCATED(pnode%eor)) THEN
        CALL Add( err_list,  &
            CODE=errSyntax,  &
            COMPONENT='C923 on ' // comp,  &
            LOCATION=.LocationOf. pnode%eor,  &
            MSG='If an "EOR=" specifier appears, an &
              &"ADVANCE=" specifier shall also appear.' )
      END IF
      IF (ALLOCATED(pnode%size)) THEN
        CALL Add( err_list,  &
            CODE=errSyntax,  &
            COMPONENT='C923 on ' // comp,  &
            LOCATION=.LocationOf. pnode%size,  &
            MSG='If a "SIZE=" specifier appears, an &
              &"ADVANCE=" specifier shall also appear.' )
      END IF
    END IF
    
    IF (ALLOCATED(pnode%decimal)) CALL fmt_check('DECIMAL', pnode%decimal)
    IF (ALLOCATED(pnode%blank))  CALL fmt_check('BLANK', pnode%blank)
    IF (ALLOCATED(pnode%pad))  CALL fmt_check('PAD', pnode%pad)
    IF (ALLOCATED(pnode%sign))  CALL fmt_check('SIGN', pnode%sign)
    IF (ALLOCATED(pnode%round))  CALL fmt_check('ROUND', pnode%round)
    
    IF (ALLOCATED(pnode%delim)) THEN
      IF (ALLOCATED(pnode%fmt)) THEN
        IF (.NOT. pnode%fmt%is_star) THEN
          CALL Add( err_list,  &
              CODE=errSyntax,  &
              COMPONENT='C930 on ' // comp,  &
              LOCATION=.LocationOf. pnode,  &
              MSG='If a "DELIM=" specifier appears and /format/ appears &
                &then /format/ must be a "*".' )
        END IF
      END IF
    END IF
    
  CONTAINS
    
    SUBROUTINE fmt_check(text, child_node)
      CHARACTER(*), INTENT(IN) :: text
      CLASS(ParseNode), INTENT(IN) :: child_node
      !*************************************************************************
      IF ( .NOT. ALLOCATED(pnode%fmt)  &
          .AND. (.NOT. ALLOCATED(pnode%nml))  &
          .AND. (.NOT. ALLOCATED(pnode%fmt_or_nml)) ) THEN
        CALL Add( err_list,  &
            CODE=errSyntax,  &
            COMPONENT='C924 on ' // comp,  &
            LOCATION=.LocationOf. child_node,  &
            MSG='If a "' // text // '=" specifier appears, a /format/ or &
              &/namelist-group-name/ shall also appear.' )
      END IF
    END SUBROUTINE fmt_check
    
  END SUBROUTINE iocs_Check
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for IOUnit.
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /io-unit/ (R901).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the first token of the 
  !! /io-unit/ in @a tlist.
  !!
  !! @param[in]     finish            Index of the last token of the 
  !! /io-unit/ in @a tlist.
  !!
  !! @param[out]    io_unit           The results of parsing the /io_unit/.  
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE parse_io_unit(tlist, start, finish, io_unit, err_list)
    
    USE Errors
    USE ErrorCodes
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(IOUnit), INTENT(OUT) :: io_unit
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /io-unit/.
    CHARACTER(*), PARAMETER :: comp = 'R901'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    IF (finish < start) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /io-unit/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    IF (Match(tlist, start, ittOperator, scck_'*')) THEN
      ! Take the /io-unit/ to be "*", which is not a valid starting point
      ! for an expression.
      io_unit%is_star = .TRUE.
      io_unit%star_location = QueryLocation(tlist(start))
      CALL ExpectEnd(tlist(:finish), start + 1, comp, err_list)
    ELSE
      ! It is either a /file-unit-number/ or a /internal-file-variable/, 
      ! both of which look like an expression.
      io_unit%is_star = .FALSE.
      CALL ParseExpr(tlist, start, finish, io_unit%expr, sub_err_list)
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
    END IF
    
  END SUBROUTINE parse_io_unit
  
  
  !*****************************************************************************
  !!
  !> Implementation of IOUnit%Write - write a text representation of the 
  !! parse node to a logical unit.
  !!
  !! @param[in]     pnode             The format.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output for the text represenation.
  
  SUBROUTINE iou_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(IOUnit), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    IF (pnode%is_star) THEN
      WRITE (unit, "('*')", ADVANCE='NO')
    ELSE
      CALL pnode%expr%Write(unit)
    END IF
    
  END SUBROUTINE iou_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of IOUnit%Query - traverse the node with a 
  !! query operation.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  SUBROUTINE iou_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(IOUnit), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%expr)) CALL pnode%expr%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE iou_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of IOUnit%Modify - traverse the node with a 
  !! modify operation.
  !!
  !! @param[in]     pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  SUBROUTINE iou_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(IOUnit), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%expr)) CALL pnode%expr%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE iou_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of IOUnit%What - return a descriptive 
  !! string for the parse node.
  !!
  !! @returns The syntax rule name for the nodes.
  
  FUNCTION iou_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'io-unit'
    
  END FUNCTION iou_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of IOUnit%GetLocation - get the start of the 
  !! source that corresponds to the ParseNode.
  !!
  !! @param[in]     pnode           The parse node of interest
  !!
  !! @returns The location in source of the start of the ParseNode.
  
  FUNCTION iou_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !-------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(IOUnit), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    IF (pnode%is_star) THEN
      loc = pnode%star_location
    ELSE
      loc = .LocationOf. pnode%expr
    END IF
    
  END FUNCTION iou_GetLocation
  
END MODULE IoControlSpecs
