! $Id: MpSubprogramStmts.f90 2161 2016-06-01 23:18:33Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the MpSubprogramStmts module.


!*******************************************************************************
!!
!> Procedures and types for parsing /mp-subprogram-stmt/'s and 
!! /end-mp-subprogram-stmt/'s.

MODULE MpSubprogramStmts
  
  USE Identifiers
  USE ParseNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseMpSubprogramStmt
  PUBLIC :: ParseEndMpSubprogramStmt
  
  !> Stores the results of parsing a /mp-subprogram-stmt/.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: MpSubprogramStmt  
    !> The identifier in the /mp-subprogram-stmt/.
    TYPE(Identifier) :: name
  CONTAINS
    PROCEDURE :: Dump => mps_Dump
    PROCEDURE :: Query => mps_Query
    PROCEDURE :: Modify => mps_Modify
    PROCEDURE, NOPASS :: What => mps_What
    PROCEDURE :: Advertise => mps_Advertise
  END TYPE MpSubprogramStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /mp-subprogram-stmt/
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    mps_stmt          The name of the separate module 
  !! procedure.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseMpSubprogramStmt(tlist, mps_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(MpSubprogramStmt), INTENT(OUT) :: mps_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for an /mp-subprogram-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R1238'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /mp-subprogram-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    CALL ExpectKeyword(tlist, 1, scck_'MODULE', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL ExpectKeyword(tlist, 2, scck_'PROCEDURE', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL ExpectName(tlist, 3, '/procedure-name/', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    mps_stmt%name = IdentifierIfortBug(tlist(3))
    
    CALL ExpectEOS(tlist, 4, comp, err_list)
    
  END SUBROUTINE ParseMpSubprogramStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /end-mp-subprogram-stmt/ (R1232).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     name              The /procedure-name/ provided on the 
  !! opening /mp-subprogram-stmt/.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseEndMpSubprogramStmt(tlist, name, err_list)
    
    USE CompilerKinds
    USE EndProgramUnitStmts
    USE Errors
    USE Identifiers
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(Identifier), INTENT(OUT), ALLOCATABLE :: name
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !***************************************************************************
    
    CALL ParseEndProgramUnitStmt( &
        tlist,  &
        scck_'PROCEDURE',  &
        'R1239',  &
        '/end-mp-subprogram-stmt/',  &
        '/procedure-name/',  &
        name,  &
        err_list )
    
  END SUBROUTINE ParseEndMpSubprogramStmt
  
  
  !*****************************************************************************
  !!
  !> Implementation of MpSubprogramStmt%Dump - write a debugging representation 
  !! of the results of parsing the statement to a logical unit.
  !!
  !! @param[in]     pnode             The MainProgram to dump.
  !!
  !! @param[in]     unit              Logical unit connected for 
  !! formatted output.
  !!
  !! @param[in]     pfx               Prefix for each record.
  
  SUBROUTINE mps_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(MpSubprogramStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    WRITE (unit, "(A,'/procedure-name/ = ',A)") pfx, .NameOf. pnode%name
    
  END SUBROUTINE mps_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of MPSubprogramStmt%Query - traverse the node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  RECURSIVE SUBROUTINE mps_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(MPSubprogramStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE mps_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of MPSubprogramStmt%Modify - traverse the node with 
  !! a modify operation.
  !!
  !! @param[in,out] pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  RECURSIVE SUBROUTINE mps_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(MPSubprogramStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE mps_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of MPSubprogramStmt%What - return a descriptive string 
  !! for the parse node.
  !!
  !! @returns 'mp-subprogram-stmt'
  
  FUNCTION mps_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'mp-subprogran-stmt'
    
  END FUNCTION mps_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of MPSubprogramStmt%Advertise - advertise the identifiers 
  !! referenced or defined by the main program.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference, from the 
  !! perspective of the host.
  !!
  !! The identifier for the procedure name is added to the scope as a result 
  !! of prepending the separate interface body for the procedure, hence this 
  !! statement advertises nothing.
  
  SUBROUTINE mps_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(MPSubprogramStmt), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(1))
    
    ident_list(1) = IdentifierNature(  &
        NAME=.NameOf. pnode%name,  &
        NATURES=[NatureDefProcedureDef],  &
        LOCATIONS=[.LocationOf. pnode%name] )
    
  END SUBROUTINE mps_Advertise
  
END MODULE MpSubprogramStmts
