! $Id: NullifyStmts.f90 2921 2019-04-26 22:08:35Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the NullifyStmts module.


!*******************************************************************************
!!
!> Procedures and types for parsing /nullify-stmt/'s (R638).

MODULE NullifyStmts
  
  USE Designators
  USE ParseNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseNullifyStmt
  
  !-----------------------------------------------------------------------------
  
  !> Results of parsing a /pointer-object/.
  !!
  !! This is a limited form of /designator/.
  TYPE, EXTENDS(TermParseNode), PUBLIC :: PointerObject
    !> The parts in the /pointer-object/.
    !!
    !! Always allocated, always size >= 1.
    TYPE(PartRef), ALLOCATABLE :: parts(:)
  CONTAINS
    PROCEDURE :: Write => po_Write
    PROCEDURE :: Query => po_Query
    PROCEDURE :: Modify => po_Modify
    PROCEDURE, NOPASS :: What => po_What
    PROCEDURE :: GetLocation => po_GetLocation
  END TYPE PointerObject
  
  !> Stores the results of parsing a /nullify-stmt/.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: NullifyStmt
    !> The objects listed in the nullify statement.
    !!
    !! Always allocated, always size >= 1.
    !!
    !! This could potentially just be an array of Designator.
    TYPE(PointerObject), ALLOCATABLE :: object_list(:)
  CONTAINS
    PROCEDURE :: Dump => nullify_Dump
    PROCEDURE :: Query => nullify_Query
    PROCEDURE :: Modify => nullify_Modify
    PROCEDURE, NOPASS :: What => nullify_What
  END TYPE NullifyStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /nullify-stmt/ (R638).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    object_list       Results of parsing the /nullify-stmt/.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseNullifyStmt(tlist, pnode, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(NullifyStmt), INTENT(OUT) :: pnode
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /nullify-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R638'
  
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect* calls.
    INTEGER :: closing        ! Index of the closing delimiter.
    
    
    INTEGER :: list_count     ! Number of value items in the object list.
    INTEGER :: item_start     ! Starting index of an item in the object list.
    INTEGER :: item_finish    ! Finishing index of an item in the object list.
    INTEGER :: item           ! Item index.
    
    ! Error list for procedure calls
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /nullify-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! NULLIFY ( ... )
    
    CALL ExpectKeyword(tlist, 1, scck_'NULLIFY', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL Expect(tlist, 2, ittDelimiter, scck_'(', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL ExpectClosing(tlist, 2, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    
    list_count = ListItemCount(tlist(3:closing-1))
    ALLOCATE(pnode%object_list(list_count))
    item_finish = 0
    DO item = 1, list_count
      CALL GetNextListItem(tlist, 3, closing - 1, item_start, item_finish)
      CALL parse_pointer_object(  &
          tlist,  &
          item_start,  &
          item_finish,  &
          pnode%object_list(item),  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
    END DO
    
    !---------------------------------------------------------------------------
    ! That's it
    
    CALL ExpectEOS(tlist, closing + 1, comp, err_list)
    
  END SUBROUTINE ParseNullifyStmt
  
  
  !*****************************************************************************
  !!
  !> Implementation of NullifyStmt%Dump - write a debugging representation 
  !! of the results of parsing a nullify statement to a logical unit.
  !!
  !! @param[in]     pnode             The ParseNode to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix for each record.
  
  SUBROUTINE nullify_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(NullifyStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Object index.
    
    !***************************************************************************
    
    WRITE (unit, "(A,'/pointer-object-list/')") pfx
    DO i = 1, SIZE(pnode%object_list)
      WRITE (unit, "(A)", ADVANCE='NO') pfx // '  '
      CALL pnode%object_list(i)%Write(unit)
      WRITE (unit, "()")
    END DO
    
  END SUBROUTINE nullify_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of NullifyStmt%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE nullify_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(NullifyStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Object index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%object_list)
      CALL pnode%object_list(i)%Query(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE nullify_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of NullifyStmt%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[in,out] pnode             The parse node element to modify.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE nullify_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(NullifyStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Object index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%object_list)
      CALL pnode%object_list(i)%Modify(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE nullify_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of NullifyStmt%What - return a descriptive string 
  !! for the parse node.
  !!
  !! @returns 'nullify-stmt'
  
  FUNCTION nullify_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'nullify-stmt'
    
  END FUNCTION nullify_What
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for PointerObject.
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /pointer-object/ (R639).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             The index of the first token in @a tlist 
  !! of the /pointer-object/.
  !!
  !! @param[in]     finish            The index of the last token in @a tlist 
  !! of the /pointer-object/.
  !!
  !! @param[out]    object            The results of parsing the 
  !! /pointer-object/.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE parse_pointer_object(tlist, start, finish, object, err_list)
    
    USE Errors
    USE ErrorCodes
    USE Identifiers
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(PointerObject), INTENT(OUT) :: object
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /pointer-object/.
    CHARACTER(*), PARAMETER :: comp = 'R639'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: part_count     ! Number of /part-ref/'s in the object.
    INTEGER :: after          ! Index of the token after the struct component.
    TYPE(Error) :: err        ! Error object for Expect calls.
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    IF (finish < start) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /pointer-object/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    part_count = CountParts(tlist, start, finish)
    
    IF (part_count > 1) THEN
      ! We must be dealing with a structure component.
      CALL ParseStructureComponent(  &
          tlist(:finish),  &
          start,  &
          object%parts,  &
          after,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
    ELSE
      ! Other options are names.
      CALL Expect(tlist, start, '/variable-name/ or /proc-pointer-name/', err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      ALLOCATE(object%parts(1))
      object%parts(1)%part_name = IdentifierIfortBug(tlist(start))
      after = start + 1
    END IF
    
    !---------------------------------------------------------------------------
    
    CALL ExpectEnd(tlist(:finish), after, comp, err_list)
    
  END SUBROUTINE parse_pointer_object
  
  
  !*****************************************************************************
  !!
  !> Implementation of PointerObject%Write - write a text representation 
  !! of the parse node to a logical unit.
  !!
  !! @param[in]     pnode             The format.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output for the text represenation.
  
  SUBROUTINE po_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(PointerObject), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: ip             ! Part index.
    
    !***************************************************************************
    
    DO ip = 1, SIZE(pnode%parts) - 1
      CALL WritePartRef(unit, pnode%parts(ip))
      WRITE (unit, "(' % ')", ADVANCE='NO')
    END DO
    CALL WritePartRef(unit, pnode%parts(ip))
    
  END SUBROUTINE po_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of PointerObject%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE po_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(PointerObject), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: ip             ! Object index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO ip = 1, SIZE(pnode%parts)
      CALL pnode%parts(ip)%Query(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE po_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of PointerObject%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[in,out] pnode             The parse node element to modify.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE po_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(PointerObject), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: ip             ! Object index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO ip = 1, SIZE(pnode%parts)
      CALL pnode%parts(ip)%Modify(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE po_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of PointerObject%What - return a descriptive string 
  !! for the parse node.
  !!
  !! @returns 'pointer-object'
  
  FUNCTION po_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'pointer-object'
    
  END FUNCTION po_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of PointerObject%GetLocation - get the start of the 
  !! source that corresponds to the ParseNode.
  !!
  !! @param[in]     pnode           The parse node of interest
  !!
  !! @returns The location in source of the start of the ParseNode.
  
  FUNCTION po_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !-------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(PointerObject), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    loc = .LocationOf. pnode%parts(1)
    
  END FUNCTION po_GetLocation
  
END MODULE NullifyStmts
