! $Id: OpenStmts.f90 2161 2016-06-01 23:18:33Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the OpenStmts module.


!*******************************************************************************
!!
!> Types and procedures for parsing /open-stmt/'s (R904).

MODULE OpenStmts
  
  USE Expressions
  USE Labels
  USE Variables
  USE ParseNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseOpenStmt
  
  !-----------------------------------------------------------------------------
  
  !> The results of parsing an /open-stmt/.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: OpenStmt
    !> The UNIT specification.
    CLASS(Expression), ALLOCATABLE :: unit
    !> The ACCESS specification.
    CLASS(Expression), ALLOCATABLE :: access
    !> The ACTION specification.
    CLASS(Expression), ALLOCATABLE :: action
    !> The ASYNCHRONOUS specification.
    CLASS(Expression), ALLOCATABLE :: asynchronous
    !> The BLANK specification.
    CLASS(Expression), ALLOCATABLE :: blank
    !> The DECIMAL specification.
    CLASS(Expression), ALLOCATABLE :: decimal
    !> The DELIM specification.
    CLASS(Expression), ALLOCATABLE :: delim
    !> The ENCODING specification.
    CLASS(Expression), ALLOCATABLE :: encoding
    !> The ERR specification.
    !!
    !! Not allocated if there was no ERR specifier.
    TYPE(Label), ALLOCATABLE :: err
    !> The FILE specification.
    !!
    !! This is a /file-name-expr/.
    CLASS(Expression), ALLOCATABLE :: file
    !> The FORM specification.
    CLASS(Expression), ALLOCATABLE :: form
    !> The IOMSG specification.
    !!
    !! This is a variable (an /iomsg-variable/).
    CLASS(Variable), ALLOCATABLE :: iomsg
    !> The IOSTAT specification.
    !!
    !! This is a variable.
    CLASS(Variable), ALLOCATABLE :: iostat
    !> The NEWUNIT specification.
    !!
    !! This is a variable.
    CLASS(Variable), ALLOCATABLE :: newunit
    !> The PAD specification.
    CLASS(Expression), ALLOCATABLE :: pad
    !> The POSITION specification.
    CLASS(Expression), ALLOCATABLE :: position
    !> The RECL specification.
    CLASS(Expression), ALLOCATABLE :: recl
    !> The ROUND specification.
    CLASS(Expression), ALLOCATABLE :: round
    !> The SIGN specification.
    CLASS(Expression), ALLOCATABLE :: sign
    !> The STATUS specification.
    CLASS(Expression), ALLOCATABLE :: status
  CONTAINS
    PROCEDURE :: Dump => open_Dump
    PROCEDURE :: Query => open_Query
    PROCEDURE :: Modify => open_Modify
    PROCEDURE, NOPASS :: What => open_What
  END TYPE OpenStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /open-stmt/ (R904).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    open_stmt         Results of parsing the /open-stmt/.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseOpenStmt(tlist, open_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(OpenStmt), INTENT(OUT) :: open_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for an /open-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R904'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: closing        ! Index of the closing delimiter.
    INTEGER :: list_count     ! Number of items in the /connect-spec-list/.
    INTEGER :: item           ! List item index.
    INTEGER :: item_start     ! Start index in tlist of a particular item.
    INTEGER :: item_finish    ! Finish index in tlist of a particular item.
    LOGICAL :: no_spec        ! Flag to indicate UNIT= was absent.
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
          MSG='Expected /open-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! OPEN...
    
    CALL ExpectKeyword(tlist, 1, scck_'OPEN', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! OPEN(...)
    
    CALL Expect(tlist, 2, ittDelimiter, scck_'(', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL ExpectClosing(tlist, 2, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! /connect-spec-list/
    
    list_count = ListItemCount(tlist(3:closing-1))
    item_finish = 0
    DO item = 1, list_count
      CALL GetNextListItem(tlist, 3, closing - 1, item_start, item_finish)
      CALL parse_connect_spec( tlist, item_start, item_finish,  &
          open_stmt, no_spec, sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      ! Continue on error.
      IF ((item /= 1) .AND. no_spec) THEN
        CALL Add( err_list, CODE=errSyntax, COMPONENT='C904 on ' // comp,  &
!            LOCATION=tlist(item_start)%location,  
            MSG='If the optional "UNIT=" is omitted, the /file-unit-number/ &
              &shall be the first item in the /connect-spec-list/.' )
        ! Continue on error.
      END IF
    END DO
    
    !---------------------------------------------------------------------------
    ! Check constraints
    
    IF ( .NOT. ALLOCATED(open_stmt%unit)  &
        .AND. (.NOT. ALLOCATED(open_stmt%newunit)) ) THEN
      CALL Add( err_list, CODE=errSyntax, COMPONENT='C904 on ' // comp,  &
!          LOCATION=tlist(item_start)%location,  
          MSG='If the "NEWUNIT=" specifier does not appear a &
            &/file-unit-number/ shall be specified.' )
      ! Continue on error.
    ELSE IF ( ALLOCATED(open_stmt%unit)  &
        .AND. ALLOCATED(open_stmt%newunit) ) THEN
      ! This test is here rather than in parse_connect_spec as the constraint 
      ! applies to the parent statement in the standard text.
      CALL Add( err_list, CODE=errSyntax, COMPONENT='C906 on ' // comp,  &
!          LOCATION=tlist(item_start)%location,  
          MSG='If a "NEWUNIT=" specifier appears a &
            &/file-unit-number/ shall not appear.' )
      ! Continue on error.
    END IF
    
    !---------------------------------------------------------------------------
    ! That's it.
    
    CALL ExpectEOS(tlist, closing + 1, comp, err_list)
    
  END SUBROUTINE ParseOpenStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /connect-spec/ (R905).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the first token of the 
  !! /connect-spec/ in @a tlist.
  !!
  !! @param[in]     finish            Index of the last token of the 
  !! /connect-spec/ in @a tlist.
  !!
  !! @param[in,out] open_stmt         Results of parsing the /connect-spec/.  
  !! Note that this is an INTENT(INOUT) argument - each spec sets a particular 
  !! component
  !!
  !! @param[out]    no_spec           Flag to indicate that the "XXX=" part of 
  !! the spec was absent.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE parse_connect_spec( tlist, start, finish, open_stmt, no_spec,  &
      err_list )
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE IoControlSpecs
    USE MatchUtils
    USE ParseUtils
    USE Scanner
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(OpenStmt), INTENT(INOUT) :: open_stmt
    LOGICAL, INTENT(OUT) :: no_spec
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /connect-spec/.
    CHARACTER(*), PARAMETER :: comp = 'R905'
    
    ! Keywords for each /connect-spec/.
    CHARACTER(*,KIND=scck), PARAMETER :: keywords(20) = [  &
        scck_'UNIT        ', scck_'ACCESS      ', scck_'ACTION      ',  &
        scck_'ASYNCHRONOUS', scck_'BLANK       ', scck_'DECIMAL     ',  &
        scck_'DELIM       ', scck_'ENCODING    ', scck_'ERR         ',  &
        scck_'FILE        ', scck_'FORM        ', scck_'IOMSG       ',  &
        scck_'IOSTAT      ', scck_'NEWUNIT     ', scck_'PAD         ',  &
        scck_'POSITION    ', scck_'RECL        ', scck_'ROUND       ',  &
        scck_'SIGN        ', scck_'STATUS      ' ]
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: opt            ! Expect option.
    INTEGER :: assign         ! Index of the =.
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    IF (finish < start) THEN
      CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
          MSG='Expected /connect-spec/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! Do we have an  argument keyword?
    
    assign = FindBare(tlist(:finish), ittAssign, scck_'=', start)
    IF (assign == 0) THEN
      ! No - this is a unit spec.
      no_spec = .TRUE.
      opt = 1
    ELSE
      no_spec = .FALSE.
      CALL ExpectKeyword(tlist, start, keywords, '/connect-spec/', opt, err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      CALL Expect(tlist, start + 1, ittAssign, scck_'=', err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    SELECT CASE (opt)
    CASE (1)      ! UNIT
      IF (ALLOCATED(open_stmt%unit)) THEN
        CALL already('UNIT')
      ELSE
        IF (assign == 0) THEN
          CALL ParseFileUnitNumber(  &
              tlist,  &
              start, finish,  &
              open_stmt%unit,  &
              sub_err_list )
        ELSE
          CALL ParseFileUnitNumber(  &
              tlist,  &
              assign + 1, finish,  &
              open_stmt%unit,  &
              sub_err_list )
        END IF
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
       
    CASE (2)      ! ACCESS
      IF (ALLOCATED(open_stmt%access)) THEN
        CALL already('ACCESS')
      ELSE
        CALL ParseExpr( &
            tlist,  &
            assign + 1, finish,  &
            open_stmt%access,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (3)      ! ACTION
      IF (ALLOCATED(open_stmt%action)) THEN
        CALL already('ACTION')
      ELSE
        CALL ParseExpr(  &
            tlist,  &
            assign + 1, finish,  &
            open_stmt%action,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (4)      ! ASYNCHRONOUS
      IF (ALLOCATED(open_stmt%asynchronous)) THEN
        CALL already('ASYNCHRONOUS')
      ELSE
        CALL ParseExpr(  &
            tlist,  &
            assign + 1, finish,  &
            open_stmt%asynchronous,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (5)      ! BLANK
      IF (ALLOCATED(open_stmt%blank)) THEN
        CALL already('BLANK')
      ELSE
        CALL ParseExpr(  &
            tlist,  &
            assign + 1, finish,  &
            open_stmt%blank,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (6)      ! DECIMAL
      IF (ALLOCATED(open_stmt%decimal)) THEN
        CALL already('DECIMAL')
      ELSE
        CALL ParseExpr(  &
            tlist,  &
            assign + 1, finish,  &
            open_stmt%decimal,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (7)      ! DELIM
      IF (ALLOCATED(open_stmt%delim)) THEN
        CALL already('DELIM')
      ELSE
        CALL ParseExpr(  &
            tlist,  &
            assign + 1, finish,  &
            open_stmt%delim,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (8)      ! ENCODING
      IF (ALLOCATED(open_stmt%encoding)) THEN
        CALL already('ENCODING')
      ELSE
        CALL ParseExpr(  &
            tlist,  &
            assign + 1, finish,  &
            open_stmt%encoding,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (9)      ! ERR
      IF (ALLOCATED(open_stmt%err)) THEN
        CALL already('ERR')
      ELSE
        ALLOCATE(open_stmt%err)
        CALL ParseLabel(  &
            tlist,  &
            assign + 1,  &
            finish,  &
            open_stmt%err,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
        IF (Failure(sub_err_list)) RETURN
      END IF
      
    CASE (10)     ! FILE
      IF (ALLOCATED(open_stmt%file)) THEN
        CALL already('FILE')
      ELSE
        ! The syntax term is actually a /file-name-expr/.
        CALL ParseExpr(  &
            tlist,  &
            assign + 1, finish,  &
            open_stmt%file,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (11)     ! FORM
      IF (ALLOCATED(open_stmt%form)) THEN
        CALL already('FORM')
      ELSE
        CALL ParseExpr(  &
            tlist,  &
            assign + 1, finish,  &
            open_stmt%form,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (12)     ! IOMSG
      IF (ALLOCATED(open_stmt%iomsg)) THEN
        CALL already('IOMSG')
      ELSE
        ! The syntax term is actually an /iomsg-variable/.
        CALL ParseVariable(  &
            tlist,  &
            assign + 1, finish,  &
            open_stmt%iomsg,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (13)     ! IOSTAT
      IF (ALLOCATED(open_stmt%iostat)) THEN
        CALL already('IOSTAT')
      ELSE
        CALL ParseVariable(  &
            tlist,  &
            assign + 1, finish,  &
            open_stmt%iostat,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (14)     ! NEWUNIT
      IF (ALLOCATED(open_stmt%newunit)) THEN
        CALL already('NEWUNIT')
      ELSE
        CALL ParseVariable(  &
            tlist,  &
            assign + 1, finish,  &
            open_stmt%newunit,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (15)     ! PAD
      IF (ALLOCATED(open_stmt%pad)) THEN
        CALL already('PAD')
      ELSE
        CALL ParseExpr(  &
            tlist,  &
            assign + 1, finish,  &
            open_stmt%pad,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (16)     ! POSITION
      IF (ALLOCATED(open_stmt%position)) THEN
        CALL already('POSITION')
      ELSE
        CALL ParseExpr(  &
            tlist,  &
            assign + 1, finish,  &
            open_stmt%position,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (17)     ! RECL
      IF (ALLOCATED(open_stmt%recl)) THEN
        CALL already('RECL')
      ELSE
        CALL ParseExpr(  &
            tlist,  &
            assign + 1, finish,  &
            open_stmt%recl,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (18)     ! ROUND
      IF (ALLOCATED(open_stmt%round)) THEN
        CALL already('ROUND')
      ELSE
        CALL ParseExpr(  &
            tlist,  &
            assign + 1, finish,  &
            open_stmt%round,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (19)     ! SIGN
      IF (ALLOCATED(open_stmt%sign)) THEN
        CALL already('SIGN')
      ELSE
        CALL ParseExpr(  &
            tlist,  &
            assign + 1, finish,  &
            open_stmt%sign,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (20)     ! STATUS
      IF (ALLOCATED(open_stmt%status)) THEN
        CALL already('STATUS')
      ELSE
        CALL ParseExpr(  &
            tlist,  &
            assign + 1, finish,  &
            open_stmt%status,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE DEFAULT
      STOP 'Unknown /connect-spec/.'
      
    END SELECT
    
  CONTAINS
    
    SUBROUTINE already(text)
      USE ErrorCodes
      !-------------------------------------------------------------------------
      ! Arguments
      CHARACTER(*), INTENT(IN) :: text
      !*************************************************************************
      CALL Add( err_list, CODE=errSyntax, COMPONENT='C903 on ' // comp,  &
!          LOCATION=tlist(start)%location,
          MSG=text // ' has already been specified for this /open-stmt/.' )
    END SUBROUTINE already
    
  END SUBROUTINE parse_connect_spec
  
  
  !*****************************************************************************
  !!
  !> Implementation of OpenStmt%Dump - write a debugging representation 
  !! of the results of parsing an open statement to a logical unit.
  !!
  !! @param[in]     pnode             The ParseNode to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix for each record.
  
  SUBROUTINE open_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(OpenStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%unit)) THEN
      WRITE (unit, "(A,'UNIT = ')", ADVANCE='NO') pfx
      CALL WriteExpr(unit, pnode%unit)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%access)) THEN
      WRITE (unit, "(A,'ACCESS = ')", ADVANCE='NO') pfx
      CALL WriteExpr(unit, pnode%access)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%action)) THEN
      WRITE (unit, "(A,'ACTION = ')", ADVANCE='NO') pfx
      CALL WriteExpr(unit, pnode%action)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%asynchronous)) THEN
      WRITE (unit, "(A,'ASYNCHRONOUS = ')", ADVANCE='NO') pfx
      CALL WriteExpr(unit, pnode%asynchronous)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%blank)) THEN
      WRITE (unit, "(A,'BLANK = ')", ADVANCE='NO') pfx
      CALL WriteExpr(unit, pnode%blank)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%decimal)) THEN
      WRITE (unit, "(A,'DECIMAL = ')", ADVANCE='NO') pfx
      CALL WriteExpr(unit, pnode%decimal)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%delim)) THEN
      WRITE (unit, "(A,'DELIM = ')", ADVANCE='NO') pfx
      CALL WriteExpr(unit, pnode%delim)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%encoding)) THEN
      WRITE (unit, "(A,'ENCODING = ')", ADVANCE='NO') pfx
      CALL WriteExpr(unit, pnode%encoding)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%err)) THEN
      WRITE (unit, "(A,'ERR = ',I0)") pfx, .ValueOf. pnode%err
    END IF
    IF (ALLOCATED(pnode%file)) THEN
      WRITE (unit, "(A,'FILE = ')", ADVANCE='NO') pfx
      CALL WriteExpr(unit, pnode%file)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%form)) THEN
      WRITE (unit, "(A,'FORM = ')", ADVANCE='NO') pfx
      CALL WriteExpr(unit, pnode%form)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%iomsg)) THEN
      WRITE (unit, "(A,'IOMSG = ')", ADVANCE='NO') pfx
      CALL WriteVariable(unit, pnode%iomsg)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%iostat)) THEN
      WRITE (unit, "(A,'IOSTAT = ')", ADVANCE='NO') pfx
      CALL WriteVariable(unit, pnode%iostat)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%newunit)) THEN
      WRITE (unit, "(A,'NEWUNIT = ')", ADVANCE='NO') pfx
      CALL WriteVariable(unit, pnode%newunit)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%pad)) THEN
      WRITE (unit, "(A,'PAD = ')", ADVANCE='NO') pfx
      CALL WriteExpr(unit, pnode%pad)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%position)) THEN
      WRITE (unit, "(A,'POSITION = ')", ADVANCE='NO') pfx
      CALL WriteExpr(unit, pnode%position)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%recl)) THEN
      WRITE (unit, "(A,'RECL = ')", ADVANCE='NO') pfx
      CALL WriteExpr(unit, pnode%recl)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%round)) THEN
      WRITE (unit, "(A,'ROUND = ')", ADVANCE='NO') pfx
      CALL WriteExpr(unit, pnode%round)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%sign)) THEN
      WRITE (unit, "(A,'SIGN = ')", ADVANCE='NO') pfx
      CALL WriteExpr(unit, pnode%sign)
      WRITE (unit, "()")
    END IF
    IF (ALLOCATED(pnode%status)) THEN
      WRITE (unit, "(A,'STATUS = ')", ADVANCE='NO') pfx
      CALL WriteExpr(unit, pnode%status)
      WRITE (unit, "()")
    END IF
    
  END SUBROUTINE open_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of OpenStmt%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE open_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(OpenStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%unit)) CALL pnode%unit%Query(visitor)
    IF (ALLOCATED(pnode%access)) CALL pnode%access%Query(visitor)
    IF (ALLOCATED(pnode%action)) CALL pnode%action%Query(visitor)
    IF (ALLOCATED(pnode%asynchronous)) CALL pnode%asynchronous%Query(visitor)
    IF (ALLOCATED(pnode%blank)) CALL pnode%blank%Query(visitor)
    IF (ALLOCATED(pnode%decimal)) CALL pnode%decimal%Query(visitor)
    IF (ALLOCATED(pnode%delim)) CALL pnode%delim%Query(visitor)
    IF (ALLOCATED(pnode%encoding)) CALL pnode%encoding%Query(visitor)
    ! err
    IF (ALLOCATED(pnode%file)) CALL pnode%file%Query(visitor)
    IF (ALLOCATED(pnode%form)) CALL pnode%form%Query(visitor)
    IF (ALLOCATED(pnode%iomsg)) CALL pnode%iomsg%Query(visitor)
    IF (ALLOCATED(pnode%iostat)) CALL pnode%iostat%Query(visitor)
    IF (ALLOCATED(pnode%newunit)) CALL pnode%newunit%Query(visitor)
    IF (ALLOCATED(pnode%pad)) CALL pnode%pad%Query(visitor)
    IF (ALLOCATED(pnode%position)) CALL pnode%position%Query(visitor)
    IF (ALLOCATED(pnode%recl)) CALL pnode%recl%Query(visitor)
    IF (ALLOCATED(pnode%round)) CALL pnode%round%Query(visitor)
    IF (ALLOCATED(pnode%sign)) CALL pnode%sign%Query(visitor)
    IF (ALLOCATED(pnode%status)) CALL pnode%status%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE open_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of OpenStmt%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[in,out] pnode             The parse node element to query.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE open_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(OpenStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%unit)) CALL pnode%unit%Modify(visitor)
    IF (ALLOCATED(pnode%access)) CALL pnode%access%Modify(visitor)
    IF (ALLOCATED(pnode%action)) CALL pnode%action%Modify(visitor)
    IF (ALLOCATED(pnode%asynchronous)) CALL pnode%asynchronous%Modify(visitor)
    IF (ALLOCATED(pnode%blank)) CALL pnode%blank%Modify(visitor)
    IF (ALLOCATED(pnode%decimal)) CALL pnode%decimal%Modify(visitor)
    IF (ALLOCATED(pnode%delim)) CALL pnode%delim%Modify(visitor)
    IF (ALLOCATED(pnode%encoding)) CALL pnode%encoding%Modify(visitor)
    ! err
    IF (ALLOCATED(pnode%file)) CALL pnode%file%Modify(visitor)
    IF (ALLOCATED(pnode%form)) CALL pnode%form%Modify(visitor)
    IF (ALLOCATED(pnode%iomsg)) CALL pnode%iomsg%Modify(visitor)
    IF (ALLOCATED(pnode%iostat)) CALL pnode%iostat%Modify(visitor)
    IF (ALLOCATED(pnode%newunit)) CALL pnode%newunit%Modify(visitor)
    IF (ALLOCATED(pnode%pad)) CALL pnode%pad%Modify(visitor)
    IF (ALLOCATED(pnode%position)) CALL pnode%position%Modify(visitor)
    IF (ALLOCATED(pnode%recl)) CALL pnode%recl%Modify(visitor)
    IF (ALLOCATED(pnode%round)) CALL pnode%round%Modify(visitor)
    IF (ALLOCATED(pnode%sign)) CALL pnode%sign%Modify(visitor)
    IF (ALLOCATED(pnode%status)) CALL pnode%status%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE open_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of OpenStmt%What - return a descriptive string 
  !! for the parse node.
  !!
  !! @returns 'open-stmt'
  
  FUNCTION open_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'open-stmt'
    
  END FUNCTION open_What
  
END MODULE OpenStmts
