! $Id: PointerStmts.f90 2904 2019-04-11 22:03:02Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the PointerStmts module.


!*******************************************************************************
!!
!> Procedures and types for parsing /pointer-stmt/'s (R550).

MODULE PointerStmts
  
  USE CompilerKinds
  USE DimensionStmts
  USE Identifiers
  USE SubscriptLists
  USE ParseNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParsePointerStmt
  
  !-----------------------------------------------------------------------------
  
  !> The results of parsing a /pointer-decl/.
  TYPE, EXTENDS(TermParseNode), PUBLIC :: PointerDecl
    !> The name of the object being declared.
    TYPE(Identifier) :: name
    !> An /array-spec/, if present, or not allocated if not.
    TYPE(DimSpec), ALLOCATABLE :: array_spec(:)
  CONTAINS
    PROCEDURE :: Write => ptrdecl_Write
    PROCEDURE :: Query => ptrdecl_Query
    PROCEDURE :: Modify => ptrdecl_Modify
    PROCEDURE, NOPASS :: What => ptrdecl_What
    PROCEDURE :: Advertise => ptrdecl_Advertise
    PROCEDURE :: GetLocation => ptrdecl_GetLocation
  END TYPE PointerDecl
  
  !> Stores the results of parsing a pointer statement.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: PointerStmt
    !> The list of entities being given the pointer attribute.
    TYPE(PointerDecl), ALLOCATABLE :: decls(:)
  CONTAINS
    PROCEDURE :: Dump => ptr_Dump
    PROCEDURE :: Query => ptr_Query
    PROCEDURE :: Modify => ptr_Modify
    PROCEDURE, NOPASS :: What => ptr_What
  END TYPE PointerStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /pointer-stmt/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    ptr_stmt          The results of parsing the statement.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParsePointerStmt(tlist, ptr_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !--------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(PointerStmt), INTENT(OUT) :: ptr_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule for a /pointer-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R550'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: opt            ! Expect option.
    INTEGER :: it             ! Current token index.
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    INTEGER :: list_count     ! Number of items in the /pointer-decl-list/.
    INTEGER :: item           ! List item index.
    INTEGER :: item_start     ! Start index in tlist of a particular item.
    INTEGER :: item_finish    ! Finish index in tlist of a particular item.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /pointer-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! POINTER...
    
    CALL ExpectKeyword(tlist, 1, scck_'POINTER', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! :: or /pointer-decl/.  A /pointer-decl/ starts with a /name/.
    
    CALL Expect(  &
        tlist, 2,  &
        [ittDeclare, ittName],  &
        [scck_'::', scck_'  '],  &
        '"::" or /pointer-decl/',  &
        opt, err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    IF (opt == 1) THEN
      it = 3
    ELSE
      it = 2
    END IF
    ! it points to the start of the /pointer-decl-list/ here.
    
    !---------------------------------------------------------------------------
    ! /pointer-decl-list/
    
    list_count = ListItemCount(tlist(it:))
    ALLOCATE(ptr_stmt%decls(list_count))
    item_finish = 0
    DO item = 1, list_count
      CALL GetNextListItem(tlist, it, SIZE(tlist), item_start, item_finish)
      CALL parse_pointer_decl(  &
          tlist,  &
          item_start, item_finish,  &
          ptr_stmt%decls(item),  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
    END DO
    
    ! That's it!
    
  END SUBROUTINE ParsePointerStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /pointer-decl/ (R551).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    decl              Results of parsing the /pointer-decl/.  
  !! Not defined on error.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE parse_pointer_decl(tlist, start, finish, decl, err_list)
    
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(PointerDecl), INTENT(OUT) :: decl
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule for a /pointer-decl/.
    CHARACTER(*), PARAMETER :: comp = 'R551'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: opt            ! Expect option.
    INTEGER :: closing        ! Index of the closing delimiter.
    
    ! Error list for ParseDeferredShapeSpecList call.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    IF (finish < start) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /pointer-decl/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    CALL ExpectName(tlist, start, '/object-name/ or /proc-entity-name/', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    decl%name = IdentifierIfortBug(tlist(start))
    
    !---------------------------------------------------------------------------
    ! Either that is it, or we have ( /deferred-shape-spec-list/)
    
    CALL Expect(  &
        tlist(:finish),  &
        start + 1,  &
        [ittEndStatement, ittDelimiter],  &
        [scck_' ', scck_'('],  &
        'end-of-statement, "," or "("',  &
        opt,  &
        err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    IF (opt == 1) RETURN    ! That's it
    
    !---------------------------------------------------------------------------
    ! /deferred-shape-spec-list/ )
    
    CALL ExpectClosing(tlist, start + 1, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL ParseDeferredShapeSpecList(  &
        tlist(start+2:closing-1),  &
        decl%array_spec,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
    !---------------------------------------------------------------------------
    ! Check nothing superfluous after the closing ")".
    
    CALL ExpectEOS(tlist(:finish), closing + 1, comp, err_list)
    
  END SUBROUTINE parse_pointer_decl
  
  
  !*****************************************************************************
  !!
  !> Implementation of PointerStmt%Dump - write a debugging representation 
  !! of the results of parsing a pointer statement to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE ptr_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(PointerStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Decl index.
    
    !***************************************************************************
    
    DO i = 1, SIZE(pnode%decls)
      IF (ALLOCATED(pnode%decls(i)%array_spec)) THEN
        WRITE (unit, "(A,'/object-name/ = ',A)") pfx,  &
            .NameOf. pnode%decls(i)%name
        WRITE ( unit, "(A,'/deferred-shape-spec-list/ = ( ')",  &
            ADVANCE='NO' ) pfx
        CALL WriteDimSpec(unit, pnode%decls(i)%array_spec)
        WRITE (unit, "(' )')")
      ELSE
        WRITE (unit, "(A,'object or proc pointer name = ',A)")  &
            pfx, .NameOf. pnode%decls(i)%name
      END IF
    END DO
    
  END SUBROUTINE ptr_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of PointerStmt%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE ptr_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(PointerStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Pointer decl index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%decls)
      CALL pnode%decls(i)%Query(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE ptr_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of PointerStmt%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[in,out] pnode             The parse node element to modify.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE ptr_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(PointerStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Pointer decl index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%decls)
      CALL pnode%decls(i)%Modify(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE ptr_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of PointerStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'pointer-stmt'.
  
  FUNCTION ptr_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'pointer-stmt'
    
  END FUNCTION ptr_What
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for bindings in PointerDecl.
  
  
  !*****************************************************************************
  !!
  !> Implementation of PointerDecl%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'pointer-decl'.
  
  FUNCTION ptrdecl_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'pointer-decl'
    
  END FUNCTION ptrdecl_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of PointerDecl%Write - write a debugging representation 
  !! of the node to a logical unit.
  !!
  !1 @param[in]     pnode             The node to write.
  !!
  !! @param[in]     unit              Unit connected for formatted output.
  
  SUBROUTINE ptrdecl_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(PointerDecl), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    WRITE (unit, "('/object-name/ = ',A)",  ADVANCE='NO') &
          .NameOf. pnode%name
    IF (ALLOCATED(pnode%array_spec)) THEN
      WRITE (unit, "(' (')", ADVANCE='NO')
      CALL WriteDimSpec(unit, pnode%array_spec)
      WRITE (unit, "(')')", ADVANCE='NO')
    END IF
    
  END SUBROUTINE ptrdecl_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of PointerDecl%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE ptrdecl_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(PointerDecl), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Array spec index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%array_spec)) THEN
      DO i = 1, SIZE(pnode%array_spec)
        CALL pnode%array_spec(i)%Query(visitor)
      END DO
    END IF
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE ptrdecl_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of PointerDecl%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[in,out] pnode             The parse node element to modify.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE ptrdecl_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(PointerDecl), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Array spec index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%array_spec)) THEN
      DO i = 1, SIZE(pnode%array_spec)
        CALL pnode%array_spec(i)%Modify(visitor)
      END DO
    END IF
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE ptrdecl_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of PointerDecl%Advertise - advertise the identifiers 
  !! referenced or defined by the construct.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE ptrdecl_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(PointerDecl), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(1))
    
    IF (ALLOCATED(pnode%array_spec)) THEN
      ident_list(1) = IdentifierNature(  &
          NAME=.NameOf. pnode%name,  &
          NATURES=[NatureDefPointer, NatureDefDimension], &
          LOCATIONS=[.LocationOf. pnode%name, .LocationOf. pnode%name] )
    ELSE
      ident_list(1) = IdentifierNature(  &
          NAME=.NameOf. pnode%name,  &
          NATURES=[NatureDefPointer], &
          LOCATIONS=[.LocationOf. pnode%name] )
    END IF
    
  END SUBROUTINE ptrdecl_Advertise
  
  
  !*****************************************************************************
  !!
  !> Implementation of PointerDecl%GetLocation - get the start of the 
  !! source that corresponds to the ParseNode.
  !!
  !! @param[in]     pnode           The parse node of interest
  !!
  !! @returns The location in source of the start of the ParseNode.
  
  FUNCTION ptrdecl_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !-------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(PointerDecl), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    loc = .LocationOf. pnode%name
    
  END FUNCTION ptrdecl_GetLocation
  
END MODULE PointerStmts
