! $Id: PrimaryExpressions.f90 2798 2019-03-22 17:18:41Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the PrimaryExpressions module.


!*******************************************************************************
!!
!> Types to store the result of parsing the various primary expressions.
!!
!! Some of these types used tokens to store the underlying data.  To 
!! allow better separation between the results of parsing and the scanner 
!! this should be changed to actually store the data that is needed (the 
!! token text, its kind (if applicable), location).

MODULE PrimaryExpressions
  
  USE ArrayConstructors
  USE CallStmts
  USE ComplexLiteralConstants
  USE DeclarationTypeSpecs
  USE Designators
  USE Expressions
  USE Identifiers
  USE ParseNodes
  USE SourceLocations
  USE StructureConstructors
  USE SubscriptLists
  USE SubstringRanges
  USE Tokens
  
  IMPLICIT NONE
  
  PRIVATE
  
  PUBLIC :: ParseFunctionRefExprFromDesignator
  PUBLIC :: ParseStructureConstructorFromDesignator
  
  !-----------------------------------------------------------------------------
  ! The primaries.
  
  !> A primary expression that is a constant that can be represented by a 
  !! single token.
  !!
  !! The name for this type is a little misleading - as 
  !! /complex-literal-constant/'s are primaries which are classed as constants, 
  !! but they cannot be represented as a single token.  Given the type is more 
  !! around the storage of information rather than the syntactic role, it 
  !! should probably be renamed to TokenExpr or similar.
  !!
  !! But then see comments for the module - perhaps this should be split 
  !! up further to avoid using the Token type as storage.
  !!
  !! This is also used for type-param-name expressions inside the type 
  !! definition.
  TYPE, EXTENDS(Expression), PUBLIC :: ConstantExpr
    !> The token that represents the constant.
    TYPE(Token) :: tok
  CONTAINS
    PROCEDURE :: Write => constant_Write
    PROCEDURE :: GetLocation => constant_GetLocation
    PROCEDURE :: Query => constant_Query
    PROCEDURE :: Modify => constant_Modify
    PROCEDURE :: Advertise => constant_Advertise
  END TYPE ConstantExpr
  
  !> A primary that is a /complex-literal-constant/.
  !!
  !! The parts of a complex constant cannot be expressions, but they can be 
  !! named constants that are defined by expressions.  Go figure.
  !!
  !! See comments for the module around not using the Token type for 
  !! storing the results of parsing.
  !!
  !! The contents of this type are the same as the contents of the 
  !! DataStmtConstantComplexLiteral type.
  TYPE, EXTENDS(Expression), PUBLIC :: ComplexLiteralExpr
    !> The location of the opening parenthesis.
    TYPE(SourceLocation) :: paren_location
    !> The results of parsing the constant.
    TYPE(ComplexLiteralConstant) :: constant
  CONTAINS
    PROCEDURE :: Write => complex_Write
    PROCEDURE :: GetLocation => complex_GetLocation
    PROCEDURE :: Query => complex_Query
    PROCEDURE :: Modify => complex_Modify
    PROCEDURE :: Advertise => complex_Advertise
  END TYPE ComplexLiteralExpr
    
  !> A primary expression that is a designator.
  !!
  !! This may also used for expressions that are a /type-param-inquiry/, 
  !! but we cannot tell them apart.
  TYPE, EXTENDS(Expression), PUBLIC :: DesignatorExpr
    TYPE(Designator) :: desig
  CONTAINS
    PROCEDURE :: Write => desig_Write
    PROCEDURE :: GetLocation => desig_GetLocation
    PROCEDURE :: Query => desig_Query
    PROCEDURE :: Modify => desig_Modify
    PROCEDURE :: Advertise => desig_Advertise
  END TYPE DesignatorExpr
  
  !> An expression that is an array constructor.
  TYPE, EXTENDS(Expression), PUBLIC :: ArrayConstructorExpr
    !> /type-spec/, if present.
    !!
    !! Not allocated if not present.
    TYPE(TypeSpec), ALLOCATABLE :: type_spec
    !> The values in the array constructor.
    !!
    !! Always allocated but may be zero length.  If zero length, then the 
    !! type_spec component will be allocated.
    TYPE(ACValueList), ALLOCATABLE :: values(:)
  CONTAINS
    PROCEDURE :: Write => array_Write
    PROCEDURE :: GetLocation => array_GetLocation
    PROCEDURE :: Query => array_Query
    PROCEDURE :: Modify => array_Modify
    PROCEDURE :: Advertise => array_Advertise
  END TYPE ArrayConstructorExpr
  
  !> An expression that is a structure constructor.
  TYPE, EXTENDS(Expression), PUBLIC :: StructureConstructorExpr
    !> The results of parsing the structure constructor.
    TYPE(StructureConstructor) :: sc
  CONTAINS
    PROCEDURE :: Write => structure_Write
    PROCEDURE :: GetLocation => structure_GetLocation
    PROCEDURE :: Query => structure_Query
    PROCEDURE :: Modify => structure_Modify
  END TYPE StructureConstructorExpr
  
  !> An expression that is a function reference.
  TYPE, EXTENDS(Expression), PUBLIC :: FunctionRefExpr
    !> The /procedure-designator/.
    !!
    !! We can't pick the /binding-name/ from the /data-ref/, or the 
    !! /procedure-component-name/ from the /scalar-variable/.
    !!
    !! Always allocated, may be zero length.
    TYPE(PartRef), ALLOCATABLE :: parts(:)
    !> The procedure name for the function reference.
    TYPE(Identifier) :: name
    !> The /actual-arg-spec-list/ for the function reference.
    TYPE(ActualArgSpec), ALLOCATABLE :: args(:)
  CONTAINS
    PROCEDURE :: Write => function_Write
    PROCEDURE :: GetLocation => function_GetLocation
    PROCEDURE :: Query => function_Query
    PROCEDURE :: Modify => function_Modify
    PROCEDURE :: Advertise => function_Advertise
  END TYPE FunctionRefExpr
  
  !> An expression that is simply another expression inside parentheses.
  !!
  !! This may seem redundant, but it is significant in terms of argument 
  !! association.
  TYPE, EXTENDS(Expression), PUBLIC :: ParenthesisedExpr
    !> The expression that is inside the parentheses.
    CLASS(Expression), ALLOCATABLE :: ex
  CONTAINS
    PROCEDURE :: Write => parens_Write
    PROCEDURE :: GetLocation => parens_GetLocation
    PROCEDURE :: Query => parens_Query
    PROCEDURE :: Modify => parens_Modify
    PROCEDURE :: Advertise => parens_Advertise
  END TYPE ParenthesisedExpr
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Implementation of ConstantExpr%Write - write a debugging representation 
  !! of the expression to a logical unit.
  !!
  !1 @param[in]     pnode             The expression to write.
  !!
  !! @param[in]     unit              Unit connected for formatted output.
  
  SUBROUTINE constant_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ConstantExpr), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    WRITE (unit, "('cnstnt: ')", ADVANCE='NO')
    
    SELECT CASE (QueryType(pnode%tok))
    CASE (ittIntegerLiteral, ittRealLiteral, ittLogicalLiteral)
      IF (HasKind(pnode%tok)) THEN
        WRITE (unit, "(A,'_',A)", ADVANCE='NO')  &
            QueryValue(pnode%tok), QueryKind(pnode%tok)
      ELSE
        WRITE (unit, "(A)", ADVANCE='NO') QueryValue(pnode%tok)
      END IF
    CASE (ittCharacterLiteral)
      IF (HasKind(pnode%tok)) THEN
        WRITE (unit, "(A,'_''',A,'''')", ADVANCE='NO')  &
            QueryKind(pnode%tok), QueryValue(pnode%tok)
      ELSE
        WRITE (unit, "('''',A,'''')", ADVANCE='NO') QueryValue(pnode%tok)
      END IF
    CASE DEFAULT
      ! Not expectecting this.
      WRITE (unit, "(A)", ADVANCE='NO') QueryRaw(pnode%tok)
    END SELECT
    
  END SUBROUTINE constant_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of ConstantExpr%GetLocation - get the location in 
  !! source of the start of the expression.
  !!
  !! @param[in]     expr              The expression.
  !!
  !! @returns The source location for the start of the expression.
  !!
  !! We just forward the call to stored token.
  
  FUNCTION constant_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Charateristics
    
    CLASS(ConstantExpr), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    loc = QueryLocation(pnode%tok)
    
  END FUNCTION constant_GetLocation
  
  
  !*****************************************************************************
  !!
  !> Implementation of ConstantExpr%Query - traverse the parse node 
  !! with a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE constant_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ConstantExpr), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE constant_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of ConstantExpr%Modify - traverse the parse node 
  !! with a modify operation.
  !!
  !! @param[inout] pnode             The parse node element to modify.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE constant_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ConstantExpr), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE constant_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of DesignatorExpr%Advertise - advertise 
  !! the identifiers referenced or defined by the expression.
  !!
  !! @param[in]     pnode             The expression to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
   SUBROUTINE constant_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ConstantExpr), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(0))
    
    CALL advertise_kind(ident_list, pnode%tok)
    
  END SUBROUTINE constant_Advertise
  
  
  !*****************************************************************************
  !!
  !> Advertise a kind specification in a token that some sort of literal 
  !! constant.
  !!
  !! @param[in,out] ident_list        The list of identifier natures.
  !!
  !! @param[in]     tok               The token to process.
  
  SUBROUTINE advertise_kind(ident_list, tok)
    
    USE CharacterTypes
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(IdentifierNature), INTENT(INOUT), ALLOCATABLE :: ident_list(:)
    TYPE(Token), INTENT(IN) :: tok
    
    !---------------------------------------------------------------------------
    ! Locals.
    
    ! Kind specifier on a value
    CHARACTER(:), ALLOCATABLE :: kind_text
    
    !***************************************************************************
    
    ! Perhaps the literal has a kind specifier that is a named constant.  
    IF (HasKind(tok)) THEN
      kind_text = QueryKind(tok)
      ! Such a constant will start with a letter.
      IF (IsLetterChar(kind_text(1:1))) THEN
        ! Kind specification would appear to be a named constant.  The 
        ! location will point to the start of the token, rather than the 
        ! kind text proper, but so be it.
        CALL MergeNatures(  &
            ident_list,  &
            IdentifierNature(  &
              NAME=kind_text,  &
              NATURES=[NatureRefDataObject],  &
              LOCATIONS=[QueryLocation(tok)] ) )
      END IF
    END IF
    
  END SUBROUTINE advertise_kind
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for bindings in ComplexLiteralExpr.
  
  
  !*****************************************************************************
  !!
  !> Implementation of ComplexLiteralExpr%Write - write a debugging 
  !! representation of the expression to a logical unit.
  !!
  !1 @param[in]     pnode             The expression to write.
  !!
  !! @param[in]     unit              Unit connected for formatted output.
  
  SUBROUTINE complex_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ComplexLiteralExpr), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    CALL pnode%constant%Write(unit)
    
  END SUBROUTINE complex_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of ComplexLiteralExpr%GetLocation - get the location in 
  !! source of the start of the expression.
  !!
  !! @param[in]     pnode             The expression.
  !!
  !! @returns The source location for the start of the expression.
  !!
  !! We just forward the call to the first stored token.  This skips the 
  !! opening parenthesis, but I don't care, do you?
  !!
  !! @todo Fix.
  
  FUNCTION complex_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !---------------------------------------------------------------------------
    ! Charateristics
    
    CLASS(ComplexLiteralExpr), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    loc = pnode%paren_location
    
  END FUNCTION complex_GetLocation
  
  
  !*****************************************************************************
  !!
  !> Implementation of ComplexLiteralExpr%Query - traverse the parse node 
  !! with a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE complex_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ComplexLiteralExpr), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%constant%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE complex_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of ComplexLiteralExpr%Modify - traverse the parse node 
  !! with a modify operation.
  !!
  !! @param[inout] pnode             The parse node element to modify.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE complex_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ComplexLiteralExpr), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%constant%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE complex_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of DesignatorExpr%Advertise - advertise 
  !! the identifiers referenced or defined by the expression.
  !!
  !! @param[in]     pnode             The expression to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE complex_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ComplexLiteralExpr), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(0))
    
  END SUBROUTINE complex_Advertise
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for bindings in DesignatorExpr.
  
  
  !*****************************************************************************
  !!
  !> Implementation of DesignatorExpr%Write - write a debugging 
  !! representation of the expression to a logical unit.
  !!
  !1 @param[in]     pnode             The expression to write.
  !!
  !! @param[in]     unit              Unit connected for formatted output.
  
  SUBROUTINE desig_Write(pnode, unit)
    
    USE Designators
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DesignatorExpr), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    WRITE (unit, "('dsg: ')", ADVANCE='NO')
    CALL WriteDesignator(unit, pnode%desig)
    
  END SUBROUTINE desig_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of DesignatorExpr%GetLocation - get the location in 
  !! source of the start of the expression.
  !!
  !! @param[in]     pnode             The expression.
  !!
  !! @returns The source location for the start of the expression.
  !!
  !! We're not sure what to do here yet - chances are that desig will get a 
  !! GetLocation interface too.
  !!
  !! @todo Fix.
  
  FUNCTION desig_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !---------------------------------------------------------------------------
    ! Charateristics
    
    CLASS(DesignatorExpr), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    loc = .LocationOf. pnode%desig
    
  END FUNCTION desig_GetLocation
  
  
  !*****************************************************************************
  !!
  !> Implementation of DesignatorExpr%Query - traverse the parse node 
  !! with a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE desig_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DesignatorExpr), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%desig%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE desig_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of DesignatorExpr%Modify - traverse the parse node 
  !! with a modify operation.
  !!
  !! @param[inout] pnode             The parse node element to modify.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE desig_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DesignatorExpr), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%desig%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE desig_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of DesignatorExpr%Advertise - advertise 
  !! the identifiers referenced or defined by the expression.
  !!
  !! @param[in]     pnode             The expression to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE desig_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DesignatorExpr), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(0))
    
  END SUBROUTINE desig_Advertise
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for bindings in ArrayConstructorExpr.
  
  
  !*****************************************************************************
  !!
  !> Implementation of ArrayConstructorExpr%Write - write a debugging 
  !! representation of the expression to a logical unit.
  !!
  !1 @param[in]     pnode             The expression to write.
  !!
  !! @param[in]     unit              Unit connected for formatted output.
  
  SUBROUTINE array_Write(pnode, unit)
    
    USE ArrayConstructors
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ArrayConstructorExpr), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    CALL WriteArrayConstructor(unit, pnode%type_spec, pnode%values)
    
  END SUBROUTINE array_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of DesignatorExpr%GetLocation - get the location in 
  !! source of the start of the expression.
  !!
  !! @param[in]     pnode             The expression.
  !!
  !! @returns The source location for the start of the expression.
  !!
  !! We just forward the call to the first stored token.  This skips the 
  !! opening bracket, as per complex literal expressions.
  !!
  !! @todo Fix.
  
  FUNCTION array_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !---------------------------------------------------------------------------
    ! Charateristics
    
    CLASS(ArrayConstructorExpr), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%type_spec)) THEN
      loc = SourceLocationIfortBug() !expr%type_spec%GetLocation()
    ELSE
      ! @todo Fix.
      loc = SourceLocationIfortBug()
    END IF
    
  END FUNCTION array_GetLocation
  
  
  !*****************************************************************************
  !!
  !> Implementation of ArrayConstructorExpr%Query - traverse the parse node 
  !! with a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  RECURSIVE SUBROUTINE array_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ArrayConstructorExpr), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Array element index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    ! The leading type spec, if relevant.
    IF (ALLOCATED(pnode%type_spec)) THEN
      CALL pnode%type_spec%Query(visitor)
    END IF
    
    ! Array values.
    DO i = 1, SIZE(pnode%values)
      CALL pnode%values(i)%Query(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE array_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of ArrayConstructorExpr%Modify - traverse the parse node 
  !! with a modify operation.
  !!
  !! @param[in]     pnode             The parse node element to modif.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  RECURSIVE SUBROUTINE array_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ArrayConstructorExpr), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Array element index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    ! The leading type spec, if relevant.
    IF (ALLOCATED(pnode%type_spec)) THEN
      CALL pnode%type_spec%Modify(visitor)
    END IF
    
    ! Array values.
    DO i = 1, SIZE(pnode%values)
      CALL pnode%values(i)%Modify(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE array_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of StructureConstructorExpr%Advertise - advertise 
  !! the identifiers referenced or defined by the expression.
  !!
  !! @param[in]     pnode             The expression to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  ! Recursive because an array constructor may contain another array 
  ! constructor.
  RECURSIVE SUBROUTINE array_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ArrayConstructorExpr), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(0))
    
  END SUBROUTINE array_Advertise
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for bindings in StructureConstructorExpr
  
  
  !*****************************************************************************
  !!
  !> Implementation of StructureConstructorExpr%Write - write a debugging 
  !! representation of the expression to a logical unit.
  !!
  !1 @param[in]     pnode             The expression to write.
  !!
  !! @param[in]     unit              Unit connected for formatted output.
  
  SUBROUTINE structure_Write(pnode, unit)
    
    USE StructureConstructors
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(StructureConstructorExpr), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    CALL pnode%sc%Write(unit)
    
  END SUBROUTINE structure_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of StructureConstructorExpr%GetLocation - get the location in 
  !! source of the start of the expression.
  !!
  !! @param[in]     pnode             The expression.
  !!
  !! @returns The source location for the start of the expression.
  !!
  !! We forward the call to the type spec.
  !!
  !! @todo Fix.
  
  FUNCTION structure_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !---------------------------------------------------------------------------
    ! Charateristics
    
    CLASS(StructureConstructorExpr), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    loc = .LocationOf. pnode%sc
    
  END FUNCTION structure_GetLocation
  
  
  !*****************************************************************************
  !!
  !> Implementation of StructureConstructorExpr%Query - traverse the parse node 
  !! with a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE structure_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(StructureConstructorExpr), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%sc%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE structure_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of StructureConstructorExpr%Modify - traverse the parse 
  !! node with a modify operation.
  !!
  !! @param[inout] pnode             The parse node element to modify.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE structure_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(StructureConstructorExpr), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%sc%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE structure_Modify
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for bindings in FunctionRefExpr.
  
  
  !*****************************************************************************
  !!
  !> Implementation of FunctionRefExpr%Write - write a debugging 
  !! representation of the expression to a logical unit.
  !!
  !1 @param[in]     pnode             The expression to write.
  !!
  !! @param[in]     unit              Unit connected for formatted output.
  
  SUBROUTINE function_Write(pnode, unit)
    
    USE CallStmts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(FunctionRefExpr), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    CALL WriteProcedureReference(unit, pnode%parts, pnode%name, pnode%args)
    
  END SUBROUTINE function_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of FunctionRefExpr%GetLocation - get the location in 
  !! source of the start of the expression.
  !!
  !! @param[in]     pnode             The expression.
  !!
  !! @returns The source location for the start of the expression.
  !!
  !! We're not sure what to do here yet - chances are that PartRef will get a 
  !! GetLocation interface too.
  
  FUNCTION function_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !---------------------------------------------------------------------------
    ! Charateristics
    
    CLASS(FunctionRefExpr), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    IF (SIZE(pnode%parts) == 0) THEN
      loc = .LocationOf. pnode%name
    ELSE
      loc = .LocationOf. pnode%parts(1)
    END IF
    
  END FUNCTION function_GetLocation
  
  
  !*****************************************************************************
  !!
  !> Implementation of FunctionRefExpr%Query - traverse the parse node 
  !! with a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE function_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(FunctionRefExpr), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Argument index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%parts)
      CALL pnode%parts(i)%Query(visitor)
    END DO
    
    ! Arguments.
    DO i = 1, SIZE(pnode%args)
      IF (ALLOCATED(pnode%args(i)%expr)) THEN
        CALL pnode%args(i)%expr%Query(visitor)
      END IF
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE function_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of FunctionRefExpr%Modify - traverse the parse node 
  !! with a modify operation.
  !!
  !! @param[inout] pnode             The parse node element to modify.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE function_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(FunctionRefExpr), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Argument index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%parts)
      CALL pnode%parts(i)%Modify(visitor)
    END DO
    
    ! Arguments.
    DO i = 1, SIZE(pnode%args)
      IF (ALLOCATED(pnode%args(i)%expr)) THEN
        CALL pnode%args(i)%expr%Modify(visitor)
      END IF
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE function_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of FunctionRefExpr%Advertise - advertise the identifiers 
  !! referenced or defined by the expression.
  !!
  !! @param[in]     expr              The expression to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  ! Recursive because a function reference may contain another function 
  ! reference.
  RECURSIVE SUBROUTINE function_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(FunctionRefExpr), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ! A multi-part reference means that the base identifier must be for 
    ! a data object.
    IF (SIZE(pnode%parts) > 0) THEN
      ALLOCATE(ident_list(1))
      ident_list(1) = IdentifierNature(  &
          NAME=.NameOf. pnode%parts(1)%part_name,  &
          NATURES=[NatureRefDataObject],  &
          LOCATIONS=[.LocationOf. pnode%parts(1)%part_name] )
    ELSE
      ! The base object could be a type (this could be a structure constructor 
      ! expression), a generic interface or a specific function.  These are 
      ! bucketed together under the category "function-like".
      ALLOCATE(ident_list(1))
      ident_list(1) = IdentifierNature(  &
          NAME=.NameOf. pnode%name,  &
          NATURES=[NatureRefFunctionLike],  &
          LOCATIONS=[.LocationOf. pnode%name] )
    END IF
    
  END SUBROUTINE function_Advertise
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for bindings in ParenthesisedExpr.
  
  
  !*****************************************************************************
  !!
  !> Implementation of ParenthesisedExpr%Write - write a debugging 
  !! representation of the expression to a logical unit.
  !!
  !1 @param[in]     pnode             The expression to write.
  !!
  !! @param[in]     unit              Unit connected for formatted output.
  
  SUBROUTINE parens_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ParenthesisedExpr), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    WRITE (unit, "('(')", ADVANCE='NO')
    CALL pnode%ex%Write(unit)
    WRITE (unit, "(')')", ADVANCE='NO')
    
  END SUBROUTINE parens_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of ParenthesisedExpr%GetLocation - get the location in 
  !! source of the start of the expression.
  !!
  !! @param[in]     expr              The expression.
  !!
  !! @returns The source location for the start of the expression.
  !!
  !! The call is forwarded to the contained expression.  This means that the 
  !! start location skips the opening parenthesis.
  !!
  !! @todo Fix.
  
  ! Recursive because you may have parenthesised expressions inside of other 
  ! parenthesised expressions.
  RECURSIVE FUNCTION parens_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !---------------------------------------------------------------------------
    ! Charateristics
    
    CLASS(ParenthesisedExpr), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    loc = pnode%ex%GetLocation()
    
  END FUNCTION parens_GetLocation
  
  
  !*****************************************************************************
  !!
  !> Implementation of ParenthesisedExpr%Query - traverse the parse node 
  !! with a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE parens_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ParenthesisedExpr), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    ! Child expression.
    CALL pnode%ex%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE parens_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of ParenthesisedExpr%Modify - traverse the parse node 
  !! with a modify operation.
  !!
  !! @param[inout] pnode             The parse node element to modify.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE parens_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ParenthesisedExpr), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    ! Child expression.
    CALL pnode%ex%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE parens_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of ParenthesisedExpr%Advertise - advertise the identifiers 
  !! referenced or defined by the expression.
  !!
  !! @param[in]     expr              The expression to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  ! Recursive because you may have parenthesised expressions inside of other 
  ! parenthesised expressions.
  RECURSIVE SUBROUTINE parens_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ParenthesisedExpr), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(0))
    
  END SUBROUTINE parens_Advertise
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures to transform from one form of expression to another.
  
  
  !*****************************************************************************
  !!
  !> Create a function reference expression from a designator.
  !!
  !! @param[in]     desig             The results of parsing the source as 
  !! a designator.
  !!
  !! @param[out]    expr              The results of reparsing the designator 
  !! as a function reference.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseFunctionRefExprFromDesignator(desig, expr, err_list)
    
    USE Errors
    USE ErrorCodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Designator), INTENT(IN) :: desig
    TYPE(FunctionRefExpr), INTENT(OUT) :: expr
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: ip             ! Part index.
    INTEGER :: is             ! Subscript index.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    IF (ALLOCATED(desig%literal)) THEN
      CALL Add( err_list, CODE=errSyntax, COMPONENT='R1219',  &
          LOCATION=desig%literal_location,  &
          MSG='A /char-literal-constant/ is not permitted at this location &
            &in a /function-reference/.' )
      RETURN
    END IF
    
    ! A designator always has one part, so this may have zero or more.
    ALLOCATE(expr%parts(SIZE(desig%part_refs)-1))
    !===========================================================================
    ! ifort bug - asssignment of dt with polymorphic components.
    ! expr%parts = desig%part_refs(:SIZE(desig%part_refs) - 1)
    DO ip = 1, SIZE(desig%part_refs) - 1
      expr%parts(ip) = desig%part_refs(ip)
!      ALLOCATE(expr%parts(ip), SOURCE=desig%part_refs(ip))
    END DO
    !===========================================================================
    
    expr%name = desig%part_refs(ip)%part_name
    
    IF (.NOT. ALLOCATED(desig%part_refs(ip)%subscripts)) THEN
      CALL Add( err_list, CODE=errSyntax, COMPONENT='R1219',  &
!          LOCATION=????
          MSG='Expected "(".' )
    END IF
    
    ! Map the subscript expressions across to functions arguments.
    ALLOCATE(expr%args(SIZE(desig%part_refs(ip)%subscripts)))
    DO is = 1, SIZE(desig%part_refs(ip)%subscripts)
      ASSOCIATE(sub => desig%part_refs(ip)%subscripts(is))
        IF (sub%colon) THEN
          CALL Add( err_list, CODE=errSyntax, COMPONENT='R1222',  &
!              LOCATION=????
              MSG='":" is not permitted here in an /actual-arg-spec/.' )
        ELSE IF (sub%star) THEN
          CALL Add( err_list, CODE=errSyntax, COMPONENT='R1222',  &
!              LOCATION=????
              MSG='"*" is not permitted here in an /actual-arg-spec/.' )
        ELSE
          ALLOCATE(expr%args(is)%expr, SOURCE=sub%finish)
        END IF
      END ASSOCIATE
    END DO
    
    IF (ALLOCATED(desig%part_refs(ip)%image_selector)) THEN
      CALL Add( err_list, CODE=errSyntax, COMPONENT='R1219',  &
!          LOCATION=????
          MSG='An image selector is not permitted at this point in a &
            &/function-reference/.' )
    END IF
    
    IF (ALLOCATED(desig%substring)) THEN
      CALL Add( err_list, CODE=errSyntax, COMPONENT='R1219',  &
!          LOCATION=????
          MSG='A substring reference is not permitted at this point in a &
            &/function-reference/.' )
    END IF
    
  END SUBROUTINE ParseFunctionRefExprFromDesignator
  
  
  !*****************************************************************************
  !!
  !> Create a structure constructor reference expression from a designator.
  !!
  !! @param[in]     desig             The results of parsing the source as 
  !! a designator.
  !!
  !! @param[out]    sc_expr           The results of reparsing the designator 
  !! as a structure constructor.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseStructureConstructorFromDesignator(desig, sc_expr, err_list)
    
    USE Designators
    USE Errors
    USE ErrorCodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Designator), INTENT(IN) :: desig
    TYPE(StructureConstructorExpr), INTENT(OUT) :: sc_expr
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Syntax rule number for a /structure-constructor/.
    CHARACTER(*), PARAMETER :: comp = 'R455'
    
    INTEGER :: is             ! Subscript index.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    IF (ALLOCATED(desig%literal)) THEN
      CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
          LOCATION=desig%literal_location,  &
          MSG='A /char-literal-constant/ is not permitted at this location &
            &in a /structure-constructor/.' )
      RETURN
    END IF
    
    ! A structure constructor does not have parts.
    IF (SIZE(desig%part_refs) > 1) THEN
      CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
!          LOCATION=???,  &
          MSG='"%" is not permitted at this location in a &
            &/structure-constructor/.' )
      RETURN
    END IF
    
    IF (.NOT. ALLOCATED(desig%part_refs(1)%subscripts)) THEN
      CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
!          LOCATION=???,  &
          MSG='Expected "(".' )
      RETURN
    END IF
    
    sc_expr%sc%type_spec%type_name = desig%part_refs(1)%part_name
    
    ! There are no type parameters.
    ALLOCATE(sc_expr%sc%type_spec%params(0))
    
    ! Map the subscript expressions across to component specifiers.
    ALLOCATE(sc_expr%sc%components(SIZE(desig%part_refs(1)%subscripts)))
    DO is = 1, SIZE(desig%part_refs(1)%subscripts)
      ASSOCIATE(sub => desig%part_refs(1)%subscripts(is))
        IF (sub%colon) THEN
          CALL Add( err_list, CODE=errSyntax, COMPONENT='R456',  &
!              LOCATION=????
              MSG='":" is not permitted here in an /component-spec/.' )
        ELSE IF (sub%star) THEN
          CALL Add( err_list, CODE=errSyntax, COMPONENT='R456',  &
!              LOCATION=????
              MSG='"*" is not permitted here in an /component-spec/.' )
        ELSE
          ALLOCATE(sc_expr%sc%components(is)%data_source, SOURCE=sub%finish)
        END IF
      END ASSOCIATE
    END DO
    
    
    IF (ALLOCATED(desig%part_refs(1)%image_selector)) THEN
      CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
!          LOCATION=????
          MSG='An image selector is not permitted at this point in a &
            &/structure-constructor/.' )
    END IF
    
    IF (ALLOCATED(desig%substring)) THEN
      CALL Add( err_list, CODE=errSyntax, COMPONENT='R1219',  &
!          LOCATION=????
          MSG='A substring reference is not permitted at this point in a &
            &/structure-constructor/.' )
    END IF
    
  END SUBROUTINE ParseStructureConstructorFromDesignator
  
  
  !*****************************************************************************
  !!
  !> Parse a structure constructor from something previously parsed as 
  !! a function reference.
  !!
  !! @param[in]     fn_expr           The results of parsing the source as a 
  !! function reference.
  !!
  !! @param[out]    sc_expr           The results of reparsing the function 
  !! reference as a structure constructor.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseStructureConstructorFromFunctionRef(  &
      fn_expr, sc_expr, err_list )
    
    USE Errors
    USE ErrorCodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(FunctionRefExpr), INTENT(IN) :: fn_expr
    TYPE(StructureConstructorExpr), INTENT(OUT) :: sc_expr
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: ia             ! Argument/component index.
    
    ! Syntax rule for a /structure-constructor/.
    CHARACTER(*), PARAMETER :: comp = 'R455'
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    ! A structure constructor does not have parts.
    IF (SIZE(fn_expr%parts) > 0) THEN
      CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
!          LOCATION=???,  &
          MSG='"%" is not permitted at this location in a &
            &/structure-constructor/.' )
      RETURN
    END IF
    
    ! The type name is what we though was previously the function name.
    sc_expr%sc%type_spec%type_name = fn_expr%name
    
    ! No type parameters were specified.
    ALLOCATE(sc_expr%sc%type_spec%params(0))
    
    ! The actual arguments are actually component specifiers.
    ALLOCATE(sc_expr%sc%components(SIZE(fn_expr%args)))
    DO ia = 1, SIZE(fn_expr%args)
      
    END DO
    
  END SUBROUTINE ParseStructureConstructorFromFunctionRef
  
END MODULE PrimaryExpressions
