! $Id: ProcedureDeclarationStmts.f90 2901 2019-04-11 20:45:38Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the ProcedureDeclarationStmts module.


!*******************************************************************************
!!
!> Types and procedures for parsing /procedure-declaration-stmt/'s and 
!! the related /proc-decl-list/ part.
!!
!! /proc-decl-list/ also appears in /proc-component-def-stmt/, hence the 
!! ComponentDefStatements module depends on this module.
!!
!! This module (simply because it was written first) also contains procedures 
!! for matching and parsing /null-init/'s.  These also appear in the syntax 
!! for /type-declaration-stmt/'s, so the TypeDeclarationStmts module also 
!! depends on this module.

MODULE ProcedureDeclarationStmts
  
  USE AccessStmts
  USE CompilerKinds
  USE DeclarationTypeSpecs
  USE Identifiers
  USE IntentSpecs
  USE LanguageBindingSpecs
  USE ParseNodes
  USE SourceLocations
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseNullInit
  PUBLIC :: MatchNullInit
  PUBLIC :: ParseProcInterface
  PUBLIC :: ParseProcDeclList
  PUBLIC :: ParseProcedureDeclarationStmt
  
  PUBLIC :: WriteProcDecl
  
  !-----------------------------------------------------------------------------
  
  !> Parent type for attribute specifications.
  TYPE, EXTENDS(ParseNode), ABSTRACT, PUBLIC :: AttrBase
    !> The starting location of the list of attributes.
    !!
    !! Currently we only have a single location for all attributes.  We 
    !! probably need to revise this.
    TYPE(SourceLocation) :: location
  CONTAINS
    PROCEDURE(ab_Dump), DEFERRED :: Dump
    PROCEDURE(ab_GetNatures), DEFERRED :: GetNatures
    PROCEDURE :: GetLocation => ab_GetLocation
  END TYPE AttrBase
  
  ! Interfaces for deferred bindings in AttrBase.
  ABSTRACT INTERFACE
    SUBROUTINE ab_Dump(pnode, unit, pfx)
      IMPORT :: AttrBase
      IMPLICIT NONE
      !-------------------------------------------------------------------------
      CLASS(AttrBase), INTENT(IN) :: pnode
      INTEGER, INTENT(IN) :: unit
      CHARACTER(*), INTENT(IN) :: pfx
    END SUBROUTINE ab_Dump
    
    SUBROUTINE ab_GetNatures(attrs, natures)
      IMPORT :: AttrBase
      IMPLICIT NONE
      !-------------------------------------------------------------------------
      CLASS(AttrBase), INTENT(IN) :: attrs
      INTEGER, INTENT(OUT), ALLOCATABLE :: natures(:)
    END SUBROUTINE ab_GetNatures
  END INTERFACE
  
  !-----------------------------------------------------------------------------
  
  !> Stores the result of parsing a /proc-interface/.
  !!
  !! We could implement this as an abstract parent class with a specialisation
  !! for an interface that is a declaration and a specialisation for an
  !! interface name.
  TYPE, EXTENDS(TermParseNode), PUBLIC :: ProcInterface
    !> The /proc-interface/, if it is a /declaration-type-spec/.
    !!
    !! Not allocated if name is allocated, otherwise always allocated.
    TYPE(DeclTypeSpec), ALLOCATABLE :: decl
    !> The /proc-interface/, if is is an /interface-name/.
    !!
    !! Not allocated if decl is allocated, otherwise always allocated. 
    TYPE(Identifier), ALLOCATABLE :: name
  CONTAINS
    PROCEDURE :: Write => pi_Write
    PROCEDURE :: Query => pi_Query
    PROCEDURE :: Modify => pi_Modify
    PROCEDURE, NOPASS :: What => pi_What
    PROCEDURE :: Advertise => pi_Advertise
    PROCEDURE :: GetLocation => pi_GetLocation
  END TYPE ProcInterface
  
  !> Attributes in a /proc-component-def-stmt/.
  TYPE, EXTENDS(AttrBase), PUBLIC :: ProcAttrSpec
    !> Procedure interface.
    !!
    !! Not allocated if no interface was specified.
    TYPE(ProcInterface), ALLOCATABLE :: proc_intf
    !> Access specification.
    INTEGER :: access = iaccNone
    !> Language binding spec, if present, not allocated if not.
    TYPE(LangSpec), ALLOCATABLE :: ls
    !> Intent specification.
    INTEGER :: intent = iintNone
    !> Optional specification.
    LOGICAL :: optional = .FALSE.
    !> Procedure has the pointer attribute.
    LOGICAL :: pointer = .FALSE.
    !> Save specification.
    LOGICAL :: save = .FALSE.
    !> Procedure has the protected attribute.
    LOGICAL :: protected = .FALSE.
  CONTAINS
    PROCEDURE :: Dump => pas_Dump
    PROCEDURE :: Query => pas_Query
    PROCEDURE :: Modify => pas_Modify
    PROCEDURE, NOPASS :: What => pas_What
    PROCEDURE :: GetNatures => pas_GetNatures
  END TYPE ProcAttrSpec
  
  !> Stores the results of parsing a /null-init/.
  !!
  !! A /null-init/ must be an reference to the intrinsic function NULL with 
  !! no arguments.  A complication is that the identifier for the intrinsic 
  !! function NULL may not actually be NULL, hence we have the ability here 
  !! to store a reference to any function reference with no arguments.
  !!
  !! Whether the identifier used to designate the NULL intrinsic is actually 
  !! the NULL intrinsic is not known until the semantic analysis phase.
  TYPE, EXTENDS(TermParseNode), PUBLIC :: NullInit
    !> The identifier used to reference the intrinsic.
    TYPE(Identifier) :: name
  CONTAINS
    PROCEDURE :: Write => null_Write
    PROCEDURE :: Query => null_Query
    PROCEDURE :: Modify => null_Modify
    PROCEDURE :: Advertise => null_Advertise
    PROCEDURE :: GetLocation => null_GetLocation
    PROCEDURE, NOPASS :: What => null_What
  END TYPE NullInit
  
  !> Stores the result of parsing a /proc-decl/.
  !!
  !! @todo Identifiers...
  TYPE, EXTENDS(TermParseNode), PUBLIC :: ProcDecl
    !> The name of the procedure.
    TYPE(Identifier) :: name
    !> The results of parsing a /null-init/.
    !!
    !! Not allocated if there was no /null-init/.  Never allocated 
    !! if @a init_target is allocated.
    TYPE(NullInit), ALLOCATABLE :: null_init
    !> Name of the initialisation target if it is not a /null-init/.
    !!
    !! Not allocated if there was no /initial-proc-target/, never allocated 
    !! if @a null_init is allocated.
    TYPE(Identifier), ALLOCATABLE :: init_target
    !> The attributes specified by the hosting statement.
    CLASS(AttrBase), POINTER :: attrs => NULL()
  CONTAINS
    PROCEDURE :: Write => pd_Write
    PROCEDURE :: Query => pd_Query
    PROCEDURE :: Modify => pd_Modify
    PROCEDURE :: GetLocation => pd_GetLocation
    PROCEDURE, NOPASS :: What => pd_What
    PROCEDURE :: Advertise => pd_Advertise
  END TYPE ProcDecl  
  
  !> Stores the results of parsing a /procedure-declaration-stmt/.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: ProcedureDeclarationStmt
    !> The attributes (possibly including an interface or type 
    !! specification) of the procedure.
    !!
    !! This is managed through a pointer so that we can reliably share it 
    !! with the ProcDecl objects.  The attrs object is owned by the 
    !! ProcedureDeclarationStmt object.
    !!
    !! Always associated after construction.
    TYPE(ProcAttrSpec), POINTER :: attrs => NULL()
    !> The procedure declarations.
    TYPE(ProcDecl), ALLOCATABLE :: decls(:)
  CONTAINS
    PROCEDURE :: Dump => procdecl_Dump
    PROCEDURE :: Query => procdecl_Query
    PROCEDURE :: Modify => procdecl_Modify
    PROCEDURE, NOPASS :: What => procdecl_What
    
    FINAL :: procdecl_Final_scalar
  END TYPE ProcedureDeclarationStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /procedure-declaration-stmt/ (R1211).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    procdecl_stmt     The results of parsing the statement.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseProcedureDeclarationStmt(tlist, procdecl_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(ProcedureDeclarationStmt), INTENT(OUT) :: procdecl_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule for a /procedure-declaration-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R1211'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: it             ! Current token.
    TYPE(Error) :: err        ! Error object for Expect* calls.
    INTEGER :: closing        ! Index of closing parenthesis.
    INTEGER :: declare        ! Index of the ::.
    INTEGER :: attr_finish    ! Last attribute token (prior to ::).
    INTEGER :: item_count     ! Number of items in the attr list.
    INTEGER :: item_start     ! Start index in tlist for a particular item.
    INTEGER :: item_finish    ! Finish index in tlist for a particular item.
    INTEGER :: item           ! Item index
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /procedure-declaration-stmt/.' )
      RETURN
    END IF
    
    ALLOCATE(procdecl_stmt%attrs)
    
    !---------------------------------------------------------------------------
    ! PROCEDURE([xxx]) ...
    
    CALL ExpectKeyword(tlist, 1, scck_'PROCEDURE', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL Expect(tlist, 2, ittDelimiter, scck_'(', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL ExpectClosing(tlist, 2, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! The interface is optional.
    IF (closing /= 3) THEN
      ALLOCATE(procdecl_stmt%attrs%proc_intf)
      CALL ParseProcInterface(  &
          tlist(3:closing-1),  &
          procdecl_stmt%attrs%proc_intf,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
    END IF
    
    it = closing + 1          ! Index after the parenthesis.
    
    !---------------------------------------------------------------------------
    ! Optional attributes.  If they are present, then there will be a comma 
    ! after PROCEDURE(xx), and a later :: must be there.
    
    IF (Match(tlist, it, ittComma, scck_',')) THEN
      declare = FindBare(tlist, ittDeclare, scck_'::', it)
      IF (declare == 0) THEN
        ! There's a syntax error somewhere - processing the remainder of the 
        ! statement should flush it out.
        attr_finish = SIZE(tlist)
      ELSE
        attr_finish = declare - 1
      END IF
      
      item_count = ListItemCount(tlist(it+1:attr_finish))
      item_finish = 0
      DO item = 1, item_count
        CALL GetNextListItem(  &
            tlist,  &
            it + 1, attr_finish,  &
            item_start,  &
            item_finish )
        CALL parse_proc_attr_spec(  &
            tlist,  &
            item_start, item_finish,  &
            procdecl_stmt%attrs,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
        IF (Failure(sub_err_list)) RETURN
      END DO
      
      IF (declare == 0) THEN
        CALL Add( err_list,  &
            CODE=errSyntax,  &
            COMPONENT=comp,  &
            LOCATION=QueryLocation(tlist(attr_finish)),  &
            MSG='Expected "::".' )
        RETURN
      END IF
      it = declare + 1
    ELSE
      ! The double colon is optional.
      IF (Match(tlist, it, ittDeclare, scck_'::')) it = it + 1
    END IF
    
    ! it here now the index of the start of the proc-decl-list.
    
    !---------------------------------------------------------------------------
    ! The /proc-decl-list/.
    
    CALL ParseProcDeclList(  &
        tlist(it:),  &
        procdecl_stmt%attrs,  &
        procdecl_stmt%decls,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
    ! Check (partially) C1221 - if BIND(C,NAME='xxx') appears then there 
    ! can only be one /proc-decl/ and no POINTER.
    IF (ALLOCATED(procdecl_stmt%attrs%ls)) THEN
      IF (ALLOCATED(procdecl_stmt%attrs%ls%name)) THEN
        IF (SIZE(procdecl_stmt%decls) > 1) THEN
          ! We'll pin the error on the first declaration in the list.
          CALL Add( err_list, CODE=errSyntax, COMPONENT='C1221 on ' // comp,  &
              LOCATION=QueryLocation(tlist(declare+1)),  &
              MSG='If the BIND attribute appears with a NAME specifier, then &
                &there may only be one /proc-decl/ in the list.' )
        END IF
        IF (procdecl_stmt%attrs%pointer) THEN
          ! We'll pin the error on the first token.
          CALL Add( err_list, CODE=errSyntax, COMPONENT='C1221 on ' // comp,  &
              LOCATION=QueryLocation(tlist(1)),  &
              MSG='If the BIND attribute appears with a NAME specifier, then &
                &the POINTER attribute may not be specified.' )
        END IF
      END IF
    END IF
    
  END SUBROUTINE ParseProcedureDeclarationStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that make up a /proc-decl-list/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     attrs             The attributes from the parent 
  !! statement.
  !!
  !! @param[out]    decls             The declarators.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseProcDeclList(tlist, attrs, decls, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    CLASS(AttrBase), INTENT(IN), TARGET :: attrs
    TYPE(ProcDecl), INTENT(OUT), ALLOCATABLE :: decls(:)
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: list_count     ! Number of items in the /implicit-spec-list/.
    INTEGER :: item           ! List item index.
    INTEGER :: item_start     ! Start index in tlist of a particular item.
    INTEGER :: item_finish    ! Finish index in tlist of a particular item.
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------    
    
    list_count = ListItemCount(tlist)
    ALLOCATE(decls(list_count))
    item_finish = 0
    DO item = 1, list_count
      CALL GetNextListItem(tlist, 1, SIZE(tlist), item_start, item_finish)
      CALL parse_proc_decl(  &
          tlist,  &
          item_start, item_finish,  &
          attrs,  &
          decls(item),  &
          sub_err_list )
      CALL Add(err_list, sub_err_list)
      IF (Failure(sub_err_list)) RETURN
    END DO
    
  END SUBROUTINE ParseProcDeclList
  
  
  !*****************************************************************************
  !!
  !> Write a debugging representation of a ProcDecl object to a logical unit.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     decl              The object to dump.
  
  SUBROUTINE WriteProcDecl(unit, decl)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    INTEGER, INTENT(IN) :: unit
    TYPE(ProcDecl), INTENT(IN) :: decl
    
    !***************************************************************************
    
    WRITE (unit, "(A)", ADVANCE='NO') .NameOf. decl%name
    IF (ALLOCATED(decl%null_init))  &
        WRITE (unit, "(' => ',A,'()')", ADVANCE='NO')  &
          .NameOf. decl%null_init%name
    IF (ALLOCATED(decl%init_target))  &
        WRITE (unit, "(' => ',A)", ADVANCE='NO')  &
          .NameOf. decl%init_target
    
  END SUBROUTINE WriteProcDecl
  
  
  !*****************************************************************************
  !!
  !> Implementation of ProcedureDeclarationStmt%Dump - write a debugging 
  !! representation of the results of parsing the statement to a 
  !! logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE procdecl_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ProcedureDeclarationStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Procedure declarator index.
    
    !***************************************************************************
    
    CALL pnode%attrs%Dump(unit, pfx)
    
    WRITE (unit, "(A,'/proc-decl-list/')") pfx
    DO i = 1, SIZE(pnode%decls)
      WRITE (unit, "(A)", ADVANCE='NO') pfx // '  '
      CALL WriteProcDecl(unit, pnode%decls(i))
      WRITE (unit, "()")
    END DO
    
  END SUBROUTINE procdecl_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of ProcedureDeclarationStmt%Query - traverse the parse 
  !! node with a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE procdecl_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ProcedureDeclarationStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Proc decl index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%attrs%Query(visitor)
    
    DO i = 1, SIZE(pnode%decls)
      CALL pnode%decls(i)%Query(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE procdecl_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of ProcedureDeclarationStmt%Modify - traverse the parse 
  !! node with a modify operation.
  !!
  !! @param[in]     pnode             The parse node element to modify.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE procdecl_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ProcedureDeclarationStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Proc decl index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%attrs%Modify(visitor)
    
    DO i = 1, SIZE(pnode%decls)
      CALL pnode%decls(i)%Modify(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE procdecl_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of ProcedureDeclarationStmt%What - return a descriptive 
  !! string for the parse node.
  !!
  !! @returns 'procedure-declaration-stmt'.
  
  FUNCTION procdecl_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'procedure-declaration-stmt'
    
  END FUNCTION procdecl_What
  
  
  !*****************************************************************************
  !!
  !> Scalar finalizer for a ProcedureDeclarationStmt object.
  !!
  !! @param[in,out] procdecl_stmt     The object to be finalized.
  
  SUBROUTINE procdecl_Final_scalar(procdecl_stmt)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(ProcedureDeclarationStmt), INTENT(INOUT) :: procdecl_stmt
    
    !***************************************************************************
    
    IF (ASSOCIATED(procdecl_stmt%attrs)) DEALLOCATE(procdecl_stmt%attrs)
    
  END SUBROUTINE procdecl_Final_scalar
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for AttrBase.
  
  
  !*****************************************************************************
  !!
  !> Implementation of AttrBase%GetLocation - get the start of the 
  !! source that corresponds to the ParseNode.
  !!
  !! @param[in]     pnode           The parse node of interest
  !!
  !! @returns The location in source of the start of the ParseNode.
  
  FUNCTION ab_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !-------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(AttrBase), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    loc = pnode%location
    
  END FUNCTION ab_GetLocation
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for ProcAttrSpec.
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /proc-attr-spec/ (R1213).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             The start of the /proc-attr-spec/ in 
  !! @a tlist.
  !!
  !! @param[in]     finish            The end of the /proc-attr-spec. in 
  !! @a tlist.
  !!
  !! @param[in,out] attrs             Structure that contains the result of 
  !! parsing all the /proc-attr-spec/'s.  Note that this is INTENT(INOUT).
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE parse_proc_attr_spec(tlist, start, finish, attrs, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(ProcAttrSpec), INTENT(INOUT) :: attrs
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /proc-attr-spec/.
    CHARACTER(*), PARAMETER :: comp = 'R1213'
    
    ! Keywords for each /proc-attr-spec/.
    !
    ! PROTECTED was added by F08 corrigendum one (interp F08/0037).
    CHARACTER(*,KIND=scck), PARAMETER :: keywords(8) = [  &
        scck_'PUBLIC     ', scck_'PRIVATE    ', scck_'BIND       ',  &
        scck_'INTENT     ', scck_'OPTIONAL   ', scck_'POINTER    ',  &
        scck_'SAVE       ', scck_'PROTECTED  ' ]
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: opt            ! Expect option.
    INTEGER :: closing        ! Index of closing parenthesis.
    
    ! Error list for child parse calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    CALL Expect(  &
        tlist,  &
        start,  &
        SPREAD(ittName, 1, SIZE(keywords)),  &
        keywords,  &
        '/proc-attr-spec/',  &
        opt, err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    SELECT CASE (opt)
    CASE (1)      ! PUBLIC
      CALL do_access(iaccPublic)
      
    CASE (2)      ! PRIVATE
      CALL do_access(iaccPrivate)
      
    CASE (3)      ! BIND
      IF (ALLOCATED(attrs%ls)) THEN
        CALL already('"BIND"')
      ELSE
        ALLOCATE(attrs%ls)
        CALL ParseLanguageBindingSpec(  &
            tlist,  &
            start, finish,  &
            attrs%ls, sub_err_list )
        CALL Add(err_list, sub_err_list)
        IF (Failure(sub_err_list)) RETURN
      END IF
      
    CASE (4)      ! INTENT
      IF (attrs%intent /= 0) THEN
        CALL already('"INTENT"')
      ELSE
        CALL Expect(tlist, start + 1, ittDelimiter, scck_'(', err)
        CALL Add(err_list, err, COMPONENT=comp)
        IF (Failure(err)) RETURN
        
        CALL ExpectClosing(tlist, start + 1, closing, err)
        CALL Add(err_list, err, COMPONENT=comp)
        IF (Failure(err)) RETURN
        
        CALL ParseIntentSpec(  &
            tlist(start+2:closing-1),  &
            attrs%intent,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
        IF (Failure(sub_err_list)) RETURN
        
        ! Make sure nothing superfluous follows
        CALL ExpectEnd(tlist(:finish), closing + 1, comp, err_list)
      END IF
      
    CASE (5)      ! OPTIONAL
      IF (attrs%optional) THEN
        CALL already('"OPTIONAL"')
      ELSE
        attrs%optional = .TRUE.
        CALL ExpectEnd(tlist(:finish), start + 1, comp, err_list)
      END IF
        
    CASE (6)      ! POINTER
      IF (attrs%pointer) THEN
        CALL already('"POINTER"')
      ELSE
        attrs%pointer = .TRUE.
        CALL ExpectEnd(tlist(:finish), start + 1, comp, err_list)
      END IF
      
    CASE (7)      ! SAVE
      IF (attrs%save) THEN
        CALL already('"SAVE"')
      ELSE
        attrs%save = .TRUE.
        CALL ExpectEnd(tlist(:finish), start + 1, comp, err_list)
      END IF
      
    CASE (8)      ! PROTECTED - see F08/0037.
      IF (attrs%protected) THEN
        CALL already('"PROTECTED"')
      ELSE
        attrs%protected = .TRUE.
        CALL ExpectEnd(tlist(:finish), start + 1, comp, err_list)
      END IF
      
    CASE DEFAULT
      STOP 'Unhandled keyword in ProcedureDeclarationStmts::&
          &parse_proc_attr_spec.'
      
    END SELECT
    
  CONTAINS
    
    ! Worker routine for handling an access specifier - tests if one has 
    ! already been specified, otherwise sets the access component 
    ! appropriately.
    !
    ! @param[in]    iacc              The access specifier to handle.
    SUBROUTINE do_access(iacc)
      !-------------------------------------------------------------------------
      ! Arguments
      INTEGER, INTENT(IN) :: iacc
      !*************************************************************************
      IF (attrs%access /= 0) THEN
        CALL already('An /access-spec/')
      ELSE
        attrs%access = iacc
      END IF
      CALL ExpectEnd(tlist(:finish), start + 1, comp, err_list)
    END SUBROUTINE do_access
    
    ! Worker routine to prepare the "already specified" error message.
    !
    ! @param[in]    text            Text describing the specifier that has 
    ! already been specified.
    SUBROUTINE already(text)
      USE ErrorCodes
      !-------------------------------------------------------------------------
      ! Arguments
      CHARACTER(*), INTENT(IN) :: text
      !*************************************************************************
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT='C514 via ' // comp,  &
          LOCATION=QueryLocation(tlist(start)),  &
          MSG=text // ' has already been specified for this &
            &/procedure-declaration-stmt/.' )
    END SUBROUTINE already
    
  END SUBROUTINE parse_proc_attr_spec
  
  
  !*****************************************************************************
  !!
  !> Implementation of AttrSpec%Dump - write a debugging 
  !! representation of the results of parsing the node to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE pas_Dump(pnode, unit, pfx)
    
    USE AccessStmts
    USE Expressions
    USE TypeBoundProcedureStmts
    USE TypeBoundGenericStmts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ProcAttrSpec), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%proc_intf)) THEN
      WRITE (unit, "(A)", ADVANCE='NO') pfx
      CALL pnode%proc_intf%Write(unit)
      WRITE (unit, "()")
    END IF
    
    IF (pnode%optional)      WRITE (unit, "(A,'OPTIONAL')") pfx
    IF (pnode%pointer)       WRITE (unit, "(A,'POINTER')") pfx
    IF (pnode%protected)     WRITE (unit, "(A,'PROTECTED')") pfx
    IF (pnode%save)          WRITE (unit, "(A,'SAVE')") pfx
    
    IF (pnode%access /= 0) THEN
      WRITE (unit, "(A,'/access-spec/ = ')", ADVANCE='NO') pfx
      CALL WriteAccess(unit, pnode%access)
      WRITE (unit, "()")
    END IF
    
    SELECT CASE (pnode%intent)
    CASE (0)  ! Do nothing
    CASE (1) ; WRITE (unit, "(A,'INTENT(IN)')") pfx
    CASE (2) ; WRITE (unit, "(A,'INTENT(OUT)')") pfx
    CASE (3) ; WRITE (unit, "(A,'INTENT(INOUT)')") pfx
    CASE DEFAULT ; STOP 'Invalid ProcAttrSpec%intent'
    END SELECT
    
    IF (ALLOCATED(pnode%ls)) THEN
      WRITE (unit, "(A,'BIND ')", ADVANCE='NO') pfx
      CALL WriteLangSpec(unit, pnode%ls)
      WRITE (unit, "()")
    END IF
    
  END SUBROUTINE pas_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of ProcAttrSpec%Query - traverse the node with a 
  !! query operation.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  RECURSIVE SUBROUTINE pas_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(ProcAttrSpec), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%ls)) CALL pnode%ls%Query(visitor)
    IF (ALLOCATED(pnode%proc_intf)) CALL pnode%proc_intf%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE pas_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of ProcAttrSpec%Modify - traverse the node with a 
  !! modify operation.
  !!
  !! @param[in]     pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  RECURSIVE SUBROUTINE pas_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(ProcAttrSpec), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%ls)) CALL pnode%ls%Modify(visitor)
    IF (ALLOCATED(pnode%proc_intf)) CALL pnode%proc_intf%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE pas_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of ProcAttrSpec%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns The syntax rule name for the nodes.
  
  FUNCTION pas_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'proc-attr-spec'
    
  END FUNCTION pas_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of ProcAttrSpec%GetNatures - get the natures specified by 
  !! the attributes in the statement.
  !!
  !! @param[in]     attrs             The attributes of the statement.
  !!
  !! @param[out]    natures           The natures specified by the attributes.
  
  SUBROUTINE pas_GetNatures(attrs, natures)
    
    USE AccessStmts
    USE IntentSpecs
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ProcAttrSpec), INTENT(IN) :: attrs
    INTEGER, INTENT(OUT), ALLOCATABLE :: natures(:)
    
    !***************************************************************************
    
    ALLOCATE(natures(0))
    IF (attrs%pointer)        natures = [natures, NatureDefPointer]
    ! Added by F08/0037
    IF (attrs%protected)      natures = [natures, NatureDefProtected]
    IF (attrs%optional)       natures = [natures, NatureDefOptional]
    IF (attrs%save)           natures = [natures, NatureDefSave]
    
    SELECT CASE (attrs%access)
    CASE (iaccPublic)
      natures = [natures, NatureDefPublic]
    CASE (iaccPrivate)
      natures = [natures, NatureDefPrivate]
    END SELECT
    
    SELECT CASE (attrs%intent)
    CASE (iintIn)           ; natures = [natures, NatureDefIntentIn]
    CASE (iintOut)          ; natures = [natures, NatureDefIntentOut]
    CASE (iintInOut)        ; natures = [natures, NatureDefIntentInOut]
    END SELECT
    
    IF (ALLOCATED(attrs%ls)) THEN
      natures = [natures, NatureDefBind]
    END IF
    
  END SUBROUTINE pas_GetNatures
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for NullInit.
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that make up a /null-init/ (R506).
  !!
  !! @param[in]     tlist             The sequence of tokens for the entire 
  !! statement.
  !!
  !! @param[in]     start             Index of the first token of the 
  !! /null-init/.
  !!
  !! @param[in]     finish            Index of the last token of the 
  !! /null-init/.
  !!
  !! @param[out]    null_init         The results of parsing the /null-init/.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseNullInit(tlist, start, finish, null_init, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(NullInit), INTENT(OUT) :: null_init
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /null-init/.
    CHARACTER(*), PARAMETER :: comp = 'R506'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: closing        ! Location of closing parenthesis.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    ! A name (typically NULL)...
    CALL ExpectName(tlist, start, '/null-init/', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! Store that name.
    null_init%name = IdentifierIfortBug(tlist(start))
    
    ! (
    CALL Expect(tlist, start + 1, ittDelimiter, scck_'(', err) 
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! )
    CALL ExpectClosing(tlist, start + 1, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! Nothing inside the parentheses.
    IF (closing /= start + 2) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT='C512 on ' // comp,  &
          LOCATION=QueryLocation(tlist(3)),  &
          MSG='The /function-reference/ in a /null-init/ must have no  &
              &arguments.' )
      RETURN
    END IF
    
    ! Nothing after the parentheses.
    CALL ExpectEnd(tlist(:finish), closing + 1, comp, err_list)
    
  END SUBROUTINE ParseNullInit
  
  
  !*****************************************************************************
  !!
  !> Tests to see if a sequence of tokens matches the first part of a 
  !! /null-init/ type of function reference - "name (".  
  !!
  !! Use ParseNullInit to then test that the rest of the tokens are still 
  !! suitable for the full /null-init/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             The index in @a tlist at which to start 
  !! the match.
  !!
  !! @returns True if the next two tokens in @a tlist from index @a start 
  !! are "NULL (", false otherwise.
  
  FUNCTION MatchNullInit(tlist, start) RESULT(b)
    
    USE CompilerKinds
    USE Tokens
    USE MatchUtils
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    
    ! Function result
    LOGICAL :: b
    
    !***************************************************************************
    
    ! We match against any name, not just NULL.
    b = Match(tlist, start, ittName)  &
        .AND. Match(tlist, start + 1, ittDelimiter, scck_'(')
    
  END FUNCTION MatchNullInit
  
  
  !*****************************************************************************
  !!
  !> Implementation of NullInit%Write - write a debugging representation 
  !! of the node to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to write.
  !!
  !! @param[in]     unit              Unit connected for formatted output.
  
  SUBROUTINE null_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(NullInit), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    WRITE (unit, "(A,'()')", ADVANCE='NO') .NameOf. pnode%name
    
  END SUBROUTINE null_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of NullInit%Query - traverse the node with a query 
  !! operation.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  RECURSIVE SUBROUTINE null_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(NullInit), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE null_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of NullInit%Query - traverse the node with a modify 
  !! operation.
  !!
  !! @param[in,out] pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  RECURSIVE SUBROUTINE null_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(NullInit), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE null_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of NullInit%What - return a descriptive 
  !! string for the parse node.
  !!
  !! @returns 'null-init'.
  
  FUNCTION null_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'null-init'
    
  END FUNCTION null_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of NullInit%GetLocation - get the start of the 
  !! source that corresponds to the ParseNode.
  !!
  !! @param[in]     pnode           The parse node of interest
  !!
  !! @returns The location in source of the start of the ParseNode.
  
  FUNCTION null_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !-------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(NullInit), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    loc = .LocationOf. pnode%name
    
  END FUNCTION null_GetLocation
  
  
  !*****************************************************************************
  !!
  !> Implementation of NullInit%Advertise - advertise the 
  !! identifiers defined or referenced by the parse node, and the nature 
  !! of the definition or reference.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE null_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    USE SourceLocations
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(NullInit), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(1))
    
    ident_list(1) = IdentifierNature(  &
        NAME=.NameOf. pnode%name,  &
        NATURES=[NatureRefFunctionLike],  &
        LOCATIONS=[.LocationOf. pnode%name] )
    
  END SUBROUTINE null_Advertise
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for ProcDecl.
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that make up a /proc-decl/ (R1214).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the first token in @a tlist 
  !! that is part of the /proc-decl/.
  !!
  !! @param[in]     finish            Index of the last token in @a tlist 
  !! that is part of the /proc-decl/.
  !!
  !! @param[in]     attrs             The attributes from the parent 
  !! statement.
  !!
  !! @param[out]    decl              The results of parsing the declaration.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE parse_proc_decl(tlist, start, finish, attrs, decl, err_list)
    
    USE CompilerKinds
    USE Errors
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    CLASS(AttrBase), INTENT(IN), TARGET :: attrs
    TYPE(ProcDecl), INTENT(OUT) :: decl
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /proc-decl/.
    CHARACTER(*), PARAMETER :: comp = 'R1214'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: opt            ! Expect option.
    
    ! Procedures for Expect matching.
    TYPE(MatchProcList) :: procs(2)
    
    ! Error list for child parse calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    ! Save the given attrs pointer.
    decl%attrs => attrs
    
    !---------------------------------------------------------------------------
    ! The procedure name.
    
    CALL ExpectName(tlist, start, '/procedure-entity-name/', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    decl%name = IdentifierIfortBug(tlist(start))
    
    ! Initialisation is optional.
    IF (finish == start) RETURN
    
    !---------------------------------------------------------------------------
    ! Initialisation (/proc-pointer-init/)
    
    CALL Expect(tlist, start + 1, ittAssociate, scck_'=>', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
     
    procs(1)%item => MatchNullInit
    procs(2)%item => MatchName
    CALL Expect(  &
        tlist,  &
        start + 2,  &
        procs,  &
        '/null-init/ or /initial-proc-target/',  &
        opt,  &
        err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! Do we have a function reference?
    IF (opt == 1) THEN
      ALLOCATE(decl%null_init)
      CALL ParseNullInit(  &
          tlist,  &
          start + 2, finish,  &
          decl%null_init,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
    ELSE
      decl%init_target = IdentifierIfortBug(tlist(start+2))
      CALL ExpectEnd(tlist(:finish), start + 3, comp, err_list)
    END IF
    
  END SUBROUTINE parse_proc_decl
  
  
  !*****************************************************************************
  !!
  !> Implementation of ProcDecl%Write - write a debugging representation 
  !! of the node to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to write.
  !!
  !! @param[in]     unit              Unit connected for formatted output.
  
  SUBROUTINE pd_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ProcDecl), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    WRITE (unit, "(A,'()')", ADVANCE='NO') .NameOf. pnode%name
    
    IF (ALLOCATED(pnode%null_init)) THEN
      WRITE (unit, "(' => ')", ADVANCE='NO')
      CALL pnode%null_init%Write(unit)
    END IF
    
    IF (ALLOCATED(pnode%init_target)) THEN
      WRITE (unit, "(' => ',A)", ADVANCE='NO') .NameOf. pnode%init_target
    END IF
    
  END SUBROUTINE pd_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of ProcDecl%Query - traverse the node with a query 
  !! operation.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  RECURSIVE SUBROUTINE pd_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(ProcDecl), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%null_init)) CALL pnode%null_init%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE pd_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of ProcDecl%Modify - traverse the node with a modify 
  !! operation.
  !!
  !! @param[in,out] pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  RECURSIVE SUBROUTINE pd_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(ProcDecl), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%null_init)) CALL pnode%null_init%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE pd_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of ProcDecl%What - return a descriptive 
  !! string for the parse node.
  !!
  !! @returns 'proc-decl'.
  
  FUNCTION pd_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'proc-decl'
    
  END FUNCTION pd_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of ProcDecl%GetLocation - get the start of the 
  !! source that corresponds to the ParseNode.
  !!
  !! @param[in]     pnode           The parse node of interest
  !!
  !! @returns The location in source of the start of the ParseNode.
  
  FUNCTION pd_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !-------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(ProcDecl), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    loc = .LocationOf. pnode%name
    
  END FUNCTION pd_GetLocation
  
  
  !*****************************************************************************
  !!
  !> Implementation of ProcDecl%Advertise - advertise the 
  !! identifiers defined or referenced by the parse node, and the nature 
  !! of the definition or reference.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE pd_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    USE SourceLocations
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ProcDecl), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! Natures from the attrs.
    INTEGER, ALLOCATABLE :: natures(:)
    
    ! Locations for the natures for each specific entity.
    TYPE(SourceLocation), ALLOCATABLE :: locations(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(1))
    
    CALL pnode%attrs%GetNatures(natures)
    
    ! A procedure declaration statement also gives the EXTERNAL attribute 
    ! to the identifier.
    natures = [natures, NatureDefExternal]
    
    IF (ALLOCATED(pnode%null_init) .OR. ALLOCATED(pnode%init_target)) THEN
      natures = [natures, NatureDefInitializer]
    END IF
    
    ALLOCATE(locations(SIZE(natures)))
    locations = .LocationOf. pnode%name
    
    ident_list(1) = IdentifierNature(  &
        NAME=.NameOf. pnode%name,  &
        NATURES=natures,  &
        LOCATIONS=locations )
    
  END SUBROUTINE pd_Advertise
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for /proc-interface/.
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that represents a /proc-interface/ (R1212).
  !!
  !! A /proc-interface/ is the thing that is found inside the parentheses 
  !! of a /procedure-declaration-stmt/.  It is either a type (for functions 
  !! with an implicit interface) or an interface name.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    proc_intf         The results of parsing the interface.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! If the token sequence is just a name, and that name matches the name 
  !! of an intrinsic type, then we treat the interface specification 
  !! to be that of an intrinsic type, in accordance with C1217.  When this 
  !! happens the results will be parsed into the decl component.
  
  SUBROUTINE ParseProcInterface(tlist, proc_intf, err_list)
    
    USE Errors
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(ProcInterface), INTENT(OUT) :: proc_intf
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /proc-interface/.
    CHARACTER(*), PARAMETER :: comp = 'R1212'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Procedures for Expect matching.
    TYPE(MatchProcList) :: procs(2)
    
    INTEGER :: opt            ! Expect option.
    TYPE(Error) :: err        ! Error object for Expect calls.
    
    ! Error list for child parse calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    ! So that we prefer to match against an intrinsic type name, 
    ! the match function for a declaration type spec is ahead of the 
    ! match function for just a name.
    procs(1)%item => MatchDeclarationTypeSpecStart
    procs(2)%item => MatchName
    CALL Expect(  &
        tlist, 1,  &
        procs,  &
        '/intrinsic-type-spec/, "TYPE", "CLASS" or /interface-name/',  &
        opt, err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    IF (opt == 1) THEN
      ALLOCATE(proc_intf%decl)
      CALL ParseDeclarationTypeSpec(  &
          tlist,  &
          .FALSE.,  &
          proc_intf%decl,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
    ELSE
      proc_intf%name = IdentifierIfortBug(tlist(1))
      CALL ExpectEnd(tlist, 2, comp, err_list)
    END IF
    
  END SUBROUTINE ParseProcInterface
  
  
  !*****************************************************************************
  !!
  !> Implementation of ProcInterface%Write - write a debugging representation 
  !! of the node to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to write.
  !!
  !! @param[in]     unit              Unit connected for formatted output.
  
  SUBROUTINE pi_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ProcInterface), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%decl)) THEN
      WRITE ( unit,  &
          "('/proc-interface/ (/declaration-type-spec/) = ')",  &
          ADVANCE='NO' )
      CALL WriteDeclTypeSpec(unit, pnode%decl)
    END IF
    IF (ALLOCATED(pnode%name)) THEN
      WRITE ( unit, "('/proc-interface/ (/name/) = ',A)", ADVANCE='NO')  &
          .NameOf. pnode%name
    END IF
    
  END SUBROUTINE pi_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of ProcInterface%Query - traverse the node with a query 
  !! operation.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  RECURSIVE SUBROUTINE pi_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(ProcInterface), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%decl)) CALL pnode%decl%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE pi_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of ProcInterface%Modify - traverse the node with a modify 
  !! operation.
  !!
  !! @param[in,out] pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  RECURSIVE SUBROUTINE pi_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(ProcInterface), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%decl)) CALL pnode%decl%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE pi_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of ProcInterface%What - return a descriptive 
  !! string for the parse node.
  !!
  !! @returns 'proc-decl'.
  
  FUNCTION pi_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'proc-interface'
    
  END FUNCTION pi_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of ProcInterface%GetLocation - get the start of the 
  !! source that corresponds to the ParseNode.
  !!
  !! @param[in]     pnode           The parse node of interest
  !!
  !! @returns The location in source of the start of the ParseNode.
  
  FUNCTION pi_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !-------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(ProcInterface), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%decl)) THEN
      loc = .LocationOf. pnode%decl
    ELSE
      loc = .LocationOf. pnode%name
    END IF
    
  END FUNCTION pi_GetLocation
  
  
  !*****************************************************************************
  !!
  !> Implementation of ProcInterface%Advertise - advertise the 
  !! identifiers defined or referenced by the parse node, and the nature 
  !! of the definition or reference.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE pi_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    USE SourceLocations
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ProcInterface), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%name)) THEN
      ALLOCATE(ident_list(1))
      ident_list(1) = IdentifierNature(  &
          NAME=.NameOf. pnode%name,  &
          NATURES=[NatureRefProcInterface],  &
          LOCATIONS=[.LocationOf. pnode%name] )
    ELSE
      ALLOCATE(ident_list(0))
    END IF
    
  END SUBROUTINE pi_Advertise
  
END MODULE ProcedureDeclarationStmts
