! $Id: ProgramUnits.f90 2889 2019-04-06 22:56:07Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the ProgramUnits module.


!*******************************************************************************
!!
!> Types and procedures for parsing the various program units and subprograms.

MODULE ProgramUnits
  
  USE BlockDataStmts
  USE ExecutionParts
  USE Identifiers
  USE FunctionStmts
  USE Labels
  USE MpSubprogramStmts
  USE ModuleStmts
  USE ParseNodes
  USE ProgramStmts
  USE SourceLocations
  USE SpecificationParts
  USE SubroutineStmts
  USE SubmoduleStmts
  
  IMPLICIT NONE
  
  PRIVATE
  
  PUBLIC :: AddProgramUnit
  
  PUBLIC :: ParseMainProgram
  PUBLIC :: ParseSubroutineSubprogram
  PUBLIC :: ParseFunctionSubprogram
  PUBLIC :: ParseSeparateModuleSubprogram
  PUBLIC :: ParseModule
  PUBLIC :: ParseSubmodule
  PUBLIC :: ParseBlockData
  
  !-----------------------------------------------------------------------------
  ! Natures for function and subroutine subprograms.
  
  !> The subprogram is defining an external procedure.
  INTEGER, PARAMETER, PUBLIC :: ispntExternal = 0
  !> The subprogram is defining a module procedure.
  INTEGER, PARAMETER, PUBLIC :: ispntModule = 1
  !> The subprogram is defining an internal procedure.
  INTEGER, PARAMETER, PUBLIC :: ispntInternal = 2
  
  !-----------------------------------------------------------------------------
  
  !> Common attributes for all program units and subprograms.
  !!
  !! All program units have a specification part.
  TYPE, EXTENDS(SCParseNode), ABSTRACT, PUBLIC :: ProgramUnit
    !> The specification part of the program unit.
    !!
    !! @todo This probably needs to be made allocatable.
    TYPE(SpecificationPart) :: spec
    
    !> The name on the closing end statement.
    !!
    !! Not allocated if no name was provided on the closing end statement.
    !!
    !! Check ensures that this is not allocated if there is no opening 
    !! statement (for main programs), and that if it is allocated the 
    !! name is the same as the name in @a prog_stmt.
    TYPE(Identifier), PRIVATE, ALLOCATABLE :: end_name
  CONTAINS
    PROCEDURE(pu_Who), DEFERRED :: Who
    PROCEDURE :: Check => pu_Check
  END TYPE ProgramUnit
  
  ! Interfaces for bindings in ProgramUnit.
  ABSTRACT INTERFACE
    !> Query the identifier for the program unit.
    !!
    !! @param[in]     pnode           The parse node for the program unit.
    !!
    !! @returns The identifier of the function or subroutine declared by 
    !! the program unit.
    FUNCTION pu_Who(pnode) RESULT(name)
      IMPORT :: ProgramUnit
      IMPLICIT NONE
      !-------------------------------------------------------------------------
      CLASS(ProgramUnit), INTENT(IN) :: pnode
      CHARACTER(:), ALLOCATABLE :: name
    END FUNCTION pu_Who
  END INTERFACE
  
  !> An element in a vector of program units.
  TYPE, PUBLIC :: ProgramUnitElement
    CLASS(ProgramUnit), ALLOCATABLE :: item
  CONTAINS
    PROCEDURE :: Dump => pue_Dump
    PROCEDURE :: Query => pue_Query
    PROCEDURE :: Modify => pue_Modify
    PROCEDURE :: What => pue_What
  END TYPE ProgramUnitElement
  
  !> A storage type for a program unit with a specification part, an 
  !! execution part and some internal programs.
  TYPE, PUBLIC, EXTENDS(ProgramUnit), ABSTRACT :: ExecProgramUnit
    !> The executable part of the main program.
    TYPE(ExecutionPart) :: exec
    !> The internal procedures of the function.
    !!
    !! Always allocated, may be length zero.
    TYPE(ProgramUnitElement), ALLOCATABLE :: internals(:)
    !> The label on the end statement, if present.
    !!
    !! Not allocated if there was no label on the end statement.
    TYPE(Label), ALLOCATABLE :: end_label
  END TYPE ExecProgramUnit
  
  !> The results of parsing a main program.
  TYPE, PUBLIC, EXTENDS(ExecProgramUnit) :: MainProgram
    !> The opening program statement of the main program, if present.
    !!
    !! Not allocated if there was no opening program statement.
    TYPE(ProgramStmt), ALLOCATABLE :: prog_stmt
  CONTAINS
    PROCEDURE :: Dump => main_Dump
    PROCEDURE :: Query => main_Query
    PROCEDURE :: Modify => main_Modify
    PROCEDURE, NOPASS :: What => main_What
    PROCEDURE :: Advertise => main_Advertise
    PROCEDURE :: Who => main_Who
    PROCEDURE :: Check => main_Check
  END TYPE MainProgram
  
  !> The results of parsing a subroutine subprogram.
  !!
  !! This is used for external subroutines that are program units and 
  !! for module and internal subroutines that are not.
  TYPE, PUBLIC, EXTENDS(ExecProgramUnit) :: SubroutineSubprogram
    !> The results of parsing the opening subroutine statement.
    TYPE(SubroutineStmt) :: sub_stmt
    !> The nature of the subroutine subprogram.  See the ispnt* constants.
    INTEGER :: nature
  CONTAINS
    PROCEDURE :: Dump => sub_Dump
    PROCEDURE :: Query => sub_Query
    PROCEDURE :: Modify => sub_Modify
    PROCEDURE, NOPASS :: What => sub_What
    PROCEDURE :: Advertise => sub_Advertise
    PROCEDURE :: Who => sub_Who
    PROCEDURE :: Check => sub_Check
  END TYPE SubroutineSubprogram
  
  !> The results of parsing a function subprogram.
  !!
  !! This is used for both external functions that are program units and 
  !! for module and internal functions that are not.
  TYPE, PUBLIC, EXTENDS(ExecProgramUnit) :: FunctionSubprogram
    !> The results of parsing the function statement.
    TYPE(FunctionStmt) :: fn_stmt
    !> The nature of the function subprogram.  See the ispnt* constants.
    INTEGER :: nature
  CONTAINS
    PROCEDURE :: Dump => fun_Dump
    PROCEDURE :: Query => fun_Query
    PROCEDURE :: Modify => fun_Modify
    PROCEDURE, NOPASS :: What => fun_What
    PROCEDURE :: Advertise => fun_Advertise
    PROCEDURE :: Who => fun_Who
    PROCEDURE :: Check => fun_Check
  END TYPE FunctionSubprogram
  
  !> The results of parsing a separate module procedure that is being 
  !! defined with a /mp-subprogram-stmt/ - i.e. a strict 
  !! /separate-module-subprogram/.
  !!
  !! Separate module procedures defined with function statments or with 
  !! subroutine statements are parsed into FunctionSubprogram and 
  !! SubroutineSubprogram respectively (syntactically they are not 
  !! /separate-module-subprogram/'s, even though semantically they are 
  !! the same).
  TYPE, PUBLIC, EXTENDS(ExecProgramUnit) :: SeparateModuleSubprogram
    !> The results of parsing the opening /mp-subprogram-stmt/ - i.e. 
    !! the name of the procedure.
    TYPE(MpSubprogramStmt) :: mps_stmt
  CONTAINS
    PROCEDURE :: Dump => sms_Dump
    PROCEDURE :: Query => sms_Query
    PROCEDURE :: Modify => sms_Modify
    PROCEDURE, NOPASS :: What => sms_What
    PROCEDURE :: Advertise => sms_Advertise
    PROCEDURE :: Who => sms_Who
    PROCEDURE :: Check => sms_Check
  END TYPE SeparateModuleSubprogram
  
  !> The common results of parsing a module or submodule.
  TYPE, PUBLIC, EXTENDS(ProgramUnit), ABSTRACT :: ModuleCommon
    !> The module procedures of the main program.
    !!
    !! Always allocated, may be length zero.
    TYPE(ProgramUnitElement), ALLOCATABLE :: internals(:)
  END TYPE ModuleCommon
  
  !> The results of parsing a module.
  TYPE, PUBLIC, EXTENDS(ModuleCommon) :: Module
    TYPE(ModuleStmt) :: mod_stmt
  CONTAINS
    PROCEDURE :: Dump => mod_Dump
    PROCEDURE, NOPASS :: What => mod_What
    PROCEDURE :: Query => mod_Query
    PROCEDURE :: Modify => mod_Modify
    PROCEDURE :: Advertise => mod_Advertise
    PROCEDURE :: Who => mod_Who
    PROCEDURE :: Check => mod_Check
  END TYPE Module
  
  !> The results of parsing a submodule.
  TYPE, PUBLIC, EXTENDS(ModuleCommon) :: Submodule
    TYPE(SubmoduleStmt) :: submod_stmt
  CONTAINS
    PROCEDURE :: Dump => sm_Dump
    PROCEDURE, NOPASS :: What => sm_What
    PROCEDURE :: Query => sm_Query
    PROCEDURE :: Modify => sm_Modify
    PROCEDURE :: Advertise => sm_Advertise
    PROCEDURE :: Who => sm_Who
    PROCEDURE :: Check => sm_Check
    
    PROCEDURE :: QueryAncestorModuleName => sm_QueryAncestorModuleName
    PROCEDURE :: QueryParentSubmoduleName => sm_QueryParentSubmoduleName
    PROCEDURE :: QueryParentIdentifier => sm_QueryParentIdentifier
  END TYPE Submodule
  
  !> The results of parsing a /block-data/ unit.
  TYPE, PUBLIC, EXTENDS(ProgramUnit) :: BlockData
    !> The opening block data statement.
    TYPE(BlockDataStmt) :: bd_stmt
  CONTAINS
    PROCEDURE :: Dump => bd_Dump
    PROCEDURE, NOPASS :: What => bd_What
    PROCEDURE :: Query => bd_Query
    PROCEDURE :: Modify => bd_Modify
    PROCEDURE :: Advertise => bd_Advertise
    PROCEDURE :: Who => bd_Who
    PROCEDURE :: Check => bd_Check
  END TYPE BlockData
  
CONTAINS
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for ProgramUnit
  
  
  !*****************************************************************************
  !!
  !> Perform some checks common to all program units.
  !!
  !! @param[in]     pnode             The parse node to check.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE pu_Check(pnode, err_list)
    
    USE Errors
    USE ErrorCodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ProgramUnit), INTENT(IN) :: pnode
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    ! The specification part of a program unit is not permitted to have an 
    ! import statement.
    IF (SIZE(pnode%spec%import_stmts) > 0) THEN
      CALL Add( err_list,  &
          CODE=errImportNotPermitted,  &
          COMPONENT='C1210',  &
          LOCATION=.LocationOf. pnode%spec%import_stmts(1),  &
          MSG='An import statement is not permitted in the specification &
            &part of a program unit.' )
    END IF
    
  END SUBROUTINE pu_Check
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for bindings in ExecProgramUnit
  
  
  !*****************************************************************************
  !!
  !> Query the child nodes of an ExecProgramUnit.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  RECURSIVE SUBROUTINE QueryExecProgramUnit(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(ExecProgramUnit), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Child node index.
    
    !***************************************************************************
    
    CALL pnode%spec%Query(visitor)
    
    CALL pnode%exec%Query(visitor)
    
    DO i = 1, SIZE(pnode%internals)
      CALL pnode%internals(i)%Query(visitor)
    END DO
    
  END SUBROUTINE QueryExecProgramUnit
  
  
  !*****************************************************************************
  !!
  !> Modify the child nodes of an ExecProgramUnit.
  !!
  !! @param[in,out] pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  RECURSIVE SUBROUTINE ModifyExecProgramUnit(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(ExecProgramUnit), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Child node index.
    
    !***************************************************************************
    
    CALL pnode%spec%Modify(visitor)
    
    CALL pnode%exec%Modify(visitor)
    
    DO i = 1, SIZE(pnode%internals)
      CALL pnode%internals(i)%Modify(visitor)
    END DO
    
  END SUBROUTINE ModifyExecProgramUnit
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for MainProgram.
  
  
  !*****************************************************************************
  !!
  !> Parse a statement tree node that corresponds to a main program.
  !!
  !! @param[in]     st                The statement tree node for the main 
  !! program.
  !!
  !! @param[in]     main_prog         The results of parsing the main program.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseMainProgram(st, main_prog, err_list)
    
    USE Errors
    USE ProgramStmts
    USE ProgUnitStNodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(MainProgramStNode), INTENT(INOUT) :: st
    TYPE(MainProgram), INTENT(OUT) :: main_prog
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    ! Program statement is optional.
    IF (ALLOCATED(st%first)) THEN 
      ALLOCATE(main_prog%prog_stmt)
      CALL ParseProgramStmt(st%first%tlist, main_prog%prog_stmt, sub_err_list)
      CALL Add(err_list, sub_err_list)
    END IF
    
    IF (ALLOCATED(st%spec)) THEN
      CALL ParseSpecificationPart(st%spec, main_prog%spec, sub_err_list)
      CALL Add(err_list, sub_err_list)
    END IF
    
    IF (ALLOCATED(st%exec)) THEN
      CALL ParseExecutionPart(st%exec, main_prog%exec, sub_err_list)
      CALL Add(err_list, sub_err_list)
    END IF
    
    IF (ALLOCATED(st%internal)) THEN
      CALL parse_internal_subprograms(  &
          st%internal,  &
          ispntInternal,  &
          main_prog%internals,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list)
    ELSE
      ALLOCATE(main_prog%internals(0))
    END IF
    
    CALL ParseEndProgramStmt(st%last%tlist, main_prog%end_name, sub_err_list)
    CALL Add(err_list, sub_err_list)
    
    IF (ALLOCATED(st%last%statement_label)) THEN
      main_prog%end_label = st%last%statement_label
    END IF
    
  END SUBROUTINE ParseMainProgram
  
  
  !*****************************************************************************
  !!
  !> Implementation of MainProgram%Dump - write a debugging representation 
  !! of the results of parsing the main program to a logical unit.
  !!
  !! @param[in]     pnode             The MainProgram to dump.
  !!
  !! @param[in]     unit              Logical unit connected for 
  !! formatted output.
  !!
  !! @param[in]     pfx               Prefix for each record.
  
  SUBROUTINE main_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(MainProgram), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%prog_stmt)) THEN
      WRITE (unit, "(A,'/',A,'/')") pfx, pnode%prog_stmt%What()
      CALL pnode%prog_stmt%Dump(unit, pfx // '  ')
    END IF
    
    WRITE (unit, "(A,'/specification-part/')") pfx
    CALL pnode%spec%Dump(unit, pfx // '  ')
    
    WRITE (unit, "(A,'/',A,'/')") pfx, pnode%exec%What()
    CALL pnode%exec%Dump(unit, pfx // '  ')
    
    IF (SIZE(pnode%internals) > 0) THEN
      WRITE (unit, "(A,'/internal-subprogram-part/')") pfx
      CALL dump_internals(pnode%internals, unit, pfx // '  ')
    END IF
    
    IF (ALLOCATED(pnode%end_name)) THEN
      WRITE (unit, "(A,'/end-program-stmt/')") pfx
      WRITE (unit, "(A,'/program-name/ = ',A)")  &
          pfx // '  ', .NameOf. pnode%end_name
    ELSE
      WRITE (unit, "(A,'/end-program-stmt/')") pfx
    END IF
    
  END SUBROUTINE main_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of MainProgram%Query - traverse the node with a 
  !! query operation.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  RECURSIVE SUBROUTINE main_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(MainProgram), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%prog_stmt)) THEN
      CALL pnode%prog_stmt%Query(visitor)
    END IF
    
    CALL QueryExecProgramUnit(pnode, visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE main_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of MainProgram%Modify - traverse the node with a 
  !! modify operation.
  !!
  !! @param[in,out] pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  RECURSIVE SUBROUTINE main_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(MainProgram), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%prog_stmt)) THEN
      CALL pnode%prog_stmt%Modify(visitor)
    END IF
    
    CALL ModifyExecProgramUnit(pnode, visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE main_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of MainProgram%What - return a descriptive string for 
  !! the type of parse node.
  !!
  !! @returns 'main-program'
  
  FUNCTION main_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'main-program'
    
  END FUNCTION main_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of MainProgram%Who - return the identifier text 
  !! for the parse node.
  !!
  !! @param[in]     pnode             The parse node to query.
  !!
  !! @returns The name of the main program, or an internal substitute if the 
  !! main program has no name.
  
  FUNCTION main_Who(pnode) RESULT(name)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(MainProgram), INTENT(IN) :: pnode
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: name
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%prog_stmt)) THEN
      name = .NameOf. pnode%prog_stmt%name
    ELSE
      name = 'unnamed main program'
    END IF
    
  END FUNCTION main_Who
  
  
  !*****************************************************************************
  !!
  !> Implementation of MainProgram%Advertise - advertise the identifiers 
  !! referenced or defined by the main program.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference, from the 
  !! perspective of the host.
  
  SUBROUTINE main_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(MainProgram), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%prog_stmt)) THEN
      ALLOCATE(ident_list(1))
      ident_list = IdentifierNature(  &
          .NameOf. pnode%prog_stmt%name,  &
          [NatureDefMainProgram],  &
          [.Locationof. pnode%prog_stmt%name] )
    ELSE
      ALLOCATE(ident_list(0))
    END IF
    
  END SUBROUTINE main_Advertise
  
  
  !*****************************************************************************
  !!
  !> Implementation of MainProgram%Check - check the parse node for 
  !! conformance with language standards.
  !!
  !! @param[in]     pnode             The parse node to check.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE main_Check(pnode, err_list)
    
    USE Errors
    USE ErrorCodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(MainProgram), INTENT(IN) :: pnode
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! Error list for child calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (ALLOCATED(pnode%end_name)) THEN
      IF (ALLOCATED(pnode%prog_stmt)) THEN
        IF (.NameOf. pnode%prog_stmt%name /= .NameOf. pnode%end_name) THEN
          CALL Add( err_list, CODE=errNameMismatch,  &
              COMPONENT='C1101 on R1101',  & 
              LOCATION=.LocationOf. pnode%end_name,  &
              MSG='The name in the /end-program-stmt/ ("'  &
                // .NameOf. pnode%end_name  &
                // '") must match the name given in the /program-stmt/ ("'  &
                // .NameOf. pnode%prog_stmt%name // '").' )
        END IF
      ELSE
        CALL Add( err_list,  &
            CODE=errNameMismatch,  &
            COMPONENT='C1101 on R1101',  &
            LOCATION=.LocationOf. pnode%end_name,  &
            MSG='A name in an /end-program-stmt/ may only be specified if a &
              &/program-stmt/ was provided at the start of the &
              &/main-program/.' )
      END IF
    END IF
    
    CALL pu_Check(pnode, sub_err_list)
    CALL Add(err_list, sub_err_list)
    
  END SUBROUTINE main_Check
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for SubroutineSubprogram
  
  
  !*****************************************************************************
  !!
  !> Parse the statement tree node for a subroutine subprogram (external, 
  !! internal or module).
  !!
  !! @param[in]     st                The statement tree node for the 
  !! subroutine.
  !!
  !! @param[in]     nature            The nature of the subroutine 
  !! subprogram:
  !! - 0: The procedure is external.
  !! - 1: The procedures is a module procedure.
  !! - 2: The procedures is an internal procedure.
  !!
  !! @param[in]     sub_prog         The results of parsing the subroutine.
  !!
  !! @param[out]    err_list          List of errors.
  
  ! Recursive because subroutine subprograms can have internal subroutines.
  RECURSIVE SUBROUTINE ParseSubroutineSubprogram( st, nature, sub_prog,  &
      err_list )
    
    USE Errors
    USE ProgUnitStNodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(SubroutineSubprogramStNode), INTENT(INOUT) :: st
    INTEGER, INTENT(IN) :: nature
    TYPE(SubroutineSubprogram), INTENT(OUT) :: sub_prog
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    ! Save nature.
    sub_prog%nature = nature
    
    CALL ParseSubroutineStmt(st%first%tlist, sub_prog%sub_stmt, sub_err_list)
    CALL Add(err_list, sub_err_list)
    
    IF (ALLOCATED(st%spec)) THEN
      CALL ParseSpecificationPart(st%spec, sub_prog%spec, sub_err_list)
      CALL Add(err_list, sub_err_list)
    END IF
    
    IF (ALLOCATED(st%exec)) THEN
      CALL ParseExecutionPart(st%exec, sub_prog%exec, sub_err_list)
      CALL Add(err_list, sub_err_list)
    END IF
    
    IF (ALLOCATED(st%internal)) THEN
      CALL parse_internal_subprograms(  &
          st%internal,  &
          ispntInternal,  &
          sub_prog%internals,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list)
    ELSE
      ALLOCATE(sub_prog%internals(0))
    END IF
    
    CALL ParseEndSubroutineStmt(st%last%tlist, sub_prog%end_name, sub_err_list)
    CALL Add(err_list, sub_err_list)
    
    IF (ALLOCATED(st%last%statement_label)) THEN
      sub_prog%end_label = st%last%statement_label
    END IF
    
  END SUBROUTINE ParseSubroutineSubprogram
  
  
  !*****************************************************************************
  !!
  !> Implementation of SubroutineSubprogram%Dump - write a debugging 
  !! representation of the results of parsing the subroutine subprogram 
  !! to a logical unit.
  !!
  !! @param[in]     pnode             The SubroutineSubprogram to dump.
  !!
  !! @param[in]     unit              Logical unit connected for 
  !! formatted output.
  !!
  !! @param[in]     pfx               Prefix for each record.
  
  ! Recursive because subroutine subprograms can have internal subroutines.
  RECURSIVE SUBROUTINE sub_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(SubroutineSubprogram), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    
    WRITE (unit, "(A,'/',A,'/')") pfx, pnode%sub_stmt%What()
    CALL pnode%sub_stmt%Dump(unit, pfx // '  ')
    
    WRITE (unit, "(A,'/specification-part/')") pfx
    CALL pnode%spec%Dump(unit, pfx // '  ')
    
    WRITE (unit, "(A,'/',A,'/')") pfx, pnode%exec%What()
    CALL pnode%exec%Dump(unit, pfx // '  ')
    
    IF (SIZE(pnode%internals) > 0) THEN
      WRITE (unit, "(A,'internal subprograms:')") pfx
      CALL dump_internals(pnode%internals, unit, pfx // '  ')
    END IF
    
    IF (ALLOCATED(pnode%end_name)) THEN
      WRITE (unit, "(A,'/end-subroutine-stmt/')") pfx
      WRITE (unit, "(A,'/subroutine-name/ = ',A)")  &
          pfx // '  ', .NameOf. pnode%end_name
    ELSE
      WRITE (unit, "(A,'/end-subroutine-stmt/')") pfx
    END IF
    
  END SUBROUTINE sub_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of SubroutineSubprogram%Query.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  RECURSIVE SUBROUTINE sub_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(SubroutineSubprogram), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%sub_stmt%Query(visitor)
    
    CALL QueryExecProgramUnit(pnode, visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE sub_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of SubroutineSubprogram%Modify.
  !!
  !! @param[in,out] pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  RECURSIVE SUBROUTINE sub_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(SubroutineSubprogram), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%sub_stmt%Modify(visitor)
    
    CALL ModifyExecProgramUnit(pnode, visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE sub_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of SubroutineSubprogram%What - return a descriptive string 
  !! for the parse node.
  !!
  !! @returns 'subroutine-subprogram'
  
  FUNCTION sub_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'subroutine-subprogram'
    
  END FUNCTION sub_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of SubroutineSubprogram%Who - return the identifier text 
  !! for the parse node.
  !!
  !! @param[in]     pnode             The parse node to query.
  !!
  !! @returns The name of the subroutine.
  
  FUNCTION sub_Who(pnode) RESULT(name)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(SubroutineSubprogram), INTENT(IN) :: pnode
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: name
    
    !***************************************************************************
    
    name = .NameOf. pnode%sub_stmt%subroutine_name
    
  END FUNCTION sub_Who
  
  
  !*****************************************************************************
  !!
  !> Implementation of SubroutineSubprogram%Advertise - advertise the 
  !! identifiers referenced or defined by the subroutine subprogram.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference, from the 
  !! perspective of the host.
  
  SUBROUTINE sub_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(SubroutineSubprogram), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(1))
    SELECT CASE (pnode%nature)
    CASE (ispntExternal)
      ident_list(1) = IdentifierNature(  &
          NAME=.NameOf. pnode%sub_stmt%subroutine_name,  &
          NATURES=[NatureDefSubroutine],  &
          LOCATIONS=[.Locationof. pnode%sub_stmt%subroutine_name] )
      
    CASE (ispntModule)
      ident_list(1) = IdentifierNature(  &
          NAME=.NameOf. pnode%sub_stmt%subroutine_name,  &
          NATURES=[NatureDefSubroutine, NatureDefModuleProcedure],  &
          LOCATIONS=[  &
            .Locationof. pnode%sub_stmt%subroutine_name,  &
            .Locationof. pnode%sub_stmt%subroutine_name ] )
      
    CASE (ispntInternal)
      ident_list(1) = IdentifierNature(  &
          NAME=.NameOf. pnode%sub_stmt%subroutine_name,  &
          NATURES=[NatureDefSubroutine, NatureDefInternalProcedure],  &
          LOCATIONS=[  &
            .Locationof. pnode%sub_stmt%subroutine_name,  &
            .Locationof. pnode%sub_stmt%subroutine_name ] )
      
    END SELECT
    
    ! @todo Check for entry statements.
    
  END SUBROUTINE sub_Advertise
  
  
  !*****************************************************************************
  !!
  !> Implementation of SubroutineSubprogram%Check - check the parse node for 
  !! conformance with language standards.
  !!
  !! @param[in]     pnode             The parse node to check.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE sub_Check(pnode, err_list)
    
    USE Errors
    USE ErrorCodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(SubroutineSubprogram), INTENT(IN) :: pnode
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! Error list for child calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (ALLOCATED(pnode%end_name)) THEN
      IF ( .NameOf. pnode%sub_stmt%subroutine_name  &
          /= .NameOf. pnode%end_name ) THEN
        CALL Add( err_list, CODE=errNameMismatch,  &
            COMPONENT='C1261 on R1236',  & 
            LOCATION=.LocationOf. pnode%end_name,  &
            MSG='The name in the /end-subroutine-stmt/ ("'  &
              // .NameOf. pnode%end_name  &
              // '") must match the name given in the /subroutine-stmt/ ("'  &
              // .NameOf. pnode%sub_stmt%subroutine_name // '").' )
      END IF
    END IF
    
    CALL pu_Check(pnode, sub_err_list)
    CALL Add(err_list, sub_err_list)
    
  END SUBROUTINE sub_Check
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for FunctionSubprogram
  
  
  !*****************************************************************************
  !!
  !> Parse the statement tree node for a function subprogram (external, 
  !! internal or module).
  !!
  !! @param[in]     st                The statement tree node for the 
  !! function.
  !!
  !! @param[in]     nature            The nature of the function subprogram:
  !! - 0: The procedure is external.
  !! - 1: The procedures is a module procedure.
  !! - 2: The procedures is an internal procedure.
  !!
  !! @param[in]     sub_prog         The results of parsing the function.
  !!
  !! @param[out]    err_list          List of errors.
  
  ! Recursive because function subprograms can have internal functions.
  RECURSIVE SUBROUTINE ParseFunctionSubprogram(st, nature, fun_prog, err_list)
    
    USE Errors
    USE ProgUnitStNodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(FunctionSubprogramStNode), INTENT(INOUT) :: st
    INTEGER, INTENT(IN) :: nature
    TYPE(FunctionSubprogram), INTENT(OUT) :: fun_prog
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    ! Save nature.
    fun_prog%nature = nature
    
    CALL ParseFunctionStmt(st%first%tlist, fun_prog%fn_stmt, sub_err_list)
    CALL Add(err_list, sub_err_list)
    
    IF (ALLOCATED(st%spec)) THEN
      CALL ParseSpecificationPart(st%spec, fun_prog%spec, sub_err_list)
      CALL Add(err_list, sub_err_list)
    END IF
    
    IF (ALLOCATED(st%exec)) THEN
      CALL ParseExecutionPart(st%exec, fun_prog%exec, sub_err_list)
      CALL Add(err_list, sub_err_list)
    END IF
    
    IF (ALLOCATED(st%internal)) THEN
      CALL parse_internal_subprograms(  &
          st%internal,  &
          ispntInternal,  &
          fun_prog%internals,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list)
    ELSE
      ALLOCATE(fun_prog%internals(0))
    END IF
    
    CALL ParseEndFunctionStmt(  &
        st%last%tlist,  &
        fun_prog%end_name,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list)
    
    IF (ALLOCATED(st%last%statement_label)) THEN
      fun_prog%end_label = st%last%statement_label
    END IF
    
  END SUBROUTINE ParseFunctionSubprogram
  
  
  !*****************************************************************************
  !!
  !> Implementation of FunctionSubprogram%Dump - write a debugging 
  !! representation of the results of parsing the function subprogram 
  !! to a logical unit.
  !!
  !! @param[in]     pnode             The FunctionSubprogram to dump.
  !!
  !! @param[in]     unit              Logical unit connected for 
  !! formatted output.
  !!
  !! @param[in]     pfx               Prefix for each record.
  
  ! Recursive because function subprograms can have internal functions.
  RECURSIVE SUBROUTINE fun_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(FunctionSubprogram), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    WRITE (unit, "(A,'/',A,'/')") pfx, pnode%fn_stmt%What()
    CALL pnode%fn_stmt%Dump(unit, pfx // '  ')
    
    WRITE (unit, "(A,'/specification-part/')") pfx
    CALL pnode%spec%Dump(unit, pfx // '  ')
    
    WRITE (unit, "(A,'/',A,'/')") pfx, pnode%exec%What()
    CALL pnode%exec%Dump(unit, pfx // '  ')
    
    IF (SIZE(pnode%internals) > 0) THEN
      WRITE (unit, "(A,'internal subprograms:')") pfx
      CALL dump_internals(pnode%internals, unit, pfx // '  ')
    END IF
    
    IF (ALLOCATED(pnode%end_name)) THEN
      WRITE (unit, "(A,'/end-function-stmt/')") pfx
      WRITE (unit, "(A,'/function-name/ = ',A)")  &
          pfx // '  ', .NameOf. pnode%end_name
    ELSE
      WRITE (unit, "(A,'/end-function-stmt/')") pfx
    END IF
    
  END SUBROUTINE fun_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of FunctionSubprogram%Query.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  RECURSIVE SUBROUTINE fun_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(FunctionSubprogram), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%fn_stmt%Query(visitor)
    
    CALL QueryExecProgramUnit(pnode, visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE fun_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of FunctionSubprogram%Modify.
  !!
  !! @param[in,out] pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  RECURSIVE SUBROUTINE fun_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(FunctionSubprogram), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%fn_stmt%Modify(visitor)
    
    CALL ModifyExecProgramUnit(pnode, visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE fun_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of FunctionSubprogram%What - return a descriptive string 
  !! for the parse node.
  !!
  !! @returns 'function-subprogram'
  
  FUNCTION fun_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'function-subprogram'
    
  END FUNCTION fun_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of FunctionSubprogram%Who - return the identifier text 
  !! for the parse node.
  !!
  !! @param[in]     pnode             The parse node to query.
  !!
  !! @returns The name of the function.
  
  FUNCTION fun_Who(pnode) RESULT(name)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(FunctionSubprogram), INTENT(IN) :: pnode
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: name
    
    !***************************************************************************
    
    name = .NameOf. pnode%fn_stmt%function_name
    
  END FUNCTION fun_Who
  
  
  !*****************************************************************************
  !!
  !> Implementation of FunctionSubprogram%Advertise - advertise the identifiers 
  !! referenced or defined by the function from the perspective of the host.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference, from the 
  !! perspective of the host.
  
  SUBROUTINE fun_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(FunctionSubprogram), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(1))
    SELECT CASE (pnode%nature)
    CASE (ispntExternal)
      ident_list(1) = IdentifierNature(  &
          NAME=.NameOf. pnode%fn_stmt%function_name,  &
          NATURES=[NatureDefFunction],  &
          LOCATIONS=[.Locationof. pnode%fn_stmt%function_name] )
      
    CASE (ispntModule)
      ident_list(1) = IdentifierNature(  &
          NAME=.NameOf. pnode%fn_stmt%function_name,  &
          NATURES=[NatureDefFunction, NatureDefModuleProcedure],  &
          LOCATIONS=[  &
            .Locationof. pnode%fn_stmt%function_name,  &
            .Locationof. pnode%fn_stmt%function_name ] )
      
    CASE (ispntInternal)
      ident_list(1) = IdentifierNature(  &
          NAME=.NameOf. pnode%fn_stmt%function_name,  &
          NATURES=[NatureDefFunction, NatureDefInternalProcedure],  &
          LOCATIONS=[  &
            .Locationof. pnode%fn_stmt%function_name,  &
            .Locationof. pnode%fn_stmt%function_name ] )
      
    END SELECT
    
    ! @todo Check for entry statements.
    
  END SUBROUTINE fun_Advertise
  
  
  !*****************************************************************************
  !!
  !> Implementation of FunctionSubprogram%Check - check the parse node for 
  !! conformance with language standards.
  !!
  !! @param[in]     pnode             The parse node to check.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE fun_Check(pnode, err_list)
    
    USE Errors
    USE ErrorCodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(FunctionSubprogram), INTENT(IN) :: pnode
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! Error list for child calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (ALLOCATED(pnode%end_name)) THEN
      IF ( .NameOf. pnode%fn_stmt%function_name  &
          /= .NameOf. pnode%end_name ) THEN
        CALL Add( err_list, CODE=errNameMismatch,  &
            COMPONENT='C1258 on R1232',  & 
            LOCATION=.LocationOf. pnode%end_name,  &
            MSG='The name in the /end-function-stmt/ ("'  &
              // .NameOf. pnode%end_name  &
              // '") must match the name given in the /function-stmt/ ("'  &
              // .NameOf. pnode%fn_stmt%function_name // '").' )
      END IF
    END IF
    
    CALL pu_Check(pnode, sub_err_list)
    CALL Add(err_list, sub_err_list)
    
  END SUBROUTINE fun_Check
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for SeparateModuleSubprogram.
  
  
  !*****************************************************************************
  !!
  !> Parse the statement tree node for a /separate-module-subprogram/.
  !!
  !! @param[in]     st                The statement tree node for the 
  !! function.
  !!
  !! @param[in]     separate_subprog  The results of parsing the separate 
  !! module procedure.
  !!
  !! @param[out]    err_list          List of errors.
  
  ! As separate module procedures cannot contain separate module procedures, 
  ! this doesn't need to be recursive.
  SUBROUTINE ParseSeparateModuleSubprogram(st, separate_subprog, err_list)
    
    USE Errors
    USE MpSubprogramStmts
    USE ProgUnitStNodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(SeparateModuleSubprogramStNode), INTENT(INOUT) :: st
    TYPE(SeparateModuleSubprogram), INTENT(OUT) :: separate_subprog
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    CALL ParseMPSubprogramStmt(  &
        st%first%tlist,  &
        separate_subprog%mps_stmt,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list)
    
    IF (ALLOCATED(st%spec)) THEN
      CALL ParseSpecificationPart(st%spec, separate_subprog%spec, sub_err_list)
      CALL Add(err_list, sub_err_list)
    END IF
    
    IF (ALLOCATED(st%exec)) THEN
      CALL ParseExecutionPart(st%exec, separate_subprog%exec, sub_err_list)
      CALL Add(err_list, sub_err_list)
    END IF
    
    IF (ALLOCATED(st%internal)) THEN
      CALL parse_internal_subprograms(  &
          st%internal,  &
          ispntInternal,  &
          separate_subprog%internals,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list)
    ELSE
      ALLOCATE(separate_subprog%internals(0))
    END IF
    
    CALL ParseEndMPSubprogramStmt(  &
        st%last%tlist,  &
        separate_subprog%end_name,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list)
    
    IF (ALLOCATED(st%last%statement_label)) THEN
      separate_subprog%end_label = st%last%statement_label
    END IF
    
  END SUBROUTINE ParseSeparateModuleSubprogram
  
  
  !*****************************************************************************
  !!
  !> Implementation of SeparateModuleProcedure%Dump - write a debugging 
  !! representation of the results of parsing the separate module procedure 
  !! to a logical unit.
  !!
  !! @param[in]     pnode             The FunctionSubprogram to dump.
  !!
  !! @param[in]     unit              Logical unit connected for 
  !! formatted output.
  !!
  !! @param[in]     pfx               Prefix for each record.
  
  ! As separate module procedures cannot contain separate module procedures, 
  ! this doesn't need to be recursive.
  SUBROUTINE sms_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(SeparateModuleSubprogram), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    WRITE (unit, "(A,'/',A,'/')") pfx, pnode%mps_stmt%What()
    CALL pnode%mps_stmt%Dump(unit, pfx // '  ')
    
    WRITE (unit, "(A,'/specification-part/')") pfx
    CALL pnode%spec%Dump(unit, pfx // '  ')
    
    WRITE (unit, "(A,'/',A,'/')") pfx, pnode%exec%What()
    CALL pnode%exec%Dump(unit, pfx // '  ')
    
    IF (SIZE(pnode%internals) > 0) THEN
      WRITE (unit, "(A,'internal subprograms:')") pfx
      CALL dump_internals(pnode%internals, unit, pfx // '  ')
    END IF
    
    IF (ALLOCATED(pnode%end_name)) THEN
      WRITE (unit, "(A,'/end-mp-subprogram-stmt/')") pfx
      WRITE (unit, "(A,'/procedure-name/ = ',A)")  &
          pfx // '  ', .NameOf. pnode%end_name
    ELSE
      WRITE (unit, "(A,'/end-mp-subprogram-stmt/')") pfx
    END IF
    
  END SUBROUTINE sms_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of SeparateModuleSubprogram%Query - traverse the node 
  !! with a query operation.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  RECURSIVE SUBROUTINE sms_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(SeparateModuleSubprogram), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%mps_stmt%Query(visitor)
    
    CALL QueryExecProgramUnit(pnode, visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE sms_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of SeparateModuleSubprogram%Modify - traverse the node 
  !! with a modify operation.
  !!
  !! @param[in,out] pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  RECURSIVE SUBROUTINE sms_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(SeparateModuleSubprogram), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%mps_stmt%Modify(visitor)
    
    CALL ModifyExecProgramUnit(pnode, visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE sms_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of SeparateModuleProcedure%What - return a descriptive 
  !! string for the parse node.
  !!
  !! @returns 'separate-module-subprogram'
  
  FUNCTION sms_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'separate-module-subprogram'
    
  END FUNCTION sms_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of SeparateModuleSubprogram%Who - return the identifier 
  !! text for the parse node.
  !!
  !! @param[in]     pnode             The parse node to query.
  !!
  !! @returns The name of the module.
  
  FUNCTION sms_Who(pnode) RESULT(name)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(SeparateModuleSubprogram), INTENT(IN) :: pnode
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: name
    
    !***************************************************************************
    
    name = .NameOf. pnode%mps_stmt%name
    
  END FUNCTION sms_Who
  
  
  !*****************************************************************************
  !!
  !> Implementation of SeparateModuleSubprogram%Advertise - advertise the 
  !! identifiers referenced or defined by the separate subprogram from the 
  !! perspective of the host.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference, from the 
  !! perspective of the host.
  
  SUBROUTINE sms_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(SeparateModuleSubprogram), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(1))
    ident_list(1) = IdentifierNature(  &
        NAME=.NameOf. pnode%mps_stmt%name,  &
        NATURES=[NatureRefSeparate],  &
        LOCATIONS=[.Locationof. pnode%mps_stmt%name] )
    
    ! @todo Check for entry statements??
    
  END SUBROUTINE sms_Advertise
  
  
  !*****************************************************************************
  !!
  !> Implementation of SeparateModuleSubprogram%Check - check the parse 
  !! node for conformance with language standards.
  !!
  !! @param[in]     pnode             The parse node to check.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE sms_Check(pnode, err_list)
    
    USE Errors
    USE ErrorCodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(SeparateModuleSubprogram), INTENT(IN) :: pnode
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! Error list for child calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (ALLOCATED(pnode%end_name)) THEN
      IF ( .NameOf. pnode%mps_stmt%name  &
          /= .NameOf. pnode%end_name ) THEN
        CALL Add( err_list, CODE=errNameMismatch,  &
            COMPONENT='C1263 on R1239',  & 
            LOCATION=.LocationOf. pnode%end_name,  &
            MSG='The name in the /end-mp-subprogram-stmt/ ("'  &
              // .NameOf. pnode%end_name  &
              // '") must match the name given in the &
              &/mp-subprogram-stmt/ ("'  &
              // .NameOf. pnode%mps_stmt%name // '").' )
      END IF
    END IF
    
    CALL pu_Check(pnode, sub_err_list)
    CALL Add(err_list, sub_err_list)
    
  END SUBROUTINE sms_Check
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for Module.
  
  
  !*****************************************************************************
  !!
  !> Parse a statement tree node that corresponds to a module.
  !!
  !! @param[in]     st                The statement tree node for the module.
  !!
  !! @param[in]     mod_unit          The results of parsing the module.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseModule(st, mod_unit, err_list)
    
    USE Errors
    USE ModuleStmts
    USE ModuleStNodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(ModuleStNode), INTENT(INOUT) :: st
    TYPE(Module), INTENT(OUT) :: mod_unit
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    CALL ParseModuleStmt(st%first%tlist, mod_unit%mod_stmt, sub_err_list)
    CALL Add(err_list, sub_err_list)
    ! Continue on error.
    
    IF (ALLOCATED(st%spec)) THEN
      CALL ParseSpecificationPart(st%spec, mod_unit%spec, sub_err_list)
      CALL Add(err_list, sub_err_list)
    END IF
    
    IF (ALLOCATED(st%mod_procs)) THEN
      CALL parse_internal_subprograms(  &
          st%mod_procs%InternalSubprogramPartStNode,  &
          ispntModule,  &
          mod_unit%internals,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list)
    ELSE
      ALLOCATE(mod_unit%internals(0))
    END IF
    
    CALL ParseEndModuleStmt(  &
        st%last%tlist,  &
        mod_unit%end_name,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list)
    
  END SUBROUTINE ParseModule
  
  
  !*****************************************************************************
  !!
  !> Implementation of Module%Dump - write a debugging representation of 
  !! the results of parsing the module to a logical unit.
  !!
  !! @param[in]     pnode             The Module to dump.
  !!
  !! @param[in]     unit              Logical unit connected for 
  !! formatted output.
  !!
  !! @param[in]     pfx               Prefix for each record.
  
  SUBROUTINE mod_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(Module), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    WRITE (unit, "(A,'/',A,'/')") pfx, pnode%mod_stmt%What()
    CALL pnode%mod_stmt%Dump(unit, pfx // '  ')
    
    WRITE (unit, "(A,'/specification-part/')") pfx
    CALL pnode%spec%Dump(unit, pfx // '  ')
    
    IF (SIZE(pnode%internals) > 0) THEN
      WRITE (unit, "(A,'/module-subprogram-part/')") pfx
      CALL dump_internals(pnode%internals, unit, pfx // '  ')
    END IF
    
    IF (ALLOCATED(pnode%end_name)) THEN
      WRITE (unit, "(A,'/end-module-stmt/')") pfx
      WRITE (unit, "(A,'/module-name/ = ',A)")  &
          pfx // '  ', .NameOf. pnode%end_name
    ELSE
      WRITE (unit, "(A,'/end-module-stmt/')") pfx
    END IF
    
  END SUBROUTINE mod_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of Module%Query.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  RECURSIVE SUBROUTINE mod_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(Module), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Child node index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%mod_stmt%Query(visitor)
    
    CALL pnode%spec%Query(visitor)
    
    DO i = 1, SIZE(pnode%internals)
      CALL pnode%internals(i)%Query(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE mod_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of Module%Modify.
  !!
  !! @param[in,out] pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  RECURSIVE SUBROUTINE mod_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(Module), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Child node index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%mod_stmt%Modify(visitor)
    
    CALL pnode%spec%Modify(visitor)
    
    DO i = 1, SIZE(pnode%internals)
      CALL pnode%internals(i)%Modify(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE mod_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of Module%What - return a descriptive string for the 
  !! parse node.
  !!
  !! @returns 'module'
  
  FUNCTION mod_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'module'
    
  END FUNCTION mod_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of Module%Who - return the identifier text 
  !! for the parse node.
  !!
  !! @param[in]     pnode             The parse node to query.
  !!
  !! @returns The name of the module.
  
  FUNCTION mod_Who(pnode) RESULT(name)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(Module), INTENT(IN) :: pnode
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: name
    
    !***************************************************************************
    
    name = .NameOf. pnode%mod_stmt%module_name
    
  END FUNCTION mod_Who
  
  
  !*****************************************************************************
  !!
  !> Implementation of Module%Advertise - advertise the identifiers 
  !! referenced or defined by the module from the perspective of the 
  !! host.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference, from the 
  !! perspective of the host.
  
  SUBROUTINE mod_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(Module), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(1))
    ident_list(1) = IdentifierNature(  &
        .NameOf. pnode%mod_stmt%module_name,  &
        [NatureDefModule],  &
        [.Locationof. pnode%mod_stmt%module_name] )
    
  END SUBROUTINE mod_Advertise
  
  
  !*****************************************************************************
  !!
  !> Implementation of Module%Check - check the parse 
  !! node for conformance with language standards.
  !!
  !! @param[in]     pnode             The parse node to check.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE mod_Check(pnode, err_list)
    
    USE Errors
    USE ErrorCodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(Module), INTENT(IN) :: pnode
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! Error list for child calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (ALLOCATED(pnode%end_name)) THEN
      IF ( .NameOf. pnode%mod_stmt%module_name  &
          /= .NameOf. pnode%end_name ) THEN
        CALL Add( err_list, CODE=errNameMismatch,  &
            COMPONENT='C1102 on R1106',  & 
            LOCATION=.LocationOf. pnode%end_name,  &
            MSG='The name in the /end-module-stmt/ ("'  &
              // .NameOf. pnode%end_name  &
              // '") must match the name given in the &
              &/module-stmt/ ("'  &
              // .NameOf. pnode%mod_stmt%module_name // '").' )
      END IF
    END IF
    
    CALL pu_Check(pnode, sub_err_list)
    CALL Add(err_list, sub_err_list)
    
  END SUBROUTINE mod_Check
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for Submodule.
  
  
  !*****************************************************************************
  !!
  !> Parse a statement tree node that corresponds to a submodule.
  !!
  !! @param[in]     st                The statement tree node for the 
  !! submodule.
  !!
  !! @param[in]     submod_unit       The results of parsing the module.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseSubmodule(st, submod_unit, err_list)
    
    USE Errors
    USE SubmoduleStmts
    USE ModuleStNodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(SubmoduleStNode), INTENT(INOUT) :: st
    TYPE(Submodule), INTENT(OUT) :: submod_unit
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    CALL ParseSubmoduleStmt(  &
        st%first%tlist,  &
        submod_unit%submod_stmt,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list)
    ! Continue on error.
    
    IF (ALLOCATED(st%spec)) THEN
      CALL ParseSpecificationPart(st%spec, submod_unit%spec, sub_err_list)
      CALL Add(err_list, sub_err_list)
    END IF
    
    IF (ALLOCATED(st%mod_procs)) THEN
      CALL parse_internal_subprograms(  &
          st%mod_procs%InternalSubprogramPartStNode,  &
          ispntModule,  &
          submod_unit%internals,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list)
    ELSE
      ALLOCATE(submod_unit%internals(0))
    END IF
    
    CALL ParseEndSubmoduleStmt(  &
        st%last%tlist,  &
        submod_unit%end_name,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list)
    
  END SUBROUTINE ParseSubmodule
  
  
  !*****************************************************************************
  !!
  !> Implementation of Submodule%Dump - write a debugging representation of 
  !! the results of parsing the submodule to a logical unit.
  !!
  !! @param[in]     pnode             The Submodule to dump.
  !!
  !! @param[in]     unit              Logical unit connected for 
  !! formatted output.
  !!
  !! @param[in]     pfx               Prefix for each record.
  
  SUBROUTINE sm_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(Submodule), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    WRITE (unit, "(A,'/',A,'/')") pfx, pnode%submod_stmt%What()
    CALL pnode%submod_stmt%Dump(unit, pfx // '  ')
    
    WRITE (unit, "(A,'/specification-part/')") pfx
    CALL pnode%spec%Dump(unit, pfx // '  ')
    
    IF (SIZE(pnode%internals) > 0) THEN
      WRITE (unit, "(A,'/module-subprogram-part/')") pfx
      CALL dump_internals(pnode%internals, unit, pfx // '  ')
    END IF
    
    IF (ALLOCATED(pnode%end_name)) THEN
      WRITE (unit, "(A,'/end-submodule-stmt/')") pfx
      WRITE (unit, "(A,'/submodule-name/ = ',A)")  &
          pfx // '  ', .NameOf. pnode%end_name
    ELSE
      WRITE (unit, "(A,'/end-submodule-stmt/')") pfx
    END IF
    
  END SUBROUTINE sm_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of Submodule%Query.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  RECURSIVE SUBROUTINE sm_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(Submodule), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Child node index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%submod_stmt%Query(visitor)
    
    CALL pnode%spec%Query(visitor)
    
    DO i = 1, SIZE(pnode%internals)
      CALL pnode%internals(i)%Query(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE sm_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of Submodule%Modify.
  !!
  !! @param[in,out] pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  RECURSIVE SUBROUTINE sm_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(Submodule), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Child node index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%submod_stmt%Modify(visitor)
    
    CALL pnode%spec%Modify(visitor)
    
    DO i = 1, SIZE(pnode%internals)
      CALL pnode%internals(i)%Modify(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE sm_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of Submodule%What - return a descriptive string for the 
  !! parse node.
  !!
  !! @returns 'submodule'
  
  FUNCTION sm_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'submodule'
    
  END FUNCTION sm_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of Submodule%Who - return the identifier text 
  !! for the parse node.
  !!
  !! @param[in]     pnode             The parse node to query.
  !!
  !! @returns The identifier of the submodule.
  
  FUNCTION sm_Who(pnode) RESULT(name)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(Submodule), INTENT(IN) :: pnode
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: name
    
    !***************************************************************************
    
    name = .NameOf. pnode%submod_stmt%ancestor_module_name  &
        // ':' // .NameOf. pnode%submod_stmt%submodule_name
    
  END FUNCTION sm_Who
  
  
  !*****************************************************************************
  !!
  !> Implementation of Module%Advertise - advertise the identifiers 
  !! referenced or defined by the submodule.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference, from the 
  !! perspective of the host.
  
  SUBROUTINE sm_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(Submodule), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ! What is advertised from the submodule subprogram is the same 
    ! as what is advertised from the submodule statement into the 
    ! subprogram.
    CALL pnode%submod_stmt%Advertise(ident_list)
    
  END SUBROUTINE sm_Advertise
  
  
  !*****************************************************************************
  !!
  !> Get the ancestor module name for a submodule.
  !!
  !! @param[in]     pnode             The parse node for the submodule.
  !!
  !! @returns The ancestor module name for a submodule.
  
  FUNCTION sm_QueryAncestorModuleName(pnode) RESULT(name)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(Submodule), INTENT(IN) :: pnode
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: name
    
    !***************************************************************************
    
    name = .NameOf. pnode%submod_stmt%ancestor_module_name
    
  END FUNCTION sm_QueryAncestorModuleName
  
  
  !*****************************************************************************
  !!
  !> Get the parent submodule name for a submodule.
  !!
  !! @param[in]     pnode             The parse node for the submodule.
  !!
  !! @returns The parent submodule name for a submodule.  If the submodule 
  !! has no parent submodule, then a zero sized string is returned.
  
  FUNCTION sm_QueryParentSubmoduleName(pnode) RESULT(name)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(Submodule), INTENT(IN) :: pnode
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: name
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%submod_stmt%parent_submodule_name)) THEN
      name = .NameOf. pnode%submod_stmt%parent_submodule_name
    ELSE
      name = ''
    END IF
    
  END FUNCTION sm_QueryParentSubmoduleName
  
  
  !*****************************************************************************
  !!
  !> Get the parent submodule name for a submodule.
  !!
  !! @param[in]     pnode             The parse node for the submodule.
  !!
  !! @returns The parent submodule name for a submodule.  If the submodule 
  !! has no parent submodule, then a zero sized string is returned.
  
  FUNCTION sm_QueryParentIdentifier(pnode) RESULT(name)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(Submodule), INTENT(IN) :: pnode
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: name
    
    !***************************************************************************
    
    name = .NameOf. pnode%submod_stmt%ancestor_module_name
    IF (ALLOCATED(pnode%submod_stmt%parent_submodule_name)) THEN
      name = name // ':' // .NameOf. pnode%submod_stmt%parent_submodule_name
    END IF
    
  END FUNCTION sm_QueryParentIdentifier
  
  
  !*****************************************************************************
  !!
  !> Implementation of Submodule%Check - check the parse 
  !! node for conformance with language standards.
  !!
  !! @param[in]     pnode             The parse node to check.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE sm_Check(pnode, err_list)
    
    USE Errors
    USE ErrorCodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(Submodule), INTENT(IN) :: pnode
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! Error list for child calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (ALLOCATED(pnode%end_name)) THEN
      IF ( .NameOf. pnode%submod_stmt%submodule_name  &
          /= .NameOf. pnode%end_name ) THEN
        CALL Add( err_list, CODE=errNameMismatch,  &
            COMPONENT='C1114 on R1119',  & 
            LOCATION=.LocationOf. pnode%end_name,  &
            MSG='The name in the /end-submodule-stmt/ ("'  &
              // .NameOf. pnode%end_name  &
              // '") must match the name given in the &
              &/submodule-stmt/ ("'  &
              // .NameOf. pnode%submod_stmt%submodule_name // '").' )
      END IF
    END IF
    
    CALL pu_Check(pnode, sub_err_list)
    CALL Add(err_list, sub_err_list)
    
  END SUBROUTINE sm_Check
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for block data.
  
  
  !*****************************************************************************
  !!
  !> Parse a statement tree node that corresponds to a /block-data/ unit.
  !!
  !! @param[in]     st                The statement tree node for the block 
  !! data unit.
  !!
  !! @param[in]     block_data        The results of parsing the /block-data/.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseBlockData(st, block_data, err_list)
    
    USE Errors
    USE BlockDataStmts
    USE BlockDataStNodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(BlockDataStNode), INTENT(INOUT) :: st
    TYPE(BlockData), INTENT(OUT) :: block_data
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    CALL ParseBlockDataStmt(st%first%tlist, block_data%bd_stmt, sub_err_list)
    CALL Add(err_list, sub_err_list)
    
    IF (ALLOCATED(st%spec)) THEN
      CALL ParseSpecificationPart(st%spec, block_data%spec, sub_err_list)
      CALL Add(err_list, sub_err_list)
    END IF
    
    CALL ParseEndBlockDataStmt(  &
        st%last%tlist,  &
        block_data%end_name,  &
        sub_err_list)
    CALL Add(err_list, sub_err_list)
    
  END SUBROUTINE ParseBlockData
  
  
  !*****************************************************************************
  !!
  !> Implementation of BlockData%Dump - write a debugging representation 
  !! of the results of parsing the block data unit to a logical unit.
  !!
  !! @param[in]     pnode             The BlockData to dump.
  !!
  !! @param[in]     unit              Logical unit connected for 
  !! formatted output.
  !!
  !! @param[in]     pfx               Prefix for each record.
  
  SUBROUTINE bd_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(BlockData), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    WRITE (unit, "(A,'/',A,'/')") pfx, pnode%bd_stmt%What()
    CALL pnode%bd_stmt%Dump(unit, pfx // '  ')
    
    WRITE (unit, "(A,'/specification-part/')") pfx
    CALL pnode%spec%Dump(unit, pfx // '  ')
    
    IF (ALLOCATED(pnode%end_name)) THEN
      WRITE (unit, "(A,'/end-block-data-stmt/')") pfx
      WRITE (unit, "(A,'/block-data-name/ = ',A)")  &
          pfx // '  ', .NameOf. pnode%end_name
    ELSE
      WRITE (unit, "(A,'/end-block-data-stmt/')") pfx
    END IF
    
  END SUBROUTINE bd_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of BlockData%What - return a descriptive string for the 
  !! parse node.
  !!
  !! @returns 'block-data'
  
  FUNCTION bd_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'block-data'
    
  END FUNCTION bd_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of BlockData%Query.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  RECURSIVE SUBROUTINE bd_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(BlockData), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%bd_stmt%Query(visitor)
    
    CALL pnode%spec%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE bd_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of BlockData%Modify.
  !!
  !! @param[in,out] pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  RECURSIVE SUBROUTINE bd_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(BlockData), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%bd_stmt%Modify(visitor)
    
    CALL pnode%spec%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE bd_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of BlockData%Who - return the identifier text 
  !! for the parse node.
  !!
  !! @param[in]     pnode             The parse node to query.
  !!
  !! @returns The name of the block data, or an internal substitute if the 
  !! block data has no name.
  
  FUNCTION bd_Who(pnode) RESULT(name)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(BlockData), INTENT(IN) :: pnode
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: name
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%bd_stmt%block_data_name)) THEN
      !=========================================================================
      ! ifort 19.0 bug - see ticket #48
      !name = .NameOf. pnode%bd_stmt%block_data_name
      !=========================================================================
      name = name_bug(pnode%bd_stmt%block_data_name)
      !=========================================================================
    ELSE
      name = 'unnamed block data'
    END IF
    
  END FUNCTION bd_Who
  
  
  !=============================================================================
  ! ifort 19.0 bug - see ticket #48.
  !=============================================================================
  function name_bug(ident)
    type(identifier), intent(in) :: ident
    character(:), allocatable :: name_bug
    
    name_bug = .NameOf. ident
  end function name_bug
  !=============================================================================
  
  
  !=============================================================================
  ! ifort 19.0 bug - see ticket #48.
  !=============================================================================
  function location_bug(ident)
    use SourceLocations
    
    type(identifier), intent(in) :: ident
    type(sourcelocation), allocatable :: location_bug
    
    location_bug = .LocationOf. ident
  end function location_bug
  !=============================================================================
  
  
  !*****************************************************************************
  !!
  !> Implementation of MainProgram%Advertise - advertise the identifiers 
  !! referenced or defined by the block data unit from the perspective 
  !! of the host.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference, from the 
  !! perspective of the host.
  
  SUBROUTINE bd_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(BlockData), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%bd_stmt%block_data_name)) THEN
      ALLOCATE(ident_list(1))
      ident_list = IdentifierNature(  &
          !=====================================================================
          ! ifort 19.0 bug - see ticket #48.
!          .NameOf. pnode%bd_stmt%block_data_name,  &
          !=====================================================================
          name_bug(pnode%bd_stmt%block_data_name),  &
          !=====================================================================
          [NatureDefBlockData],  &
          !=====================================================================
          ! ifort 19.0 bug - see ticket #48.
          ![.Locationof. pnode%bd_stmt%block_data_name] )
          !=====================================================================
          [location_bug(pnode%bd_stmt%block_data_name)] )
          !=====================================================================
    ELSE
      ALLOCATE(ident_list(0))
    END IF
    
  END SUBROUTINE bd_Advertise
  
  
  !*****************************************************************************
  !!
  !> Implementation of BlockData%Check - check the parse 
  !! node for conformance with language standards.
  !!
  !! @param[in]     pnode             The parse node to check.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE bd_Check(pnode, err_list)
    
    USE Errors
    USE ErrorCodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(BlockData), INTENT(IN) :: pnode
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! Error list for child calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (ALLOCATED(pnode%end_name)) THEN
      IF (ALLOCATED(pnode%bd_stmt%block_data_name)) THEN
        !=======================================================================
        ! ifort 19.0 bug - see ticket #48.
        !IF ( .NameOf. pnode%bd_stmt%block_data_name  &
        !=======================================================================
        IF (name_bug(pnode%bd_stmt%block_data_name)  &
        !=======================================================================
            /= .NameOf. pnode%end_name ) THEN
          CALL Add( err_list, CODE=errNameMismatch,  &
              COMPONENT='C1115 on R1122',  & 
              LOCATION=.LocationOf. pnode%end_name,  &
              MSG='The name in the /end-block-data-stmt/ ("'  &
                // .NameOf. pnode%end_name  &
                // '") must match the name given in the /block-data-stmt/ ("'  &
                !===============================================================
                ! ifort 19.0 bug - see ticket #48.
                !// .NameOf. pnode%bd_stmt%block_data_name // '").' )
                !===============================================================
                // name_bug(pnode%bd_stmt%block_data_name) // '").' )
                !===============================================================
        END IF
      ELSE
        CALL Add( err_list,  &
            CODE=errNameMismatch,  &
            COMPONENT='C1115 on R1122',  &
            LOCATION=.LocationOf. pnode%end_name,  &
            MSG='A name in an /end-block-data-stmt/ may only be specified if &
              &a /block-data-name/ was provided in the corresponding &
              &/block-data-stmt/.' )
      END IF
    END IF
    
    CALL pu_Check(pnode, sub_err_list)
    CALL Add(err_list, sub_err_list)
    
  END SUBROUTINE bd_Check
  
 
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Common procedures for the various program units and ProgramUnitElement.
  
  
  !*****************************************************************************
  !!
  !> Implementation of ProgramUnitElement%QueryDump - forward the Dump call 
  !! to the contained program unit.
  !!
  !! @param[in]     pnode             The parse node element being queried.
  !!
  !! @param[in]     unit              Unit connected for formatted output.
  !!
  !! @param[in]     pfx               Prefix to write before each record.
  
  SUBROUTINE pue_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ProgramUnitElement), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    CALL pnode%item%Dump(unit, pfx)
    
  END SUBROUTINE pue_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of ProgramUnitElement%Query - forward the query call to 
  !! the contained program unit.
  !!
  !! @param[in]     pnode             The parse node element being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  SUBROUTINE pue_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ProgramUnitElement), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL pnode%item%Query(visitor)
    
  END SUBROUTINE pue_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of ProgramUnitElement%Modify - forward the modify call to 
  !! the contained program unit.
  !!
  !! @param[in,out] pnode             The parse node element being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  SUBROUTINE pue_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ProgramUnitElement), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL pnode%item%Modify(visitor)
    
  END SUBROUTINE pue_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of ProgramUnitElement%What - forward the What call to 
  !! the contained program unit.
  !!
  !! @param[in]     pnode             The parse node element being queried.
  !!
  !! @returns A descriptive string for the parse node.
  
  FUNCTION pue_What(pnode) RESULT(s)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ProgramUnitElement), INTENT(IN) :: pnode
    
    CHARACTER(:), ALLOCATABLE :: s
    
    !***************************************************************************
    
    s = pnode%item%What()
    
  END FUNCTION pue_What
  
  
  !*****************************************************************************
  !!
  !> Parse the internal or module subprograms.
  !!
  !! @param[in]     st                The tree node for the internal or 
  !! module procedure part.
  !!
  !! @param[in]     nature            The nature of the internal or 
  !! module subprograms.  This should either be ispntModule or ispntInternal.
  !!
  !! @param[out]    sub_progs         The subprograms.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! This also parses the /contains-stmt/.
  !!
  !! We don't enforce uniqueness of names - that is left to the Advertise 
  !! pass when attempts to add the procedure name to the entity table 
  !! will cause an error due to the name clash.
  
  SUBROUTINE parse_internal_subprograms(st, nature, sub_progs, err_list)
    
    USE ContainsStmts
    USE Errors
    USE ProgUnitStNodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(InternalSubprogramPartStNode), INTENT(INOUT) :: st
    INTEGER, INTENT(IN) :: nature
    TYPE(ProgramUnitElement), INTENT(OUT), ALLOCATABLE :: sub_progs(:)
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables.
    
    INTEGER :: i              ! Subprogram index.
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    ! Object for parsing an internal subroutine subprogram.
    TYPE(SubroutineSubprogram), ALLOCATABLE :: subroutine_subprogram
    
    ! Object for parsing an internal function subprogram.
    TYPE(FunctionSubprogram), ALLOCATABLE :: function_subprogram
    
    ! Object for parsing a separate module subprogram.
    TYPE(SeparateModuleSubprogram), ALLOCATABLE :: separate_module_subprogram
    
    ! Temporary for both types of program unit.
    CLASS(ProgramUnit), ALLOCATABLE :: item
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    CALL ParseContainsStmt(st%contains%tlist, sub_err_list)
    CALL Add(err_list, sub_err_list)
    ! Continue on error.
    
    ALLOCATE(sub_progs(0))
    DO i = 1, SIZE(st%subprograms)
      ! Pattern is to allocate a type specific to the subprogram, parse the 
      ! subprogram into that type, if sucessful move that type into a 
      ! generic ProgramUnit item (otherwise deallocate it on failure) then 
      ! at the end add that program unit item to the array of internal 
      ! subprograms.
      SELECT TYPE (sp => st%subprograms(i)%item)
      TYPE IS (SubroutineSubprogramStNode)
        ALLOCATE(subroutine_subprogram)
        CALL ParseSubroutineSubprogram(  &
            sp,  &
            nature,  &
            subroutine_subprogram,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list)
        IF (Failure(err_list)) THEN
          DEALLOCATE(subroutine_subprogram)
        ELSE
          CALL MOVE_ALLOC(subroutine_subprogram, item)
        END IF
        
      TYPE IS (FunctionSubprogramStNode)
        ALLOCATE(function_subprogram)
        CALL ParseFunctionSubprogram(  &
            sp,  &
            nature,  &
            function_subprogram,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list)
        IF (Failure(err_list)) THEN
          DEALLOCATE(function_subprogram)
        ELSE
          CALL MOVE_ALLOC(function_subprogram, item)
        END IF
        
      TYPE IS (SeparateModuleSubprogramStNode)
        ALLOCATE(separate_module_subprogram)
        CALL ParseSeparateModuleSubprogram(  &
            sp,  &
            separate_module_subprogram,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list)
        IF (Failure(err_list)) THEN
          DEALLOCATE(separate_module_subprogram)
        ELSE
          CALL MOVE_ALLOC(separate_module_subprogram, item)
        END IF
        
      CLASS DEFAULT
        STOP 'Internal error in ProgramUnits%%parse_internal_subprograms, &
            &unknown internal subprogram type.'
        
      END SELECT
      
      IF (ALLOCATED(item)) THEN
        CALL item%Init(st%subprograms(i)%item)
        CALL AddProgramUnit(sub_progs, item)
      END IF
    END DO
    
  END SUBROUTINE parse_internal_subprograms
  
  
  !*****************************************************************************
  !!
  !! Can also be used for module procedures.
  !!
  !! @param[in]     internals         The vector of internal or module 
  !! subprograms to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix for each record.
  
  SUBROUTINE dump_internals(internals, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(ProgramUnitElement), INTENT(IN) :: internals(:)
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: ip             ! Subprogram index.
    
    !***************************************************************************
    
    DO ip = 1, SIZE(internals)
      WRITE (unit, "(A,'/',A,'/')") pfx, internals(ip)%item%What()
      CALL internals(ip)%item%Dump(unit, pfx // '  ')
    END DO
    
  END SUBROUTINE dump_internals
  
  
  !*****************************************************************************
  !!
  !> Add a program unit to a list of program units by moving it.
  !!
  !! @param[in]     units             The list of program units to be added 
  !! to.  Must be allocated on input.
  !!
  !! @param[in]     item              The item to be added.  This will be 
  !! unallocated after this call.
  
  SUBROUTINE AddProgramUnit(units, item)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(ProgramUnitElement), INTENT(INOUT), ALLOCATABLE :: units(:)
    CLASS(ProgramUnit), INTENT(INOUT), ALLOCATABLE :: item
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Temporary for growing array.
    TYPE(ProgramUnitElement), ALLOCATABLE :: tmp(:)
    
    INTEGER :: i              ! Array index.
    
    !***************************************************************************
    
    ALLOCATE(tmp(LBOUND(units,1):UBOUND(units,1)+1))
    DO i = LBOUND(units, 1), UBOUND(units, 1)
      CALL MOVE_ALLOC(units(i)%item, tmp(i)%item)
    END DO
    
    CALL MOVE_ALLOC(item, tmp(UBOUND(tmp,1))%item)
    CALL MOVE_ALLOC(tmp, units)
    
  END SUBROUTINE AddProgramUnit
  
END MODULE ProgramUnits
