! $Id: SelectTypeStmts.f90 2888 2019-04-06 22:55:48Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the SelectTypeStmts module.


!*******************************************************************************
!!
!> Procedures for parsing /select-type-stmt/'s, /type-guard-stmt/'s and 
!! /end-select-type-stmt/'s.
!!
!! There is syntactic similarity between the /association/ in an 
!! /associate-stmt/ and the bit that appears in the parentheses of a 
!! /select-type-stmt/ (in fact, the /selector/ syntax part is the same), 
!! but because the /associate-name/ is optional in the /select-type-stmt/ 
!! variant we parse things using separate procedures.

MODULE SelectTypeStmts
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseSelectTypeStmt
  PUBLIC :: ParseEndSelectTypeStmt
  PUBLIC :: ParseTypeGuardStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /select-type-stmt/ (R847).
  !!
  !! @param[in]     tlist             The sequence of tokens that make up the 
  !! /select-type-stmt/.
  !!
  !! @param[out]    associate_name    The /associate-name/, if present 
  !! (not allocated if no associate name was present in the statement).
  !!
  !! @param[out]    selector          The /selector/ (expression or variable).
  !!
  !! @param[out]    construct_name    The construct name, if provided.  Not 
  !! allocated if the statement had no construct name.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseSelectTypeStmt( tlist, associate_name, selector,  &
      construct_name, err_list )
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE Expressions
    USE Identifiers
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(Identifier), INTENT(OUT), ALLOCATABLE :: associate_name
    CLASS(Expression), INTENT(OUT), ALLOCATABLE :: selector
    TYPE(Identifier), INTENT(OUT), ALLOCATABLE :: construct_name
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /select-type-stmt/ for error reporting.
    CHARACTER(*), PARAMETER :: comp = 'R847'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: it             ! Current token index.
    INTEGER :: closing        ! Closing delimited index.
    INTEGER :: associate      ! Index of the associate "=>") token
    
    ! Error list for child Parse calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /select-type-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! Check for construct name
    
    IF (Match(tlist, 1, ittName) .AND. Match(tlist, 2, ittColon)) THEN
      it = 3
      !=========================================================================
      ! ifort bug - assignment of dt with polymorphic component.
      !construct_name = IdentifierIfortBug(tlist(1))
      ALLOCATE(construct_name, SOURCE=IdentifierIfortBug(tlist(1)))
      !=========================================================================
    ELSE
      it = 1
      ! Leave name unallocated.
    END IF
    
    !---------------------------------------------------------------------------
    ! SELECT TYPE.. or SELECTTYPE... - either form allowed by 3.3.2.2.
    
    CALL ExpectSplittableKeyword(  &
        tlist,  &
        (it),  &
        scck_'SELECT',  &
        scck_'TYPE',  &
        it, err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! SELECT TYPE (...)
    
    CALL Expect(tlist, it, ittDelimiter, scck_'(', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! Find closing parenthesis.
    CALL ExpectClosing(tlist, it, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! Set it to the start of the bit in parentheses
    it = it + 1
    
    !---------------------------------------------------------------------------
    ! The bit inside the parentheses.
    
    ! Do we have an associate name?
    associate = FindBare(tlist, ittAssociate, scck_'=>', it)
    IF (associate /= 0) THEN
      ! We appear to have an associate name.
      CALL ExpectName(tlist, it, '/associate-name/', err)
      CALL Add(err_list, err, COMPONENT=comp)
      ! Carry on (bar getting associate name value) in the event of error.
      IF (.NOT. Failure(err)) THEN
        !=======================================================================
        ! ifort bug - assignment of derived type with polymorphic component.
        !associate_name = IdentifierIfortBug(tlist(it))
        ALLOCATE(associate_name, SOURCE=IdentifierIfortBug(tlist(it)))
        !=======================================================================
      END IF
      
      ! Make sure there's nothing extra after the name.
      CALL Expect(tlist, it + 1, ittAssociate, scck_'=>', err)
      CALL Add(err_list, err, COMPONENT=comp)
      ! Carry on in the event of error.
      
      it = associate + 1
    END IF
    ! it here is the start of the selector.
    
    ! Parse the selector as a general expression.
    CALL ParseExpr(tlist, it, closing - 1, selector, sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    
    ! Thats it!
    CALL ExpectEOS(tlist, closing + 1, comp, err_list)
    
  END SUBROUTINE ParseSelectTypeStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /end-select-type-stmt/ (R849).
  !!
  !! @param[in]     tlist             The sequence of tokens that make up the 
  !! /end-select-type-stmt/.
  !!
  !! @param[in]     construct_name    The /construct-name/ in the opening 
  !! /select-type-stmt/ if one was present, not allocated if a construct name 
  !! was not present.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! Essentially identical to SelectCaseStmts::ParseEndSelectStmt.
  
  SUBROUTINE ParseEndSelectTypeStmt(tlist, construct_name, err_list)
    
    USE CompilerKinds
    USE ConstructNames
    USE Errors
    USE ErrorCodes
    USE Identifiers
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(Identifier), INTENT(IN), ALLOCATABLE :: construct_name
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local components
    
    ! Syntax rule number for an /end-select-type-stmt/ for error reporting.
    CHARACTER(*), PARAMETER :: comp = 'R849'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: it             ! Token index.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /end-select-type-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    ! ENDSELECT or END SELECT
    CALL ExpectSplittableKeyword(tlist, 1, scck_'END', scck_'SELECT', it, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! Construct name
    
    CALL CheckConstructNameEnd(  &
        tlist,  &
        it,  &
        construct_name,  &
        '/select-type-stmt/',  &
        '/end-select-type-stmt/',  &
        'C843 on ' // comp,  &
        err_list )
    
  END SUBROUTINE ParseEndSelectTypeStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /type-guard-stmt/.
  !!
  !! @param[in]     tlist             The sequence of tokens that make up the 
  !! /case-stmt/.
  !!
  !! @param[in]     construct_name    The /construct-name/ in the opening 
  !! /select-type-stmt/ if one was present, not allocated if a construct name 
  !! was not present.
  !!
  !! @param[out]    polymorphic       Flag that indicates the type guard 
  !! was for a polymorphic object.
  !!
  !! @param[out]    type_spec         Type spec of a CLASS IS or TYPE IS 
  !! guard statement.  Not allocated if a CLASS DEFAULT style of statement.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseTypeGuardStmt( tlist, construct_name, polymorphic,  &
        type_spec, err_list )
    
    USE CompilerKinds
    USE ConstructNames
    USE DeclarationTypeSpecs
    USE Errors
    USE ErrorCodes
    USE Identifiers
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(Identifier), INTENT(IN), ALLOCATABLE :: construct_name
    LOGICAL, INTENT(OUT) :: polymorphic
    TYPE(TypeSpec), INTENT(OUT), ALLOCATABLE :: type_spec
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for error reporting.
    CHARACTER(*), PARAMETER :: comp = 'R848'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: it             ! Token index.
    INTEGER :: opt            ! Option for Expect calls.
    INTEGER :: closing        ! Index of closing parenthesis.
    
    ! Error list for child Parse calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /type-guard-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! TYPE or CLASS
    
    CALL ExpectKeyword(  &
        tlist,  &
        1,  &
        [scck_'TYPE ', scck_'CLASS'],  &
        '"TYPE" or "CLASS"',  &
        opt,  &
        err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    polymorphic = opt == 2
    
    ! What's permitted next depends on the leading keyword.  We use opt to 
    ! track use of CLASS DEFAULT.
    IF (polymorphic) THEN
      ! Can have CLASS IS or CLASS DEFAULT
      CALL ExpectKeyword(  &
          tlist,  &
          2,  &
          [scck_'IS     ', scck_'DEFAULT'],  &
          '"IS" or "DEFAULT"',  &
          opt,  &
          err )
    ELSE
      ! IS
      opt = 1
      CALL ExpectKeyword(tlist, 2, scck_'IS', err)
    END IF
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! ( /type-spec/ )      (if opt == 1)
    
    IF (opt == 1) THEN
      CALL Expect(tlist, 3, ittDelimiter, scck_'(', err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      CALL ExpectClosing(tlist, 3, closing, err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      ALLOCATE(type_spec)
      CALL ParseTypeSpec(tlist(4:closing-1), type_spec, sub_err_list)
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
      it = closing + 1
    ELSE
      it = 3
    END IF
    
    !---------------------------------------------------------------------------
    ! Optional construct name.
    
    CALL CheckConstructNameIntermediate(  &
        tlist,  &
        it,  &
        construct_name,  &
        '/select-type-stmt',  &
        '/type-guard-stmt/',  &
        'C843 on ' // comp,  &
        err_list )
    
  END SUBROUTINE ParseTypeGuardStmt
  
END MODULE SelectTypeStmts
