! $Id: SourceUnits.f90 1311 2014-09-09 03:59:39Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the SourceFiles module.


!*******************************************************************************
!!
!> Types and procedures for manipulating source units.
!!
!! A source unit is like a source file, except it is designated by a logical 
!! unit (which may be connected to a file) rather than by a filename.
!!
!! Source units might be handy for supplying source from the console or 
!! similar.

MODULE SourceUnits
  
  USE Sources
  
  IMPLICIT NONE
  
  PUBLIC :: ConnectSourceUnit
  
  !-----------------------------------------------------------------------------
  
  TYPE, PUBLIC, EXTENDS(Source) :: SourceUnit
    PRIVATE    ! Can't be constructed outside of this module.
    !> The unit that is the sauce of source.
    INTEGER :: unit
  CONTAINS
    PROCEDURE :: ReadLine => sourceunit_ReadLine
    PROCEDURE :: Reset => sourceunit_Reset
  END TYPE SourceUnit
  
  !-----------------------------------------------------------------------------
  
  !> Open a source unit.
  INTERFACE ConnectSourceUnit
    MODULE PROCEDURE ConnectSourceUnit_
  END INTERFACE ConnectSourceUnit
  
  !> Component for underlying Fortran processor reporting errors.
  CHARACTER(*), PARAMETER :: comp = 'Fortran runtime'
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Create a SourceUnit object.
  !!
  !! @param[in]     unit              The unit that the source unit object 
  !! is connected to.
  !!
  !! @param[out]    source_unit       The source unit object.
  !!
  !! @param[in]     name              An optional descriptive name for the unit.
  
  SUBROUTINE ConnectSourceUnit_(unit, source_unit, name)
    
    USE CompilerKinds
    USE Errors
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    INTEGER, INTENT(IN) :: unit
    TYPE(SourceUnit), INTENT(OUT), POINTER :: source_unit
    CHARACTER(*), INTENT(IN), OPTIONAL :: name
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! The new source object prior to addition to the global list.
    CLASS(Source), ALLOCATABLE :: new_source_file
    
    ! The name to use for the unit.
    CHARACTER(:,KIND=fnck), ALLOCATABLE :: local_name
    
    ! The new source object, after addition to the global list.
    CLASS(Source), POINTER :: list_ptr
    
    !***************************************************************************
    
    ! Set default for the name if not present.
    IF (PRESENT(name)) THEN
      local_name = name         ! Kind conversion.
    ELSE
      CALL get_unit_name(unit, local_name)
    END IF
    
    ! Create the source object and add it to the persistent list.
!    ALLOCATE(new_source_file, SOURCE=SourceUnit( UNIT=unit, NAME=local_name,  &
!        LINE=0, PARENT=local_parent ))
    ALLOCATE(SourceUnit :: new_source_file)
    CALL AddSource(new_source_file, list_ptr)
    
    SELECT TYPE (list_ptr)
    TYPE IS (SourceUnit)
      source_unit => list_ptr
      source_unit%name = local_name
      source_unit%unit = unit
    END SELECT
    
  END SUBROUTINE ConnectSourceUnit_
  
  
  !*****************************************************************************
  !!
  !> Implementation of SourceUnit%ReadLine.
  !!
  !! @param[in,out] src               The source unit object.
  !!
  !! @param[out]    line              The line of source.
  !!
  !! @param[out]    err               Error object.
  
  SUBROUTINE sourceunit_ReadLine(src, line, err)
    
    USE CompilerKinds
    USE Errors
    USE FileUtilities
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(SourceUnit), INTENT(INOUT) :: src
    CHARACTER(:,KIND=scck), INTENT(OUT), ALLOCATABLE :: line
    TYPE(Error), INTENT(OUT) :: err
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: stat           ! Error code.
    CHARACTER(256) :: msg      ! Error message.
    
    !***************************************************************************
    
    CALL GetLine(src%unit, line, stat, msg)
    IF (stat == 0) THEN
      src%line = src%line + 1
    ELSE IF (stat < 0) THEN
      DEALLOCATE(line)
    ELSE IF (stat /= 0) THEN
      err = ConstructError( CODE=stat, COMPONENT=comp,  &
          MSG='Error reading from ' // src%name // ': ' // TRIM(msg) )
    END IF
    
  END SUBROUTINE sourceunit_ReadLine
  
  
  !*****************************************************************************
  !!
  !> Implementation of SourceUnit%Reset - reset the source unit to its 
  !! initial state.
  !!
  !! @param[in,out] src               The source unit object.
  !!
  !! This is practically a no-op for this type.  All we do is reset the 
  !! line count.
  
  SUBROUTINE sourceunit_Reset(src)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(SourceUnit), INTENT(INOUT) :: src
    
    !***************************************************************************
    
    src%line = 0
    
  END SUBROUTINE sourceunit_Reset
  
  
  !*****************************************************************************
  !!
  !> Get an appropriate name for a unit.
  !!
  !! @param[in]     unit              The unit being queried.
  !!
  !! @param[out]    name              A name for the unit.
  
  SUBROUTINE get_unit_name(unit, name)
    
    USE CompilerKinds
    USE CharUtils
    
    USE, INTRINSIC :: ISO_FORTRAN_ENV, ONLY: INPUT_UNIT
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    INTEGER, INTENT(IN) :: unit
    CHARACTER(:,KIND=fnck), ALLOCATABLE :: name
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    LOGICAL :: named          ! Is the unit connected to a file with a name?
    
    !***************************************************************************
    
    IF (unit == INPUT_UNIT) THEN
      name = 'INPUT_UNIT'   ! Why not!
    ELSE
      INQUIRE(UNIT=unit, NAMED=named)
      IF (named) THEN
        ALLOCATE(CHARACTER(256) :: name)
        INQUIRE(UNIT=unit, NAME=name)
        name = TRIM(name)
      ELSE
        name = ToString(unit)
      END IF
    END IF
    
  END SUBROUTINE get_unit_name
  
END MODULE SourceUnits
