! $Id: SpecificationParts.f90 2908 2019-04-13 00:28:11Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the SpecificationParts module.


!*******************************************************************************
!!
!> Procedures and types for parsing specification parts.

MODULE SpecificationParts
  
  USE Identifiers
  USE ImplicitStmts
  USE ImportStmts
  USE ParseNodes
  USE Strings
  USE UseStmts
  
  IMPLICIT NONE
  
  PRIVATE
  
  PUBLIC :: ParseSpecificationPart
  
  !-----------------------------------------------------------------------------
  
  ! The results of parsing a specification part.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: SpecificationPart
    !> The results of parsing the use statements.
    !!
    !! Always allocated after ParseSpecificationPart, may be zero sized.
    TYPE(UseStmt), ALLOCATABLE :: use_stmts(:)
    !> The results of parsing the import statements.
    !!
    !! Always allocated after ParseSpecificationPart, may be zero sized.
    TYPE(ImportStmt), ALLOCATABLE :: import_stmts(:)
    !> List of parse nodes for child constructs and statements other 
    !! that those stored above.
    !!
    !! This includes all statements from the /declaration-construct/ part 
    !! and from the /implicit-part/ (i.e. we merge together the two separate 
    !! stores of the relevant StNode).
    TYPE(SCParseNodeElement), ALLOCATABLE :: children(:)
  CONTAINS
    PROCEDURE :: Dump => spec_Dump
    PROCEDURE :: Query => spec_Query
    PROCEDURE :: Modify => spec_Modify
    PROCEDURE, NOPASS :: What => spec_What
    PROCEDURE :: Advertise => spec_Advertise
    PROCEDURE :: Check => spec_Check
  END TYPE SpecificationPart
  
  ! Some specification part constructs can themselves contain a specification 
  ! part (namely interface bodies), which cause a source dependency issue, 
  ! so we use external procedures to do the actual parsing.
  INTERFACE
    RECURSIVE SUBROUTINE ParseImplicitPartStmtListExternal( stmts,  &
        children, err_list )
      USE Errors
      USE ParseNodes
      USE StNodes
      IMPLICIT NONE
      !-------------------------------------------------------------------------
      TYPE(StNodeList), INTENT(INOUT) :: stmts(:)
      TYPE(SCParseNodeElement), INTENT(OUT), ALLOCATABLE :: children(:)
      TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    END SUBROUTINE ParseImplicitPartStmtListExternal
  
    RECURSIVE SUBROUTINE ParseDeclarationConstructListExternal( stmts,  &
        children, err_list )
      USE Errors
      USE ParseNodes
      USE StNodes
      IMPLICIT NONE
      !-------------------------------------------------------------------------
      TYPE(StNodeList), INTENT(INOUT) :: stmts(:)
      TYPE(SCParseNodeElement), INTENT(OUT), ALLOCATABLE :: children(:)
      TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    END SUBROUTINE ParseDeclarationConstructListExternal
  END INTERFACE
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a SpecStNode.
  !!
  !! @param[in]     st                The statement tree node for the 
  !! specification part to parse.
  !!
  !! @param[out]    spec              The results of the parse.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! Recursive because specification parts may contain other specification 
  !! parts (via interface bodies).
  !!
  !! This also includes the call to the Init binding to intially define 
  !! the parent component.
  
  RECURSIVE SUBROUTINE ParseSpecificationPart(st, spec, err_list)
    
    USE Errors
    USE ModuleStNodes
    USE InterfaceBodyStNodes
    USE SpecStNodes
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(SpecStNode), INTENT(INOUT) :: st
    TYPE(SpecificationPart), INTENT(OUT) :: spec
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    ! Initially define the ParseNode parent component.
    CALL spec%Init(st)
    
    !---------------------------------------------------------------------------
    ! Process use and import statements.
    
    CALL process_use_and_import_statements(st, spec, sub_err_list)
    CALL Add(err_list, sub_err_list)
    
    !---------------------------------------------------------------------------
    ! Statements in the /implicit-part/.
    
    ALLOCATE(spec%children(0))
    CALL ParseImplicitPartStmtListExternal(  &
        st%implicit_part_nodes,  &
        spec%children,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list)
    
    !---------------------------------------------------------------------------
    ! Statements in the /declaration-construct/ part.
    
    CALL ParseDeclarationConstructListExternal(  &
        st%children,  &
        spec%children,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list)
    
  END SUBROUTINE ParseSpecificationPart
  
  
  !*****************************************************************************
  !!
  !> Worker routine for ParseSpecificationPart to handle the use_stmts 
  !! component of the SpecStNode object.
  !!
  !! @param[in]     st                The statement tree node for the 
  !! specification part.
  !!
  !! @param[in,out] spec              The results of parsing the specification 
  !! part.
  !!
  !! @param[in,out] err_list          Progressive list of errors.  Note that 
  !! this is INOUT.
  
  SUBROUTINE process_use_and_import_statements(st, spec, err_list)
    
    USE Errors
    USE ImportStmts
    USE SpecStNodes
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(SpecStNode), INTENT(INOUT) :: st
    TYPE(SpecificationPart), INTENT(INOUT) :: spec
    TYPE(Error), INTENT(INOUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Error object for child procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    INTEGER :: i              ! Statement index.
    INTEGER :: i_use          ! Use statement index.
    INTEGER :: i_import       ! Import statement index.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    ! Count the number of use statements, progressively add to the list of 
    ! import names.
    ALLOCATE(spec%use_stmts(COUNT(st%use_stmts%ist == istUse)))
    ALLOCATE(spec%import_stmts(COUNT(st%use_stmts%ist == istImport)))
    
    ! Statements are ordered use-stmts, followed by import statements, so 
    ! don't have to keep separate tallies, but we are lazy.
    i_use = 1
    i_import = 1
    DO i = 1, SIZE(st%use_stmts)
      SELECT CASE (st%use_stmts(i)%ist)
      CASE (istUse)
        CALL ParseUseStmt(  &
            st%use_stmts(i)%tlist,  &
            spec%use_stmts(i_use),  &
            sub_err_list )
        CALL Add(err_list, sub_err_list)
        i_use = i_use + 1
        
      CASE (istImport)
        CALL ParseImportStmt(  &
            st%use_stmts(i)%tlist,  &
            spec%import_stmts(i_import),  &
            sub_err_list )
        CALL Add(err_list, sub_err_list)
        i_import = i_import + 1
        
      CASE DEFAULT
        STOP 'Internal error in &
            &SpecificationParts%%process_use_and_import_statements, &
            &unknown statement type in the first part of the specification &
            &section.'
      END SELECT
    END DO
    
  END SUBROUTINE process_use_and_import_statements
  
  
  !*****************************************************************************
  !!
  !> Implementation of SpecificationPart%Check.
  !!
  !! @param[in]     pnode             The parse node being checked.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE spec_Check(pnode, err_list)
    
    USE Errors
    USE ErrorCodes
    USE ParameterStmts
    USE SourceLocations
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(SpecificationPart), INTENT(IN) :: pnode
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Implicit statement status.
    !
    ! 0 - Not seen.
    ! 1 - type spec form
    ! 2 - IMPLICIT NONE.
    INTEGER :: impl
    
    ! The combined letter specs that arise from each statement.
    TYPE(LetterSpec), ALLOCATABLE :: all_impl_letters(:)
    
    INTEGER :: i              ! Child construct index.
    
    ! Location of the first PARAMETER statement.
    TYPE(SourceLocation), ALLOCATABLE :: first_parameter_location
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    impl = 0
    ALLOCATE(all_impl_letters(0))
    
    DO i = 1, SIZE(pnode%children)
      SELECT TYPE (child => pnode%children(i)%item)
      CLASS IS (ImplicitStmt)
        IF (SIZE(child%specs) == 0) THEN
          ! Implicit none.
          IF (impl == 1) THEN
            CALL Add( err_list,  &
                CODE=errSyntax,  &
                COMPONENT='C582 on R560',  &
                LOCATION=child%GetLocation(),  &
                MSG='If IMPLICIT NONE is specified in a scoping unit &
                  &there shall be no other IMPLICIT statements in &
                  &the scoping unit.' )
          ELSE IF (impl == 2) THEN
            CALL Add( err_list,  &
                CODE=errSyntax,  &
                COMPONENT='C582 on R560',  &
                LOCATION=child%GetLocation(),  &
                MSG='There are multiple IMPLICIT NONE statements in this &
                  &scoping unit.' )
          ELSE
            impl = 2
          END IF
          IF (ALLOCATED(first_parameter_location)) THEN
            CALL Add( err_list,  &
                CODE=errSyntax,  &
                COMPONENT='C582 on R560',  &
                LOCATION=first_parameter_location,  &
                MSG='This parameter statement precedes the IMPLICIT NONE &
                  &statement at ' // .ToString. child%GetLocation() // '.' )
          END IF
        ELSE
          IF (impl == 2) THEN
            CALL Add( err_list,  &
                CODE=errSyntax,  &
                COMPONENT='C582 on R560',  &
                LOCATION=child%GetLocation(),  &
                MSG='If IMPLICIT NONE is specified in a scoping unit &
                  &there shall be no other IMPLICIT statements in &
                  &the scoping unit.' )
          ELSE
            ! We deliberately don't test for overlap within the 
            ! statement here, to avoid a repeated error message 
            ! (that condition is handled by the testing code for the 
            ! statement itself.
            CALL test_impl_overlap(all_impl_letters, child%specs, err_list)
            CALL combine_impl_specs(all_impl_letters, child%specs)
            impl = 1
          END IF
        END IF
        
      CLASS IS (ParameterStmt)
        first_parameter_location = child%GetLocation()
        
      END SELECT
    END DO
    
  END SUBROUTINE spec_Check
  
  
  !*****************************************************************************
  !!
  !> Test whether an implicit spec overlaps existing letter specs.
  !!
  !! @param[in]     letters           The existing letter specs.
  !!
  !! @param[in]     spec              The spec to test against.
  !!
  !! @param[in,out] err_list          List of errors.  Note this is INOUT.
  
  SUBROUTINE test_impl_overlap(letters, specs, err_list)
    
    USE Errors
    USE ErrorCodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(LetterSpec), INTENT(IN) :: letters(:)
    TYPE(ImplicitSpec), INTENT(IN) :: specs(:)
    TYPE(Error), INTENT(INOUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: ileft
    INTEGER :: ispec
    INTEGER :: iright
    
    !***************************************************************************
    
    DO ileft = 1, SIZE(letters)
      DO ispec = 1, SIZE(specs)
        DO iright = 1, SIZE(specs(ispec)%letters)
          IF (TestLetterSpecsOverlap(  &
              letters(ileft),  &
              specs(ispec)%letters(iright) )) THEN
            CALL Add( err_list,  &
                CODE=errOverlappingImplicitSpecs,  &
                COMPONENT='5.5p2',  &
                LOCATION=specs(ispec)%letters(iright)%GetLocation(),  &
                MSG='There is an overlap between the letters in this &
                  &/letter-spec/ and the /letter-spec/''s at ' &
                  // .ToString. letters(ileft)%GetLocation()  &
                  // '.' )
            
          END IF
        END DO
      END DO
    END DO
    
  END SUBROUTINE test_impl_overlap
  
  
  !*****************************************************************************
  !!
  !> Add a set of letter specs to an existing set of letters.
  !!
  !! @param[in,out] letters           The existing set of letters.
  !!
  !! @param[in]     specs             The specs to add.
  
  SUBROUTINE combine_impl_specs(letters, specs)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(LetterSpec), INTENT(INOUT), ALLOCATABLE :: letters(:)
    TYPE(ImplicitSpec), INTENT(IN) :: specs(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: is
    
    !***************************************************************************
    
    DO is = 1, SIZE(specs)
      letters = [letters, specs(is)%letters]
    END DO
    
  END SUBROUTINE combine_impl_specs
  
  
  !*****************************************************************************
  !!
  !> Implementation of SpecificationPart%Dump - write a debugging 
  !! representation of the results of parsing a specification part to a 
  !! logical unit.
  !!
  !! @param[in]     pnode             The specification part to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix for each record.
  
  ! Recursive because specification parts may contain other specification 
  ! parts (via interface bodies).
  RECURSIVE SUBROUTINE spec_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(SpecificationPart), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Child node index.
    
    !***************************************************************************
    
    !---------------------------------------------------------------------------
    ! Use statements.
    
    DO i = 1, SIZE(pnode%use_stmts)
      WRITE (unit, "(A,'/use-stmt/')") pfx
      CALL pnode%use_stmts(i)%Dump(unit, pfx // '  ')
    END DO
    
    !---------------------------------------------------------------------------
    ! import statements.
    
    DO i = 1, SIZE(pnode%import_stmts)
      WRITE (unit, "(A,'/import-stmt/')") pfx
      CALL pnode%import_stmts(i)%Dump(unit, pfx // '  ')
    END DO
    
    !---------------------------------------------------------------------------
    ! Other statements.
    
    DO i = 1, SIZE(pnode%children)
      WRITE (unit, "(A,'/',A,'/')") pfx, pnode%Children(i)%item%What()
      CALL pnode%Children(i)%item%Dump(unit, pfx // '  ')
    END DO
    
  END SUBROUTINE spec_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of SpecificationPart%What - return a descriptive string 
  !! for the parse node.
  !!
  !! @returns 'specification-part'.
  
  FUNCTION spec_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'specification-part'
    
  END FUNCTION spec_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of SpecificationPart%Query.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  RECURSIVE SUBROUTINE spec_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(SpecificationPart), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Child node index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%use_stmts)
      CALL pnode%use_stmts(i)%Query(visitor)
    END DO
    
    DO i = 1, SIZE(pnode%import_stmts)
      CALL pnode%import_stmts(i)%Query(visitor)
    END DO
    
    DO i = 1, SIZE(pnode%children)
      CALL pnode%children(i)%Query(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE spec_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of SpecificationPart%Modify.
  !!
  !! @param[in,out] pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  RECURSIVE SUBROUTINE spec_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(SpecificationPart), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Child node index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%use_stmts)
      CALL pnode%use_stmts(i)%Modify(visitor)
    END DO
    
    DO i = 1, SIZE(pnode%import_stmts)
      CALL pnode%import_stmts(i)%Modify(visitor)
    END DO
    
    DO i = 1, SIZE(pnode%children)
      CALL pnode%children(i)%Modify(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE spec_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of SpecificationPart%Advertise - advertise the identifiers 
  !! referenced or defined by the part.
  !!
  !! @param[in]     pnode             The node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE spec_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(SpecificationPart), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(0))
    
  END SUBROUTINE spec_Advertise
  
END MODULE SpecificationParts
