! $Id: StopStmts.f90 2161 2016-06-01 23:18:33Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the StopStmts module.


!*******************************************************************************
!!
!> Procedures for parsing /stop-stmt/'s and /error-stop-stmt/'s.

MODULE StopStmts
  
  USE Expressions
  USE ParseNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseStopStmt
  PUBLIC :: ParseErrorStopStmt
  
  !-----------------------------------------------------------------------------
  
  !> Stores the results of parsing a /stop-stmt/
  TYPE, EXTENDS(SCParseNode), PUBLIC :: StopStmt
    !> The stop code, if any provided on the statement.  Not allocated if 
    !! no stop code provided.
    CLASS(Expression), ALLOCATABLE :: stop_code
  CONTAINS
    PROCEDURE :: Dump => stop_Dump
    PROCEDURE :: Query => stop_Query
    PROCEDURE :: Modify => stop_Modify
    PROCEDURE, NOPASS :: What => stop_What
  END TYPE StopStmt
  
  !> Stores the results of parsing an /error-stop-stmt/.
  TYPE, EXTENDS(StopStmt), PUBLIC :: ErrorStopStmt
  CONTAINS
    PROCEDURE, NOPASS :: What => errstop_What
  END TYPE ErrorStopStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a /stop-stmt/ (R855).
  !!
  !! @param[in]     tlist             The sequence of tokens that make up the 
  !! statement.
  !!
  !! @param[out]    stop_stmt         The results of parsing the statement.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseStopStmt(tlist, stop_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE Expressions
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(StopStmt), INTENT(OUT) :: stop_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /stop-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R855'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    
    ! Error list for child Parse calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
          MSG='Expected /stop-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! STOP ...
    
    CALL Expect(tlist, 1, ittName, scck_'STOP', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! Stop code is optional
    IF (2 > SIZE(tlist)) RETURN
    
    !---------------------------------------------------------------------------
    ! The stop code - an expression.
    
    CALL ParseExpr(tlist, 2, SIZE(tlist), stop_stmt%stop_code, sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    
  END SUBROUTINE ParseStopStmt
  
  
  !*****************************************************************************
  !!
  !> Parse an /error-stop-stmt/ (R856).
  !!
  !! @param[in]     tlist             The sequence of tokens that make up the 
  !! statement.
  !!
  !! @param[out]    stop_stmt         The results of parsing the statement.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseErrorStopStmt(tlist, stop_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE Expressions
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(ErrorStopStmt), INTENT(OUT) :: stop_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /stop-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R856'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    
    ! Error list for child Parse calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
          MSG='Expected /error-stop-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! ERROR STOP ...
    
    CALL Expect(tlist, 1, ittName, scck_'ERROR', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL Expect(tlist, 2, ittName, scck_'STOP', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! Stop code is optional
    IF (3 > SIZE(tlist)) RETURN
    
    !---------------------------------------------------------------------------
    ! The stop code - an expression.
    
    CALL ParseExpr(tlist, 3, SIZE(tlist), stop_stmt%stop_code, sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    
  END SUBROUTINE ParseErrorStopStmt
  
  
  !*****************************************************************************
  !!
  !> Implementation of StopStmt%Dump - write a debugging representation 
  !! of the results of parsing a stop statement to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE stop_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(StopStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%stop_code)) THEN
      WRITE (unit, "(A,'/stop-code/ = ')", ADVANCE='NO') pfx
      CALL WriteExpr(unit, pnode%stop_code)
      WRITE (unit, "()")
    END IF
    
  END SUBROUTINE stop_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of StopStmt%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE stop_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(StopStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%stop_code)) THEN
      CALL pnode%stop_code%Query(visitor)
    END IF
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE stop_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of StopStmt%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[in,out] pnode             The parse node element to modify.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE stop_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(StopStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%stop_code)) THEN
      CALL pnode%stop_code%Modify(visitor)
    END IF
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE stop_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of StopStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'stop-stmt'.
  
  FUNCTION stop_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'stop-stmt'
    
  END FUNCTION stop_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of ErrorStopStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'error-stop-stmt'.
  
  FUNCTION errstop_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'error-stop-stmt'
    
  END FUNCTION errstop_What
  
END MODULE StopStmts
