! $Id: SyncStmts.f90 2795 2019-03-22 09:15:08Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the SyncStmts module.


!*******************************************************************************
!!
!> Procedures for parsing /sync-all-stmt/'s, /sync-image-stmt/'s and 
!! /sync-memory-stmt/'s.
!!
!! ParseSyncStat is also used by the /unlock-stmt/ parsing procedure.

MODULE SyncStmts
  
  USE Expressions
  USE ParseNodes
  USE Variables
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseSyncAllStmt
  PUBLIC :: ParseSyncImagesStmt
  PUBLIC :: ParseSyncMemoryStmt
  
  PUBLIC :: ParseSyncStat
  PUBLIC :: WriteSyncStat
  
  !> Results of parsing a /sync-stmt/.
  TYPE, EXTENDS(SCParseNode), ABSTRACT, PUBLIC :: SyncStat
    !> The STAT specification.
    !!
    !! Not allocated if there was no such specification.
    CLASS(Variable), ALLOCATABLE :: stat
    !> The ERRMSG specification.
    !!
    !! Not allocated if there was no such specification.
    CLASS(Variable), ALLOCATABLE :: errmsg
  CONTAINS
    PROCEDURE :: Dump => sync_Dump
    PROCEDURE :: Query => sync_Query
    PROCEDURE :: Modify => sync_Modify
  END TYPE SyncStat
  
  TYPE, EXTENDS(SyncStat), PUBLIC :: SyncAllStmt
  CONTAINS
    PROCEDURE, NOPASS :: What => syncall_What
  END TYPE SyncAllStmt
  
  TYPE, EXTENDS(SyncStat), PUBLIC :: SyncImagesStmt
    !> The image set expression.  If not allocated the set is "*".
    CLASS(Expression), ALLOCATABLE :: image_set
  CONTAINS
    PROCEDURE :: Dump => syncimages_Dump
    PROCEDURE :: Query => syncimages_Query
    PROCEDURE :: Modify => syncimages_Modify
    PROCEDURE, NOPASS :: What => syncimages_What
  END TYPE SyncImagesStmt
  
  TYPE, EXTENDS(SyncStat), PUBLIC :: SyncMemoryStmt
  CONTAINS
    PROCEDURE, NOPASS :: What => syncmem_What
  END TYPE SyncMemoryStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a /sync-all-stmt/ (R858).
  !!
  !! @param[in]     tlist             The sequence of tokens that make up the 
  !! statement.
  !!
  !! @param[out]    sync__all_stmt    The results of parsing the statement.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseSyncAllStmt(tlist, sync_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(SyncAllStmt), INTENT(OUT) :: sync_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /sync-all-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R858'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: closing        ! Index of the closing parenthesis.
    INTEGER :: list_count     ! Number of items in the /sync-stat-list/.
    INTEGER :: item           ! List item index.
    INTEGER :: item_start     ! Start index in tlist of a particular item.
    INTEGER :: item_finish    ! Finish index in tlist of a particular item.
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
          MSG='Expected /sync-all-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! SYNC ALL ...
    
    CALL ExpectKeyword(tlist, 1, scck_'SYNC', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL ExpectKeyword(tlist, 2, scck_'ALL', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! That can be it.  If so, return leaving all components of sync_stmt 
    ! unallocated.
    IF (2 == SIZE(tlist)) RETURN
    
    !---------------------------------------------------------------------------
    ! Optional (...)
    
    CALL Expect(tlist, 3, ittDelimiter, scck_'(', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL ExpectClosing(tlist, 3, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! Optional /sync-stat-list/.
    
    IF (closing > 4) THEN
      list_count = ListItemCount(tlist(4:closing-1))
      item_finish = 0
      DO item = 1, list_count
        CALL GetNextListItem(tlist, 4, closing - 1, item_start, item_finish)
        CALL ParseSyncStat( tlist, item_start, item_finish,  &
            sync_stmt, sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
        ! Continue on error.
      END DO
    END IF
    
    ! Nothing else!
    CALL ExpectEOS(tlist, closing + 1, comp, err_list)
    
  END SUBROUTINE ParseSyncAllStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a /sync-all-stmt/ (R858).
  !!
  !! @param[in]     tlist             The sequence of tokens that make up the 
  !! statement.
  !!
  !! @param[out]    sync_stat         The results of parsing the 
  !! /sync-stat-list/.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseSyncImagesStmt(tlist, sync_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE Expressions
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(SyncImagesStmt), INTENT(OUT) :: sync_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /sync-images-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R860'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: closing        ! Index of the closing parenthesis.
    INTEGER :: list_count     ! Number of items in the /sync-stat-list/.
    INTEGER :: item           ! List item index.
    INTEGER :: item_start     ! Start index in tlist of a particular item.
    INTEGER :: item_finish    ! Finish index in tlist of a particular item.
    LOGICAL :: got_set        ! Flag for the image set.
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /sync-images-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! SYNC IMAGES ...
    
    CALL ExpectKeyword(tlist, 1, scck_'SYNC', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL ExpectKeyword(tlist, 2, scck_'IMAGES', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! ... ( /image-set/ ... )
    
    CALL Expect(tlist, 3, ittDelimiter, scck_'(', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL ExpectClosing(tlist, 3, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    got_set = .FALSE.
    
    list_count = ListItemCount(tlist(4:closing-1))
    item_finish = 0
    DO item = 1, list_count
      CALL GetNextListItem(tlist, 4, closing - 1, item_start, item_finish)
      IF (.NOT. got_set) THEN
        IF (item_start > item_finish) THEN
          CALL Add( err_list,  &
              CODE=errSyntax,  &
              COMPONENT=comp,  &
              LOCATION=QueryLocation(tlist(item_start)),  &
              MSG='Expected /image-set/ ("*" or /int-expr/).' )
        ELSE IF (Match(tlist, item_start, ittOperator, scck_'*')) THEN
          CALL ExpectEnd(tlist(:item_finish), item_start + 1, comp, err_list)
        ELSE
          CALL ParseExpr(  &
              tlist,  &
              item_start, item_finish,  &
              sync_stmt%image_set,  &
              sub_err_list )
          CALL Add(err_list, sub_err_list, COMPONENT=comp)
        END IF
        got_set = .TRUE.
      ELSE
        CALL ParseSyncStat(  &
            tlist,  &
            item_start, item_finish,  &
            sync_stmt,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      ! Continue on error.
    END DO
    
    ! Nothing else!
    CALL ExpectEOS(tlist, closing + 1, comp, err_list)
    
  END SUBROUTINE ParseSyncImagesStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a /sync-memory-stmt/ (R862).
  !!
  !! @param[in]     tlist             The sequence of tokens that make up the 
  !! statement.
  !!
  !! @param[out]    sync_stmt         The results of parsing the statement.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseSyncMemoryStmt(tlist, sync_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(SyncMemoryStmt), INTENT(OUT) :: sync_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /sync-memory-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R862'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: closing        ! Index of the closing parenthesis.
    INTEGER :: list_count     ! Number of items in the /sync-stat-list/.
    INTEGER :: item           ! List item index.
    INTEGER :: item_start     ! Start index in tlist of a particular item.
    INTEGER :: item_finish    ! Finish index in tlist of a particular item.
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /sync-memory-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! SYNC ALL ...
    
    CALL ExpectKeyword(tlist, 1, scck_'SYNC', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL ExpectKeyword(tlist, 2, scck_'MEMORY', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! That can be it.  If so, return leaving all components of sync_stmt 
    ! unallocated.
    IF (2 == SIZE(tlist)) RETURN
    
    !---------------------------------------------------------------------------
    ! Optional (...)
    
    CALL Expect(tlist, 3, ittDelimiter, scck_'(', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL ExpectClosing(tlist, 3, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! Optional /sync-stat-list/.
    
    IF (closing > 4) THEN
      list_count = ListItemCount(tlist(4:closing-1))
      item_finish = 0
      DO item = 1, list_count
        CALL GetNextListItem(tlist, 4, closing - 1, item_start, item_finish)
        CALL ParseSyncStat(  &
            tlist,  &
            item_start, item_finish,  &
            sync_stmt,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
        ! Continue on error.
      END DO
    END IF
    
    ! Nothing else!
    CALL ExpectEOS(tlist, closing + 1, comp, err_list)
    
  END SUBROUTINE ParseSyncMemoryStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /sync-stat/ (R859).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the first token of the 
  !! /sync-stat/ in @a tlist.
  !!
  !! @param[in]     finish            Index of the last token of the 
  !! /sync-stat/ in @a tlist.
  !!
  !! @param[in,out] sync_stat         Results of parsing the /sync-stat/.  
  !! Note that this is an INTENT(INOUT) argument - each spec sets a particular 
  !! component
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseSyncStat( tlist, start, finish, sync_stat,  &
      err_list )
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE IoControlSpecs
    USE ParseUtils
    USE Scanner
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    CLASS(SyncStat), INTENT(INOUT) :: sync_stat
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /sync-stat/.
    CHARACTER(*), PARAMETER :: comp = 'R859'
    
    ! Keywords for each /sync-stat/.
    CHARACTER(*,KIND=scck), PARAMETER :: keywords(2) = [  &
        scck_'STAT  ', scck_'ERRMSG' ]
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: opt            ! Expect option.
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    IF (finish < start) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          LOCATION=SafeLocation(tlist, finish),  &
          MSG='Expected /sync-stat/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! keyword ...
    
    CALL ExpectKeyword(tlist, start, keywords, '/sync-stat/', opt, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! ... = ...
    CALL Expect(tlist, start + 1, ittAssign, scck_'=', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! stat specific thing (currently both variables).
    
    SELECT CASE (opt)
    CASE (1)      ! STAT
      IF (ALLOCATED(sync_stat%stat)) THEN
        CALL already('STAT')
      ELSE
        CALL ParseVariable(  &
            tlist, &
            start + 2, finish,  &
            sync_stat%stat,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (2)      ! ERRMSG
      IF (ALLOCATED(sync_stat%errmsg)) THEN
        CALL already('ERRMSG')
      ELSE
        CALL ParseVariable(  &
            tlist,  &
            start + 2, finish,  &
            sync_stat%errmsg,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE DEFAULT
      STOP 'Unknown /sync-stat/.'
      
    END SELECT
    
  CONTAINS
    
    SUBROUTINE already(text)
      USE ErrorCodes
      !-------------------------------------------------------------------------
      ! Arguments
      CHARACTER(*), INTENT(IN) :: text
      !*************************************************************************
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT='C851 on ' // comp,  &
          LOCATION=QueryLocation(tlist(start)),  &
          MSG=text // ' has already been specified for this statement.' )
    END SUBROUTINE already
    
  END SUBROUTINE ParseSyncStat
  
  
  !*****************************************************************************
  !!
  !> Write a debugging representation of a /sync-stat-list/ to a formatted 
  !! file.
  !!
  !! @param[in]     log_unit          Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     sync_stat         The /sync-stat-list/ to output.
  
  SUBROUTINE WriteSyncStat(log_unit, sync_stat)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    INTEGER, INTENT(IN) :: log_unit
    CLASS(SyncStat), INTENT(IN) :: sync_stat
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    LOGICAL :: need_comma     ! Flag that a separating comma is required.
    
    !***************************************************************************
    
    need_comma = .FALSE.
    
    IF (ALLOCATED(sync_stat%stat)) THEN
      IF (need_comma) WRITE (log_unit, "(', ')", ADVANCE='NO')
      WRITE (log_unit, "('STAT=')", ADVANCE='NO')
      CALL WriteVariable(log_unit, sync_stat%stat)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(sync_stat%errmsg)) THEN
      IF (need_comma) WRITE (log_unit, "(', ')", ADVANCE='NO')
      WRITE (log_unit, "('ERRMSG=')", ADVANCE='NO')
      CALL WriteVariable(log_unit, sync_stat%errmsg)
      need_comma = .TRUE.
    END IF
    
  END SUBROUTINE WriteSyncStat
  
  
  !*****************************************************************************
  !!
  !> Implementation of SyncStat%Dump - write a debugging representation 
  !! of the results of parsing a sync all or sync memory statement to 
  !! a logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE sync_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(SyncStat), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    WRITE (unit, "(A,'/sync-stat-list/ = ')", ADVANCE='NO') pfx
    CALL WriteSyncStat(unit, pnode)
    WRITE (unit, "()")
    
  END SUBROUTINE sync_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of SyncStat%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE sync_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(SyncStat), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%stat)) CALL pnode%stat%Query(visitor)
    IF (ALLOCATED(pnode%errmsg)) CALL pnode%errmsg%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE sync_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of SyncStat%Modif - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[in,out] pnode             The parse node element to query.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE sync_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(SyncStat), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%stat)) CALL pnode%stat%Modify(visitor)
    IF (ALLOCATED(pnode%errmsg)) CALL pnode%errmsg%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE sync_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of SyncAllStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'sync-all-stmt'.
  
  FUNCTION syncall_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'sync-all-stmt'
    
  END FUNCTION syncall_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of SyncImages%Dump - write a debugging representation 
  !! of the results of parsing a wait statement to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE syncimages_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(SyncImagesStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    WRITE (unit, "(A,'/image-set/ = ')", ADVANCE='NO') pfx
    IF (ALLOCATED(pnode%image_set)) THEN
      CALL WriteExpr(unit, pnode%image_set)
    ELSE
      WRITE (unit, "('*')", ADVANCE='NO')
    END IF
    WRITE (unit, "()")
    
    CALL sync_Dump(pnode, unit, pfx)
    
  END SUBROUTINE syncimages_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of SyncImagesStmt%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE syncimages_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(SyncImagesStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%stat)) CALL pnode%stat%Query(visitor)
    IF (ALLOCATED(pnode%errmsg)) CALL pnode%errmsg%Query(visitor)
    IF (ALLOCATED(pnode%image_set)) CALL pnode%image_set%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE syncimages_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of SyncImagesStmt%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[in,out  pnode             The parse node element to modify.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE syncimages_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(SyncImagesStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%stat)) CALL pnode%stat%Modify(visitor)
    IF (ALLOCATED(pnode%errmsg)) CALL pnode%errmsg%Modify(visitor)
    IF (ALLOCATED(pnode%image_set)) CALL pnode%image_set%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE syncimages_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of SyncImagesStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'sync-images-stmt'.
  
  FUNCTION syncimages_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'sync-images-stmt'
    
  END FUNCTION syncimages_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of SyncMemoryStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'sync-memory-stmt'.
  
  FUNCTION syncmem_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'sync-memory-stmt'
    
  END FUNCTION syncmem_What
  
END MODULE SyncStmts
