! $Id: TypeParamDefStmts.f90 2795 2019-03-22 09:15:08Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the TypeParamDefStmts module.


!*******************************************************************************
!!
!> Procedures and type for parsing /type-param-def-stmt/'s.

MODULE TypeParamDefStmts
  
  USE CompilerKinds
  USE Expressions
  USE Identifiers
  USE ParseNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseTypeParamDefStmt
  
  !-----------------------------------------------------------------------------
  
  !> The type parameter is a kind type parameter.
  INTEGER, PARAMETER, PUBLIC :: itattrKind = 1
  !> The type parameter is a length type parameter.
  INTEGER, PARAMETER, PUBLIC :: itattrLen = 2
  
  !> Stores the results of parsing a /type-param-decl/.
  TYPE, EXTENDS(TermParseNode), PUBLIC :: TypeParamDecl
    !> The name of the type parameter.
    TYPE(Identifier) :: name
    !> The default value for the type parameter.
    !!
    !! Not allocated if no default parameter value was provided.
    CLASS(Expression), ALLOCATABLE :: default
    !> The attributes for the parameter from the parent statement.
    !!
    !! The object for the parent statement owns the associated object.
    INTEGER, POINTER :: attr => NULL()
  CONTAINS
    PROCEDURE :: Write => tpdecl_Write
    PROCEDURE :: Query => tpdecl_Query
    PROCEDURE :: Modify => tpdecl_Modify
    PROCEDURE, NOPASS :: What => tpdecl_What
    PROCEDURE :: Advertise => tpdecl_Advertise
    PROCEDURE :: GetLocation  => tpdecl_GetLocation
  END TYPE TypeParamDecl
  
  !> Represents a /type-param-def-stmt/.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: TypeParamDefStmt
    !> The expression for the kind of the parameter, if present.
    !!
    !! Not allocated if not present.
    CLASS(Expression), ALLOCATABLE :: kind_expr
    
    !> The parameter type - an itattr* value.
    !!
    !! This is a pointer to allow aliasing with the TypeParamDecl type.  
    !! The TypeParamDefStmt object owns the associated object.  Its lifetime 
    !! is maanged with the scalar finalizer.
    INTEGER, POINTER :: attr => NULL()
    
    !> The type parameter declarations.
    !!
    !! Always allocated, may be zero sized.
    TYPE(TypeParamDecl), ALLOCATABLE :: params(:)
  CONTAINS
    PROCEDURE :: Dump => tp_Dump
    PROCEDURE :: Query => tp_Query
    PROCEDURE :: Modify => tp_Modify
    PROCEDURE, NOPASS :: What => tp_What
    PROCEDURE :: Advertise => tp_Advertise
    
    FINAL :: tp_final_scalar, tp_final_rank1
  END TYPE TypeParamDefStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a type-param-def-stmt.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    tpd_stmt          The results of parsing the statement.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseTypeParamDefStmt(tlist, tpd_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(TypeParamDefStmt), INTENT(OUT) :: tpd_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /type-param-def-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R431'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Error object and list for procedure calls.
    TYPE(Error) :: err
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    INTEGER :: it           ! Current token index.
    INTEGER :: closing      ! Index of closing parenthesis.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /type-param-decl-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! The equivalent of the type spec
    
    ! We roll our own Expect here to provide a better error message.
    IF (.NOT. Match(tlist, 1, ittName, scck_'INTEGER')) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          LOCATION=QueryLocation(tlist(1)),  &
          MSG='The type specified in a /type-param-def-stmt/ must be &
            &"INTEGER".' )
    END IF
    
    !---------------------------------------------------------------------------
    ! The kind-selector.
    
    IF (Match(tlist, 2, ittDelimiter, scck_'(')) THEN
      CALL ExpectClosing(tlist, 2, closing, err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN    ! Too hard to recover from mismatch.
      
      ! KIND= specifier is optional.
      IF (Match(tlist, 3, ittName, scck_'KIND')) THEN
        CALL Expect(tlist, 4, ittAssign, scck_'=', err)
        CALL Add(err_list, err, COMPONENT=comp)
        it = 5
      ELSE
        it = 3
      END IF
      
      CALL ParseExpr(tlist, it, closing - 1, tpd_stmt%kind_expr, sub_err_list)
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
      
      it = closing + 1
    ELSE
      it = 2
    END IF
    
    CALL Expect(tlist, it, ittComma, scck_',', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    it = it + 1
    
    ALLOCATE(tpd_stmt%attr)
    CALL ExpectKeyword(  &
        tlist, it,  &
        [scck_'KIND', scck_'LEN '],  &
        '/type-param-attr-spec/',  &
        tpd_stmt%attr,  &
        err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    it = it + 1
    
    CALL Expect(tlist, it, ittDeclare, scck_'::', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    it = it + 1
    
    !---------------------------------------------------------------------------
    ! The list of parameters, with optional default values.
    
    CALL parse_type_param_decl_list(  &
        tlist(it:),  &
        tpd_stmt%attr,  &
        tpd_stmt%params,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
     
  END SUBROUTINE ParseTypeParamDefStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that make up a /type-param-decl-list/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     attr              The attributes of the type 
  !! parameter declaration in the parent statement.
  !!
  !! @param[out]    decls             The type parameter declarations.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE parse_type_param_decl_list(tlist, attr, decls, err_list)
    
    USE Errors
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN), POINTER :: attr
    TYPE(TypeParamDecl), INTENT(OUT), ALLOCATABLE :: decls(:)
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: list_count     ! Number of items in the /implicit-spec-list/.
    INTEGER :: item           ! List item index.
    INTEGER :: item_start     ! Start index in tlist of a particular item.
    INTEGER :: item_finish    ! Finish index in tlist of a particular item.
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------    
    
    list_count = ListItemCount(tlist)
    ALLOCATE(decls(list_count))
    item_finish = 0
    DO item = 1, list_count
      CALL GetNextListItem(tlist, 1, SIZE(tlist), item_start, item_finish)
      CALL parse_type_param_decl(  &
          tlist(item_start:item_finish),  &
          attr,  &
          decls(item),  &
          sub_err_list )
      CALL Add(err_list, sub_err_list)
      IF (Failure(sub_err_list)) RETURN
    END DO
    
  END SUBROUTINE parse_type_param_decl_list
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that make up a /type-param-decl/ (R432).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     attr              The attributes of the type 
  !! parameter declaration in the parent statement.
  !!
  !! @param[out]    decl              The type parameter declaration.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE parse_type_param_decl(tlist, attr, decl, err_list)
    
    USE CompilerKinds
    USE Errors
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN), POINTER :: attr
    TYPE(TypeParamDecl), INTENT(OUT) :: decl
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /type-param-decl/.
    CHARACTER(*), PARAMETER :: comp = 'R432'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------    
    
    ! Save the attributes for later nature reporting.
    decl%attr => attr
    
    CALL Expectname(tlist, 1, '/type-param-name/', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    decl%name = IdentifierIfortBug(tlist(1))
    
    !---------------------------------------------------------------------------
    ! Either an "initialiser" or that's it.
    
    IF (Match(tlist, 2, ittAssign, scck_'=')) THEN 
      CALL ParseExpr(tlist, 3, SIZE(tlist), decl%default, sub_err_list)
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
    ELSE
      CALL ExpectEnd(tlist, 2, comp, err_list)
    END IF
    
  END SUBROUTINE parse_type_param_decl
  
  
  !*****************************************************************************
  !!
  !> Implementation of TypeParamDefStmt%Dump - write a debugging 
  !! representation of the results of parsing the statement to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE tp_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(TypeParamDefStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Parameter declaration index.
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%kind_expr)) THEN
      WRITE (unit, "(A,'kind-expr = ')", ADVANCE='NO') pfx
      CALL pnode%kind_expr%Write(unit)
      WRITE (unit, "()")
    END IF
    
    WRITE (unit, "(A,'/type-param-attr-spec/ = ')", ADVANCE='NO') pfx
    SELECT CASE (pnode%attr)
    CASE (itattrKind) ; WRITE (unit, "(A)", ADVANCE='NO') 'KIND'
    CASE (itattrLen)  ; WRITE (unit, "(A)", ADVANCE='NO') 'LEN'
    CASE DEFAULT
      STOP 'Internal error in TypeParamDefStmts%%tp_Dump, &
          &unknown parameter attribute.'
    END SELECT
    WRITE (unit, "()")
    
    WRITE (unit, "(A,'/type-param-decl-list/')") pfx
    DO i = 1, SIZE(pnode%params)
      WRITE (unit, "(A)", ADVANCE='NO') pfx // '  '
      CALL pnode%params(i)%Write(unit)
      WRITE (unit, "()")
    END DO
    
  END SUBROUTINE tp_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of TypeParamDefStmt%Query - traverse the node with a 
  !! query operation.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  RECURSIVE SUBROUTINE tp_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(TypeParamDefStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: ip             ! Type param decl index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%kind_expr)) THEN
      CALL pnode%kind_expr%Query(visitor)
    END IF
    
    DO ip = 1, SIZE(pnode%params)
      CALL pnode%params(ip)%Query(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE tp_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of TypeParamDefStmt%Modify - traverse the node with a 
  !! modify operation.
  !!
  !! @param[in,out] pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  RECURSIVE SUBROUTINE tp_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(TypeParamDefStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: ip             ! Type param decl index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%kind_expr)) THEN
      CALL pnode%kind_expr%Modify(visitor)
    END IF
    
    DO ip = 1, SIZE(pnode%params)
      CALL pnode%params(ip)%Modify(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE tp_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of TypeParamDefStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'type-param-def-stmt'.
  
  FUNCTION tp_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'type-param-def-stmt'
    
  END FUNCTION tp_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of TypeParamDefStmt%Advertise - advertise the identifiers 
  !! referenced or defined by the block.
  !!
  !! @param[in]     pnode             The node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE tp_Advertise(pnode, ident_list)
    
    USE AccessStmts
    USE IdentifierNatures
    USE SourceLocations
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(TypeParamDefStmt), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(1))
    ident_list(1) = IdentifierNature(  &
        NAME='INTEGER',  &
        NATURES=[NatureRefType],  &
        LOCATIONS=[pnode%GetLocation()] )
    
  END SUBROUTINE tp_Advertise
  
  
  !*****************************************************************************
  !!
  !> Finalizer for a TypeParamDefStmt scalar.
  !!
  !! @param[in,out] scalar            The object to be finalized.
  
  SUBROUTINE tp_final_scalar(scalar)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(TypeParamDefStmt), INTENT(INOUT) :: scalar
    
    !***************************************************************************
    
    IF (ASSOCIATED(scalar%attr)) DEALLOCATE(scalar%attr)
    
  END SUBROUTINE tp_final_scalar
  
  
  !*****************************************************************************
  !!
  !> Finalizer for a TypeParamDefStmt rank one array.
  !!
  !! @param[in,out] array             The array object to be finalized.
  !!
  !! Provided this becaues we might have an array of type parameter 
  !! statement objects in a type definition object.
  
  SUBROUTINE tp_final_rank1(array)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(TypeParamDefStmt), INTENT(INOUT) :: array(:)
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Element index.
    
    !***************************************************************************
    
    DO i = 1, SIZE(array)
      CALL tp_final_scalar(array(i))
    END DO
    
  END SUBROUTINE tp_final_rank1
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for TypeParamDecl
  
  
  !*****************************************************************************
  !!
  !> Implementation of TypeParamDecl%Write - write a debugging representation 
  !! of the node to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to write.
  !!
  !! @param[in]     unit              Unit connected for formatted output.
  
  SUBROUTINE tpdecl_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(TypeParamDecl), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    WRITE (unit, "('/type-param-name/ = ',A)", ADVANCE='NO')  &
        .NameOf. pnode%name
    
    IF (ALLOCATED(pnode%default)) THEN
      WRITE (unit, "(', default value = ')", ADVANCE='NO')
      CALL pnode%default%Write(unit)
    END IF
    
  END SUBROUTINE tpdecl_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of TypeParamDecl%Query - traverse the node with a 
  !! query operation.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  RECURSIVE SUBROUTINE tpdecl_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(TypeParamDecl), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%default)) THEN
      CALL pnode%default%Query(visitor)
    END IF
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE tpdecl_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of TypeParamDecl%Modify - traverse the node with a 
  !! modify operation.
  !!
  !! @param[in,out] pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  RECURSIVE SUBROUTINE tpdecl_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(TypeParamDecl), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%default)) THEN
      CALL pnode%default%Modify(visitor)
    END IF
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE tpdecl_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of TypeParamDecl%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'type-param-def-stmt'.
  
  FUNCTION tpdecl_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'type-param-decl'
    
  END FUNCTION tpdecl_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of TypeParamDecl%Advertise - advertise the identifiers 
  !! referenced or defined by the block.
  !!
  !! @param[in]     pnode             The node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE tpdecl_Advertise(pnode, ident_list)
    
    USE AccessStmts
    USE IdentifierNatures
    USE SourceLocations
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(TypeParamDecl), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: attr_nature    ! The KIND or LEN nature.
    
    !***************************************************************************
    
    SELECT CASE (pnode%attr)
    CASE (itattrKind)   ; attr_nature = NatureDefKindParam
    CASE (itattrLen)    ; attr_nature = NatureDefLenParam
    CASE DEFAULT
      STOP 'Internal error in TypeParamDefStmts%%tpdecl_Advertise, &
          &unknown parameter attribute.'
    END SELECT
    
    ALLOCATE(ident_list(1))
    ident_list(1) = IdentifierNature(  &
        NAME=.NameOf. pnode%name,  &
        NATURES=[NatureRefTypeParameter, NatureDefHasType, attr_nature],  &
        LOCATIONS=[ .LocationOf. pnode%name, .LocationOf. pnode%name,  &
          .LocationOf. pnode%name ] )
    
  END SUBROUTINE tpdecl_Advertise
  
  
  !*****************************************************************************
  !!
  !> Implementation of TypeParamDecl%GetLocation - get the start of the 
  !! source that corresponds to the ParseNode.
  !!
  !! @param[in]     pnode           The parse node of interest
  !!
  !! @returns The location in source of the start of the ParseNode.
  
  FUNCTION tpdecl_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !-------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(TypeParamDecl), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    loc = .LocationOf. pnode%name
    
  END FUNCTION tpdecl_GetLocation
  
END MODULE TypeParamDefStmts
